Notebook[{

Cell[CellGroupData[{
Cell["precession v2: equation generator", \
"Title",ExpressionUUID->"e37cf413-9409-4e4c-ac4c-0c46d6988182"],

Cell["Davide Gerosa", \
"Subtitle",ExpressionUUID->"66fc7894-e1ab-4b53-acf5-67a3cf76bf25"],

Cell[TextData[StyleBox["This is a notebook to derive the equations describing \
the dynamics of precessing black-hole binaries at 2PN.", "Subsection"]], \
"Subtitle",ExpressionUUID->"30a0f08e-27ee-431c-b458-c77e3fff6173"],

Cell["\<\
Export expressions in a format that python and numpy are going to like. Use a \
modified version of the package by Gustavo Wiederhecker  \
https://github.com/gwiederhecker/ToPython, supplemented by a set of custom \
transformation for the precession code.  See also  \
https://mathematica.stackexchange.com/questions/10970/how-to-clear-all-\
variables-except-one\
\>", "Text",ExpressionUUID->"5b3d9249-312b-4d20-b5fc-bdd73d44f4b3"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CurrentValue", "[", 
    RowBox[{"$FrontEnd", ",", "\"\<NotebookAutoSave\>\""}], "]"}], "=", 
   "True"}], ";"}]], "Input",
 CellLabel->"In[1]:=",ExpressionUUID->"3b37a92e-ae6b-48d7-8ab0-64f0dce354e7"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"ToPython", "[", 
    RowBox[{"x_", ",", 
     RowBox[{"numpyprefix_", ":", "\"\<np\>\""}]}], "]"}], ":=", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"expression", "=", "x"}], ",", "greekrule", ",", "PythonForm", 
       ",", 
       RowBox[{"numpypre", "=", "numpyprefix"}], ",", "lp", ",", "rp", ",", 
       "a", ",", "b"}], "}"}], ",", "\n", 
     RowBox[{"(*", 
      RowBox[{
       RowBox[{"By", " ", "Gustavo", " ", "Wiederhecker"}], ",", " ", 
       RowBox[{"minor", " ", "edits", " ", "by", " ", "Davide", " ", 
        RowBox[{"Gerosa", "."}]}]}], "*)"}], "\n", 
     RowBox[{
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{"numpyprefix", "==", "\"\<\>\""}], ",", 
        RowBox[{"sep", "=", "\"\<\>\""}], ",", 
        RowBox[{"sep", "=", "\"\<.\>\""}]}], "]"}], ";", "\n", 
      RowBox[{"lp", "=", "\"\<(\>\""}], ";", "\n", 
      RowBox[{"rp", "=", "\"\<)\>\""}], ";", "\n", 
      RowBox[{
       RowBox[{"PythonForm", "[", 
        RowBox[{"Rational", "[", 
         RowBox[{"a_", ",", "b_"}], "]"}], "]"}], ":=", 
       RowBox[{
        RowBox[{"PythonForm", "[", "a", "]"}], "<>", "\"\</\>\"", "<>", 
        RowBox[{"PythonForm", "[", "b", "]"}]}]}], ";", "\n", 
      RowBox[{
       RowBox[{"PythonForm", "[", 
        RowBox[{"Complex", "[", 
         RowBox[{"a_", ",", "b_"}], "]"}], "]"}], ":=", 
       RowBox[{"\"\<complex\>\"", "<>", "lp", "<>", 
        RowBox[{"PythonForm", "[", "a", "]"}], "<>", "\"\<,\>\"", "<>", 
        RowBox[{"PythonForm", "[", "b", "]"}], "<>", "rp"}]}], ";", "\n", 
      RowBox[{
       RowBox[{"PythonForm", "[", 
        RowBox[{"Times", "[", 
         RowBox[{"a_", ",", "b_"}], "]"}], "]"}], ":=", 
       RowBox[{
        RowBox[{"PythonForm", "[", "a", "]"}], "<>", "\"\< * \>\"", "<>", 
        RowBox[{"PythonForm", "[", "b", "]"}]}]}], ";", "\n", 
      RowBox[{
       RowBox[{"PythonForm", "[", 
        RowBox[{"Plus", "[", 
         RowBox[{"a_", ",", "b_"}], "]"}], "]"}], ":=", 
       RowBox[{"lp", "<>", 
        RowBox[{"PythonForm", "[", "a", "]"}], "<>", "\"\< + \>\"", "<>", 
        RowBox[{"PythonForm", "[", "b", "]"}], "<>", "rp"}]}], ";", "\n", 
      RowBox[{
       RowBox[{"PythonForm", "[", 
        RowBox[{"h_", "[", "args__", "]"}], "]"}], ":=", 
       RowBox[{"numpypre", "<>", "sep", "<>", 
        RowBox[{"ToLowerCase", "[", 
         RowBox[{"PythonForm", "[", "h", "]"}], "]"}], "<>", "lp", "<>", 
        RowBox[{"PythonForm", "[", "args", "]"}], "<>", "rp"}]}], ";", "\n", 
      RowBox[{
       RowBox[{"PythonForm", "[", 
        RowBox[{"Power", "[", 
         RowBox[{"a_", ",", "b_"}], "]"}], "]"}], ":=", 
       RowBox[{"lp", "<>", 
        RowBox[{"PythonForm", "[", "a", "]"}], "<>", "rp", "<>", "\"\<**\>\"",
         "<>", "lp", "<>", 
        RowBox[{"PythonForm", "[", "b", "]"}], "<>", "rp"}]}], ";", "\n", 
      RowBox[{
       RowBox[{"PythonForm", "[", "a_ListQ", "]"}], ":=", 
       RowBox[{"numpypre", "<>", "sep", "<>", "\"\<array\>\"", "<>", 
        RowBox[{"StringReplace", "[", 
         RowBox[{
          RowBox[{"ToString", "[", "a", "]"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"\"\<{\>\"", "->", " ", "\"\<[\>\""}], ",", 
            RowBox[{"\"\<}\>\"", "->", " ", "\"\<]\>\""}]}], "}"}]}], 
         "]"}]}]}], ";", "\n", 
      RowBox[{
       RowBox[{"PythonForm", "[", "Arg", "]"}], "=", 
       RowBox[{"numpypre", "<>", "sep", "<>", "\"\<angle\>\""}]}], ";", "\n", 
      
      RowBox[{
       RowBox[{"PythonForm", "[", "Csc", "]"}], ":=", 
       RowBox[{
       "\"\<1/\>\"", "<>", "numpypre", "<>", "sep", "<>", "\"\<sin\>\""}]}], 
      ";", "\n", 
      RowBox[{
       RowBox[{"PythonForm", "[", "Sec", "]"}], ":=", 
       RowBox[{
       "\"\<1/\>\"", "<>", "numpypre", "<>", "sep", "<>", "\"\<cos\>\""}]}], 
      ";", "\n", 
      RowBox[{
       RowBox[{"PythonForm", "[", "Cot", "]"}], ":=", 
       RowBox[{
       "\"\<1/\>\"", "<>", "numpypre", "<>", "sep", "<>", "\"\<tan\>\""}]}], 
      ";", "\n", 
      RowBox[{
       RowBox[{"PythonForm", "[", "Csch", "]"}], ":=", 
       RowBox[{
       "\"\<1/\>\"", "<>", "numpypre", "<>", "sep", "<>", "\"\<sinh\>\""}]}], 
      ";", "\n", 
      RowBox[{
       RowBox[{"PythonForm", "[", "Sech", "]"}], ":=", 
       RowBox[{
       "\"\<1/\>\"", "<>", "numpypre", "<>", "sep", "<>", "\"\<cosh\>\""}]}], 
      ";", "\n", 
      RowBox[{
       RowBox[{"PythonForm", "[", "Coth", "]"}], ":=", 
       RowBox[{
       "\"\<1/\>\"", "<>", "numpypre", "<>", "sep", "<>", "\"\<tanh\>\""}]}], 
      ";", "\n", 
      RowBox[{
       RowBox[{"PythonForm", "[", 
        RowBox[{"List", "[", "args__", "]"}], "]"}], ":=", 
       RowBox[{
       "numpypre", "<>", "sep", "<>", "\"\<array\>\"", "<>", "lp", "<>", 
        "\"\<[\>\"", "<>", 
        RowBox[{"Table", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"PythonForm", "[", 
            RowBox[{
             RowBox[{"{", "args", "}"}], "[", 
             RowBox[{"[", "ii", "]"}], "]"}], "]"}], "<>", "\"\<,\>\""}], ",", 
          RowBox[{"{", 
           RowBox[{"ii", ",", "1", ",", 
            RowBox[{"Length", "@", 
             RowBox[{"{", "args", "}"}]}]}], "}"}]}], "]"}], "<>", 
        "\"\<]\>\"", "<>", "rp"}]}], ";", "\n", 
      RowBox[{
       RowBox[{"PythonForm", "[", "\[Pi]", "]"}], "=", 
       RowBox[{"numpypre", "<>", "sep", "<>", "\"\<pi\>\""}]}], ";", "\n", 
      RowBox[{
       RowBox[{"PythonForm", "[", "E", "]"}], "=", 
       RowBox[{"numpypre", "<>", "sep", "<>", "\"\<e\>\""}]}], ";", "\n", 
      RowBox[{
       RowBox[{"PythonForm", "[", "r_Real", "]"}], ":=", 
       RowBox[{"Block", "[", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"a", "=", 
           RowBox[{"MantissaExponent", "[", "r", "]"}]}], "}"}], ",", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"r", ">=", "0"}], ",", 
           RowBox[{
            RowBox[{"ToString", "[", 
             RowBox[{"N", "[", 
              RowBox[{
               RowBox[{"a", "[", 
                RowBox[{"[", "1", "]"}], "]"}], ",", "6"}], "]"}], "]"}], 
            "<>", "\"\<e\>\"", "<>", 
            RowBox[{"ToString", "[", 
             RowBox[{"a", "[", 
              RowBox[{"[", "2", "]"}], "]"}], "]"}]}], ",", 
           RowBox[{"\"\<(\>\"", "<>", 
            RowBox[{"ToString", "[", 
             RowBox[{"N", "[", 
              RowBox[{
               RowBox[{"a", "[", 
                RowBox[{"[", "1", "]"}], "]"}], ",", "6"}], "]"}], "]"}], 
            "<>", "\"\<e\>\"", "<>", 
            RowBox[{"ToString", "[", 
             RowBox[{"a", "[", 
              RowBox[{"[", "2", "]"}], "]"}], "]"}], "<>", "\"\<)\>\""}]}], 
          "]"}]}], "]"}]}], ";", "\n", 
      RowBox[{"greekrule", "=", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"\"\<\[Alpha]\>\"", "->", "\"\<alpha\>\""}], ",", 
         RowBox[{"\"\<\[Beta]\>\"", "->", "\"\<beta\>\""}], ",", 
         RowBox[{"\"\<\[Gamma]\>\"", "->", "\"\<gamma\>\""}], ",", 
         RowBox[{"\"\<\[Delta]\>\"", "->", "\"\<delta\>\""}], ",", 
         RowBox[{"\"\<\[CurlyEpsilon]\>\"", "->", "\"\<curlyepsilon\>\""}], 
         ",", 
         RowBox[{"\"\<\[Zeta]\>\"", "->", "\"\<zeta\>\""}], ",", 
         RowBox[{"\"\<\[Eta]\>\"", "->", "\"\<eta\>\""}], ",", 
         RowBox[{"\"\<\[Theta]\>\"", "->", "\"\<theta\>\""}], ",", 
         RowBox[{"\"\<\[Iota]\>\"", "->", "\"\<iota\>\""}], ",", 
         RowBox[{"\"\<\[Kappa]\>\"", "->", "\"\<kappa\>\""}], ",", 
         RowBox[{"\"\<\[Lambda]\>\"", "->", "\"\<lambda\>\""}], ",", 
         RowBox[{"\"\<\[Mu]\>\"", "->", "\"\<mu\>\""}], ",", 
         RowBox[{"\"\<\[Nu]\>\"", "->", "\"\<nu\>\""}], ",", 
         RowBox[{"\"\<\[Xi]\>\"", "->", "\"\<xi\>\""}], ",", 
         RowBox[{"\"\<\[Omicron]\>\"", "->", "\"\<omicron\>\""}], ",", 
         RowBox[{"\"\<\[Pi]\>\"", "->", "\"\<pi\>\""}], ",", 
         RowBox[{"\"\<\[Rho]\>\"", "->", "\"\<rho\>\""}], ",", 
         RowBox[{"\"\<\[FinalSigma]\>\"", "->", "\"\<finalsigma\>\""}], ",", 
         RowBox[{"\"\<\[Sigma]\>\"", "->", "\"\<sigma\>\""}], ",", 
         RowBox[{"\"\<\[Tau]\>\"", "->", "\"\<tau\>\""}], ",", 
         RowBox[{"\"\<\[Upsilon]\>\"", "->", "\"\<upsilon\>\""}], ",", 
         RowBox[{"\"\<\[CurlyPhi]\>\"", "->", "\"\<curlyphi\>\""}], ",", 
         RowBox[{"\"\<\[Chi]\>\"", "->", "\"\<chi\>\""}], ",", 
         RowBox[{"\"\<\[Psi]\>\"", "->", "\"\<psi\>\""}], ",", 
         RowBox[{"\"\<\[Omega]\>\"", "->", "\"\<omega\>\""}], ",", 
         RowBox[{"\"\<\[CapitalAlpha]\>\"", "->", "\"\<Alpha\>\""}], ",", 
         RowBox[{"\"\<\[CapitalBeta]\>\"", "->", "\"\<Beta\>\""}], ",", 
         RowBox[{"\"\<\[CapitalGamma]\>\"", "->", "\"\<Gamma\>\""}], ",", 
         RowBox[{"\"\<\[CapitalDelta]\>\"", "->", "\"\<Delta\>\""}], ",", 
         RowBox[{"\"\<\[CapitalEpsilon]\>\"", "->", "\"\<CurlyEpsilon\>\""}], 
         ",", 
         RowBox[{"\"\<\[CapitalZeta]\>\"", "->", "\"\<Zeta\>\""}], ",", 
         RowBox[{"\"\<\[CapitalEta]\>\"", "->", "\"\<Eta\>\""}], ",", 
         RowBox[{"\"\<\[CapitalTheta]\>\"", "->", "\"\<Theta\>\""}], ",", 
         RowBox[{"\"\<\[CapitalIota]\>\"", "->", "\"\<Iota\>\""}], ",", 
         RowBox[{"\"\<\[CapitalKappa]\>\"", "->", "\"\<Kappa\>\""}], ",", 
         RowBox[{"\"\<\[CapitalLambda]\>\"", "->", "\"\<Lambda\>\""}], ",", 
         RowBox[{"\"\<\[CapitalMu]\>\"", "->", "\"\<Mu\>\""}], ",", 
         RowBox[{"\"\<\[CapitalNu]\>\"", "->", "\"\<Nu\>\""}], ",", 
         RowBox[{"\"\<\[CapitalXi]\>\"", "->", "\"\<Xi\>\""}], ",", 
         RowBox[{"\"\<\[CapitalOmicron]\>\"", "->", "\"\<Omicron\>\""}], ",", 
         
         RowBox[{"\"\<\[CapitalPi]\>\"", "->", "\"\<Pi\>\""}], ",", 
         RowBox[{"\"\<\[CapitalRho]\>\"", "->", "\"\<Rho\>\""}], ",", 
         RowBox[{"\"\<\[CapitalSigma]\>\"", "->", "\"\<Sigma\>\""}], ",", 
         RowBox[{"\"\<\[CapitalTau]\>\"", "->", "\"\<Tau\>\""}], ",", 
         RowBox[{"\"\<\[CapitalUpsilon]\>\"", "->", "\"\<Upsilon\>\""}], ",", 
         
         RowBox[{"\"\<\[CapitalPhi]\>\"", "->", "\"\<CurlyPhi\>\""}], ",", 
         RowBox[{"\"\<\[CapitalChi]\>\"", "->", "\"\<Chi\>\""}], ",", 
         RowBox[{"\"\<\[CapitalPsi]\>\"", "->", "\"\<Psi\>\""}], ",", 
         RowBox[{"\"\<\[CapitalOmega]\>\"", "->", "\"\<Omega\>\""}]}], 
        "}"}]}], ";", "\n", 
      RowBox[{
       RowBox[{"PythonForm", "[", "allOther_", "]"}], ":=", 
       RowBox[{"StringReplace", "[", 
        RowBox[{
         RowBox[{"ToString", "[", 
          RowBox[{"allOther", ",", "FortranForm"}], "]"}], ",", "greekrule"}],
         "]"}]}], ";", "\n", 
      RowBox[{"PythonForm", "[", "expression", "]"}]}]}], "]"}]}], 
  ";"}]], "Input",
 CellLabel->"In[4]:=",ExpressionUUID->"aad61bc4-3eba-4de9-bb02-e95bc0048690"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"precessionform", "[", 
    RowBox[{"label_", ",", "expression_"}], "]"}], ":=", 
   RowBox[{"\"\<# Machine generated with \>\"", "<>", 
    RowBox[{"Last", "@", 
     RowBox[{"FileNameSplit", "@", 
      RowBox[{"NotebookFileName", "[", "]"}]}]}], "<>", "\"\<\\n\>\"", "<>", 
    " ", "label", "<>", "\"\< = \>\"", "<>", 
    RowBox[{"StringReplace", "[", 
     RowBox[{
      RowBox[{"ToPython", "[", "expression", "]"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"\"\<(q)\>\"", "\[Rule]", "\"\<q\>\""}], ",", 
        RowBox[{"\"\<(L)\>\"", "\[Rule]", "\"\<L\>\""}], ",", 
        RowBox[{"\"\<(S)\>\"", "\[Rule]", "\"\<S\>\""}], ",", 
        RowBox[{"\"\<(J)\>\"", "\[Rule]", "\"\<J\>\""}], ",", 
        RowBox[{"\"\<(chi1)\>\"", "\[Rule]", "\"\<chi1\>\""}], ",", 
        RowBox[{"\"\<(chi2)\>\"", "\[Rule]", "\"\<chi2\>\""}], ",", 
        RowBox[{"\"\<(u)\>\"", "->", "\"\<u\>\""}], ",", 
        RowBox[{"\"\<(kappa)\>\"", "->", "\"\<kappa\>\""}], ",", 
        RowBox[{"\"\<(chieff)\>\"", "->", "\"\<chieff\>\""}], ",", 
        RowBox[{"\"\<(2)\>\"", "->", "\"\<2\>\""}], ",", 
        RowBox[{"\"\<(3)\>\"", "->", "\"\<3\>\""}], ",", 
        RowBox[{"\"\<(4)\>\"", "->", "\"\<4\>\""}], ",", 
        RowBox[{"\"\<(5)\>\"", "->", "\"\<5\>\""}], ",", 
        RowBox[{"\"\<(6)\>\"", "->", "\"\<6\>\""}], ",", 
        RowBox[{"\"\<(7)\>\"", "->", "\"\<7\>\""}], ",", 
        RowBox[{"\"\<(8)\>\"", "->", "\"\<8\>\""}], ",", 
        RowBox[{"\"\<(9)\>\"", "->", "\"\<9\>\""}], ",", 
        RowBox[{"\"\<(10)\>\"", "->", "\"\<10\>\""}]}], "}"}]}], "]"}]}]}], 
  ";"}]], "Input",
 CellLabel->"In[5]:=",ExpressionUUID->"a14bde24-f4dc-47e8-94cb-0f86ba28c24d"],

Cell[CellGroupData[{

Cell["(i) S as a function of \[Delta]\[Chi]", \
"Section",ExpressionUUID->"f7fa412e-204d-42ec-a430-d4433e6126c8"],

Cell[BoxData[
 RowBox[{"Clear", "@@", 
  RowBox[{"DeleteCases", "[", 
   RowBox[{
    RowBox[{"Names", "@", "\"\<`*\>\""}], ",", 
    RowBox[{"\"\<ToPython\>\"", "|", "\"\<precessionform\>\""}]}], 
   "]"}]}]], "Input",
 CellLabel->"In[6]:=",ExpressionUUID->"0a061175-54a9-4ceb-9f4b-4f0312514933"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"S1", " ", "=", " ", 
  RowBox[{"\[Chi]1", " ", "/", " ", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}], "\[IndentingNewLine]", 
 RowBox[{"S2", " ", "=", " ", 
  RowBox[{"\[Chi]2", " ", 
   RowBox[{
    RowBox[{"q", "^", "2"}], " ", "/", " ", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}]}], "Input",
 CellLabel->"In[7]:=",ExpressionUUID->"578bae74-440a-4797-bd31-f85e68a407e1"],

Cell[BoxData[
 FractionBox["\[Chi]1", 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{"1", "+", "q"}], ")"}], "2"]]], "Output",
 CellLabel->"Out[7]=",ExpressionUUID->"3721ee39-c0f4-42e1-92b6-9c46c634cff4"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   SuperscriptBox["q", "2"], " ", "\[Chi]2"}], 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{"1", "+", "q"}], ")"}], "2"]]], "Output",
 CellLabel->"Out[8]=",ExpressionUUID->"64f6cc7d-cebd-4850-8b53-f168c5c4f848"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"cos\[Theta]1S", " ", "=", " ", 
  RowBox[{
   RowBox[{"1", "/", 
    RowBox[{"(", 
     RowBox[{"2", 
      RowBox[{"(", 
       RowBox[{"1", "-", "q"}], ")"}], " ", "S1"}], ")"}]}], 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"J", "^", "2"}], "-", 
        RowBox[{"L", "^", "2"}], "-", "Ss"}], ")"}], "/", "L"}], " ", "-", 
     " ", 
     RowBox[{"2", "q", " ", 
      RowBox[{"\[Chi]eff", "/", 
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}]}]}]}], 
    ")"}]}]}], "\[IndentingNewLine]", 
 RowBox[{"cos\[Theta]2S", " ", "=", " ", 
  RowBox[{
   RowBox[{"q", "/", 
    RowBox[{"(", 
     RowBox[{"2", 
      RowBox[{"(", 
       RowBox[{"1", "-", "q"}], ")"}], " ", "S2"}], ")"}]}], 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"J", "^", "2"}], "-", 
         RowBox[{"L", "^", "2"}], "-", "Ss"}], ")"}]}], "/", "L"}], " ", "+", 
     " ", 
     RowBox[{"2", " ", 
      RowBox[{"\[Chi]eff", "/", 
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}]}]}]}], ")"}]}]}]}], "Input",
 CellLabel->"In[9]:=",ExpressionUUID->"1bc2d91e-dd12-49aa-a2cf-c9c6a7c058e2"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["J", "2"], "-", 
       SuperscriptBox["L", "2"], "-", "Ss"}], "L"], "-", 
     FractionBox[
      RowBox[{"2", " ", "q", " ", "\[Chi]eff"}], 
      RowBox[{"1", "+", "q"}]]}], ")"}]}], 
  RowBox[{"2", " ", 
   RowBox[{"(", 
    RowBox[{"1", "-", "q"}], ")"}], " ", "\[Chi]1"}]]], "Output",
 CellLabel->"Out[9]=",ExpressionUUID->"3dbff839-d7da-4e63-97aa-cd7f047b7bbf"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     FractionBox[
      RowBox[{
       RowBox[{"-", 
        SuperscriptBox["J", "2"]}], "+", 
       SuperscriptBox["L", "2"], "+", "Ss"}], "L"], "+", 
     FractionBox[
      RowBox[{"2", " ", "\[Chi]eff"}], 
      RowBox[{"1", "+", "q"}]]}], ")"}]}], 
  RowBox[{"2", " ", 
   RowBox[{"(", 
    RowBox[{"1", "-", "q"}], ")"}], " ", "q", " ", "\[Chi]2"}]]], "Output",
 CellLabel->
  "Out[10]=",ExpressionUUID->"af366576-3fbe-4f63-b590-ddb6792f2ed3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"deltachi", " ", "=", " ", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"\[Chi]1", " ", "cos\[Theta]1x"}], " ", "-", " ", 
      RowBox[{"q", " ", "\[Chi]2", " ", "cos\[Theta]2x"}]}], ")"}], "/", 
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}]}], "//", "Simplify"}]}]], "Input",
 CellLabel->"In[12]:=",ExpressionUUID->"2073ac7a-0670-4436-93bd-7e8c23955121"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{"cos\[Theta]1x", " ", "\[Chi]1"}], "-", 
   RowBox[{"cos\[Theta]2x", " ", "q", " ", "\[Chi]2"}]}], 
  RowBox[{"1", "+", "q"}]]], "Output",
 CellLabel->
  "Out[12]=",ExpressionUUID->"b07189fb-be49-47b6-b4b4-39aed27b1075"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Sssol", " ", "=", 
  RowBox[{
   RowBox[{"Ss", "/.", " ", 
    RowBox[{
     RowBox[{"Solve", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"deltachi", "/.", 
           RowBox[{"cos\[Theta]1x", "->", "cos\[Theta]1S"}]}], "/.", 
          RowBox[{"cos\[Theta]2x", "->", "cos\[Theta]2S"}]}], ")"}], "==", 
        "\[Delta]\[Chi]"}], ",", " ", "Ss"}], "]"}], "[", 
     RowBox[{"[", "1", "]"}], "]"}]}], "//", "Simplify"}]}]], "Input",
 CellLabel->"In[13]:=",ExpressionUUID->"a3bb4955-ce6f-47e5-a16d-0441931a5d5c"],

Cell[BoxData[
 RowBox[{
  SuperscriptBox["J", "2"], "-", 
  FractionBox[
   RowBox[{"L", " ", 
    RowBox[{"(", 
     RowBox[{"L", "+", 
      RowBox[{"L", " ", "q"}], "+", "\[Delta]\[Chi]", "-", 
      RowBox[{"q", " ", "\[Delta]\[Chi]"}], "+", "\[Chi]eff", "+", 
      RowBox[{"q", " ", "\[Chi]eff"}]}], ")"}]}], 
   RowBox[{"1", "+", "q"}]]}]], "Output",
 CellLabel->
  "Out[13]=",ExpressionUUID->"27233892-0973-4b81-88b4-45e1f131942b"]
}, Open  ]],

Cell["Rewrite in a nicer form.", \
"Text",ExpressionUUID->"98a66384-a030-485b-97a2-d966ca0178e9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Ss", "=", " ", 
  RowBox[{
   RowBox[{"q", " ", "/", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}], " ", 
   RowBox[{"Sqrt", "[", "r", "]"}], 
   RowBox[{"(", " ", 
    RowBox[{
     RowBox[{"2", " ", "\[Kappa]"}], " ", "-", "\[Chi]eff", " ", "-", 
     RowBox[{"\[Delta]\[Chi]", " ", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], "/", 
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}]}]}]}], ")"}]}]}]], "Input",
 CellLabel->"In[14]:=",ExpressionUUID->"7a08650f-d358-48a9-859a-1409cd9ae6b0"],

Cell[BoxData[
 FractionBox[
  RowBox[{"q", " ", 
   SqrtBox["r"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], " ", "\[Delta]\[Chi]"}], 
       RowBox[{"1", "+", "q"}]]}], "+", 
     RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}], 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{"1", "+", "q"}], ")"}], "2"]]], "Output",
 CellLabel->
  "Out[14]=",ExpressionUUID->"7a180f7e-d99a-4be4-9aa0-a04bd7d9d1ed"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"L", "=", " ", 
  RowBox[{
   RowBox[{"q", " ", "/", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}], " ", 
   RowBox[{"Sqrt", "[", "r", "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{"\[Kappa]", " ", "=", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"J", "^", "2"}], " ", "-", " ", 
     RowBox[{"L", "^", "2"}]}], ")"}], " ", "/", " ", 
   RowBox[{"(", 
    RowBox[{"2", "L"}], ")"}]}]}]}], "Input",
 CellLabel->"In[15]:=",ExpressionUUID->"7e5b6069-3b75-4fb5-a71e-505c4ba5f169"],

Cell[BoxData[
 FractionBox[
  RowBox[{"q", " ", 
   SqrtBox["r"]}], 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{"1", "+", "q"}], ")"}], "2"]]], "Output",
 CellLabel->"Out[15]=",ExpressionUUID->"6595c28a-073b-462f-b98d-5c4c75bcb156"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     SuperscriptBox["J", "2"], "-", 
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "2"], " ", "r"}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "4"]]}], ")"}]}], 
  RowBox[{"2", " ", "q", " ", 
   SqrtBox["r"]}]]], "Output",
 CellLabel->
  "Out[16]=",ExpressionUUID->"661ec5ec-4159-47e8-b506-87b5ede592c1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Sssol", "-", " ", "Ss"}], "//", "Simplify"}]], "Input",
 CellLabel->"In[17]:=",ExpressionUUID->"da680eec-38b0-4074-a995-3a0f91a29785"],

Cell[BoxData["0"], "Output",
 CellLabel->
  "Out[17]=",ExpressionUUID->"1f73929a-8d9d-4ab5-a3fc-cc0b3702ca70"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["(ii) Spin angles", \
"Section",ExpressionUUID->"da95115e-8cb2-44b8-9311-fea72ad6140f"],

Cell[BoxData[
 RowBox[{"Clear", "@@", 
  RowBox[{"DeleteCases", "[", 
   RowBox[{
    RowBox[{"Names", "@", "\"\<`*\>\""}], ",", 
    RowBox[{"\"\<ToPython\>\"", "|", "\"\<precessionform\>\""}]}], 
   "]"}]}]], "Input",
 CellLabel->"In[18]:=",ExpressionUUID->"c12910bf-2817-488c-9e3d-4466c2ce86de"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"chieff", " ", "=", " ", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"\[Chi]1", " ", "cos\[Theta]1x"}], " ", "+", " ", 
      RowBox[{"q", " ", "\[Chi]2", " ", "cos\[Theta]2x"}]}], ")"}], "/", 
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}]}], "//", 
   "Simplify"}]}], "\[IndentingNewLine]", 
 RowBox[{"deltachi", " ", "=", " ", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"\[Chi]1", " ", "cos\[Theta]1x"}], " ", "-", " ", 
      RowBox[{"q", " ", "\[Chi]2", " ", "cos\[Theta]2x"}]}], ")"}], "/", 
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}]}], "//", "Simplify"}]}]}], "Input",
 CellLabel->"In[19]:=",ExpressionUUID->"c3d7346c-4b36-4054-84ae-91c967c15855"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{"cos\[Theta]1x", " ", "\[Chi]1"}], "+", 
   RowBox[{"cos\[Theta]2x", " ", "q", " ", "\[Chi]2"}]}], 
  RowBox[{"1", "+", "q"}]]], "Output",
 CellLabel->"Out[19]=",ExpressionUUID->"90626a9f-9741-45d9-81d8-a489dbbb5285"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{"cos\[Theta]1x", " ", "\[Chi]1"}], "-", 
   RowBox[{"cos\[Theta]2x", " ", "q", " ", "\[Chi]2"}]}], 
  RowBox[{"1", "+", "q"}]]], "Output",
 CellLabel->
  "Out[20]=",ExpressionUUID->"9c23a67e-8e60-494c-9186-7e9c71149b9b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"Solve", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"chieff", "==", "\[Chi]eff"}], ",", 
       RowBox[{"deltachi", "==", "\[Delta]\[Chi]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"cos\[Theta]1x", ",", "cos\[Theta]2x"}], "}"}]}], "]"}], "[", 
   RowBox[{"[", "1", "]"}], "]"}], "//", 
  "Simplify"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"cos\[Theta]1sol", "=", " ", 
   RowBox[{"cos\[Theta]1x", "/.", 
    RowBox[{"Out", "[", 
     RowBox[{"-", "1"}], "]"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"cos\[Theta]2sol", " ", "=", 
   RowBox[{"cos\[Theta]2x", "/.", 
    RowBox[{"Out", "[", 
     RowBox[{"-", "2"}], "]"}]}]}], ";"}]}], "Input",
 CellLabel->"In[21]:=",ExpressionUUID->"8bfe1d55-5f3e-4da6-a706-697d8a532300"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"cos\[Theta]1x", "\[Rule]", 
    FractionBox[
     RowBox[{
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"\[Delta]\[Chi]", "+", "\[Chi]eff"}], ")"}]}], 
     RowBox[{"2", " ", "\[Chi]1"}]]}], ",", 
   RowBox[{"cos\[Theta]2x", "\[Rule]", 
    FractionBox[
     RowBox[{
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "\[Delta]\[Chi]"}], "+", "\[Chi]eff"}], ")"}]}], 
     RowBox[{"2", " ", "q", " ", "\[Chi]2"}]]}]}], "}"}]], "Output",
 CellLabel->
  "Out[21]=",ExpressionUUID->"c97b4da2-77bd-4164-ac7b-93dfeca8d568"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"S1", " ", "=", " ", 
  RowBox[{"\[Chi]1", " ", "/", " ", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}], "\[IndentingNewLine]", 
 RowBox[{"S2", " ", "=", " ", 
  RowBox[{"\[Chi]2", " ", 
   RowBox[{
    RowBox[{"q", "^", "2"}], " ", "/", " ", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], "\[IndentingNewLine]", 
 RowBox[{"L", "=", 
  RowBox[{
   RowBox[{"q", " ", "/", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}], " ", 
   RowBox[{"Sqrt", "[", "r", "]"}]}]}]}], "Input",
 CellLabel->"In[24]:=",ExpressionUUID->"30f84188-f24d-4915-b3b6-81f7748cd188"],

Cell[BoxData[
 FractionBox["\[Chi]1", 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{"1", "+", "q"}], ")"}], "2"]]], "Output",
 CellLabel->"Out[24]=",ExpressionUUID->"ff779d26-4940-4c1c-8511-02c458340883"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   SuperscriptBox["q", "2"], " ", "\[Chi]2"}], 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{"1", "+", "q"}], ")"}], "2"]]], "Output",
 CellLabel->"Out[25]=",ExpressionUUID->"6b6a34bc-6fd2-43ee-87c4-fe961a4ce197"],

Cell[BoxData[
 FractionBox[
  RowBox[{"q", " ", 
   SqrtBox["r"]}], 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{"1", "+", "q"}], ")"}], "2"]]], "Output",
 CellLabel->
  "Out[26]=",ExpressionUUID->"adac81a7-0582-4018-8513-0cef9f2f9621"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Ss", "=", " ", 
  RowBox[{
   RowBox[{"q", " ", "/", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}], " ", 
   RowBox[{"Sqrt", "[", "r", "]"}], 
   RowBox[{"(", " ", 
    RowBox[{
     RowBox[{"2", " ", "\[Kappa]"}], " ", "-", "\[Chi]eff", " ", "-", 
     RowBox[{"\[Delta]\[Chi]", " ", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], "/", 
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}]}]}]}], ")"}]}]}]], "Input",
 CellLabel->"In[27]:=",ExpressionUUID->"d77f6c6d-0abf-466b-938f-bf0ed8eb6f6d"],

Cell[BoxData[
 FractionBox[
  RowBox[{"q", " ", 
   SqrtBox["r"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], " ", "\[Delta]\[Chi]"}], 
       RowBox[{"1", "+", "q"}]]}], "+", 
     RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}], 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{"1", "+", "q"}], ")"}], "2"]]], "Output",
 CellLabel->
  "Out[27]=",ExpressionUUID->"ad140fd9-0a98-4374-af1f-89dda2294d80"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cos\[Theta]12sol", "=", " ", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{"Ss", " ", "-", 
      RowBox[{"S1", "^", "2"}], "-", 
      RowBox[{"S2", "^", "2"}]}], ")"}], " ", "/", " ", 
    RowBox[{"(", 
     RowBox[{"2", "S1", " ", "S2"}], ")"}]}], "//", "Simplify"}]}]], "Input",
 CellLabel->"In[28]:=",ExpressionUUID->"48f7297f-0bfb-4999-9a22-c000c923c62c"],

Cell[BoxData[
 RowBox[{"-", 
  RowBox[{
   FractionBox["1", 
    RowBox[{"2", " ", 
     SuperscriptBox["q", "2"], " ", "\[Chi]1", " ", "\[Chi]2"}]], 
   RowBox[{"(", 
    RowBox[{
     SuperscriptBox["\[Chi]1", "2"], "+", 
     RowBox[{
      SuperscriptBox["q", "4"], " ", 
      SuperscriptBox["\[Chi]2", "2"]}], "+", 
     RowBox[{"2", " ", 
      SuperscriptBox["q", "2"], " ", 
      SqrtBox["r"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", "\[Kappa]"}], "+", "\[Chi]eff"}], ")"}]}], 
     "+", 
     RowBox[{
      SuperscriptBox["q", "3"], " ", 
      SqrtBox["r"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "\[Delta]\[Chi]"}], "-", 
        RowBox[{"2", " ", "\[Kappa]"}], "+", "\[Chi]eff"}], ")"}]}], "+", 
     RowBox[{"q", " ", 
      SqrtBox["r"], " ", 
      RowBox[{"(", 
       RowBox[{"\[Delta]\[Chi]", "-", 
        RowBox[{"2", " ", "\[Kappa]"}], "+", "\[Chi]eff"}], ")"}]}]}], 
    ")"}]}]}]], "Output",
 CellLabel->
  "Out[28]=",ExpressionUUID->"e94e9d9d-a3d9-427f-9560-bc49ed6b4db5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cos\[CapitalDelta]\[CapitalPhi]sol", " ", "=", " ", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{"cos\[Theta]12sol", " ", "-", " ", 
     RowBox[{"cos\[Theta]1sol", " ", "cos\[Theta]2sol"}]}], ")"}], " ", "/", 
   " ", 
   RowBox[{"Sqrt", "[", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "-", 
       RowBox[{"cos\[Theta]1sol", "^", "2"}]}], ")"}], 
     RowBox[{"(", 
      RowBox[{"1", "-", 
       RowBox[{"cos\[Theta]2sol", "^", "2"}]}], ")"}]}], "]"}]}]}]], "Input",
 CellLabel->"In[29]:=",ExpressionUUID->"d0f899e6-0675-4ec9-ad21-e0f6bbff98ca"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "\[Delta]\[Chi]"}], "+", "\[Chi]eff"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"\[Delta]\[Chi]", "+", "\[Chi]eff"}], ")"}]}], 
      RowBox[{"4", " ", "q", " ", "\[Chi]1", " ", "\[Chi]2"}]]}], "-", 
    RowBox[{
     FractionBox["1", 
      RowBox[{"2", " ", 
       SuperscriptBox["q", "2"], " ", "\[Chi]1", " ", "\[Chi]2"}]], 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox["\[Chi]1", "2"], "+", 
       RowBox[{
        SuperscriptBox["q", "4"], " ", 
        SuperscriptBox["\[Chi]2", "2"]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox["q", "2"], " ", 
        SqrtBox["r"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "2"}], " ", "\[Kappa]"}], "+", "\[Chi]eff"}], ")"}]}],
        "+", 
       RowBox[{
        SuperscriptBox["q", "3"], " ", 
        SqrtBox["r"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "\[Delta]\[Chi]"}], "-", 
          RowBox[{"2", " ", "\[Kappa]"}], "+", "\[Chi]eff"}], ")"}]}], "+", 
       RowBox[{"q", " ", 
        SqrtBox["r"], " ", 
        RowBox[{"(", 
         RowBox[{"\[Delta]\[Chi]", "-", 
          RowBox[{"2", " ", "\[Kappa]"}], "+", "\[Chi]eff"}], ")"}]}]}], 
      ")"}]}]}], ")"}], "/", 
  RowBox[{"(", 
   SqrtBox[
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "\[Delta]\[Chi]"}], "+", "\[Chi]eff"}], ")"}], 
          "2"]}], 
        RowBox[{"4", " ", 
         SuperscriptBox["q", "2"], " ", 
         SuperscriptBox["\[Chi]2", "2"]}]]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"\[Delta]\[Chi]", "+", "\[Chi]eff"}], ")"}], "2"]}], 
        RowBox[{"4", " ", 
         SuperscriptBox["\[Chi]1", "2"]}]]}], ")"}]}]], ")"}]}]], "Output",
 CellLabel->
  "Out[29]=",ExpressionUUID->"15226ef6-2241-48f9-8d75-eb61566d6771"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cos\[Theta]Lsol", " ", "=", " ", 
  RowBox[{
   RowBox[{
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["J", "2"], "+", " ", 
       SuperscriptBox["L", "2"], "-", "Ss"}], 
      RowBox[{"2", "J", " ", "L"}]], "/.", 
     RowBox[{"J", "->", 
      RowBox[{"L", " ", 
       RowBox[{"Sqrt", "[", 
        RowBox[{"1", "+", " ", 
         RowBox[{"2", 
          RowBox[{"\[Kappa]", "/", "L"}]}]}], "]"}]}]}]}], "//", "Simplify"}],
    "//", "Simplify"}]}]], "Input",
 CellLabel->"In[30]:=",ExpressionUUID->"1493a5af-48ab-4bf8-89c6-613fffe45f33"],

Cell[BoxData[
 FractionBox[
  RowBox[{"\[Delta]\[Chi]", "+", "\[Chi]eff", "+", 
   RowBox[{"2", " ", "q", " ", 
    RowBox[{"(", 
     RowBox[{
      SqrtBox["r"], "+", "\[Chi]eff"}], ")"}]}], "+", 
   RowBox[{
    SuperscriptBox["q", "2"], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "\[Delta]\[Chi]"}], "+", "\[Chi]eff"}], ")"}]}]}], 
  RowBox[{"2", " ", "q", " ", 
   SqrtBox["r"], " ", 
   SqrtBox[
    RowBox[{"1", "+", 
     FractionBox[
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}], 
      RowBox[{"q", " ", 
       SqrtBox["r"]}]]}]]}]]], "Output",
 CellLabel->
  "Out[30]=",ExpressionUUID->"cf815eb3-969e-4731-8c03-f3ee420d6220"]
}, Open  ]],

Cell["Rewrite in a nicer form", \
"Text",ExpressionUUID->"418d249e-c71a-489e-a2b8-267300f96c70"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cos\[Theta]1", "=", 
  FractionBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"\[Chi]eff", "+", "\[Delta]\[Chi]"}], ")"}]}], 
   RowBox[{"2", " ", "\[Chi]1"}]]}]], "Input",
 CellLabel->"In[31]:=",ExpressionUUID->"9c14c815-266e-4ad7-b437-faeceb00e4f1"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{"(", 
    RowBox[{"1", "+", "q"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{"\[Delta]\[Chi]", "+", "\[Chi]eff"}], ")"}]}], 
  RowBox[{"2", " ", "\[Chi]1"}]]], "Output",
 CellLabel->
  "Out[31]=",ExpressionUUID->"aca2e8e2-86df-443f-8ef4-d90b5b09f7ba"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cos\[Theta]2", "=", 
  FractionBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"\[Chi]eff", "-", "\[Delta]\[Chi]"}], ")"}]}], 
   RowBox[{"2", " ", "q", " ", "\[Chi]2"}]]}]], "Input",
 CellLabel->"In[32]:=",ExpressionUUID->"6ebd315e-6f49-4bad-b31c-86aa7f2491e6"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{"(", 
    RowBox[{"1", "+", "q"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "\[Delta]\[Chi]"}], "+", "\[Chi]eff"}], ")"}]}], 
  RowBox[{"2", " ", "q", " ", "\[Chi]2"}]]], "Output",
 CellLabel->
  "Out[32]=",ExpressionUUID->"36ef66a1-740a-444d-8398-5280f6727561"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cos\[Theta]12", " ", "=", 
   RowBox[{
    FractionBox["1", 
     RowBox[{"2", " ", "\[Chi]1", " ", "\[Chi]2", " ", 
      SuperscriptBox["q", "2"]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"q", " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", "\[Kappa]", " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}]}], "-", " ", 
         RowBox[{"\[Chi]eff", 
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}]}], "-", 
         RowBox[{"\[Delta]\[Chi]", 
          RowBox[{"(", 
           RowBox[{"1", "-", "q"}], ")"}]}]}], ")"}], " ", 
       SqrtBox["r"]}], " ", "-", " ", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["\[Chi]1", "2"], "+", 
        RowBox[{
         SuperscriptBox["\[Chi]2", "2"], "  ", 
         SuperscriptBox["q", "4"]}]}], ")"}]}], ")"}]}]}], " ", 
  "\[IndentingNewLine]"}]], "Input",
 CellLabel->"In[33]:=",ExpressionUUID->"fc978199-cdb3-4fd9-9185-a8c85103e457"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{"-", 
    SuperscriptBox["\[Chi]1", "2"]}], "-", 
   RowBox[{
    SuperscriptBox["q", "4"], " ", 
    SuperscriptBox["\[Chi]2", "2"]}], "+", 
   RowBox[{"q", " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], " ", 
    SqrtBox["r"], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"1", "-", "q"}], ")"}], " ", "\[Delta]\[Chi]"}], ")"}]}], 
      "+", 
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], " ", "\[Kappa]"}], "-", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], " ", "\[Chi]eff"}]}], ")"}]}]}], 
  RowBox[{"2", " ", 
   SuperscriptBox["q", "2"], " ", "\[Chi]1", " ", "\[Chi]2"}]]], "Output",
 CellLabel->
  "Out[33]=",ExpressionUUID->"bd64a0cf-1b77-44b0-aaf1-26a9322f2188"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cos\[CapitalDelta]\[CapitalPhi]", "=", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"2", 
      RowBox[{"(", 
       RowBox[{"q", " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", "\[Kappa]", " ", 
           RowBox[{"(", 
            RowBox[{"1", "+", "q"}], ")"}]}], "-", 
          RowBox[{"\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{"1", "+", "q"}], ")"}]}], "-", 
          RowBox[{"\[Delta]\[Chi]", " ", 
           RowBox[{"(", 
            RowBox[{"1", "-", "q"}], ")"}]}]}], ")"}], " ", 
        SqrtBox["r"]}], ")"}]}], " ", "-", " ", 
     RowBox[{"2", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["\[Chi]1", "2"], "+", 
        RowBox[{
         SuperscriptBox["\[Chi]2", "2"], " ", 
         SuperscriptBox["q", "4"]}]}], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"\[Chi]eff", "^", "2"}], "-", 
        RowBox[{"\[Delta]\[Chi]", "^", "2"}]}], ")"}], "q", " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"]}]}], ")"}], "/", 
   RowBox[{"(", 
    RowBox[{"q", 
     RowBox[{"\[Sqrt]", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"4", 
           SuperscriptBox["\[Chi]1", "2"]}], "-", 
          RowBox[{
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"\[Chi]eff", "+", "\[Delta]\[Chi]"}], ")"}], "2"], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "2"]}]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"4", " ", 
           SuperscriptBox["\[Chi]2", "2"], " ", 
           SuperscriptBox["q", "2"]}], "-", 
          RowBox[{
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"\[Chi]eff", "-", "\[Delta]\[Chi]"}], ")"}], "2"], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "2"]}]}], ")"}]}], ")"}]}]}], 
    ")"}]}]}]], "Input",
 CellLabel->"In[34]:=",ExpressionUUID->"2657781d-c118-4fe4-8a76-33a8d39a8460"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox["\[Chi]1", "2"], "+", 
       RowBox[{
        SuperscriptBox["q", "4"], " ", 
        SuperscriptBox["\[Chi]2", "2"]}]}], ")"}]}], "+", 
    RowBox[{"2", " ", "q", " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], " ", 
     SqrtBox["r"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{"1", "-", "q"}], ")"}], " ", "\[Delta]\[Chi]"}], ")"}]}], 
       "+", 
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], " ", "\[Kappa]"}], "-", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], " ", "\[Chi]eff"}]}], ")"}]}], "-", 
    RowBox[{"q", " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", 
        SuperscriptBox["\[Delta]\[Chi]", "2"]}], "+", 
       SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}]}], ")"}], "/", 
  RowBox[{"(", 
   RowBox[{"q", " ", 
    SqrtBox[
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"4", " ", 
         SuperscriptBox["q", "2"], " ", 
         SuperscriptBox["\[Chi]2", "2"]}], "-", 
        RowBox[{
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "\[Delta]\[Chi]"}], "+", "\[Chi]eff"}], ")"}], 
          "2"]}]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"4", " ", 
         SuperscriptBox["\[Chi]1", "2"]}], "-", 
        RowBox[{
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"\[Delta]\[Chi]", "+", "\[Chi]eff"}], ")"}], "2"]}]}], 
       ")"}]}]]}], ")"}]}]], "Output",
 CellLabel->
  "Out[34]=",ExpressionUUID->"9adb3f61-f613-4760-a2f4-145b9499337a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cos\[Theta]L", "=", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{"1", "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], " ", 
      FractionBox[
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{"1", "-", "q"}], ")"}], "\[Delta]\[Chi]"}], "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "\[Chi]eff"}]}], 
       RowBox[{"2", " ", "q", " ", 
        SqrtBox["r"]}]]}]}], ")"}], 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "+", 
      FractionBox[
       RowBox[{"2", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}], 
       RowBox[{"q", " ", 
        SqrtBox["r"]}]]}], ")"}], 
    RowBox[{
     RowBox[{"-", "1"}], "/", "2"}]]}]}]], "Input",
 CellLabel->"In[35]:=",ExpressionUUID->"3486c149-7339-4a39-9fa0-a37e1787432c"],

Cell[BoxData[
 FractionBox[
  RowBox[{"1", "+", 
   FractionBox[
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], " ", "\[Delta]\[Chi]"}], "+", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], " ", "\[Chi]eff"}]}], ")"}]}], 
    RowBox[{"2", " ", "q", " ", 
     SqrtBox["r"]}]]}], 
  SqrtBox[
   RowBox[{"1", "+", 
    FractionBox[
     RowBox[{"2", " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}], 
     RowBox[{"q", " ", 
      SqrtBox["r"]}]]}]]]], "Output",
 CellLabel->
  "Out[35]=",ExpressionUUID->"1d24f6e9-5a44-4403-aaa0-2fcf489c104c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"cos\[Theta]1", "-", "cos\[Theta]1sol"}], "\[IndentingNewLine]", 
 RowBox[{"cos\[Theta]2", "-", "cos\[Theta]2sol"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"cos\[Theta]12", "-", "cos\[Theta]12sol"}], "//", 
  "Simplify"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"cos\[CapitalDelta]\[CapitalPhi]sol", "^", "2"}], "-", 
   RowBox[{"cos\[CapitalDelta]\[CapitalPhi]", "^", "2"}]}], "//", 
  "Simplify"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"cos\[Theta]Lsol", "-", "cos\[Theta]L"}], "//", 
  "Simplify"}]}], "Input",
 CellLabel->"In[36]:=",ExpressionUUID->"087ccd1b-3164-4ff3-93cb-19d74d98ebb3"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[36]=",ExpressionUUID->"26dfae2f-3cc4-457b-bc30-31cda1775bdd"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[37]=",ExpressionUUID->"27e2bd51-3a4b-4fce-b45d-bd1fc10dcac6"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[38]=",ExpressionUUID->"f4c50246-e4e4-442b-8128-ba127a59a166"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[39]=",ExpressionUUID->"e6488b1f-3fb3-4c3e-ae9c-f036a8424bf9"],

Cell[BoxData["0"], "Output",
 CellLabel->
  "Out[40]=",ExpressionUUID->"73a61e3f-45bb-49d9-8829-df261ed0babd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"precessionform", "[", 
  RowBox[{"\"\<costheta12\>\"", ",", "cos\[Theta]12"}], "]"}]], "Input",
 CellLabel->"In[41]:=",ExpressionUUID->"5d36a83b-6e09-42c2-9364-024a39493f04"],

Cell[BoxData["\<\"# Machine generated with eq_generator.nb\\ncostheta12 = 1/2 \
* q**(-2) * chi1**(-1) * chi2**(-1) * (-1 * chi1**2 + (-1 * q**4 * chi2**2 + \
q * (1 + q) * (r)**(1/2) * (-1 * (1 + -1 * q) * deltachi + (2 * (1 + q) * \
kappa + -1 * (1 + q) * chieff))))\"\>"], "Output",
 CellLabel->
  "Out[41]=",ExpressionUUID->"4ab0262d-ee39-40b5-aae4-fb31f58e5703"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"precessionform", "[", 
  RowBox[{"\"\<cosdeltaphi\>\"", ",", "cos\[CapitalDelta]\[CapitalPhi]"}], 
  "]"}]], "Input",
 CellLabel->"In[42]:=",ExpressionUUID->"728c47a3-d2aa-48c8-88bb-648d678245c9"],

Cell[BoxData["\<\"# Machine generated with eq_generator.nb\\ncosdeltaphi = \
q**(-1) * ((4 * q**2 * chi2**2 + -1 * ((1 + q))**2 * ((-1 * deltachi + \
chieff))**2) * (4 * chi1**2 + -1 * ((1 + q))**2 * ((deltachi + \
chieff))**2))**(-1/2) * (-2 * (chi1**2 + q**4 * chi2**2) + (2 * q * (1 + q) * \
(r)**(1/2) * (-1 * (1 + -1 * q) * deltachi + (2 * (1 + q) * kappa + -1 * (1 + \
q) * chieff)) + -1 * q * ((1 + q))**2 * (-1 * (deltachi)**2 + \
chieff**2)))\"\>"], "Output",
 CellLabel->
  "Out[42]=",ExpressionUUID->"d5bb0b8f-36b5-47ec-9403-50d45c70634e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"precessionform", "[", 
  RowBox[{"\"\<costhetaL\>\"", ",", "cos\[Theta]L"}], "]"}]], "Input",
 CellLabel->"In[43]:=",ExpressionUUID->"77bc07be-a1f0-4802-b69b-3414aca160e1"],

Cell[BoxData["\<\"# Machine generated with eq_generator.nb\\ncosthetaL = ((1 \
+ 2 * q**(-1) * ((1 + q))**2 * (r)**(-1/2) * kappa))**(-1/2) * (1 + 1/2 * \
q**(-1) * (1 + q) * (r)**(-1/2) * ((1 + -1 * q) * deltachi + (1 + q) * \
chieff))\"\>"], "Output",
 CellLabel->
  "Out[43]=",ExpressionUUID->"a6b07797-181e-48d5-a333-f13fcc1abfb5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["(iii) Cubic polynomial (d\[Delta]\[Chi]/dt)^2", \
"Section",ExpressionUUID->"688a53e1-82f3-480d-9d7c-75188acb3d1d"],

Cell[BoxData[
 RowBox[{"Clear", "@@", 
  RowBox[{"DeleteCases", "[", 
   RowBox[{
    RowBox[{"Names", "@", "\"\<`*\>\""}], ",", 
    RowBox[{"\"\<ToPython\>\"", "|", "\"\<precessionform\>\""}]}], 
   "]"}]}]], "Input",
 CellLabel->"In[44]:=",ExpressionUUID->"3cd49c9d-cb04-46dc-b4da-8f6933cd8fd4"],

Cell["Prefactor", \
"Text",ExpressionUUID->"df929736-36ee-4527-8159-4a67b748bf37"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"S1", " ", "=", " ", 
   RowBox[{"\[Chi]1", " ", "/", " ", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"S2", " ", "=", " ", 
   RowBox[{"\[Chi]2", " ", 
    RowBox[{
     RowBox[{"q", "^", "2"}], " ", "/", " ", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"L", "=", 
   RowBox[{
    RowBox[{"q", " ", "/", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}], " ", 
    RowBox[{"Sqrt", "[", "r", "]"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"J", "=", 
   RowBox[{"L", " ", 
    RowBox[{"Sqrt", "[", 
     RowBox[{"L", "+", 
      RowBox[{"2", "\[Kappa]"}]}], " ", "]"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[Eta]", "=", 
   RowBox[{"q", " ", "/", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}], 
  ";"}], "\[IndentingNewLine]"}], "Input",
 CellLabel->"In[45]:=",ExpressionUUID->"143cb5c3-24c0-4e52-8d74-01c36b2150c6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Ss", "=", " ", 
  RowBox[{
   RowBox[{"q", " ", "/", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}], " ", 
   RowBox[{"Sqrt", "[", "r", "]"}], 
   RowBox[{"(", " ", 
    RowBox[{
     RowBox[{"2", " ", "\[Kappa]"}], " ", "-", "\[Chi]eff", " ", "-", 
     RowBox[{"\[Delta]\[Chi]", " ", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], "/", 
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}]}]}]}], ")"}]}]}]], "Input",
 CellLabel->"In[50]:=",ExpressionUUID->"28f0a9de-019e-4ca2-a9a4-19c7d23e8db9"],

Cell[BoxData[
 FractionBox[
  RowBox[{"q", " ", 
   SqrtBox["r"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], " ", "\[Delta]\[Chi]"}], 
       RowBox[{"1", "+", "q"}]]}], "+", 
     RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}], 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{"1", "+", "q"}], ")"}], "2"]]], "Output",
 CellLabel->
  "Out[50]=",ExpressionUUID->"3563c005-dacf-454a-8177-8f64a8ee025f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"prefactorsol", " ", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"-", " ", 
      RowBox[{"(", 
       RowBox[{"3", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{"1", "-", 
            RowBox[{"q", "^", "2"}]}], ")"}], "/", " ", "q"}], ")"}], "S1", 
        " ", "S2", " ", 
        RowBox[{
         RowBox[{"\[Eta]", "^", "6"}], " ", "/", " ", 
         RowBox[{"L", "^", "5"}]}], " ", 
        RowBox[{"(", 
         RowBox[{"1", "-", 
          RowBox[{"\[Eta]", " ", 
           RowBox[{"\[Chi]eff", "/", " ", "L"}]}]}], ")"}]}], ")"}]}], " ", 
     "/", "  ", 
     RowBox[{"D", "[", 
      RowBox[{"Ss", ",", "\[Delta]\[Chi]"}], "]"}]}], "//", "Simplify"}]}], 
  "\[IndentingNewLine]"}]], "Input",
 CellLabel->"In[51]:=",ExpressionUUID->"8e692311-7410-40f3-bab6-6185e78f2c4d"],

Cell[BoxData[
 FractionBox[
  RowBox[{"3", " ", "q", " ", "\[Chi]1", " ", "\[Chi]2", " ", 
   RowBox[{"(", 
    RowBox[{
     SqrtBox["r"], "-", "\[Chi]eff"}], ")"}]}], 
  RowBox[{
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
   SuperscriptBox["r", 
    RowBox[{"7", "/", "2"}]]}]]], "Output",
 CellLabel->
  "Out[51]=",ExpressionUUID->"79a8fbbb-a861-482b-9038-2b6174cf2ba2"]
}, Open  ]],

Cell["Rewrite in a nicer form.", \
"Text",ExpressionUUID->"36f45dad-74b3-4999-b079-2561b6037aba"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"prefactor", " ", "=", 
  RowBox[{
   FractionBox[
    RowBox[{"3", " ", "q"}], 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], "2"]], "\[Chi]1", " ", "\[Chi]2", 
   SuperscriptBox[
    RowBox[{"(", "r", ")"}], 
    RowBox[{"-", "3"}]], " ", 
   RowBox[{"(", 
    RowBox[{"1", "-", 
     FractionBox[
      RowBox[{"\[Chi]eff", " "}], 
      SqrtBox["r"]]}], ")"}]}]}]], "Input",
 CellLabel->"In[52]:=",ExpressionUUID->"25b27335-57f1-4b10-9a54-b0f1765feb57"],

Cell[BoxData[
 FractionBox[
  RowBox[{"3", " ", "q", " ", "\[Chi]1", " ", "\[Chi]2", " ", 
   RowBox[{"(", 
    RowBox[{"1", "-", 
     FractionBox["\[Chi]eff", 
      SqrtBox["r"]]}], ")"}]}], 
  RowBox[{
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
   SuperscriptBox["r", "3"]}]]], "Output",
 CellLabel->
  "Out[52]=",ExpressionUUID->"3f2f68a6-62c6-4952-ab6f-f6a9d3e8512e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"prefactorsol", "-", "prefactor"}], "//", "Simplify"}]], "Input",
 CellLabel->"In[53]:=",ExpressionUUID->"9a14fee4-7159-462c-b236-163edf88a7e7"],

Cell[BoxData["0"], "Output",
 CellLabel->
  "Out[53]=",ExpressionUUID->"1ab0c22b-9739-456e-b5c9-25cf3d976a8d"]
}, Open  ]],

Cell["Angles", "Text",ExpressionUUID->"6c5a47d6-0a6a-4e1d-8cd4-cb9051081582"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cos\[CapitalDelta]\[CapitalPhi]s", "=", " ", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"cos\[Theta]12", " ", "-", " ", 
       RowBox[{"cos\[Theta]1", " ", "cos\[Theta]2"}]}], ")"}], "^", "2"}], 
    " ", "/", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"1", "-", 
        RowBox[{"cos\[Theta]1", " ", "^", "2"}]}], ")"}], 
      RowBox[{"(", 
       RowBox[{"1", "-", 
        RowBox[{"cos\[Theta]2", "^", "2"}]}], ")"}]}], ")"}]}]}], 
  "\[IndentingNewLine]"}]], "Input",
 CellLabel->"In[54]:=",ExpressionUUID->"5902fa5c-9d43-4c3f-8aa3-0a36a2d57934"],

Cell[BoxData[
 FractionBox[
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{"cos\[Theta]12", "-", 
     RowBox[{"cos\[Theta]1", " ", "cos\[Theta]2"}]}], ")"}], "2"], 
  RowBox[{
   RowBox[{"(", 
    RowBox[{"1", "-", 
     SuperscriptBox["cos\[Theta]1", "2"]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{"1", "-", 
     SuperscriptBox["cos\[Theta]2", "2"]}], ")"}]}]]], "Output",
 CellLabel->
  "Out[54]=",ExpressionUUID->"875ef68d-3768-43c6-9d40-bbde242d7667"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"threesines", " ", "=", " ", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"cos\[Theta]1", "^", "2"}]}], ")"}], 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"cos\[Theta]2", "^", "2"}]}], ")"}], 
    RowBox[{"(", 
     RowBox[{"1", "-", "cos\[CapitalDelta]\[CapitalPhi]s"}], ")"}]}], "//", 
   "Simplify"}]}]], "Input",
 CellLabel->"In[55]:=",ExpressionUUID->"f537021f-6cba-47e2-9df6-5fc53804b98e"],

Cell[BoxData[
 RowBox[{"1", "-", 
  SuperscriptBox["cos\[Theta]1", "2"], "-", 
  SuperscriptBox["cos\[Theta]12", "2"], "+", 
  RowBox[{
  "2", " ", "cos\[Theta]1", " ", "cos\[Theta]12", " ", "cos\[Theta]2"}], "-", 
  
  SuperscriptBox["cos\[Theta]2", "2"]}]], "Output",
 CellLabel->
  "Out[55]=",ExpressionUUID->"16054c5e-0676-4b24-8aa4-1918760df055"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"cos\[Theta]1", "=", 
  FractionBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"\[Chi]eff", "+", "\[Delta]\[Chi]"}], ")"}]}], 
   RowBox[{"2", " ", "\[Chi]1"}]]}], "\n", 
 RowBox[{"cos\[Theta]2", "=", 
  FractionBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"\[Chi]eff", "-", "\[Delta]\[Chi]"}], ")"}]}], 
   RowBox[{"2", " ", "q", " ", "\[Chi]2"}]]}], "\n", 
 RowBox[{"cos\[Theta]12", " ", "=", 
  RowBox[{
   FractionBox["1", 
    RowBox[{"2", " ", "\[Chi]1", " ", "\[Chi]2", " ", 
     SuperscriptBox["q", "2"]}]], 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"q", " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"2", " ", "\[Kappa]", " ", 
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}]}], "-", " ", 
        RowBox[{"\[Chi]eff", 
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}]}], "-", 
        RowBox[{"\[Delta]\[Chi]", 
         RowBox[{"(", 
          RowBox[{"1", "-", "q"}], ")"}]}]}], ")"}], " ", 
      SqrtBox["r"]}], " ", "-", " ", 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox["\[Chi]1", "2"], "+", 
       RowBox[{
        SuperscriptBox["\[Chi]2", "2"], "  ", 
        SuperscriptBox["q", "4"]}]}], ")"}]}], ")"}]}], " "}]}], "Input",
 CellLabel->"In[56]:=",ExpressionUUID->"24b61c7e-6425-4289-addc-26722ab22c7d"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{"(", 
    RowBox[{"1", "+", "q"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{"\[Delta]\[Chi]", "+", "\[Chi]eff"}], ")"}]}], 
  RowBox[{"2", " ", "\[Chi]1"}]]], "Output",
 CellLabel->"Out[56]=",ExpressionUUID->"fd28fc2e-cc1c-49e9-93fa-fca3ca6c3b02"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{"(", 
    RowBox[{"1", "+", "q"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "\[Delta]\[Chi]"}], "+", "\[Chi]eff"}], ")"}]}], 
  RowBox[{"2", " ", "q", " ", "\[Chi]2"}]]], "Output",
 CellLabel->"Out[57]=",ExpressionUUID->"cd7fc01f-73e4-461f-8346-d6dbfb248752"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", 
   RowBox[{"2", " ", 
    SuperscriptBox["q", "2"], " ", "\[Chi]1", " ", "\[Chi]2"}]], 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"-", 
     SuperscriptBox["\[Chi]1", "2"]}], "-", 
    RowBox[{
     SuperscriptBox["q", "4"], " ", 
     SuperscriptBox["\[Chi]2", "2"]}], "+", 
    RowBox[{"q", " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], " ", 
     SqrtBox["r"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{"1", "-", "q"}], ")"}], " ", "\[Delta]\[Chi]"}], ")"}]}], 
       "+", 
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], " ", "\[Kappa]"}], "-", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], " ", "\[Chi]eff"}]}], ")"}]}]}], 
   ")"}]}]], "Output",
 CellLabel->
  "Out[58]=",ExpressionUUID->"7f365a4a-2925-4992-8a2e-99a97e9dc8d2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cubic", " ", "=", " ", 
  RowBox[{
   RowBox[{
    RowBox[{"prefactor", "^", "2"}], " ", "threesines"}], "//", 
   "Simplify"}]}]], "Input",
 CellLabel->"In[59]:=",ExpressionUUID->"cccfbf75-7c0f-47ce-bc01-46f978e667b8"],

Cell[BoxData[
 RowBox[{"-", 
  RowBox[{
   FractionBox["1", 
    RowBox[{"4", " ", 
     SuperscriptBox["q", "2"], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "4"], " ", 
     SuperscriptBox["r", "6"]}]], 
   RowBox[{"9", " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", 
       FractionBox["\[Chi]eff", 
        SqrtBox["r"]]}], ")"}], "2"], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "4"}], " ", 
       SuperscriptBox["q", "4"], " ", 
       SuperscriptBox["\[Chi]1", "2"], " ", 
       SuperscriptBox["\[Chi]2", "2"]}], "+", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox["\[Chi]1", "2"], "+", 
         RowBox[{
          SuperscriptBox["q", "4"], " ", 
          SuperscriptBox["\[Chi]2", "2"]}], "-", 
         RowBox[{"q", " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], " ", 
          SqrtBox["r"], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", "\[Delta]\[Chi]"}],
             "+", 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"1", "+", "q"}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}]}], 
           ")"}]}]}], ")"}], "2"], "+", 
      RowBox[{
       SuperscriptBox["q", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
       SuperscriptBox["\[Chi]1", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"\[Delta]\[Chi]", "-", "\[Chi]eff"}], ")"}], "2"]}], "-", 
      RowBox[{"q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox["\[Chi]1", "2"], "+", 
         RowBox[{
          SuperscriptBox["q", "4"], " ", 
          SuperscriptBox["\[Chi]2", "2"]}], "-", 
         RowBox[{"q", " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], " ", 
          SqrtBox["r"], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", "\[Delta]\[Chi]"}],
             "+", 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"1", "+", "q"}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}]}], 
           ")"}]}]}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"\[Delta]\[Chi]", "-", "\[Chi]eff"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"\[Delta]\[Chi]", "+", "\[Chi]eff"}], ")"}]}], "+", 
      RowBox[{
       SuperscriptBox["q", "4"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
       SuperscriptBox["\[Chi]2", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"\[Delta]\[Chi]", "+", "\[Chi]eff"}], ")"}], "2"]}]}], 
     ")"}]}]}]}]], "Output",
 CellLabel->
  "Out[59]=",ExpressionUUID->"d1ccf54f-d1d4-42ca-8397-f95a4ad3ff64"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"coeffs", " ", "=", " ", 
  RowBox[{"Simplify", "[", 
   RowBox[{
    RowBox[{"CoefficientList", "[", 
     RowBox[{"cubic", ",", "\[Delta]\[Chi]"}], "]"}], ",", 
    RowBox[{
     RowBox[{"q", ">", "0"}], "&&", 
     RowBox[{"q", "<=", "1"}], "&&", 
     RowBox[{"r", ">", "0"}]}]}], "]"}]}]], "Input",
 CellLabel->"In[60]:=",ExpressionUUID->"0baa5728-a027-4208-8ba2-8522926b873c"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", 
    RowBox[{
     FractionBox["1", 
      RowBox[{"4", " ", 
       SuperscriptBox["q", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "4"], " ", 
       SuperscriptBox["r", "7"]}]], 
     RowBox[{"9", " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         SqrtBox["r"], "-", "\[Chi]eff"}], ")"}], "2"], " ", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["\[Chi]1", "4"], "+", 
        RowBox[{
         SuperscriptBox["q", "8"], " ", 
         SuperscriptBox["\[Chi]2", "4"]}], "+", 
        RowBox[{"q", " ", 
         SuperscriptBox["\[Chi]1", "2"], " ", 
         RowBox[{"(", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], "+", 
           RowBox[{
            SqrtBox["r"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "4"}], " ", "\[Kappa]"}], "+", 
              RowBox[{"2", " ", "\[Chi]eff"}]}], ")"}]}]}], ")"}]}], "+", 
        RowBox[{
         SuperscriptBox["q", "7"], " ", 
         SuperscriptBox["\[Chi]2", "2"], " ", 
         RowBox[{"(", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], "+", 
           RowBox[{
            SqrtBox["r"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "4"}], " ", "\[Kappa]"}], "+", 
              RowBox[{"2", " ", "\[Chi]eff"}]}], ")"}]}]}], ")"}]}], "+", 
        RowBox[{
         SuperscriptBox["q", "2"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"3", " ", 
            SuperscriptBox["\[Chi]1", "2"], " ", 
            SuperscriptBox["\[Chi]eff", "2"]}], "+", 
           RowBox[{"r", " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "2"}], " ", "\[Kappa]"}], "+", "\[Chi]eff"}], 
              ")"}], "2"]}], "-", 
           RowBox[{
            SqrtBox["r"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}], " ", 
            
            RowBox[{"(", 
             RowBox[{
              RowBox[{"4", " ", 
               SuperscriptBox["\[Chi]1", "2"]}], "+", 
              SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}]}], ")"}]}], "+", 
        RowBox[{
         SuperscriptBox["q", "4"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            SuperscriptBox["\[Chi]2", "2"], " ", 
            SuperscriptBox["\[Chi]eff", "2"]}], "+", 
           RowBox[{"6", " ", 
            SqrtBox["r"], " ", 
            SuperscriptBox["\[Chi]eff", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", "\[Kappa]"}], "+", "\[Chi]eff"}], 
             ")"}]}], "+", 
           RowBox[{"6", " ", "r", " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "2"}], " ", "\[Kappa]"}], "+", "\[Chi]eff"}], 
              ")"}], "2"]}], "+", 
           RowBox[{
            SuperscriptBox["\[Chi]1", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}]}], ")"}]}], "+", 
        RowBox[{
         SuperscriptBox["q", "6"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"3", " ", 
            SuperscriptBox["\[Chi]2", "2"], " ", 
            SuperscriptBox["\[Chi]eff", "2"]}], "+", 
           RowBox[{"r", " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "2"}], " ", "\[Kappa]"}], "+", "\[Chi]eff"}], 
              ")"}], "2"]}], "-", 
           RowBox[{
            SqrtBox["r"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}], " ", 
            
            RowBox[{"(", 
             RowBox[{
              RowBox[{"4", " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}]}], ")"}]}], "+", 
        RowBox[{
         SuperscriptBox["q", "3"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"3", " ", 
            SuperscriptBox["\[Chi]1", "2"], " ", 
            SuperscriptBox["\[Chi]eff", "2"]}], "+", 
           RowBox[{"4", " ", "r", " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "2"}], " ", "\[Kappa]"}], "+", "\[Chi]eff"}], 
              ")"}], "2"]}], "-", 
           RowBox[{"2", " ", 
            SqrtBox["r"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}], " ", 
            
            RowBox[{"(", 
             RowBox[{
              SuperscriptBox["\[Chi]1", "2"], "+", 
              RowBox[{"2", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], "+", 
        
        RowBox[{
         SuperscriptBox["q", "5"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"3", " ", 
            SuperscriptBox["\[Chi]2", "2"], " ", 
            SuperscriptBox["\[Chi]eff", "2"]}], "+", 
           RowBox[{"4", " ", "r", " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "2"}], " ", "\[Kappa]"}], "+", "\[Chi]eff"}], 
              ")"}], "2"]}], "-", 
           RowBox[{"2", " ", 
            SqrtBox["r"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}], " ", 
            
            RowBox[{"(", 
             RowBox[{
              SuperscriptBox["\[Chi]2", "2"], "+", 
              RowBox[{"2", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}]}], 
       ")"}]}]}]}], ",", 
   RowBox[{
    FractionBox["1", 
     RowBox[{"4", " ", "q", " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "3"], " ", 
      SuperscriptBox["r", "7"]}]], 
    RowBox[{"9", " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        SqrtBox["r"], "-", "\[Chi]eff"}], ")"}], "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "2"}], " ", 
        SqrtBox["r"], " ", 
        SuperscriptBox["\[Chi]1", "2"]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox["q", "5"], " ", 
        SqrtBox["r"], " ", 
        SuperscriptBox["\[Chi]2", "2"]}], "+", 
       RowBox[{
        SuperscriptBox["q", "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"4", " ", "r", " ", "\[Kappa]"}], "-", 
          RowBox[{"2", " ", "r", " ", "\[Chi]eff"}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["\[Chi]1", "2"], " ", "\[Chi]eff"}], "-", 
          RowBox[{
           SqrtBox["r"], " ", 
           SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox["q", "3"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "4"}], " ", "r", " ", "\[Kappa]"}], "+", 
          RowBox[{"2", " ", "r", " ", "\[Chi]eff"}], "-", 
          RowBox[{"2", " ", 
           SuperscriptBox["\[Chi]2", "2"], " ", "\[Chi]eff"}], "+", 
          RowBox[{
           SqrtBox["r"], " ", 
           SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
       RowBox[{"q", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"r", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"4", " ", "\[Kappa]"}], "-", 
             RowBox[{"2", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["\[Chi]1", "2"], " ", "\[Chi]eff"}], "+", 
          RowBox[{
           SqrtBox["r"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", 
              SuperscriptBox["\[Chi]1", "2"]}], "-", 
             SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox["q", "4"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "2"}], " ", 
           SuperscriptBox["\[Chi]2", "2"], " ", "\[Chi]eff"}], "+", 
          RowBox[{"r", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "4"}], " ", "\[Kappa]"}], "+", 
             RowBox[{"2", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
          RowBox[{
           SqrtBox["r"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "2"}], " ", 
              SuperscriptBox["\[Chi]2", "2"]}], "+", 
             SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}]}], ")"}]}]}], 
      ")"}]}]}], ",", 
   RowBox[{
    FractionBox["1", 
     RowBox[{"4", " ", "q", " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
      SuperscriptBox["r", "7"]}]], 
    RowBox[{"9", " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        SqrtBox["r"], "-", "\[Chi]eff"}], ")"}], "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox["\[Chi]1", "2"], "+", 
       RowBox[{
        SuperscriptBox["q", "4"], " ", 
        SuperscriptBox["\[Chi]2", "2"]}], "-", 
       RowBox[{"q", " ", 
        RowBox[{"(", 
         RowBox[{"r", "+", 
          SuperscriptBox["\[Chi]1", "2"], "+", 
          RowBox[{
           SqrtBox["r"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}]}], 
         ")"}]}], "-", 
       RowBox[{
        SuperscriptBox["q", "3"], " ", 
        RowBox[{"(", 
         RowBox[{"r", "+", 
          SuperscriptBox["\[Chi]2", "2"], "+", 
          RowBox[{
           SqrtBox["r"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}]}], 
         ")"}]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox["q", "2"], " ", 
        SqrtBox["r"], " ", 
        RowBox[{"(", 
         RowBox[{
          SqrtBox["r"], "-", 
          RowBox[{"2", " ", "\[Kappa]"}], "+", "\[Chi]eff"}], ")"}]}]}], 
      ")"}]}]}], ",", 
   RowBox[{"-", 
    FractionBox[
     RowBox[{"9", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         SqrtBox["r"], "-", "\[Chi]eff"}], ")"}], "2"]}], 
     RowBox[{"4", " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], " ", 
      SuperscriptBox["r", 
       RowBox[{"13", "/", "2"}]]}]]}]}], "}"}]], "Output",
 CellLabel->
  "Out[60]=",ExpressionUUID->"4f46034d-c64f-4856-8b62-c9ff5382f2cc"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[CapitalAlpha]2sol", " ", "=", " ", 
  RowBox[{
   RowBox[{
    RowBox[{"coeffs", "[", 
     RowBox[{"[", "4", "]"}], "]"}], "/", 
    RowBox[{"(", 
     RowBox[{"1", "-", "q"}], ")"}]}], "//", "Simplify"}]}]], "Input",
 CellLabel->"In[61]:=",ExpressionUUID->"ce1e85c5-d1bf-46aa-b36d-5c2270a8b875"],

Cell[BoxData[
 FractionBox[
  RowBox[{"9", " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{
      SqrtBox["r"], "-", "\[Chi]eff"}], ")"}], "2"]}], 
  RowBox[{"4", " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "q"}], ")"}], " ", 
   SuperscriptBox["r", 
    RowBox[{"13", "/", "2"}]]}]]], "Output",
 CellLabel->
  "Out[61]=",ExpressionUUID->"188cd252-ada8-4207-8ffd-7a6d0045738b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[CapitalAlpha]", " ", "=", " ", 
  RowBox[{
   RowBox[{"3", "/", "2"}], " ", 
   RowBox[{"1", "/", 
    RowBox[{"Sqrt", "[", 
     RowBox[{"1", "+", "q"}], "]"}]}], " ", 
   RowBox[{"r", "^", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "11"}], "/", "4"}], " ", ")"}]}], 
   RowBox[{"(", 
    RowBox[{"1", "-", 
     RowBox[{"\[Chi]eff", " ", "/", 
      RowBox[{"Sqrt", "[", "r", "]"}]}]}], ")"}]}]}]], "Input",
 CellLabel->"In[62]:=",ExpressionUUID->"6524f61b-b43e-40f6-90c2-802329ecce02"],

Cell[BoxData[
 FractionBox[
  RowBox[{"3", " ", 
   RowBox[{"(", 
    RowBox[{"1", "-", 
     FractionBox["\[Chi]eff", 
      SqrtBox["r"]]}], ")"}]}], 
  RowBox[{"2", " ", 
   SqrtBox[
    RowBox[{"1", "+", "q"}]], " ", 
   SuperscriptBox["r", 
    RowBox[{"11", "/", "4"}]]}]]], "Output",
 CellLabel->
  "Out[62]=",ExpressionUUID->"fdf9cf90-edd1-484c-8eff-d33ed84f084e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"\[CapitalAlpha]2sol", "-", 
   RowBox[{"\[CapitalAlpha]", "^", "2"}]}], "//", "Simplify"}]], "Input",
 CellLabel->"In[63]:=",ExpressionUUID->"d5e5f849-250e-416c-b910-e3de10f68226"],

Cell[BoxData["0"], "Output",
 CellLabel->
  "Out[63]=",ExpressionUUID->"1d08d64f-f708-4fcf-8150-317c29757efb"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"r", " ", "=", " ", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "^", "2"}], " ", "/", " ", 
      RowBox[{"(", 
       RowBox[{"2", " ", "u", " ", "q"}], " ", ")"}]}], ")"}], "^", "2"}]}], 
  ";"}]], "Input",
 CellLabel->"In[64]:=",ExpressionUUID->"419f2221-82f2-44b5-82a0-87e87ee96933"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cubic", " ", "=", 
  RowBox[{"Simplify", "[", 
   RowBox[{"cubic", ",", " ", 
    RowBox[{
     RowBox[{"q", ">", "0"}], "&&", 
     RowBox[{"q", "<", "1"}], "&&", 
     RowBox[{"u", ">", "0"}]}]}], "]"}]}]], "Input",
 CellLabel->"In[65]:=",ExpressionUUID->"173877c9-93af-4e3a-bcc6-8436dace4e55"],

Cell[BoxData[
 RowBox[{"-", 
  RowBox[{
   FractionBox["1", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], "28"]], 
   RowBox[{"9216", " ", 
    SuperscriptBox["q", "10"], " ", 
    SuperscriptBox["u", "12"], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", "q", " ", "u", " ", "\[Chi]eff"}], 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "2"]]}], ")"}], "2"], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "4"}], " ", 
       SuperscriptBox["q", "4"], " ", 
       SuperscriptBox["\[Chi]1", "2"], " ", 
       SuperscriptBox["\[Chi]2", "2"]}], "+", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox["\[Chi]1", "2"], "+", 
         RowBox[{
          SuperscriptBox["q", "4"], " ", 
          SuperscriptBox["\[Chi]2", "2"]}], "-", 
         FractionBox[
          RowBox[{
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "3"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", 
              "\[Delta]\[Chi]"}], "+", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{"1", "+", "q"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}]}],
             ")"}]}], 
          RowBox[{"2", " ", "u"}]]}], ")"}], "2"], "+", 
      RowBox[{
       SuperscriptBox["q", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
       SuperscriptBox["\[Chi]1", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"\[Delta]\[Chi]", "-", "\[Chi]eff"}], ")"}], "2"]}], "-", 
      RowBox[{"q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox["\[Chi]1", "2"], "+", 
         RowBox[{
          SuperscriptBox["q", "4"], " ", 
          SuperscriptBox["\[Chi]2", "2"]}], "-", 
         FractionBox[
          RowBox[{
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "3"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", 
              "\[Delta]\[Chi]"}], "+", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{"1", "+", "q"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}]}],
             ")"}]}], 
          RowBox[{"2", " ", "u"}]]}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"\[Delta]\[Chi]", "-", "\[Chi]eff"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"\[Delta]\[Chi]", "+", "\[Chi]eff"}], ")"}]}], "+", 
      RowBox[{
       SuperscriptBox["q", "4"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
       SuperscriptBox["\[Chi]2", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"\[Delta]\[Chi]", "+", "\[Chi]eff"}], ")"}], "2"]}]}], 
     ")"}]}]}]}]], "Output",
 CellLabel->
  "Out[65]=",ExpressionUUID->"600d096c-f749-4cbc-be85-f5e824aa3fa5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"coeffs", " ", "=", " ", 
  RowBox[{"Simplify", "[", 
   RowBox[{
    RowBox[{"CoefficientList", "[", 
     RowBox[{"cubic", ",", "\[Delta]\[Chi]"}], "]"}], ",", 
    RowBox[{
     RowBox[{"q", ">", "0"}], "&&", 
     RowBox[{"q", "<=", "1"}], "&&", 
     RowBox[{"r", ">", "0"}]}]}], "]"}]}]], "Input",
 CellLabel->"In[66]:=",ExpressionUUID->"15704ca5-f6f6-4b70-8947-e96e8d5710fe"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    FractionBox["1", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "28"]], 
    RowBox[{"2304", " ", 
     SuperscriptBox["q", "10"], " ", 
     SuperscriptBox["u", "10"], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "-", 
        FractionBox[
         RowBox[{"2", " ", "q", " ", "u", " ", "\[Chi]eff"}], 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "2"]]}], ")"}], "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "4"}], " ", 
        SuperscriptBox["u", "2"], " ", 
        SuperscriptBox["\[Chi]1", "4"]}], "+", 
       RowBox[{"8", " ", 
        SuperscriptBox["q", "4"], " ", 
        SuperscriptBox["u", "2"], " ", 
        SuperscriptBox["\[Chi]1", "2"], " ", 
        SuperscriptBox["\[Chi]2", "2"]}], "-", 
       RowBox[{"4", " ", 
        SuperscriptBox["q", "8"], " ", 
        SuperscriptBox["u", "2"], " ", 
        SuperscriptBox["\[Chi]2", "4"]}], "+", 
       RowBox[{"4", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "4"], " ", "u", " ", 
        SuperscriptBox["\[Chi]1", "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}], "+", 
       RowBox[{"4", " ", 
        SuperscriptBox["q", "4"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "4"], " ", "u", " ", 
        SuperscriptBox["\[Chi]2", "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}], "-", 
       RowBox[{"4", " ", "q", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
        SuperscriptBox["u", "2"], " ", 
        SuperscriptBox["\[Chi]1", "2"], " ", 
        SuperscriptBox["\[Chi]eff", "2"]}], "-", 
       RowBox[{"4", " ", 
        SuperscriptBox["q", "2"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
        SuperscriptBox["u", "2"], " ", 
        SuperscriptBox["\[Chi]1", "2"], " ", 
        SuperscriptBox["\[Chi]eff", "2"]}], "-", 
       RowBox[{"4", " ", 
        SuperscriptBox["q", "4"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
        SuperscriptBox["u", "2"], " ", 
        SuperscriptBox["\[Chi]2", "2"], " ", 
        SuperscriptBox["\[Chi]eff", "2"]}], "-", 
       RowBox[{"4", " ", 
        SuperscriptBox["q", "5"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
        SuperscriptBox["u", "2"], " ", 
        SuperscriptBox["\[Chi]2", "2"], " ", 
        SuperscriptBox["\[Chi]eff", "2"]}], "+", 
       RowBox[{"2", " ", "q", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "6"], " ", "u", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}], " ", 
        SuperscriptBox["\[Chi]eff", "2"]}], "-", 
       RowBox[{
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "8"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", "2"}], " ", "\[Kappa]"}], "+", "\[Chi]eff"}], ")"}], 
         "2"]}]}], ")"}]}]}], ",", 
   RowBox[{
    FractionBox["1", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "30"]], 
    RowBox[{"4608", " ", 
     SuperscriptBox["q", "10"], " ", 
     SuperscriptBox["u", "10"], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", 
        SuperscriptBox["q", "2"], "+", 
        RowBox[{"q", " ", 
         RowBox[{"(", 
          RowBox[{"2", "-", 
           RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], ")"}], 
      "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "2"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "5"], " ", "\[Kappa]"}], "+", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "5"], " ", "\[Chi]eff"}], "+", 
       RowBox[{"4", " ", 
        SuperscriptBox["q", "2"], " ", 
        SuperscriptBox["u", "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox["\[Chi]1", "2"], "-", 
          RowBox[{
           SuperscriptBox["q", "2"], " ", 
           SuperscriptBox["\[Chi]2", "2"]}]}], ")"}], " ", "\[Chi]eff"}], "+", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", 
          SuperscriptBox["q", "2"]}], ")"}], " ", "u", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", 
           SuperscriptBox["\[Chi]1", "2"]}], "+", 
          RowBox[{"q", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", 
              SuperscriptBox["q", "3"], " ", 
              SuperscriptBox["\[Chi]2", "2"]}], "+", 
             SuperscriptBox["\[Chi]eff", "2"], "+", 
             RowBox[{"2", " ", "q", " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{
              SuperscriptBox["q", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}]}], 
      ")"}]}]}], ",", 
   RowBox[{"-", 
    RowBox[{
     FractionBox["1", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "30"]], 
     RowBox[{"2304", " ", 
      SuperscriptBox["q", "10"], " ", 
      SuperscriptBox["u", "10"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", 
         SuperscriptBox["q", "2"], "+", 
         RowBox[{"q", " ", 
          RowBox[{"(", 
           RowBox[{"2", "-", 
            RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], ")"}], 
       "2"], " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", 
        SuperscriptBox["q", "6"], "+", 
        RowBox[{
         SuperscriptBox["q", "2"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", 
           RowBox[{"16", " ", "u", " ", "\[Kappa]"}], "+", 
           RowBox[{"4", " ", 
            SuperscriptBox["u", "2"], " ", 
            SuperscriptBox["\[Chi]1", "2"]}], "-", 
           RowBox[{"8", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
        RowBox[{
         SuperscriptBox["q", "4"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", 
           RowBox[{"16", " ", "u", " ", "\[Kappa]"}], "+", 
           RowBox[{"4", " ", 
            SuperscriptBox["u", "2"], " ", 
            SuperscriptBox["\[Chi]2", "2"]}], "-", 
           RowBox[{"8", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
        RowBox[{"4", " ", 
         SuperscriptBox["q", "3"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", 
           RowBox[{"6", " ", "u", " ", "\[Kappa]"}], "-", 
           RowBox[{"3", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
        RowBox[{"q", " ", 
         RowBox[{"(", 
          RowBox[{"2", "+", 
           RowBox[{"4", " ", "u", " ", "\[Kappa]"}], "-", 
           RowBox[{"4", " ", 
            SuperscriptBox["u", "2"], " ", 
            SuperscriptBox["\[Chi]1", "2"]}], "-", 
           RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
        RowBox[{
         SuperscriptBox["q", "5"], " ", 
         RowBox[{"(", 
          RowBox[{"2", "+", 
           RowBox[{"4", " ", "u", " ", "\[Kappa]"}], "-", 
           RowBox[{"4", " ", 
            SuperscriptBox["u", "2"], " ", 
            SuperscriptBox["\[Chi]2", "2"]}], "-", 
           RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
       ")"}]}]}]}], ",", 
   RowBox[{"-", 
    FractionBox[
     RowBox[{"4608", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", 
      SuperscriptBox["q", "11"], " ", 
      SuperscriptBox["u", "11"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", 
         SuperscriptBox["q", "2"], "+", 
         RowBox[{"q", " ", 
          RowBox[{"(", 
           RowBox[{"2", "-", 
            RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], ")"}], 
       "2"]}], 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "27"]]}]}], "}"}]], "Output",
 CellLabel->
  "Out[66]=",ExpressionUUID->"54adfc83-afa4-45dc-89e8-bcd85d984e46"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"overall", " ", "=", " ", 
  RowBox[{
   RowBox[{
    RowBox[{"coeffs", "[", 
     RowBox[{"[", "4", "]"}], "]"}], "/", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"1", "-", "q"}], ")"}], "u"}], ")"}]}], "//", 
   "FullSimplify"}]}]], "Input",
 CellLabel->"In[67]:=",ExpressionUUID->"a2ceb419-a67b-42ed-8bc6-4b1ced80660d"],

Cell[BoxData[
 FractionBox[
  RowBox[{"4608", " ", 
   SuperscriptBox["q", "11"], " ", 
   SuperscriptBox["u", "10"], " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"], "-", 
      RowBox[{"2", " ", "q", " ", "u", " ", "\[Chi]eff"}]}], ")"}], "2"]}], 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{"1", "+", "q"}], ")"}], "27"]]], "Output",
 CellLabel->
  "Out[67]=",ExpressionUUID->"5846018b-668f-43d1-b14a-dd417834be27"]
}, Open  ]],

Cell["\<\
Some simplification by hand to get a nicer form. Define the following  \
shortcuts:\
\>", "Text",ExpressionUUID->"5ed431d6-f21c-4550-b8e2-097c267685d1"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"\[Kappa]tr", "  ", "=", " ", 
  RowBox[{"t\[Kappa]", "->", 
   RowBox[{
    RowBox[{"2", " ", "\[Kappa]"}], "-", 
    "\[Chi]eff"}]}]}], "\[IndentingNewLine]", 
 RowBox[{"\[Chi]tr", " ", "=", " ", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"t\[Chi]1", "->", 
     RowBox[{
      SuperscriptBox["\[Chi]1", "2"], "+", 
      RowBox[{
       SuperscriptBox["q", "4"], " ", 
       SuperscriptBox["\[Chi]2", "2"]}]}]}], ",", 
    RowBox[{"t\[Chi]2", "->", " ", 
     RowBox[{
      SuperscriptBox["\[Chi]1", "2"], " ", 
      SuperscriptBox["\[Chi]2", "2"]}]}], ",", 
    RowBox[{"t\[Chi]3", "->", 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox["\[Chi]1", "2"], " ", "+", 
       RowBox[{
        SuperscriptBox["q", "2"], 
        SuperscriptBox["\[Chi]2", "2"]}]}], ")"}]}], ",", 
    RowBox[{"t\[Chi]4", "->", 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox["\[Chi]1", "2"], " ", "-", 
       RowBox[{
        SuperscriptBox["q", "2"], 
        SuperscriptBox["\[Chi]2", "2"]}]}], ")"}]}]}], "}"}]}]}], "Input",
 CellLabel->"In[68]:=",ExpressionUUID->"89cedf39-89a0-4d4c-904a-366d9d680bf1"],

Cell[BoxData[
 RowBox[{"t\[Kappa]", "\[Rule]", 
  RowBox[{
   RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}]}]], "Output",
 CellLabel->"Out[68]=",ExpressionUUID->"02596f15-3772-4ebd-9293-d2774b708be7"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"t\[Chi]1", "\[Rule]", 
    RowBox[{
     SuperscriptBox["\[Chi]1", "2"], "+", 
     RowBox[{
      SuperscriptBox["q", "4"], " ", 
      SuperscriptBox["\[Chi]2", "2"]}]}]}], ",", 
   RowBox[{"t\[Chi]2", "\[Rule]", 
    RowBox[{
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "2"]}]}], ",", 
   RowBox[{"t\[Chi]3", "\[Rule]", 
    RowBox[{
     SuperscriptBox["\[Chi]1", "2"], "+", 
     RowBox[{
      SuperscriptBox["q", "2"], " ", 
      SuperscriptBox["\[Chi]2", "2"]}]}]}], ",", 
   RowBox[{"t\[Chi]4", "\[Rule]", 
    RowBox[{
     SuperscriptBox["\[Chi]1", "2"], "-", 
     RowBox[{
      SuperscriptBox["q", "2"], " ", 
      SuperscriptBox["\[Chi]2", "2"]}]}]}]}], "}"}]], "Output",
 CellLabel->
  "Out[69]=",ExpressionUUID->"1abb6332-0618-4780-8164-7ce0d45d8ed8"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"mycubic", "=", " ", 
   RowBox[{
    RowBox[{"-", 
     FractionBox["1", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "28"]]}], "9216", " ", 
    SuperscriptBox["q", "10"], " ", 
    SuperscriptBox["u", "12"], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", "q", " ", "u", " ", "\[Chi]eff"}], 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "2"]]}], ")"}], "2"], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "4"}], " ", 
       SuperscriptBox["q", "4"], " ", "t\[Chi]2"}], "+", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"t\[Chi]1", "-", 
         FractionBox[
          RowBox[{
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "3"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", 
              "\[Delta]\[Chi]"}], "+", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{"1", "+", "q"}], ")"}], " ", "t\[Kappa]"}]}], ")"}]}], 
          
          RowBox[{"2", " ", "u"}]]}], ")"}], "2"], "-", 
      RowBox[{"q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{"t\[Chi]1", "-", 
         FractionBox[
          RowBox[{
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "3"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", 
              "\[Delta]\[Chi]"}], "+", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{"1", "+", "q"}], ")"}], " ", "t\[Kappa]"}]}], ")"}]}], 
          
          RowBox[{"2", " ", "u"}]]}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"\[Delta]\[Chi]", "-", "\[Chi]eff"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"\[Delta]\[Chi]", "+", "\[Chi]eff"}], ")"}]}], "+", 
      RowBox[{
       SuperscriptBox["q", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], "t\[Chi]3", 
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox["\[Delta]\[Chi]", "2"], "+", 
         SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}], "-", 
      RowBox[{"2", 
       SuperscriptBox["q", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "t\[Chi]4", "  ", 
       "\[Delta]\[Chi]", " ", "\[Chi]eff"}]}], ")"}]}]}], ";"}]], "Input",
 CellLabel->"In[70]:=",ExpressionUUID->"a2bb14f3-f381-4709-8d0d-ea4dbf009797"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"mycubic", "-", "cubic"}], "/.", "\[Kappa]tr"}], "/.", 
   "\[Chi]tr"}], "//", "Simplify"}]], "Input",
 CellLabel->"In[71]:=",ExpressionUUID->"c83cb8e7-ba1d-4cf0-aaec-68b66b15fc87"],

Cell[BoxData["0"], "Output",
 CellLabel->
  "Out[71]=",ExpressionUUID->"6d8fb429-7d49-436b-b213-ce96bc7d0748"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"myc", " ", "=", " ", 
   RowBox[{"Simplify", "[", 
    RowBox[{"CoefficientList", "[", 
     RowBox[{"mycubic", ",", "\[Delta]\[Chi]"}], "]"}], "]"}]}], 
  ";"}]], "Input",
 CellLabel->"In[72]:=",ExpressionUUID->"cf9c4fdd-2b26-4427-bc0d-e478bd506434"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"myc", "[", 
    RowBox[{"[", "4", "]"}], "]"}], "/", "overall"}], " ", "//", 
  "FullSimplify"}]], "Input",
 CellLabel->"In[73]:=",ExpressionUUID->"2a9e3b60-f8cc-4009-b5ea-a0229fab4193"],

Cell[BoxData[
 RowBox[{"u", "-", 
  RowBox[{"q", " ", "u"}]}]], "Output",
 CellLabel->
  "Out[73]=",ExpressionUUID->"25ac73cb-ff56-482a-9ab8-3a3f202628eb"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"myc", "[", 
    RowBox[{"[", "3", "]"}], "]"}], "/", "overall"}], " ", "//", 
  "FullSimplify"}]], "Input",
 CellLabel->"In[74]:=",ExpressionUUID->"f9a7b9fe-7753-4fce-bd7e-29f40de28bec"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   FractionBox[
    RowBox[{
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "q"}], ")"}], "2"], " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}]}], 
    RowBox[{"2", " ", "q"}]]}], "-", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{"1", "+", "q"}], ")"}], " ", "t\[Kappa]", " ", "u"}], "+", 
  FractionBox[
   RowBox[{"2", " ", 
    RowBox[{"(", 
     RowBox[{"t\[Chi]1", "-", 
      RowBox[{"q", " ", "t\[Chi]3"}]}], ")"}], " ", 
    SuperscriptBox["u", "2"]}], 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], "3"]]}]], "Output",
 CellLabel->
  "Out[74]=",ExpressionUUID->"8173e042-932a-48da-8e35-1e8055379f6c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Sigma]2", "=", " ", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}]}], 
      RowBox[{"2", " ", "q"}]]}], "-", 
    RowBox[{"u", 
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}], " ", "+", 
    " ", 
    RowBox[{
     SuperscriptBox["u", "2"], 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox["\[Chi]1", "2"], "-", 
         RowBox[{
          SuperscriptBox["q", "3"], " ", 
          SuperscriptBox["\[Chi]2", "2"]}]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "3"]]}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"\[Sigma]2", "-", 
     RowBox[{
      RowBox[{"coeffs", "[", 
       RowBox[{"[", "3", "]"}], "]"}], "/", "overall"}]}], " ", "/.", 
    "\[Kappa]tr"}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"(*", 
   RowBox[{
    RowBox[{
     RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
    "TeXForm"}], "*)"}]}]}], "Input",
 CellLabel->"In[75]:=",ExpressionUUID->"f1ea76a9-88a7-4854-b218-27a72d5615f2"],

Cell[BoxData["0"], "Output",
 CellLabel->
  "Out[76]=",ExpressionUUID->"5ccb1bac-260c-4c13-ba85-abb13ebea8a0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CoefficientList", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"myc", "[", 
      RowBox[{"[", "2", "]"}], "]"}], "/", "overall"}], " ", ",", "u"}], 
   "]"}], "//", "Simplify"}]], "Input",
 CellLabel->"In[77]:=",ExpressionUUID->"42ff626b-d4fe-4e98-a57f-9abaf4e8edfa"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", 
    FractionBox[
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "t\[Kappa]"}], "q"]}], ",", 
   FractionBox[
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"2", " ", "t\[Chi]1"}], "+", 
       RowBox[{"q", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
        SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], 
    RowBox[{"q", " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "2"]}]], ",", 
   FractionBox[
    RowBox[{"4", " ", "q", " ", "t\[Chi]4", " ", "\[Chi]eff"}], 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], "3"]]}], "}"}]], "Output",
 CellLabel->
  "Out[77]=",ExpressionUUID->"1ea1203f-b5a3-4e2e-aedc-e8658523df53"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Sigma]1", "=", " ", 
   RowBox[{
    RowBox[{
     FractionBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], " "}], "q"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}], "-", 
    RowBox[{"u", " ", 
     FractionBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], " "}], 
      RowBox[{"q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"]}]], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox["\[Chi]1", "2"], "+", 
          RowBox[{
           SuperscriptBox["q", "4"], " ", 
           SuperscriptBox["\[Chi]2", "2"]}]}], ")"}]}], "+", 
       RowBox[{"q", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
        SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
    RowBox[{"4", " ", 
     SuperscriptBox["u", "2"], 
     FractionBox[
      RowBox[{" ", "q"}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "3"]], " ", 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox["\[Chi]1", "2"], " ", "-", 
       RowBox[{
        SuperscriptBox["q", "2"], 
        SuperscriptBox["\[Chi]2", "2"]}]}], ")"}], " ", "\[Chi]eff"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"\[Sigma]1", "-", " ", 
     RowBox[{
      RowBox[{"coeffs", "[", 
       RowBox[{"[", "2", "]"}], "]"}], "/", "overall"}]}], " ", "/.", 
    "\[Kappa]tr"}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"(*", 
   RowBox[{
    RowBox[{
     RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
    "TeXForm"}], "*)"}]}]}], "Input",
 CellLabel->"In[78]:=",ExpressionUUID->"04c9d24b-df0f-4ed4-a372-07e1f3faf0f2"],

Cell[BoxData["0"], "Output",
 CellLabel->
  "Out[79]=",ExpressionUUID->"4911101c-8706-4212-ba36-d66479ef0a8b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CoefficientList", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"myc", "[", 
      RowBox[{"[", "1", "]"}], "]"}], "/", "overall"}], ",", "u"}], "]"}], "//",
   "FullSimplify"}]], "Input",
 CellLabel->"In[80]:=",ExpressionUUID->"abcc302c-3300-468c-bf0b-e373ba0ca82c"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", 
    FractionBox[
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "3"], " ", 
      SuperscriptBox["t\[Kappa]", "2"]}], 
     RowBox[{"2", " ", "q"}]]}], ",", 
   RowBox[{"t\[Kappa]", " ", 
    RowBox[{"(", 
     RowBox[{
      FractionBox[
       RowBox[{"2", " ", "t\[Chi]1"}], 
       RowBox[{"q", "+", 
        SuperscriptBox["q", "2"]}]], "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], " ", 
       SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], ",", 
   FractionBox[
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      SuperscriptBox["t\[Chi]1", "2"]}], "+", 
     RowBox[{"8", " ", 
      SuperscriptBox["q", "4"], " ", "t\[Chi]2"}], "-", 
     RowBox[{"2", " ", "q", " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
      RowBox[{"(", 
       RowBox[{"t\[Chi]1", "+", 
        RowBox[{"q", " ", "t\[Chi]3"}]}], ")"}], " ", 
      SuperscriptBox["\[Chi]eff", "2"]}]}], 
    RowBox[{"q", " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "5"]}]]}], "}"}]], "Output",
 CellLabel->
  "Out[80]=",ExpressionUUID->"4ffb878b-5fc3-472b-9a5d-43a740616dc6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Sigma]0", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "3"], " "}], 
       RowBox[{"2", " ", "q"}]]}], 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}], "2"]}], 
    "+", 
    RowBox[{
     FractionBox["u", 
      RowBox[{"q", 
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}]}]], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox["\[Chi]1", "2"], "+", 
          RowBox[{
           SuperscriptBox["q", "4"], " ", 
           SuperscriptBox["\[Chi]2", "2"]}]}], ")"}]}], "+", 
       RowBox[{"q", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
        SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{"2", 
       SuperscriptBox["u", "2"]}], 
      RowBox[{"q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "5"]}]], 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox["\[Chi]1", "2"], "-", 
          RowBox[{
           SuperscriptBox["q", "4"], " ", 
           SuperscriptBox["\[Chi]2", "2"]}]}], ")"}], "2"], "+", 
       RowBox[{"q", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "3"], "  ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox["\[Chi]1", "2"], "+", 
          RowBox[{
           SuperscriptBox["q", "3"], " ", 
           SuperscriptBox["\[Chi]2", "2"]}]}], ")"}], " ", 
        SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"\[Sigma]0", "-", " ", 
     RowBox[{
      RowBox[{"coeffs", "[", 
       RowBox[{"[", "1", "]"}], "]"}], "/", "overall"}]}], " ", "/.", 
    "\[Kappa]tr"}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"(*", 
   RowBox[{
    RowBox[{
     RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
    "TeXForm"}], "*)"}]}]}], "Input",
 CellLabel->"In[81]:=",ExpressionUUID->"d911bbf1-7370-4c3b-a4fd-1540fbddb5ea"],

Cell[BoxData["0"], "Output",
 CellLabel->
  "Out[82]=",ExpressionUUID->"eca19b8f-7e1d-4d91-9702-6d424c96404f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"precessionform", "[", 
  RowBox[{"\"\<coeff2\>\"", ",", "\[Sigma]2"}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"precessionform", "[", 
  RowBox[{"\"\<coeff1\>\"", ",", "\[Sigma]1"}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"precessionform", "[", 
  RowBox[{"\"\<coeff0\>\"", ",", "\[Sigma]0"}], "]"}]}], "Input",
 CellLabel->"In[83]:=",ExpressionUUID->"2da7d4b5-d236-4ad7-a794-08c84a3679fc"],

Cell[BoxData["\<\"# Machine generated with eq_generator.nb\\ncoeff2 = (-1/2 * \
((1 + -1 * q))**2 * q**(-1) * (1 + q) + (2 * (1 + -1 * q) * ((1 + q))**(-3) * \
u**2 * (chi1**2 + -1 * q**3 * chi2**2) + -1 * (1 + q) * u * (2 * kappa + -1 * \
chieff)))\"\>"], "Output",
 CellLabel->"Out[83]=",ExpressionUUID->"58113180-fe50-49f5-ac6f-454a8cb382d1"],

Cell[BoxData["\<\"# Machine generated with eq_generator.nb\\ncoeff1 = ((1 + \
-1 * q) * q**(-1) * ((1 + q))**2 * (2 * kappa + -1 * chieff) + (4 * q * ((1 + \
q))**(-3) * u**2 * (chi1**2 + -1 * q**2 * chi2**2) * chieff + -1 * (1 + -1 * \
q) * q**(-1) * ((1 + q))**(-2) * u * (2 * (chi1**2 + q**4 * chi2**2) + q * \
((1 + q))**2 * chieff**2)))\"\>"], "Output",
 CellLabel->"Out[84]=",ExpressionUUID->"661f8b4a-0d63-4971-8b11-1f8efc379f7d"],

Cell[BoxData["\<\"# Machine generated with eq_generator.nb\\ncoeff0 = (-1/2 * \
q**(-1) * ((1 + q))**3 * ((2 * kappa + -1 * chieff))**2 + (q**(-1) * ((1 + \
q))**(-1) * u * (2 * kappa + -1 * chieff) * (2 * (chi1**2 + q**4 * chi2**2) + \
q * ((1 + q))**2 * chieff**2) + -2 * q**(-1) * ((1 + q))**(-5) * u**2 * \
(((chi1**2 + -1 * q**4 * chi2**2))**2 + q * ((1 + q))**3 * (chi1**2 + q**3 * \
chi2**2) * chieff**2)))\"\>"], "Output",
 CellLabel->
  "Out[85]=",ExpressionUUID->"16979bfd-d0bf-47c3-84c0-60d755a04e4d"]
}, Open  ]],

Cell["Large separation and equal mass limit", \
"Text",ExpressionUUID->"e260723f-30da-4e19-a2f1-a8ed0c3524b6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"largesep", " ", "=", " ", 
  RowBox[{
   RowBox[{"Limit", "[", 
    RowBox[{
     RowBox[{"cubic", "/", "overall"}], ",", 
     RowBox[{"u", "->", "0"}]}], "]"}], "//", "Simplify"}]}]], "Input",
 CellLabel->"In[86]:=",ExpressionUUID->"16002e8e-8a68-4834-9e26-a7c718aad26c"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", "\[Delta]\[Chi]"}], "+", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}]}], 
      ")"}], "2"]}], 
   RowBox[{"2", " ", "q"}]]}]], "Output",
 CellLabel->
  "Out[86]=",ExpressionUUID->"46d38f53-bf45-40e4-b281-d27d4eba6a29"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Solve", "[", 
   RowBox[{
    RowBox[{"largesep", "==", "0"}], ",", "\[Delta]\[Chi]"}], "]"}], "//", 
  "Simplify"}]], "Input",
 CellLabel->"In[87]:=",ExpressionUUID->"594470e1-b9c6-4ecc-ab01-8c32956a80e3"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"\[Delta]\[Chi]", "\[Rule]", 
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}], 
       RowBox[{
        RowBox[{"-", "1"}], "+", "q"}]]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"\[Delta]\[Chi]", "\[Rule]", 
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}], 
       RowBox[{
        RowBox[{"-", "1"}], "+", "q"}]]}]}], "}"}]}], "}"}]], "Output",
 CellLabel->
  "Out[87]=",ExpressionUUID->"1d519e7a-a58e-44fc-80e1-550700b9f21b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"equalmass", " ", "=", " ", 
  RowBox[{
   RowBox[{"cubic", "/.", 
    RowBox[{"q", "->", "1"}]}], "//", "Simplify"}]}]], "Input",
 CellLabel->"In[88]:=",ExpressionUUID->"d8205260-0b36-43eb-a4be-9208084eafb6"],

Cell[BoxData[
 RowBox[{"-", 
  RowBox[{
   FractionBox["1", "1048576"], 
   RowBox[{"9", " ", 
    SuperscriptBox["u", "10"], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "2"}], "+", 
       RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}], "2"], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"64", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "2"}], " ", "\[Kappa]"}], "+", "\[Chi]eff"}], ")"}], 
        "2"]}], "-", 
      RowBox[{"16", " ", "u", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "2"}], " ", 
          SuperscriptBox["\[Delta]\[Chi]", "2"]}], "+", 
         SuperscriptBox["\[Chi]1", "2"], "+", 
         SuperscriptBox["\[Chi]2", "2"], "+", 
         RowBox[{"2", " ", 
          SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
      RowBox[{
       SuperscriptBox["u", "2"], " ", 
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox["\[Chi]1", "4"], "+", 
         SuperscriptBox["\[Chi]2", "4"], "+", 
         RowBox[{"8", " ", 
          SuperscriptBox["\[Chi]2", "2"], " ", "\[Chi]eff", " ", 
          RowBox[{"(", 
           RowBox[{"\[Delta]\[Chi]", "+", "\[Chi]eff"}], ")"}]}], "-", 
         RowBox[{"2", " ", 
          SuperscriptBox["\[Chi]1", "2"], " ", 
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox["\[Chi]2", "2"], "+", 
            RowBox[{"4", " ", 
             RowBox[{"(", 
              RowBox[{"\[Delta]\[Chi]", "-", "\[Chi]eff"}], ")"}], " ", 
             "\[Chi]eff"}]}], ")"}]}]}], ")"}]}]}], ")"}]}]}]}]], "Output",
 CellLabel->
  "Out[88]=",ExpressionUUID->"185e3eb7-ac0f-441b-b10c-68c51b3f4a4f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"tau", " ", "=", 
  RowBox[{"Simplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"4", "*", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"\[Pi]", "/", "2"}], ")"}], "/", 
       RowBox[{"Sqrt", "[", 
        RowBox[{"-", " ", 
         RowBox[{
          RowBox[{"CoefficientList", "[", 
           RowBox[{"equalmass", ",", "\[Delta]\[Chi]"}], "]"}], "[", 
          RowBox[{"[", "3", "]"}], "]"}]}], "]"}]}]}], "/.", 
     RowBox[{"u", "->", 
      RowBox[{"2", "/", 
       RowBox[{"Sqrt", "[", "\[Rho]", "]"}]}]}]}], ",", 
    RowBox[{"\[Rho]", ">", "0"}]}], "]"}]}]], "Input",
 CellLabel->"In[89]:=",ExpressionUUID->"b6b00072-3eec-443a-85ec-fd85c261e15f"],

Cell[BoxData[
 FractionBox[
  RowBox[{"4", " ", "\[Pi]", " ", 
   SuperscriptBox["\[Rho]", 
    RowBox[{"13", "/", "4"}]]}], 
  RowBox[{"3", " ", 
   SqrtBox[
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        SqrtBox["\[Rho]"], "-", "\[Chi]eff"}], ")"}], "2"]}]]}]]], "Output",
 CellLabel->
  "Out[89]=",ExpressionUUID->"001113c8-8286-4780-b615-f062bf2eb4b2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["(iv) Spin-orbit resonances", \
"Section",ExpressionUUID->"0d2ab530-1275-44ed-bb47-80cffee14671"],

Cell[BoxData[
 RowBox[{"Clear", "@@", 
  RowBox[{"DeleteCases", "[", 
   RowBox[{
    RowBox[{"Names", "@", "\"\<`*\>\""}], ",", 
    RowBox[{"\"\<ToPython\>\"", "|", "\"\<precessionform\>\""}]}], 
   "]"}]}]], "Input",
 CellLabel->"In[90]:=",ExpressionUUID->"1f2803c8-7368-4ca2-b06d-8d72445b3ebd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cubic", " ", "=", 
  RowBox[{
   RowBox[{"-", 
    FractionBox["1", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "28"]]}], "9216", " ", 
   SuperscriptBox["q", "10"], " ", 
   SuperscriptBox["u", "12"], " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "-", 
      FractionBox[
       RowBox[{"2", " ", "q", " ", "u", " ", "\[Chi]eff"}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"]]}], ")"}], "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "4"}], " ", 
      SuperscriptBox["q", "4"], " ", 
      SuperscriptBox["\[Chi]1", "2"], " ", 
      SuperscriptBox["\[Chi]2", "2"]}], "+", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["\[Chi]1", "2"], "+", 
        RowBox[{
         SuperscriptBox["q", "4"], " ", 
         SuperscriptBox["\[Chi]2", "2"]}], "-", 
        FractionBox[
         RowBox[{
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"1", "+", "q"}], ")"}], "3"], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", "\[Delta]\[Chi]"}],
             "+", 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"1", "+", "q"}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}]}], 
           ")"}]}], 
         RowBox[{"2", " ", "u"}]]}], ")"}], "2"], "+", 
     RowBox[{
      SuperscriptBox["q", "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
      SuperscriptBox["\[Chi]1", "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"\[Delta]\[Chi]", "-", "\[Chi]eff"}], ")"}], "2"]}], "-", 
     RowBox[{"q", " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["\[Chi]1", "2"], "+", 
        RowBox[{
         SuperscriptBox["q", "4"], " ", 
         SuperscriptBox["\[Chi]2", "2"]}], "-", 
        FractionBox[
         RowBox[{
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"1", "+", "q"}], ")"}], "3"], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", "\[Delta]\[Chi]"}],
             "+", 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"1", "+", "q"}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}]}], 
           ")"}]}], 
         RowBox[{"2", " ", "u"}]]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"\[Delta]\[Chi]", "-", "\[Chi]eff"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"\[Delta]\[Chi]", "+", "\[Chi]eff"}], ")"}]}], "+", 
     RowBox[{
      SuperscriptBox["q", "4"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
      SuperscriptBox["\[Chi]2", "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"\[Delta]\[Chi]", "+", "\[Chi]eff"}], ")"}], "2"]}]}], 
    ")"}]}]}]], "Input",
 CellLabel->"In[91]:=",ExpressionUUID->"8000ab20-86ba-4b3a-b3d8-40f556e1590e"],

Cell[BoxData[
 RowBox[{"-", 
  RowBox[{
   FractionBox["1", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], "28"]], 
   RowBox[{"9216", " ", 
    SuperscriptBox["q", "10"], " ", 
    SuperscriptBox["u", "12"], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", "q", " ", "u", " ", "\[Chi]eff"}], 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "2"]]}], ")"}], "2"], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "4"}], " ", 
       SuperscriptBox["q", "4"], " ", 
       SuperscriptBox["\[Chi]1", "2"], " ", 
       SuperscriptBox["\[Chi]2", "2"]}], "+", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox["\[Chi]1", "2"], "+", 
         RowBox[{
          SuperscriptBox["q", "4"], " ", 
          SuperscriptBox["\[Chi]2", "2"]}], "-", 
         FractionBox[
          RowBox[{
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "3"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", 
              "\[Delta]\[Chi]"}], "+", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{"1", "+", "q"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}]}],
             ")"}]}], 
          RowBox[{"2", " ", "u"}]]}], ")"}], "2"], "+", 
      RowBox[{
       SuperscriptBox["q", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
       SuperscriptBox["\[Chi]1", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"\[Delta]\[Chi]", "-", "\[Chi]eff"}], ")"}], "2"]}], "-", 
      RowBox[{"q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox["\[Chi]1", "2"], "+", 
         RowBox[{
          SuperscriptBox["q", "4"], " ", 
          SuperscriptBox["\[Chi]2", "2"]}], "-", 
         FractionBox[
          RowBox[{
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "3"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", 
              "\[Delta]\[Chi]"}], "+", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{"1", "+", "q"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}]}],
             ")"}]}], 
          RowBox[{"2", " ", "u"}]]}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"\[Delta]\[Chi]", "-", "\[Chi]eff"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"\[Delta]\[Chi]", "+", "\[Chi]eff"}], ")"}]}], "+", 
      RowBox[{
       SuperscriptBox["q", "4"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
       SuperscriptBox["\[Chi]2", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"\[Delta]\[Chi]", "+", "\[Chi]eff"}], ")"}], "2"]}]}], 
     ")"}]}]}]}]], "Output",
 CellLabel->
  "Out[91]=",ExpressionUUID->"9a23cc19-9459-4c48-aeff-707c5d89a2fd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"overallcubic", " ", "=", 
  FractionBox[
   RowBox[{"4608", " ", 
    SuperscriptBox["q", "11"], " ", 
    SuperscriptBox["u", "10"], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], "-", 
       RowBox[{"2", " ", "q", " ", "u", " ", "\[Chi]eff"}]}], ")"}], "2"]}], 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], "27"]]}]], "Input",
 CellLabel->"In[92]:=",ExpressionUUID->"c69a891a-6158-4831-b627-98548b2830ed"],

Cell[BoxData[
 FractionBox[
  RowBox[{"4608", " ", 
   SuperscriptBox["q", "11"], " ", 
   SuperscriptBox["u", "10"], " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"], "-", 
      RowBox[{"2", " ", "q", " ", "u", " ", "\[Chi]eff"}]}], ")"}], "2"]}], 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{"1", "+", "q"}], ")"}], "27"]]], "Output",
 CellLabel->
  "Out[92]=",ExpressionUUID->"2f21718f-c4c7-47e2-8406-0a3b3668e80b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"FactorInteger", "[", "4608", "]"}], 
  "\[IndentingNewLine]"}]], "Input",
 CellLabel->"In[93]:=",ExpressionUUID->"266e7a46-8502-429b-954e-a773b682caa9"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"2", ",", "9"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"3", ",", "2"}], "}"}]}], "}"}]], "Output",
 CellLabel->
  "Out[93]=",ExpressionUUID->"0d44eb57-a73a-4b1e-8b26-cf133b32af21"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"discr", " ", "=", " ", 
  RowBox[{"Discriminant", "[", 
   RowBox[{
    RowBox[{"cubic", " ", "/", "overallcubic"}], ",", "\[Delta]\[Chi]"}], 
   "]"}]}]], "Input",
 CellLabel->"In[94]:=",ExpressionUUID->"28c64dc4-b39e-4cc8-82b6-a267e1dd75f2"],

Cell[BoxData[
 TemplateBox[<|"shortenedBoxes" -> TagBox[
     FractionBox[
      RowBox[{"4", " ", 
        SuperscriptBox["u", "2"], " ", 
        RowBox[{"(", 
          RowBox[{
            SuperscriptBox["\[Kappa]", "4"], "+", 
            RowBox[{"18", " ", "q", " ", 
              SuperscriptBox["\[Kappa]", "4"]}], "+", 
            RowBox[{"151", " ", 
              SuperscriptBox["q", "2"], " ", 
              SuperscriptBox["\[Kappa]", "4"]}], "+", 
            RowBox[{"780", " ", 
              SuperscriptBox["q", "3"], " ", 
              SuperscriptBox["\[Kappa]", "4"]}], "+", 
            TemplateBox[{"2966"}, "OutputSizeLimit`Skeleton"], "+", 
            RowBox[{"1456", " ", 
              SuperscriptBox["q", "16"], " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "6"]}], "+", 
            RowBox[{"224", " ", 
              SuperscriptBox["q", "17"], " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "6"]}], "+", 
            RowBox[{"16", " ", 
              SuperscriptBox["q", "18"], " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "6"]}]}], ")"}]}], 
      RowBox[{
        SuperscriptBox["q", "2"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "14"]}]], Short[#, 5]& ], "line" -> 
   94, "sessionID" -> 30211309614211285111, "byteCount" -> 716680, "size" -> 
   5, "stored" -> False, "expr" -> Missing["NotStored"], "wrap" -> 
   OutputSizeLimit`Defer, "version" -> 1|>,
  "OutputSizeLimitTemplate"]], "Output",
 CellLabel->
  "Out[94]=",ExpressionUUID->"ea8da00b-2b57-433a-8bf6-48b3084bb05b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"coeffs", " ", "=", " ", 
  RowBox[{
   RowBox[{"CoefficientList", "[", 
    RowBox[{"discr", ",", "\[Kappa]"}], "]"}], "//", "Simplify"}]}]], "Input",
 
 CellLabel->"In[95]:=",ExpressionUUID->"8f35a1ea-08b4-40c0-b812-b1a2fc53be6d"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", 
    RowBox[{
     FractionBox["1", 
      RowBox[{
       SuperscriptBox["q", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "14"]}]], 
     RowBox[{"4", " ", 
      SuperscriptBox["u", "2"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", "q"}], ")"}], "2"], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "12"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "q"}], ")"}], "2"]}], " ", 
            SuperscriptBox["\[Chi]1", "2"]}], "+", 
           RowBox[{
            SuperscriptBox["q", "2"], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
            SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", "2"}], " ", 
            SuperscriptBox["q", "3"], " ", 
            SuperscriptBox["\[Chi]2", "2"]}], "+", 
           RowBox[{
            SuperscriptBox["q", "4"], " ", 
            SuperscriptBox["\[Chi]2", "2"]}], "-", 
           SuperscriptBox["\[Chi]eff", "2"], "-", 
           RowBox[{"2", " ", "q", " ", 
            SuperscriptBox["\[Chi]eff", "2"]}], "+", 
           RowBox[{
            SuperscriptBox["q", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              SuperscriptBox["\[Chi]2", "2"], "-", 
              SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}]}], ")"}]}], "-", 
        RowBox[{"16", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", "q"}], ")"}], "2"], " ", "q", " ", 
         SuperscriptBox["u", "6"], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox["\[Chi]1", "4"], "-", 
            RowBox[{
             SuperscriptBox["q", "3"], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", "q"}], ")"}], " ", 
             SuperscriptBox["\[Chi]1", "2"], " ", 
             SuperscriptBox["\[Chi]2", "2"]}], "+", 
            RowBox[{
             SuperscriptBox["q", "7"], " ", 
             SuperscriptBox["\[Chi]2", "4"]}]}], ")"}], "2"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", 
              SuperscriptBox["\[Chi]1", "2"]}], ")"}]}], "+", 
           RowBox[{"q", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               SuperscriptBox["q", "3"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              SuperscriptBox["\[Chi]eff", "2"], "+", 
              RowBox[{"2", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}]}], ")"}]}]}], 
          ")"}]}], "-", 
        RowBox[{"8", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", "q", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "3"], " ", 
         SuperscriptBox["u", "5"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            SuperscriptBox["\[Chi]1", "2"]}], "+", 
           RowBox[{
            SuperscriptBox["q", "4"], " ", 
            SuperscriptBox["\[Chi]2", "2"]}]}], ")"}], " ", "\[Chi]eff", " ", 
         
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"(", 
             RowBox[{"5", "-", 
              RowBox[{"13", " ", "q"}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["q", "2"]}]}], ")"}], " ", 
            SuperscriptBox["\[Chi]1", "6"]}], "+", 
           RowBox[{
            SuperscriptBox["q", "8"], " ", 
            SuperscriptBox["\[Chi]2", "4"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"8", " ", 
               SuperscriptBox["q", "2"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              RowBox[{"5", " ", 
               SuperscriptBox["q", "4"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "-", 
              RowBox[{"8", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "-", 
              RowBox[{"12", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "13"}], " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{
            SuperscriptBox["q", "4"], " ", 
            SuperscriptBox["\[Chi]1", "2"], " ", 
            SuperscriptBox["\[Chi]2", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", 
                SuperscriptBox["q", "5"]}], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{"8", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "-", 
              RowBox[{"2", " ", 
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], "-", 
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], "-", 
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"7", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "-", 
           RowBox[{
            SuperscriptBox["\[Chi]1", "4"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", 
               SuperscriptBox["q", "5"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "6"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "-", 
              RowBox[{"4", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "7"}], " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"8", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], "+", 
                 RowBox[{"12", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}]}], 
          ")"}]}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "11"], " ", "u", " ", "\[Chi]eff", 
         " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "q"}], ")"}], "4"], " ", 
            SuperscriptBox["\[Chi]1", "4"]}], "-", 
           RowBox[{
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "q"}], ")"}], "2"], " ", "q", " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], " ", 
            SuperscriptBox["\[Chi]1", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "10"}], " ", 
               SuperscriptBox["q", "3"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              RowBox[{"5", " ", 
               SuperscriptBox["q", "4"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "-", 
              RowBox[{"5", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "-", 
              RowBox[{"8", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"5", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "-", 
                 RowBox[{"3", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{
            SuperscriptBox["q", "3"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               SuperscriptBox["q", "8"], " ", 
               SuperscriptBox["\[Chi]2", "4"]}], "-", 
              RowBox[{"4", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "-", 
              RowBox[{"20", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "+", 
              RowBox[{"5", " ", 
               SuperscriptBox["q", "3"], " ", 
               SuperscriptBox["\[Chi]eff", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], "-", 
                 RowBox[{"8", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{"3", " ", 
               SuperscriptBox["q", "6"], " ", 
               SuperscriptBox["\[Chi]2", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "7"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "4"}], " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{"5", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"3", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"40", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "4"], "-", 
                 RowBox[{"6", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"20", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "-", 
              RowBox[{"2", " ", 
               SuperscriptBox["q", "5"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{"5", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}]}], ")"}]}]}], 
          ")"}]}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "7"], " ", 
         SuperscriptBox["u", "3"], " ", "\[Chi]eff", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "q"}], ")"}], "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "+", 
              RowBox[{"25", " ", "q"}], "+", 
              RowBox[{"12", " ", 
               SuperscriptBox["q", "2"]}]}], ")"}], " ", 
            SuperscriptBox["\[Chi]1", "6"]}], "+", 
           RowBox[{"q", " ", 
            SuperscriptBox["\[Chi]1", "4"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"85", " ", 
               SuperscriptBox["q", "6"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "-", 
              RowBox[{"32", " ", 
               SuperscriptBox["q", "7"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "-", 
              RowBox[{"4", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{"48", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"83", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "-", 
                 RowBox[{"40", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], "+", 
                 RowBox[{"7", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "5"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "103"}], " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"20", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
              RowBox[{
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"37", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"84", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{
            SuperscriptBox["q", "3"], " ", 
            SuperscriptBox["\[Chi]1", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "37"}], " ", 
               SuperscriptBox["q", "8"], " ", 
               SuperscriptBox["\[Chi]2", "4"]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "9"], " ", 
               SuperscriptBox["\[Chi]2", "4"]}], "+", 
              RowBox[{"16", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "+", 
              RowBox[{"32", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "+", 
              RowBox[{"16", " ", 
               SuperscriptBox["q", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], "-", 
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "7"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"83", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{"16", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "6"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "103"}], " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{"24", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "5"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"85", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "-", 
                 RowBox[{"8", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"32", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"3", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"16", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "-", 
              RowBox[{"8", " ", 
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"14", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{
            SuperscriptBox["q", "7"], " ", 
            SuperscriptBox["\[Chi]2", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", 
                SuperscriptBox["q", "8"]}], " ", 
               SuperscriptBox["\[Chi]2", "4"]}], "-", 
              RowBox[{"32", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "-", 
              RowBox[{"112", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "7"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"27", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "-", 
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "6"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "39"}], " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{"48", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"5", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"32", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "-", 
              RowBox[{"8", " ", 
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"5", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"3", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "-", 
                 RowBox[{"21", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"8", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "5"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "4"], "+", 
                 RowBox[{"28", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"16", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}]}], ")"}]}]}], 
          ")"}]}], "+", 
        RowBox[{
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "4"], " ", 
         SuperscriptBox["u", "4"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "q"}], ")"}], "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "+", 
              RowBox[{"20", " ", "q"}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["q", "2"]}]}], ")"}], " ", 
            SuperscriptBox["\[Chi]1", "8"]}], "-", 
           RowBox[{"4", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", "q", " ", 
            SuperscriptBox["\[Chi]1", "6"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "11"}], " ", 
               SuperscriptBox["q", "5"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["q", "6"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "-", 
              RowBox[{"8", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{"20", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"30", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "-", 
                 RowBox[{"22", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
              RowBox[{"8", " ", 
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"5", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"42", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{
            SuperscriptBox["q", "10"], " ", 
            SuperscriptBox["\[Chi]2", "4"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", 
                SuperscriptBox["q", "8"]}], " ", 
               SuperscriptBox["\[Chi]2", "4"]}], "-", 
              RowBox[{"96", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "-", 
              RowBox[{"288", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "7"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"22", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "-", 
                 RowBox[{"32", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"11", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"26", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "-", 
              RowBox[{"8", " ", 
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"7", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"16", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "6"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "33"}], " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{"112", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"16", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "5"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "4"], "+", 
                 RowBox[{"22", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"8", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "4"], "-", 
                 RowBox[{"25", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"24", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{"4", " ", 
            SuperscriptBox["q", "6"], " ", 
            SuperscriptBox["\[Chi]1", "2"], " ", 
            SuperscriptBox["\[Chi]2", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"5", " ", 
               SuperscriptBox["q", "9"], " ", 
               SuperscriptBox["\[Chi]2", "4"]}], "+", 
              RowBox[{"24", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "+", 
              RowBox[{"56", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "+", 
              RowBox[{"12", " ", 
               SuperscriptBox["q", "3"], " ", 
               SuperscriptBox["\[Chi]eff", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], "-", 
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["q", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "2"}], " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "7"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"62", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "-", 
                 RowBox[{"6", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "8"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "37"}], " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "8"}], " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{"52", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"8", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "6"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "41"}], " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "-", 
                 RowBox[{"38", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"24", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "5"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"19", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "-", 
                 RowBox[{"6", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"56", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{"2", " ", 
            SuperscriptBox["q", "2"], " ", 
            SuperscriptBox["\[Chi]1", "4"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", 
               SuperscriptBox["q", "9"], " ", 
               SuperscriptBox["\[Chi]2", "4"]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "10"], " ", 
               SuperscriptBox["\[Chi]2", "4"]}], "-", 
              RowBox[{"8", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "+", 
              RowBox[{"16", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], "+", 
                 RowBox[{"24", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "8"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"53", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "-", 
                 RowBox[{"32", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "7"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "110"}], " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{"24", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "6"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"53", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{"104", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"48", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "4"], "-", 
                 RowBox[{"19", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"26", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "12"}], " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"64", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "-", 
              RowBox[{"2", " ", 
               SuperscriptBox["q", "5"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "4"], "+", 
                 RowBox[{"6", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"72", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}]}], ")"}]}]}], 
          ")"}]}], "+", 
        RowBox[{
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "8"], " ", 
         SuperscriptBox["u", "2"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "q"}], ")"}], "4"], " ", 
            SuperscriptBox["\[Chi]1", "6"]}], "-", 
           RowBox[{
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "q"}], ")"}], "2"], " ", 
            SuperscriptBox["\[Chi]1", "4"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"4", " ", 
               SuperscriptBox["q", "5"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              RowBox[{"10", " ", 
               SuperscriptBox["q", "6"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              SuperscriptBox["\[Chi]eff", "2"], "-", 
              RowBox[{"38", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"26", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "-", 
                 RowBox[{"86", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
              RowBox[{
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"39", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"23", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
              RowBox[{
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], "+", 
                 RowBox[{"102", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{
            SuperscriptBox["q", "2"], " ", 
            SuperscriptBox["\[Chi]1", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "28"}], " ", 
               SuperscriptBox["q", "9"], " ", 
               SuperscriptBox["\[Chi]2", "4"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "10"], " ", 
               SuperscriptBox["\[Chi]2", "4"]}], "+", 
              RowBox[{"32", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "+", 
              RowBox[{"72", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "+", 
              RowBox[{"48", " ", 
               SuperscriptBox["q", "3"], " ", 
               SuperscriptBox["\[Chi]eff", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], "-", 
                 RowBox[{"5", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "8"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"92", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "-", 
                 RowBox[{"22", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
              RowBox[{"12", " ", 
               SuperscriptBox["q", "7"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"9", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "-", 
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["q", "5"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "-", 
                 RowBox[{"12", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"11", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "6"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"37", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{"22", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"8", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "-", 
              RowBox[{"2", " ", 
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"11", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"20", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "-", 
              RowBox[{"2", " ", 
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"5", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "-", 
                 RowBox[{"11", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"120", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{
            SuperscriptBox["q", "4"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "16"}], " ", 
               SuperscriptBox["\[Chi]eff", "6"]}], "-", 
              RowBox[{"96", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "6"]}], "-", 
              RowBox[{"8", " ", 
               SuperscriptBox["q", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], "+", 
                 RowBox[{"30", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "9"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "6"], "-", 
                 RowBox[{"10", " ", 
                  SuperscriptBox["\[Chi]2", "4"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "10"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "6"], "-", 
                 RowBox[{
                  SuperscriptBox["\[Chi]2", "4"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "7"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "6"], "+", 
                 RowBox[{"20", " ", 
                  SuperscriptBox["\[Chi]2", "4"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"18", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "8"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"6", " ", 
                  SuperscriptBox["\[Chi]2", "6"]}], "+", 
                 RowBox[{"25", " ", 
                  SuperscriptBox["\[Chi]2", "4"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"32", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"23", " ", 
                  SuperscriptBox["\[Chi]2", "4"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"240", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}], "-", 
                 RowBox[{"240", " ", 
                  SuperscriptBox["\[Chi]eff", "6"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "6"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "6"], "-", 
                 RowBox[{"47", " ", 
                  SuperscriptBox["\[Chi]2", "4"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"40", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}], "-", 
                 RowBox[{"16", " ", 
                  SuperscriptBox["\[Chi]eff", "6"]}]}], ")"}]}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["q", "5"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"5", " ", 
                  SuperscriptBox["\[Chi]2", "4"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"30", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}], "-", 
                 RowBox[{"12", " ", 
                  SuperscriptBox["\[Chi]eff", "6"]}]}], ")"}]}], "-", 
              RowBox[{"8", " ", 
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"11", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}], "+", 
                 RowBox[{"40", " ", 
                  SuperscriptBox["\[Chi]eff", "6"]}]}], ")"}]}]}], ")"}]}]}], 
          ")"}]}]}], ")"}]}]}]}], ",", 
   RowBox[{"-", 
    RowBox[{
     FractionBox["1", 
      RowBox[{
       SuperscriptBox["q", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "10"]}]], 
     RowBox[{"8", " ", 
      SuperscriptBox["u", "2"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", "q"}], ")"}], "2"]}], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "11"], " ", "\[Chi]eff", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "q"}], ")"}], "2"], " ", 
            SuperscriptBox["\[Chi]1", "2"]}], "+", 
           RowBox[{"q", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               SuperscriptBox["q", "4"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "-", 
              SuperscriptBox["\[Chi]eff", "2"], "-", 
              RowBox[{"3", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], "-", 
                 RowBox[{"3", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
              RowBox[{
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}]}], ")"}]}]}], 
          ")"}]}], "+", 
        RowBox[{"8", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", "q", " ", 
         SuperscriptBox["u", "5"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            SuperscriptBox["\[Chi]1", "2"]}], "+", 
           RowBox[{
            SuperscriptBox["q", "3"], " ", 
            SuperscriptBox["\[Chi]2", "2"]}]}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"(", 
             RowBox[{"5", "-", 
              RowBox[{"7", " ", "q"}], "+", 
              RowBox[{"2", " ", 
               SuperscriptBox["q", "2"]}]}], ")"}], " ", 
            SuperscriptBox["\[Chi]1", "6"]}], "+", 
           RowBox[{"q", " ", 
            SuperscriptBox["\[Chi]1", "4"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "7"}], " ", 
               SuperscriptBox["q", "2"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "-", 
              RowBox[{"2", " ", 
               SuperscriptBox["q", "4"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              RowBox[{"2", " ", 
               SuperscriptBox["q", "5"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{"6", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"7", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "-", 
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "-", 
           RowBox[{
            SuperscriptBox["q", "4"], " ", 
            SuperscriptBox["\[Chi]1", "2"], " ", 
            SuperscriptBox["\[Chi]2", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"7", " ", 
               SuperscriptBox["q", "5"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              RowBox[{"2", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{"4", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "-", 
              RowBox[{"2", " ", 
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], "-", 
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "7"}], " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{"2", " ", 
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], "+", 
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{
            SuperscriptBox["q", "8"], " ", 
            SuperscriptBox["\[Chi]2", "4"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"5", " ", 
               SuperscriptBox["q", "4"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "-", 
              RowBox[{"2", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{"2", " ", 
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], "+", 
                 RowBox[{"3", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "7"}], " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}]}], 
          ")"}]}], "+", 
        RowBox[{"4", " ", "q", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "3"], " ", 
         SuperscriptBox["u", "4"], " ", "\[Chi]eff", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"(", 
             RowBox[{"20", "-", 
              RowBox[{"49", " ", "q"}], "+", 
              RowBox[{"41", " ", 
               SuperscriptBox["q", "2"]}], "-", 
              RowBox[{"12", " ", 
               SuperscriptBox["q", "3"]}]}], ")"}], " ", 
            SuperscriptBox["\[Chi]1", "6"]}], "+", 
           RowBox[{
            SuperscriptBox["q", "4"], " ", 
            SuperscriptBox["\[Chi]1", "2"], " ", 
            SuperscriptBox["\[Chi]2", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "3"}], " ", 
               SuperscriptBox["q", "6"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{"4", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"22", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "-", 
                 RowBox[{"28", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "14"}], " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"7", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "5"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"3", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"8", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{
            SuperscriptBox["q", "8"], " ", 
            SuperscriptBox["\[Chi]2", "4"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"20", " ", 
               SuperscriptBox["q", "5"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              RowBox[{"12", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{"4", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "-", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"3", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"41", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "49"}], " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"12", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{"q", " ", 
            SuperscriptBox["\[Chi]1", "4"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"22", " ", 
               SuperscriptBox["q", "5"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "-", 
              RowBox[{"8", " ", 
               SuperscriptBox["q", "6"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              RowBox[{"12", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{"4", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "-", 
              RowBox[{"2", " ", 
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"7", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "-", 
                 RowBox[{"6", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"3", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
              RowBox[{
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"3", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"16", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}]}], 
          ")"}]}], "-", 
        RowBox[{
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "8"], " ", "u", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "q"}], ")"}], "4"], " ", 
            SuperscriptBox["\[Chi]1", "4"]}], "+", 
           RowBox[{
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "q"}], ")"}], "2"], " ", 
            SuperscriptBox["\[Chi]1", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "14"}], " ", 
               SuperscriptBox["q", "5"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "6"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "-", 
              SuperscriptBox["\[Chi]eff", "2"], "+", 
              RowBox[{"16", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"26", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"7", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "14"}], " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"32", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], "+", 
                 RowBox[{"42", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{
            SuperscriptBox["q", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "8"}], " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "-", 
              RowBox[{"56", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "8"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "4"], "-", 
                 RowBox[{
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "7"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "4"}], " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{"18", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "4"], "-", 
                 RowBox[{"15", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"152", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"7", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"152", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{"2", " ", 
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"9", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"104", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "6"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"6", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{"9", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"8", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "-", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "5"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "4"], "+", 
                 RowBox[{"9", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"14", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}]}], ")"}]}]}], 
          ")"}]}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "4"], " ", 
         SuperscriptBox["u", "3"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "q"}], ")"}], "2"]}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "+", 
              RowBox[{"15", " ", "q"}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "2"]}]}], ")"}], " ", 
            SuperscriptBox["\[Chi]1", "6"]}], "-", 
           RowBox[{"q", " ", 
            SuperscriptBox["\[Chi]1", "4"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"10", " ", 
               SuperscriptBox["q", "6"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "7"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "-", 
              RowBox[{"24", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{"16", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"143", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "-", 
                 RowBox[{"32", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
              RowBox[{"5", " ", 
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"17", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"12", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "5"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "87"}], " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"20", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"15", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"32", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{
            SuperscriptBox["q", "2"], " ", 
            SuperscriptBox["\[Chi]1", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "15"}], " ", 
               SuperscriptBox["q", "9"], " ", 
               SuperscriptBox["\[Chi]2", "4"]}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "+", 
              RowBox[{"24", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "8"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"85", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "-", 
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "7"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "143"}], " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{"24", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
              RowBox[{"2", " ", 
               SuperscriptBox["q", "5"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"5", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{"48", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"44", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "-", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "4"], "+", 
                 RowBox[{"5", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"30", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"3", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"10", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "6"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"87", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "-", 
                 RowBox[{"20", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"24", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "4"}], " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"40", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{
            SuperscriptBox["q", "6"], " ", 
            SuperscriptBox["\[Chi]2", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               SuperscriptBox["q", "8"], " ", 
               SuperscriptBox["\[Chi]2", "4"]}], "+", 
              RowBox[{"24", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "+", 
              RowBox[{"88", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "-", 
              RowBox[{"20", " ", 
               SuperscriptBox["q", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], "-", 
                 RowBox[{"6", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "7"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "17"}], " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{"24", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{"16", " ", 
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"5", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "6"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"27", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "-", 
                 RowBox[{"16", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"8", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "5"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "7"}], " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "-", 
                 RowBox[{"32", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"24", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "4"}], " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{"60", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"40", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}]}], ")"}]}]}], 
          ")"}]}], "+", 
        RowBox[{
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "7"], " ", 
         SuperscriptBox["u", "2"], " ", "\[Chi]eff", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "q"}], ")"}], "2"]}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "3"}], "+", 
              RowBox[{"49", " ", "q"}], "+", 
              RowBox[{"16", " ", 
               SuperscriptBox["q", "2"]}]}], ")"}], " ", 
            SuperscriptBox["\[Chi]1", "4"]}], "-", 
           RowBox[{"2", " ", "q", " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], " ", 
            SuperscriptBox["\[Chi]1", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"22", " ", 
               SuperscriptBox["q", "4"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "-", 
              RowBox[{"15", " ", 
               SuperscriptBox["q", "5"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "6"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "-", 
              RowBox[{"4", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{"6", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], "-", 
                 RowBox[{"7", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
              RowBox[{
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"15", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"38", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{
            SuperscriptBox["q", "3"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"3", " ", 
               SuperscriptBox["q", "8"], " ", 
               SuperscriptBox["\[Chi]2", "4"]}], "+", 
              RowBox[{"16", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "+", 
              RowBox[{"80", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "+", 
              RowBox[{"160", " ", 
               SuperscriptBox["q", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "4"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "6"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"85", " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "-", 
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "7"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "55"}], " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{"8", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "5"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "17"}], " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{"44", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"16", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"19", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"40", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "16"}], " ", 
                  SuperscriptBox["\[Chi]2", "4"]}], "+", 
                 RowBox[{"132", " ", 
                  SuperscriptBox["\[Chi]2", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"80", " ", 
                  SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}]}], ")"}]}]}], 
          ")"}]}]}], ")"}]}]}]}], ",", 
   RowBox[{
    FractionBox["1", 
     RowBox[{
      SuperscriptBox["q", "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "6"]}]], 
    RowBox[{"4", " ", 
     SuperscriptBox["u", "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "16"}], " ", "q", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "10"}], "+", 
          RowBox[{"18", " ", "q"}], "-", 
          RowBox[{"9", " ", 
           SuperscriptBox["q", "2"]}], "+", 
          SuperscriptBox["q", "3"]}], ")"}], " ", 
        SuperscriptBox["u", "4"], " ", 
        SuperscriptBox["\[Chi]1", "6"]}], "+", 
       SuperscriptBox["\[Chi]eff", "2"], "+", 
       RowBox[{"4", " ", "q", " ", 
        SuperscriptBox["\[Chi]eff", "2"], " ", 
        RowBox[{"(", 
         RowBox[{"3", "+", 
          RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox["q", "14"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"6", " ", 
           SuperscriptBox["u", "2"], " ", 
           SuperscriptBox["\[Chi]2", "4"]}], "+", 
          SuperscriptBox["\[Chi]eff", "2"], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]2", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             RowBox[{"6", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], ")"}]}], 
       "+", 
       RowBox[{
        SuperscriptBox["q", "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["\[Chi]2", "2"]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"59", "+", 
             RowBox[{"144", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"16", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{"4", " ", 
        SuperscriptBox["q", "13"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"40", " ", 
           SuperscriptBox["u", "4"], " ", 
           SuperscriptBox["\[Chi]2", "6"]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"3", "+", 
             RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
          RowBox[{"12", " ", 
           SuperscriptBox["u", "2"], " ", 
           SuperscriptBox["\[Chi]2", "4"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             RowBox[{"5", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]2", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "-", 
             RowBox[{"4", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"24", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], "-", 
       RowBox[{"4", " ", 
        SuperscriptBox["q", "3"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           SuperscriptBox["\[Chi]2", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"1", "+", 
             RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
          RowBox[{"2", " ", 
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"19", "+", 
             RowBox[{"126", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"24", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox["q", "4"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "2"}], " ", 
           SuperscriptBox["\[Chi]2", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"1", "+", 
             RowBox[{"43", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"4", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"201", "+", 
             RowBox[{"3920", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"976", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{"4", " ", 
        SuperscriptBox["q", "5"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"13", "+", 
             RowBox[{"2430", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"704", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]2", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"3", "-", 
             RowBox[{"72", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"10", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}], "-", 
       RowBox[{"4", " ", 
        SuperscriptBox["q", "7"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           SuperscriptBox["u", "2"], " ", 
           SuperscriptBox["\[Chi]2", "4"], " ", 
           RowBox[{"(", 
            RowBox[{"19", "+", 
             RowBox[{"8", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
          RowBox[{"4", " ", 
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "27"}], "+", 
             RowBox[{"1218", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"392", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
          RowBox[{"2", " ", 
           SuperscriptBox["\[Chi]2", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             RowBox[{"20", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"169", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"32", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox["q", "12"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "288"}], " ", 
           SuperscriptBox["u", "4"], " ", 
           SuperscriptBox["\[Chi]2", "6"]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"59", "+", 
             RowBox[{"144", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"16", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["u", "2"], " ", 
           SuperscriptBox["\[Chi]2", "4"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "67"}], "+", 
             RowBox[{"204", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"48", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["\[Chi]2", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "-", 
             RowBox[{"95", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"296", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"48", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{"4", " ", 
        SuperscriptBox["q", "9"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", 
           SuperscriptBox["u", "2"], " ", 
           SuperscriptBox["\[Chi]2", "4"], " ", 
           RowBox[{"(", 
            RowBox[{"13", "+", 
             RowBox[{"6", " ", "u", " ", "\[Chi]eff"}], "-", 
             RowBox[{"8", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"13", "+", 
             RowBox[{"2430", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"704", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["\[Chi]2", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             RowBox[{"70", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"379", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"100", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{"4", " ", 
        SuperscriptBox["q", "11"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"36", " ", 
           SuperscriptBox["u", "4"], " ", 
           SuperscriptBox["\[Chi]2", "6"]}], "+", 
          RowBox[{
           SuperscriptBox["u", "2"], " ", 
           SuperscriptBox["\[Chi]2", "4"], " ", 
           RowBox[{"(", 
            RowBox[{"5", "-", 
             RowBox[{"16", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"24", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"19", "+", 
             RowBox[{"126", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"24", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]2", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"3", "-", 
             RowBox[{"102", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"406", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"112", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox["q", "8"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", 
           SuperscriptBox["u", "2"], " ", 
           SuperscriptBox["\[Chi]2", "4"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "53"}], "+", 
             RowBox[{"28", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "261"}], "+", 
             RowBox[{"16416", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"5152", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{"4", " ", 
           SuperscriptBox["\[Chi]2", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "7"}], "+", 
             RowBox[{"189", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"614", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"120", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox["q", "6"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "8"}], " ", 
           SuperscriptBox["u", "2"], " ", 
           SuperscriptBox["\[Chi]2", "4"]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "261"}], "+", 
             RowBox[{"16416", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"5152", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]2", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"17", "-", 
             RowBox[{"338", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"424", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"128", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox["q", "10"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "16"}], " ", 
           SuperscriptBox["u", "4"], " ", 
           SuperscriptBox["\[Chi]2", "6"]}], "-", 
          RowBox[{"2", " ", 
           SuperscriptBox["u", "2"], " ", 
           SuperscriptBox["\[Chi]2", "4"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "121"}], "+", 
             RowBox[{"136", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"40", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"201", "+", 
             RowBox[{"3920", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"976", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]2", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"17", "-", 
             RowBox[{"148", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"2680", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"832", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox["u", "2"], " ", 
        SuperscriptBox["\[Chi]1", "4"], " ", 
        RowBox[{"(", 
         RowBox[{"3", "-", 
          RowBox[{"4", " ", 
           SuperscriptBox["q", "8"]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["q", "7"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "19"}], "+", 
             RowBox[{"36", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]2", "2"]}], "-", 
             RowBox[{"8", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
          RowBox[{"24", " ", "q", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             RowBox[{"5", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["q", "3"], " ", 
           RowBox[{"(", 
            RowBox[{"5", "-", 
             RowBox[{"16", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"24", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["q", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "67"}], "+", 
             RowBox[{"204", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"48", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{"4", " ", 
           SuperscriptBox["q", "5"], " ", 
           RowBox[{"(", 
            RowBox[{"13", "+", 
             RowBox[{"6", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"9", " ", 
                 SuperscriptBox["\[Chi]2", "2"]}], "-", 
                SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["q", "6"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "53"}], "+", 
             RowBox[{"28", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "27"}], " ", 
                 SuperscriptBox["\[Chi]2", "2"]}], "+", 
                SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}]}], ")"}]}], "-", 
          RowBox[{
           SuperscriptBox["q", "4"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "121"}], "+", 
             RowBox[{"136", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"18", " ", 
                 SuperscriptBox["\[Chi]2", "2"]}], "+", 
                RowBox[{"5", " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}]}], 
         ")"}]}], "-", 
       RowBox[{
        SuperscriptBox["\[Chi]1", "2"], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", 
          SuperscriptBox["q", "12"], "-", 
          RowBox[{"6", " ", "u", " ", "\[Chi]eff"}], "+", 
          RowBox[{
           SuperscriptBox["q", "11"], " ", 
           RowBox[{"(", 
            RowBox[{"4", "+", 
             RowBox[{"60", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]2", "2"]}], "+", 
             RowBox[{"8", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
          RowBox[{"q", " ", 
           RowBox[{"(", 
            RowBox[{"4", "+", 
             RowBox[{"16", " ", "u", " ", "\[Chi]eff"}], "-", 
             RowBox[{"96", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["q", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"2", "+", 
             RowBox[{"190", " ", "u", " ", "\[Chi]eff"}], "-", 
             RowBox[{"592", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "-", 
             RowBox[{"96", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["q", "10"], " ", 
           RowBox[{"(", 
            RowBox[{"2", "+", 
             RowBox[{"288", " ", 
              SuperscriptBox["u", "4"], " ", 
              SuperscriptBox["\[Chi]2", "4"]}], "+", 
             RowBox[{"86", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"24", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]2", "2"], " ", "\[Chi]eff"}], "+", 
             RowBox[{"4", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox["\[Chi]2", "2"], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "-", 
          RowBox[{"4", " ", 
           SuperscriptBox["q", "3"], " ", 
           RowBox[{"(", 
            RowBox[{"3", "-", 
             RowBox[{"102", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"112", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}], "+", 
             RowBox[{
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "15"}], " ", 
                 SuperscriptBox["\[Chi]2", "2"]}], "+", 
                RowBox[{"406", " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "-", 
          RowBox[{"4", " ", 
           SuperscriptBox["q", "9"], " ", 
           RowBox[{"(", 
            RowBox[{"3", "-", 
             RowBox[{"72", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["u", "3"], " ", "\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "3"}], " ", 
                 SuperscriptBox["\[Chi]2", "2"]}], "+", 
                SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}], "+", 
             RowBox[{"2", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"29", " ", 
                 SuperscriptBox["\[Chi]2", "2"]}], "+", 
                RowBox[{"5", " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"24", " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"6", " ", 
                 SuperscriptBox["\[Chi]2", "4"]}], "-", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{
           SuperscriptBox["q", "4"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "17"}], "+", 
             RowBox[{"148", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"4", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox["\[Chi]2", "2"], "-", 
                RowBox[{"670", " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["u", "3"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"3", " ", 
                 SuperscriptBox["\[Chi]2", "2"], " ", "\[Chi]eff"}], "-", 
                RowBox[{"104", " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{"8", " ", 
           SuperscriptBox["q", "5"], " ", 
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"70", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"12", " ", 
              SuperscriptBox["u", "4"], " ", 
              SuperscriptBox["\[Chi]2", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "-", 
             RowBox[{
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"29", " ", 
                 SuperscriptBox["\[Chi]2", "2"]}], "+", 
                RowBox[{"379", " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"4", " ", 
              SuperscriptBox["u", "3"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"3", " ", 
                 SuperscriptBox["\[Chi]2", "2"], " ", "\[Chi]eff"}], "-", 
                RowBox[{"25", " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{"4", " ", 
           SuperscriptBox["q", "6"], " ", 
           RowBox[{"(", 
            RowBox[{"7", "-", 
             RowBox[{"189", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["u", "4"], " ", 
              SuperscriptBox["\[Chi]2", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "-", 
             RowBox[{
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox["\[Chi]2", "2"], "+", 
                RowBox[{"614", " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"6", " ", 
              SuperscriptBox["u", "3"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"7", " ", 
                 SuperscriptBox["\[Chi]2", "2"], " ", "\[Chi]eff"}], "-", 
                RowBox[{"20", " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{
           SuperscriptBox["q", "8"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "17"}], "+", 
             RowBox[{"338", " ", "u", " ", "\[Chi]eff"}], "-", 
             RowBox[{"4", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                SuperscriptBox["\[Chi]2", "2"], "+", 
                RowBox[{"106", " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"16", " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"27", " ", 
                 SuperscriptBox["\[Chi]2", "4"]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox["\[Chi]2", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["u", "3"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"21", " ", 
                 SuperscriptBox["\[Chi]2", "2"], " ", "\[Chi]eff"}], "-", 
                RowBox[{"16", " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}], 
          "-", 
          RowBox[{"8", " ", 
           SuperscriptBox["q", "7"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             RowBox[{"20", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "43"}], " ", 
                 SuperscriptBox["\[Chi]2", "2"]}], "+", 
                RowBox[{"169", " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"2", " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"9", " ", 
                 SuperscriptBox["\[Chi]2", "4"]}], "+", 
                RowBox[{"8", " ", 
                 SuperscriptBox["\[Chi]2", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
             RowBox[{"8", " ", 
              SuperscriptBox["u", "3"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"3", " ", 
                 SuperscriptBox["\[Chi]2", "2"], " ", "\[Chi]eff"}], "-", 
                RowBox[{"4", " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}]}], 
         ")"}]}]}], ")"}]}]}], ",", 
   RowBox[{"-", 
    RowBox[{
     FractionBox["1", 
      RowBox[{
       SuperscriptBox["q", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"]}]], 
     RowBox[{"8", " ", 
      SuperscriptBox["u", "2"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", "q"}], ")"}], "2"], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "8"], " ", "\[Chi]eff"}], "+", 
        RowBox[{"8", " ", "q", " ", 
         SuperscriptBox["u", "3"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"(", 
             RowBox[{"10", "-", 
              RowBox[{"12", " ", "q"}], "+", 
              RowBox[{"3", " ", 
               SuperscriptBox["q", "2"]}]}], ")"}], " ", 
            SuperscriptBox["\[Chi]1", "4"]}], "-", 
           RowBox[{"2", " ", "q", " ", 
            SuperscriptBox["\[Chi]1", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"6", " ", 
               SuperscriptBox["q", "2"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              RowBox[{"6", " ", 
               SuperscriptBox["q", "4"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "-", 
              RowBox[{"2", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "-", 
              RowBox[{"3", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "11"}], " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}]}], ")"}]}], "+", 
           
           RowBox[{
            SuperscriptBox["q", "4"], " ", 
            SuperscriptBox["\[Chi]2", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"10", " ", 
               SuperscriptBox["q", "4"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "-", 
              RowBox[{"2", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "3"}], " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}], "+", 
              RowBox[{"3", " ", 
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], "+", 
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}]}], 
          ")"}]}], "+", 
        RowBox[{"4", " ", "q", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "3"], " ", 
         SuperscriptBox["u", "2"], " ", "\[Chi]eff", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "20"}], "+", 
                RowBox[{"3", " ", "q"}], "+", 
                SuperscriptBox["q", "2"]}], ")"}], " ", 
              SuperscriptBox["\[Chi]1", "2"]}], ")"}]}], "+", 
           RowBox[{"q", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"20", " ", 
               SuperscriptBox["q", "4"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{"12", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "-", 
              RowBox[{
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["\[Chi]2", "2"], "-", 
                 RowBox[{"12", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "3"}], " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}]}], 
          ")"}]}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "4"], " ", "u", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "q"}], ")"}], "2"]}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "+", 
              RowBox[{"5", " ", "q"}]}], ")"}], " ", 
            SuperscriptBox["\[Chi]1", "2"]}], "+", 
           RowBox[{"q", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               SuperscriptBox["q", "5"], " ", 
               SuperscriptBox["\[Chi]2", "2"]}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{"40", " ", "q", " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{
               SuperscriptBox["q", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "7"}], " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"8", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"11", " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"40", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["q", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "5"}], " ", 
                  SuperscriptBox["\[Chi]2", "2"]}], "+", 
                 RowBox[{"64", " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}]}], 
          ")"}]}]}], ")"}]}]}]}], ",", 
   RowBox[{
    FractionBox["1", 
     SuperscriptBox["q", "2"]], 
    RowBox[{"4", " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
     SuperscriptBox["u", "2"], " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", 
       SuperscriptBox["q", "6"], "+", 
       RowBox[{"q", " ", 
        RowBox[{"(", 
         RowBox[{"2", "+", 
          RowBox[{"80", " ", 
           SuperscriptBox["u", "2"], " ", 
           SuperscriptBox["\[Chi]1", "2"]}], "+", 
          RowBox[{"40", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox["q", "5"], " ", 
        RowBox[{"(", 
         RowBox[{"2", "+", 
          RowBox[{"80", " ", 
           SuperscriptBox["u", "2"], " ", 
           SuperscriptBox["\[Chi]2", "2"]}], "+", 
          RowBox[{"40", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
       RowBox[{"4", " ", 
        SuperscriptBox["q", "3"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", 
          RowBox[{"60", " ", "u", " ", "\[Chi]eff"}], "+", 
          RowBox[{"8", " ", 
           SuperscriptBox["u", "2"], " ", 
           SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox["q", "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", 
          RowBox[{"160", " ", "u", " ", "\[Chi]eff"}], "+", 
          RowBox[{"16", " ", 
           SuperscriptBox["u", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "3"}], " ", 
              SuperscriptBox["\[Chi]1", "2"]}], "+", 
             SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox["q", "4"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", 
          RowBox[{"160", " ", "u", " ", "\[Chi]eff"}], "+", 
          RowBox[{"16", " ", 
           SuperscriptBox["u", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "3"}], " ", 
              SuperscriptBox["\[Chi]2", "2"]}], "+", 
             SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}]}], ")"}]}]}], 
      ")"}]}]}], ",", 
   RowBox[{"-", 
    FractionBox[
     RowBox[{"64", " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "6"], " ", 
      SuperscriptBox["u", "3"]}], "q"]}]}], "}"}]], "Output",
 CellLabel->
  "Out[95]=",ExpressionUUID->"9f4a800c-cd2b-4e0b-b234-13827f09e141"]
}, Open  ]],

Cell["\<\
Some simplification by hand to get a nicer form . Define the following  \
shortcuts.\
\>", "Text",ExpressionUUID->"4f7012f1-bd8b-4244-8057-a56c807789ba"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[Chi]tr", " ", "=", " ", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"t\[Chi]1", "->", 
     RowBox[{
      SuperscriptBox["\[Chi]1", "2"], "+", 
      RowBox[{
       SuperscriptBox["q", "4"], " ", 
       SuperscriptBox["\[Chi]2", "2"]}]}]}], ",", 
    RowBox[{"t\[Chi]2", "->", " ", 
     RowBox[{
      SuperscriptBox["\[Chi]1", "2"], " ", 
      SuperscriptBox["\[Chi]2", "2"]}]}], ",", 
    RowBox[{"t\[Chi]3", "->", 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox["\[Chi]1", "2"], " ", "+", 
       RowBox[{
        SuperscriptBox["q", "2"], 
        SuperscriptBox["\[Chi]2", "2"]}]}], ")"}]}], ",", 
    RowBox[{"t\[Chi]4", "->", 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox["\[Chi]1", "2"], " ", "-", 
       RowBox[{
        SuperscriptBox["q", "2"], 
        SuperscriptBox["\[Chi]2", "2"]}]}], ")"}]}]}], "}"}]}]], "Input",
 CellLabel->"In[96]:=",ExpressionUUID->"49bf7069-fa2f-424f-89a2-6f906cba24b4"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"t\[Chi]1", "\[Rule]", 
    RowBox[{
     SuperscriptBox["\[Chi]1", "2"], "+", 
     RowBox[{
      SuperscriptBox["q", "4"], " ", 
      SuperscriptBox["\[Chi]2", "2"]}]}]}], ",", 
   RowBox[{"t\[Chi]2", "\[Rule]", 
    RowBox[{
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "2"]}]}], ",", 
   RowBox[{"t\[Chi]3", "\[Rule]", 
    RowBox[{
     SuperscriptBox["\[Chi]1", "2"], "+", 
     RowBox[{
      SuperscriptBox["q", "2"], " ", 
      SuperscriptBox["\[Chi]2", "2"]}]}]}], ",", 
   RowBox[{"t\[Chi]4", "\[Rule]", 
    RowBox[{
     SuperscriptBox["\[Chi]1", "2"], "-", 
     RowBox[{
      SuperscriptBox["q", "2"], " ", 
      SuperscriptBox["\[Chi]2", "2"]}]}]}]}], "}"}]], "Output",
 CellLabel->
  "Out[96]=",ExpressionUUID->"67380497-739a-4bb2-9409-2f9bb477dd87"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"mycubic", "=", " ", 
   RowBox[{
    RowBox[{"-", 
     FractionBox["1", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "28"]]}], "9216", " ", 
    SuperscriptBox["q", "10"], " ", 
    SuperscriptBox["u", "12"], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"2", " ", "q", " ", "u", " ", "\[Chi]eff"}], 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "2"]]}], ")"}], "2"], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "4"}], " ", 
       SuperscriptBox["q", "4"], " ", "t\[Chi]2"}], "+", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"t\[Chi]1", "-", 
         FractionBox[
          RowBox[{
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "3"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", 
              "\[Delta]\[Chi]"}], "+", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{"1", "+", "q"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}]}],
             ")"}]}], 
          RowBox[{"2", " ", "u"}]]}], ")"}], "2"], "-", 
      RowBox[{"q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{"t\[Chi]1", "-", 
         FractionBox[
          RowBox[{
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "3"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", 
              "\[Delta]\[Chi]"}], "+", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{"1", "+", "q"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}]}],
             ")"}]}], 
          RowBox[{"2", " ", "u"}]]}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"\[Delta]\[Chi]", "-", "\[Chi]eff"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"\[Delta]\[Chi]", "+", "\[Chi]eff"}], ")"}]}], "+", 
      RowBox[{
       SuperscriptBox["q", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], "t\[Chi]3", 
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox["\[Delta]\[Chi]", "2"], "+", 
         SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}], "-", 
      RowBox[{"2", 
       SuperscriptBox["q", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "t\[Chi]4", "  ", 
       "\[Delta]\[Chi]", " ", "\[Chi]eff"}]}], ")"}]}]}], ";"}]], "Input",
 CellLabel->"In[97]:=",ExpressionUUID->"e32b1925-8d08-4bfa-ae80-e6d400e4a854"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{" ", 
  RowBox[{"overalldiscr", " ", "=", " ", 
   RowBox[{
    RowBox[{"coeffs", "[", 
     RowBox[{"[", "6", "]"}], "]"}], " ", "/", 
    RowBox[{"(", 
     RowBox[{"-", " ", "u"}], ")"}]}]}]}]], "Input",
 CellLabel->"In[98]:=",ExpressionUUID->"055a1882-e076-4785-82be-2df0d1db7a6a"],

Cell[BoxData[
 FractionBox[
  RowBox[{"64", " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], "6"], " ", 
   SuperscriptBox["u", "2"]}], "q"]], "Output",
 CellLabel->
  "Out[98]=",ExpressionUUID->"326a068e-82be-4835-a123-d972d338f363"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"myc", "=", 
  RowBox[{
   RowBox[{"CoefficientList", "[", " ", 
    RowBox[{
     RowBox[{"Discriminant", "[", 
      RowBox[{
       RowBox[{"mycubic", "/", "overallcubic"}], ",", "\[Delta]\[Chi]"}], 
      "]"}], ",", "\[Kappa]"}], "]"}], "//", "Simplify"}]}]], "Input",
 CellLabel->"In[99]:=",ExpressionUUID->"8c874a9d-cf29-4b98-84fb-dc135d43e177"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    FractionBox["1", 
     RowBox[{
      SuperscriptBox["q", "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "14"]}]], 
    RowBox[{"2", " ", 
     SuperscriptBox["u", "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"2", " ", 
        SuperscriptBox["t\[Chi]1", "4"], " ", 
        SuperscriptBox["u", "4"]}], "-", 
       RowBox[{"2", " ", "t\[Chi]1", " ", 
        RowBox[{"(", 
         RowBox[{"t\[Chi]3", "+", "t\[Chi]4"}], ")"}], " ", "u", " ", 
        "\[Chi]eff"}], "+", 
       RowBox[{"4", " ", 
        SuperscriptBox["t\[Chi]1", "3"], " ", 
        SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "-", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"t\[Chi]3", "+", "t\[Chi]4"}], ")"}], " ", 
        SuperscriptBox["\[Chi]eff", "2"]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox["t\[Chi]1", "2"], " ", 
        SuperscriptBox["u", "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "t\[Chi]3"}], "+", 
          SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox["q", "20"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", "t\[Chi]2"}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "t\[Chi]3"}], "+", "t\[Chi]4", "+", 
             RowBox[{"2", " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox["q", "19"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"6", " ", "t\[Chi]2", " ", 
           RowBox[{"(", 
            RowBox[{"1", "+", 
             RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", 
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"7", "+", 
                RowBox[{"4", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"5", "+", 
                RowBox[{"6", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{"5", "+", 
                RowBox[{"6", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
            ")"}]}]}], ")"}]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox["q", "18"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"t\[Chi]2", " ", 
           RowBox[{"(", 
            RowBox[{"9", "+", 
             RowBox[{"96", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"48", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"89", "+", 
                RowBox[{"128", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"16", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{"20", "+", 
                RowBox[{"70", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"24", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"21", "+", 
                RowBox[{"70", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"24", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}]}], 
         ")"}]}], "+", 
       RowBox[{"2", " ", "q", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"16", " ", 
           SuperscriptBox["t\[Chi]1", "5"], " ", 
           SuperscriptBox["u", "6"]}], "-", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"t\[Chi]3", "+", "t\[Chi]4"}], ")"}], " ", 
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"5", "+", 
             RowBox[{"6", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["t\[Chi]1", "4"], " ", 
           SuperscriptBox["u", "4"], " ", 
           RowBox[{"(", 
            RowBox[{"1", "+", 
             RowBox[{"20", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
          RowBox[{"4", " ", 
           SuperscriptBox["t\[Chi]1", "3"], " ", 
           SuperscriptBox["u", "3"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"5", " ", "t\[Chi]3", " ", "u"}], "-", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{"3", "+", 
                RowBox[{"8", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
            ")"}]}], "-", 
          RowBox[{"2", " ", "t\[Chi]1", " ", "u", " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{"3", "+", 
                RowBox[{"11", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"4", "+", 
                RowBox[{"11", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
            ")"}]}], "-", 
          RowBox[{"2", " ", 
           SuperscriptBox["t\[Chi]1", "2"], " ", 
           SuperscriptBox["u", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"2", "+", 
                RowBox[{"15", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"11", " ", "t\[Chi]4", " ", "u"}], "-", 
                RowBox[{"2", " ", "\[Chi]eff", " ", 
                 RowBox[{"(", 
                  RowBox[{"3", "+", 
                   RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox["q", "17"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"8", " ", "t\[Chi]2", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "2"}], "+", 
             RowBox[{"36", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"51", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{"35", "+", 
                RowBox[{"358", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"284", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"32", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "-", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"45", "+", 
                RowBox[{"366", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"284", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"32", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
             RowBox[{"2", " ", "\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"t\[Chi]1", " ", "u"}], "+", 
                RowBox[{"8", " ", "\[Chi]eff", " ", 
                 RowBox[{"(", 
                  RowBox[{"21", "+", 
                   RowBox[{"60", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"14", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox["q", "2"], " ", 
        RowBox[{"(", 
         RowBox[{"t\[Chi]2", "-", 
          RowBox[{
           SuperscriptBox["t\[Chi]1", "4"], " ", 
           SuperscriptBox["u", "4"], " ", 
           RowBox[{"(", 
            RowBox[{"1", "+", 
             RowBox[{"48", " ", "t\[Chi]3", " ", 
              SuperscriptBox["u", "2"]}], "-", 
             RowBox[{"160", " ", "u", " ", "\[Chi]eff"}], "-", 
             RowBox[{"16", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["t\[Chi]1", "3"], " ", 
           SuperscriptBox["u", "3"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "4"}], " ", "t\[Chi]3", " ", "u", " ", 
              RowBox[{"(", 
               RowBox[{"5", "+", 
                RowBox[{"12", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{"13", "-", 
                RowBox[{"44", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"]}], "+", 
                RowBox[{"144", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"16", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "-", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"5", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"]}], "-", 
             RowBox[{
              SuperscriptBox["t\[Chi]4", "2"], " ", 
              SuperscriptBox["u", "2"]}], "-", 
             SuperscriptBox["\[Chi]eff", "2"], "+", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{"20", "+", 
                RowBox[{"70", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"24", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"21", "+", 
                RowBox[{"4", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"]}], "+", 
                RowBox[{"70", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"24", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "-", 
          RowBox[{"2", " ", "t\[Chi]1", " ", "u", " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"4", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"]}], "-", 
             RowBox[{"4", " ", 
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"3", "+", 
                RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"2", " ", "t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"6", "+", 
                RowBox[{"2", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"]}], "+", 
                RowBox[{"48", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"23", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{"5", "+", 
                RowBox[{"88", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"46", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "-", 
          RowBox[{
           SuperscriptBox["t\[Chi]1", "2"], " ", 
           SuperscriptBox["u", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"8", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"]}], "+", 
             RowBox[{"2", " ", "t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", 
                RowBox[{"90", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"70", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"44", " ", "t\[Chi]4", " ", "u", " ", 
                 RowBox[{"(", 
                  RowBox[{"2", "+", 
                   RowBox[{"3", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
                
                RowBox[{"\[Chi]eff", " ", 
                 RowBox[{"(", 
                  RowBox[{"59", "+", 
                   RowBox[{"144", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"16", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox["q", "12"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"432", " ", 
           SuperscriptBox["t\[Chi]2", "2"], " ", 
           SuperscriptBox["u", "4"]}], "+", 
          RowBox[{
           SuperscriptBox["t\[Chi]1", "2"], " ", 
           SuperscriptBox["u", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "t\[Chi]3"}], "+", 
             RowBox[{
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"59", "+", 
                RowBox[{"144", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"16", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{"6", " ", "t\[Chi]2", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "11"}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "4"]}], "+", 
             RowBox[{"176", " ", "u", " ", "\[Chi]eff"}], "-", 
             RowBox[{"96", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "-", 
             RowBox[{"744", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "-", 
             RowBox[{"336", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "4"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "-", 
             RowBox[{"1632", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}], "+", 
             RowBox[{"4", " ", "t\[Chi]1", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "12"}], "+", 
                RowBox[{"65", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"8", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"2", " ", "t\[Chi]3", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "17"}], "+", 
                RowBox[{"16", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"208", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{"2", " ", "t\[Chi]1", " ", "u", " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "4"}], " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"]}], "+", 
             RowBox[{"4", " ", 
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"91", "+", 
                RowBox[{"2442", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"880", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"2", " ", "t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"9", "+", 
                RowBox[{"2", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"]}], "-", 
                RowBox[{"368", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"838", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                RowBox[{"144", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "25"}], "+", 
                RowBox[{"352", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"1608", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"320", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "16"}], " ", 
              SuperscriptBox["t\[Chi]3", "3"], " ", 
              SuperscriptBox["u", "4"]}], "+", 
             RowBox[{
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "125"}], "+", 
                RowBox[{"8", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"]}], "+", 
                RowBox[{"920", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"572", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"286", " ", 
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "7"}], "+", 
                RowBox[{"320", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"168", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox["t\[Chi]4", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "39"}], "+", 
                RowBox[{"1760", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"804", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{"325", "-", 
                RowBox[{"5044", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"18304", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"10384", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"39", "+", 
                RowBox[{"8", " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "4"]}], "+", 
                RowBox[{"1248", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"23672", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                RowBox[{"8624", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}], "-", 
                RowBox[{"4", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"11", "+", 
                   RowBox[{"722", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"54", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}]}], ")"}]}], "-", 
       RowBox[{"4", " ", 
        SuperscriptBox["q", "5"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", 
            SuperscriptBox["t\[Chi]1", "4"]}], " ", 
           SuperscriptBox["u", "4"], " ", 
           RowBox[{"(", 
            RowBox[{"1", "+", 
             RowBox[{"20", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
          RowBox[{"4", " ", "t\[Chi]2", " ", 
           RowBox[{"(", 
            RowBox[{"2", "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["t\[Chi]1", "3"], " ", 
              SuperscriptBox["u", "6"]}], "-", 
             RowBox[{"36", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"9", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "-", 
             RowBox[{"51", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "-", 
             RowBox[{"8", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}], "-", 
             RowBox[{"3", " ", "t\[Chi]3", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"2", "+", 
                RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"6", " ", 
              SuperscriptBox["t\[Chi]1", "2"], " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "5"}], "+", 
                RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
             RowBox[{"6", " ", "t\[Chi]1", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", 
                RowBox[{"t\[Chi]3", " ", 
                 SuperscriptBox["u", "2"]}], "+", 
                RowBox[{"15", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["t\[Chi]1", "3"], " ", 
           SuperscriptBox["u", "3"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"t\[Chi]3", " ", "u", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "5"}], "+", 
                RowBox[{"96", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"8", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
             RowBox[{"2", " ", "\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "7"}], "+", 
                RowBox[{"22", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"]}], "+", 
                RowBox[{"360", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"80", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "-", 
          RowBox[{"2", " ", 
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "16"}], " ", 
              SuperscriptBox["t\[Chi]4", "3"], " ", 
              SuperscriptBox["u", "5"], " ", "\[Chi]eff"}], "+", 
             RowBox[{"4", " ", 
              SuperscriptBox["t\[Chi]3", "3"], " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "5"}], "+", 
                RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"4", " ", 
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"21", "+", 
                RowBox[{"60", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"14", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox["t\[Chi]4", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", 
                RowBox[{"286", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"82", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "31"}], "+", 
                RowBox[{"405", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"1255", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"400", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "47"}], "+", 
                RowBox[{"34", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"66", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"4", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "3"}], "+", 
                   RowBox[{"4", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
               ")"}]}], "+", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"11", "-", 
                RowBox[{"544", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"1277", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                RowBox[{"392", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}], "-", 
                RowBox[{"4", " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "4"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "2"}], "+", 
                   RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
                RowBox[{"8", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "4"}], "+", 
                   RowBox[{"41", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "-", 
          RowBox[{"t\[Chi]1", " ", "u", " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "16"}], " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"2", "+", 
                RowBox[{"t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"]}], "-", 
                RowBox[{"29", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"16", " ", 
              SuperscriptBox["t\[Chi]3", "3"], " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "2"}], "+", 
                RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"16", " ", 
              SuperscriptBox["t\[Chi]4", "2"], " ", 
              SuperscriptBox["u", "3"], " ", "\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{"11", "+", 
                RowBox[{"3", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"4", " ", 
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"91", "+", 
                RowBox[{"860", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"264", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{"17", "+", 
                RowBox[{"121", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"3656", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                RowBox[{"1080", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"60", "-", 
                RowBox[{"771", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"16", " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "5"], " ", "\[Chi]eff"}], "-", 
                RowBox[{"4000", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                RowBox[{"904", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}], "+", 
                RowBox[{"8", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"4", "+", 
                   RowBox[{"91", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"4", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "-", 
          RowBox[{"2", " ", 
           SuperscriptBox["t\[Chi]1", "2"], " ", 
           SuperscriptBox["u", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"4", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"2", "+", 
                RowBox[{"27", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"2", "-", 
                RowBox[{"105", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"940", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"180", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"8", " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "4"], " ", "\[Chi]eff"}], "+", 
                RowBox[{"t\[Chi]4", " ", "u", " ", 
                 RowBox[{"(", 
                  RowBox[{"77", "-", 
                   RowBox[{"462", " ", "u", " ", "\[Chi]eff"}], "-", 
                   RowBox[{"100", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                RowBox[{"\[Chi]eff", " ", 
                 RowBox[{"(", 
                  RowBox[{"13", "+", 
                   RowBox[{"2430", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"704", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}]}], ")"}]}], "-", 
       RowBox[{"2", " ", 
        SuperscriptBox["q", "8"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"432", " ", 
           SuperscriptBox["t\[Chi]2", "2"], " ", 
           SuperscriptBox["u", "4"]}], "-", 
          RowBox[{"2", " ", 
           SuperscriptBox["t\[Chi]1", "3"], " ", 
           SuperscriptBox["u", "3"], " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{"13", "+", 
             RowBox[{"144", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"16", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{"4", " ", "t\[Chi]2", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "29"}], "-", 
             RowBox[{"16", " ", 
              SuperscriptBox["t\[Chi]3", "3"], " ", 
              SuperscriptBox["u", "6"]}], "+", 
             RowBox[{"360", " ", "u", " ", "\[Chi]eff"}], "-", 
             RowBox[{"144", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "-", 
             RowBox[{"732", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "-", 
             RowBox[{"504", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "4"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "-", 
             RowBox[{"280", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}], "+", 
             RowBox[{"12", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", 
                RowBox[{"8", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"6", " ", 
              SuperscriptBox["t\[Chi]1", "2"], " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{"5", "+", 
                RowBox[{"16", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"3", " ", "t\[Chi]3", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"17", "-", 
                RowBox[{"16", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"48", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "-", 
                RowBox[{"208", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
             RowBox[{"12", " ", "t\[Chi]1", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"4", "-", 
                RowBox[{"35", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"68", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"4", " ", "t\[Chi]3", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", 
                   RowBox[{"6", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "-", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "16"}], " ", 
              SuperscriptBox["t\[Chi]3", "4"], " ", 
              SuperscriptBox["u", "6"]}], "+", 
             RowBox[{"160", " ", 
              SuperscriptBox["t\[Chi]3", "3"], " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", 
                RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"364", " ", 
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"3", "+", 
                RowBox[{"96", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"44", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"3", " ", 
              SuperscriptBox["t\[Chi]4", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"9", "-", 
                RowBox[{"704", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"56", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"249", "+", 
                RowBox[{"16", " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "4"]}], "+", 
                RowBox[{"5232", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"616", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{"325", "-", 
                RowBox[{"5044", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"18304", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"10384", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"39", "+", 
                RowBox[{"1248", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"23672", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                RowBox[{"8624", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}], "-", 
                RowBox[{"80", " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "4"], " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "+", 
                   RowBox[{"4", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
                
                RowBox[{"12", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"33", "+", 
                   RowBox[{"244", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"28", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "+", 
          RowBox[{"2", " ", "t\[Chi]1", " ", "u", " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"8", " ", 
              SuperscriptBox["t\[Chi]3", "3"], " ", 
              SuperscriptBox["u", "4"]}], "-", 
             RowBox[{"8", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"7", "+", 
                RowBox[{"t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"]}], "+", 
                RowBox[{"131", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"99", "-", 
                RowBox[{"1728", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"8232", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"2240", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}], "+", 
                RowBox[{"8", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"7", "+", 
                   RowBox[{"91", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"5", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
             "-", 
             RowBox[{"4", " ", "\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "3"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "22"}], "+", 
                   RowBox[{"15", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
                RowBox[{"11", " ", "\[Chi]eff", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "13"}], "+", 
                   RowBox[{"690", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"288", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["t\[Chi]1", "2"], " ", 
           SuperscriptBox["u", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"8", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"]}], "+", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "17"}], "+", 
                RowBox[{"240", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"2288", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"288", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "-", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"8", " ", "t\[Chi]4", " ", "u", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "22"}], "+", 
                   RowBox[{"231", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"40", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                RowBox[{"\[Chi]eff", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "261"}], "+", 
                   RowBox[{"16416", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"5152", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox["q", "4"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", 
           SuperscriptBox["t\[Chi]1", "4"], " ", 
           SuperscriptBox["u", "4"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             RowBox[{"160", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"16", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{"6", " ", "t\[Chi]2", " ", 
           RowBox[{"(", 
            RowBox[{"3", "+", 
             RowBox[{"4", " ", "t\[Chi]3", " ", 
              SuperscriptBox["u", "2"]}], "+", 
             RowBox[{"40", " ", 
              SuperscriptBox["t\[Chi]1", "2"], " ", 
              SuperscriptBox["u", "4"]}], "+", 
             RowBox[{"32", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"16", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"8", " ", "t\[Chi]1", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"4", "+", 
                RowBox[{"5", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
            ")"}]}], "-", 
          RowBox[{"8", " ", 
           SuperscriptBox["t\[Chi]1", "3"], " ", 
           SuperscriptBox["u", "3"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{"7", "-", 
                RowBox[{"568", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"120", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"4", " ", "t\[Chi]3", " ", "u", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "5"}], "+", 
                RowBox[{"36", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{"4", " ", "t\[Chi]1", " ", "u", " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "8"}], " ", 
              SuperscriptBox["t\[Chi]3", "3"], " ", 
              SuperscriptBox["u", "4"]}], "-", 
             RowBox[{"4", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"13", "+", 
                RowBox[{"2", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"]}], "-", 
                RowBox[{"26", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["t\[Chi]4", "2"], " ", 
              SuperscriptBox["u", "3"], " ", "\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{"22", "+", 
                RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"4", " ", 
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"49", "+", 
                RowBox[{"206", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"48", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "25"}], "+", 
                RowBox[{"352", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"1608", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"320", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
             RowBox[{"2", " ", "t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"9", "-", 
                RowBox[{"368", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"838", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                RowBox[{"144", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}], "+", 
                RowBox[{"2", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "3"}], "+", 
                   RowBox[{"78", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "32"}], " ", 
              SuperscriptBox["t\[Chi]3", "3"], " ", 
              SuperscriptBox["u", "4"]}], "+", 
             RowBox[{"2", " ", 
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"89", "+", 
                RowBox[{"128", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"16", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
             RowBox[{"2", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"119", "+", 
                RowBox[{"8", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"]}], "+", 
                RowBox[{"8", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"44", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"2", " ", 
              SuperscriptBox["t\[Chi]4", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"11", "+", 
                RowBox[{"352", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"52", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "5"}], "+", 
                RowBox[{"2040", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"3072", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"672", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "77"}], "+", 
                RowBox[{"16", " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "4"]}], "-", 
                RowBox[{"2240", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"3088", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                RowBox[{"672", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}], "+", 
                RowBox[{"8", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "25"}], "+", 
                   RowBox[{"86", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["t\[Chi]1", "2"], " ", 
           SuperscriptBox["u", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "16"}], " ", 
              SuperscriptBox["t\[Chi]3", "3"], " ", 
              SuperscriptBox["u", "4"]}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", 
                RowBox[{"36", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"16", " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "4"], " ", "\[Chi]eff"}], "-", 
                RowBox[{"8", " ", "t\[Chi]4", " ", "u", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "22"}], "+", 
                   RowBox[{"231", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"40", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                RowBox[{"\[Chi]eff", " ", 
                 RowBox[{"(", 
                  RowBox[{"201", "+", 
                   RowBox[{"3920", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"976", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
             "+", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"17", "-", 
                RowBox[{"240", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"2288", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"16", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"13", " ", "t\[Chi]4", " ", "\[Chi]eff"}], "-", 
                   RowBox[{"18", " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox["q", "3"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"6", " ", "t\[Chi]2", " ", 
           RowBox[{"(", 
            RowBox[{"1", "+", 
             RowBox[{"4", " ", "t\[Chi]1", " ", 
              SuperscriptBox["u", "2"]}], "+", 
             RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
          RowBox[{"4", " ", 
           SuperscriptBox["t\[Chi]1", "4"], " ", 
           SuperscriptBox["u", "4"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             RowBox[{"60", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{"4", " ", 
           SuperscriptBox["t\[Chi]1", "3"], " ", 
           SuperscriptBox["u", "3"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"12", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "3"]}], "+", 
             RowBox[{"t\[Chi]3", " ", "u", " ", 
              RowBox[{"(", 
               RowBox[{"5", "-", 
                RowBox[{"96", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"8", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"4", " ", "\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{"1", "-", 
                RowBox[{"11", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"]}], "+", 
                RowBox[{"68", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"12", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "-", 
          RowBox[{"2", " ", "t\[Chi]1", " ", "u", " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "44"}], " ", 
              SuperscriptBox["t\[Chi]4", "2"], " ", 
              SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "-", 
             RowBox[{"8", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "3"}], "+", 
                RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
             RowBox[{"8", " ", 
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"8", "+", 
                RowBox[{"15", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "5"}], "+", 
                RowBox[{"275", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"408", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"40", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"12", "+", 
                RowBox[{"361", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"416", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"40", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}], "+", 
                RowBox[{"4", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"4", "-", 
                   RowBox[{"13", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "-", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"19", "+", 
                RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
             RowBox[{"2", " ", 
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"7", "+", 
                RowBox[{"4", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
             RowBox[{"2", " ", 
              SuperscriptBox["t\[Chi]4", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"3", "+", 
                RowBox[{"22", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{"35", "+", 
                RowBox[{"358", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"284", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"32", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"45", "+", 
                RowBox[{"366", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"284", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"32", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}], "-", 
                RowBox[{"8", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "4"}], "+", 
                   RowBox[{"5", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["t\[Chi]1", "2"], " ", 
           SuperscriptBox["u", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"4", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "2"}], "+", 
                RowBox[{"9", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"4", " ", "\[Chi]eff", " ", 
                 RowBox[{"(", 
                  RowBox[{"19", "+", 
                   RowBox[{"126", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"24", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
                RowBox[{"t\[Chi]4", " ", "u", " ", 
                 RowBox[{"(", 
                  RowBox[{"33", "+", 
                   RowBox[{"396", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"40", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
             "+", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"6", "-", 
                RowBox[{"195", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"424", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"13", " ", "t\[Chi]4", " ", "\[Chi]eff"}], "-", 
                   RowBox[{"6", " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}]}], ")"}]}], "-", 
       RowBox[{"4", " ", 
        SuperscriptBox["q", "7"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "2"}], " ", 
           SuperscriptBox["t\[Chi]1", "3"], " ", 
           SuperscriptBox["u", "3"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "5"}], " ", "t\[Chi]3", " ", "u"}], "+", 
             RowBox[{"4", " ", "\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", 
                RowBox[{"68", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"12", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{"t\[Chi]1", " ", "u", " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "16"}], " ", 
              SuperscriptBox["t\[Chi]3", "3"], " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "2"}], "+", 
                RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
             RowBox[{"40", " ", 
              SuperscriptBox["t\[Chi]4", "2"], " ", 
              SuperscriptBox["u", "3"], " ", "\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "11"}], "+", 
                RowBox[{"4", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
             RowBox[{"16", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"7", "+", 
                RowBox[{"t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"]}], "+", 
                RowBox[{"85", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
             RowBox[{"88", " ", "u", " ", 
              SuperscriptBox["\[Chi]eff", "3"], " ", 
              RowBox[{"(", 
               RowBox[{"227", "+", 
                RowBox[{"90", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{"55", "-", 
                RowBox[{"1089", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"3888", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"1680", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"44", "-", 
                RowBox[{"1143", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"16", " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "5"], " ", "\[Chi]eff"}], "+", 
                RowBox[{"7872", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"2128", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{"12", " ", "t\[Chi]2", " ", 
           RowBox[{"(", 
            RowBox[{"1", "+", 
             RowBox[{"18", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"9", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "-", 
             RowBox[{"122", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "-", 
             RowBox[{"36", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "4"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "-", 
             RowBox[{"56", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}], "+", 
             RowBox[{"4", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", 
                RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"4", " ", 
              SuperscriptBox["t\[Chi]1", "2"], " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{"5", "+", 
                RowBox[{"6", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]3", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"6", "-", 
                RowBox[{"13", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"12", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "-", 
                RowBox[{"64", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]1", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"17", "+", 
                RowBox[{"8", " ", 
                 SuperscriptBox["t\[Chi]3", "2"], " ", 
                 SuperscriptBox["u", "4"]}], "-", 
                RowBox[{"40", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"24", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "+", 
                RowBox[{"32", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"2", " ", "t\[Chi]3", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "-", 
                   RowBox[{"16", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "-", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "16"}], " ", 
              SuperscriptBox["t\[Chi]3", "4"], " ", 
              SuperscriptBox["u", "6"]}], "-", 
             RowBox[{"160", " ", 
              SuperscriptBox["t\[Chi]4", "3"], " ", 
              SuperscriptBox["u", "5"], " ", "\[Chi]eff"}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["t\[Chi]3", "3"], " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "11"}], "+", 
                RowBox[{"15", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"28", " ", 
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"23", "+", 
                RowBox[{"260", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"104", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox["t\[Chi]4", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "19"}], "+", 
                RowBox[{"330", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"432", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
             RowBox[{"2", " ", "t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "5"}], "-", 
                RowBox[{"507", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"4251", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"2112", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"35", "+", 
                RowBox[{"16", " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "4"]}], "+", 
                RowBox[{"1410", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"176", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"8", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "7"}], "+", 
                   RowBox[{"20", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
               ")"}]}], "-", 
             RowBox[{"2", " ", "t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "47"}], "+", 
                RowBox[{"364", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"4703", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"1848", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "4"], " ", 
                 RowBox[{"(", 
                  RowBox[{"2", "+", 
                   RowBox[{"15", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
                RowBox[{"2", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"44", "+", 
                   RowBox[{"489", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"48", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["t\[Chi]1", "2"], " ", 
           SuperscriptBox["u", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"2", "-", 
                RowBox[{"9", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"t\[Chi]4", " ", "u", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "77"}], "+", 
                   RowBox[{"462", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"100", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                RowBox[{"4", " ", "\[Chi]eff", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "27"}], "+", 
                   RowBox[{"1218", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"392", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
             "+", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"2", "-", 
                RowBox[{"105", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"940", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"13", " ", "t\[Chi]4", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"90", " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox["q", "10"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"432", " ", 
           SuperscriptBox["t\[Chi]2", "2"], " ", 
           SuperscriptBox["u", "4"]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["t\[Chi]1", "3"], " ", 
           SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "+", 
          RowBox[{"6", " ", "t\[Chi]2", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "11"}], "+", 
             RowBox[{"20", " ", 
              SuperscriptBox["t\[Chi]1", "2"], " ", 
              SuperscriptBox["u", "4"]}], "+", 
             RowBox[{"176", " ", "u", " ", "\[Chi]eff"}], "-", 
             RowBox[{"744", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "-", 
             RowBox[{"1632", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}], "-", 
             RowBox[{"8", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", 
                RowBox[{"8", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"8", " ", "t\[Chi]3", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"7", "-", 
                RowBox[{"14", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"12", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "+", 
                RowBox[{"10", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"8", " ", "t\[Chi]1", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"4", "-", 
                RowBox[{"35", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"6", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "-", 
                RowBox[{"68", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"2", " ", "t\[Chi]3", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "+", 
                   RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "+", 
          RowBox[{"2", " ", "t\[Chi]1", " ", "u", " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"8", " ", 
              SuperscriptBox["t\[Chi]4", "2"], " ", 
              SuperscriptBox["u", "3"], " ", "\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{"22", "+", 
                RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"52", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", 
                RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"44", " ", 
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "13"}], "+", 
                RowBox[{"690", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"288", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{"33", "-", 
                RowBox[{"968", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"5094", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"1920", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "-", 
             RowBox[{"2", " ", "t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "22"}], "+", 
                RowBox[{"16", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"3225", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"832", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}], "+", 
                RowBox[{"2", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "3"}], "+", 
                   RowBox[{"78", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["t\[Chi]1", "2"], " ", 
           SuperscriptBox["u", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "2"}], " ", "t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", 
                RowBox[{"90", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"70", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"44", " ", "t\[Chi]4", " ", "u", " ", 
                 RowBox[{"(", 
                  RowBox[{"2", "+", 
                   RowBox[{"3", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
                
                RowBox[{"\[Chi]eff", " ", 
                 RowBox[{"(", 
                  RowBox[{"201", "+", 
                   RowBox[{"3920", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"976", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"256", " ", 
              SuperscriptBox["t\[Chi]4", "3"], " ", 
              SuperscriptBox["u", "5"], " ", "\[Chi]eff"}], "+", 
             RowBox[{"32", " ", 
              SuperscriptBox["t\[Chi]3", "3"], " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "5"}], "+", 
                RowBox[{"3", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"3", " ", 
              SuperscriptBox["t\[Chi]4", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"9", "-", 
                RowBox[{"704", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"56", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"286", " ", 
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "7"}], "+", 
                RowBox[{"320", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"168", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"249", "+", 
                RowBox[{"5232", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"616", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                RowBox[{"16", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "7"}], "+", 
                   RowBox[{"16", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
               ")"}]}], "-", 
             RowBox[{"2", " ", "t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"16", " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "4"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", 
                   RowBox[{"3", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
                
                RowBox[{"6", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"33", "+", 
                   RowBox[{"244", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"28", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                RowBox[{"11", " ", 
                 RowBox[{"(", 
                  RowBox[{"13", "-", 
                   RowBox[{"234", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"1128", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                   RowBox[{"288", " ", 
                    SuperscriptBox["u", "3"], " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}]}], ")"}]}], "-", 
       RowBox[{"4", " ", 
        SuperscriptBox["q", "9"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"432", " ", 
           SuperscriptBox["t\[Chi]2", "2"], " ", 
           SuperscriptBox["u", "4"]}], "-", 
          RowBox[{"2", " ", 
           SuperscriptBox["t\[Chi]1", "3"], " ", 
           SuperscriptBox["u", "3"], " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{"3", "+", 
             RowBox[{"8", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
          RowBox[{"t\[Chi]1", " ", "u", " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"16", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"2", "-", 
                RowBox[{"29", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
             RowBox[{"16", " ", 
              SuperscriptBox["t\[Chi]4", "2"], " ", 
              SuperscriptBox["u", "3"], " ", "\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{"11", "+", 
                RowBox[{"3", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
             RowBox[{"66", " ", 
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "13"}], "+", 
                RowBox[{"528", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"224", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{"55", "-", 
                RowBox[{"1089", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"3888", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"1680", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"44", "-", 
                RowBox[{"1143", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"7872", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"2128", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}], "+", 
                RowBox[{"8", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"4", "+", 
                   RowBox[{"91", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"4", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "+", 
          RowBox[{"12", " ", "t\[Chi]2", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "6"}], "+", 
             RowBox[{"44", " ", "u", " ", "\[Chi]eff"}], "-", 
             RowBox[{"42", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "+", 
             RowBox[{
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "-", 
             RowBox[{"84", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "4"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"120", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}], "+", 
             RowBox[{"2", " ", 
              SuperscriptBox["t\[Chi]1", "2"], " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "5"}], "+", 
                RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", 
                RowBox[{"3", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
             RowBox[{"2", " ", "t\[Chi]3", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"2", "-", 
                RowBox[{"7", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"32", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"2", " ", "t\[Chi]1", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"t\[Chi]3", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox["u", "2"], "-", 
                   RowBox[{"16", " ", 
                    SuperscriptBox["u", "3"], " ", "\[Chi]eff"}]}], ")"}]}], 
                "+", 
                RowBox[{"2", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "7"}], "+", 
                   RowBox[{"35", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"6", " ", "t\[Chi]4", " ", 
                    SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "+", 
                   RowBox[{"50", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["t\[Chi]1", "2"], " ", 
           SuperscriptBox["u", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "6"}], "+", 
                RowBox[{"195", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"424", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"24", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "-", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"t\[Chi]4", " ", "u", " ", 
                 RowBox[{"(", 
                  RowBox[{"33", "+", 
                   RowBox[{"396", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"40", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                RowBox[{"2", " ", "\[Chi]eff", " ", 
                 RowBox[{"(", 
                  RowBox[{"13", "+", 
                   RowBox[{"2430", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"704", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "160"}], " ", 
              SuperscriptBox["t\[Chi]4", "3"], " ", 
              SuperscriptBox["u", "5"], " ", "\[Chi]eff"}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["t\[Chi]3", "3"], " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{"11", "-", 
                RowBox[{"15", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"12", " ", 
              SuperscriptBox["t\[Chi]4", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "3"}], "+", 
                RowBox[{"154", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"14", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
             RowBox[{"104", " ", 
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", 
                RowBox[{"308", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"154", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{"195", "-", 
                RowBox[{"2288", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"6006", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"3696", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
             RowBox[{"4", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "39"}], "-", 
                RowBox[{"798", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"154", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"2", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "7"}], "+", 
                   RowBox[{"20", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
               ")"}]}], "+", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"8", " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "4"], " ", 
                 RowBox[{"(", 
                  RowBox[{"2", "+", 
                   RowBox[{"15", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
                RowBox[{"7", " ", 
                 RowBox[{"(", 
                  RowBox[{"13", "-", 
                   RowBox[{"286", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"1782", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                   RowBox[{"528", " ", 
                    SuperscriptBox["u", "3"], " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{"4", " ", 
        SuperscriptBox["q", "11"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"864", " ", 
           SuperscriptBox["t\[Chi]2", "2"], " ", 
           SuperscriptBox["u", "4"]}], "+", 
          RowBox[{"2", " ", "t\[Chi]2", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "55"}], "+", 
             RowBox[{"594", " ", "u", " ", "\[Chi]eff"}], "-", 
             RowBox[{"252", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "-", 
             RowBox[{"1656", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "-", 
             RowBox[{"504", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "4"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "-", 
             RowBox[{"3360", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}], "-", 
             RowBox[{"24", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", 
                RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"6", " ", "t\[Chi]1", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "17"}], "+", 
                RowBox[{"2", " ", "t\[Chi]3", " ", 
                 SuperscriptBox["u", "2"]}], "+", 
                RowBox[{"40", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"32", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"12", " ", "t\[Chi]3", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"2", "-", 
                RowBox[{"7", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"6", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "+", 
                RowBox[{"32", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{"t\[Chi]1", " ", "u", " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"44", " ", 
              SuperscriptBox["t\[Chi]4", "2"], " ", 
              SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "3"}], "+", 
                RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"88", " ", "u", " ", 
              SuperscriptBox["\[Chi]eff", "3"], " ", 
              RowBox[{"(", 
               RowBox[{"227", "+", 
                RowBox[{"90", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "17"}], "-", 
                RowBox[{"121", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"3656", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"1080", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "-", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "60"}], "+", 
                RowBox[{"771", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"4000", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"904", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}], "+", 
                RowBox[{"4", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "4"}], "+", 
                   RowBox[{"13", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["t\[Chi]1", "2"], " ", 
           SuperscriptBox["u", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "t\[Chi]3"}], " ", 
              RowBox[{"(", 
               RowBox[{"2", "+", 
                RowBox[{"15", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"11", " ", "t\[Chi]4", " ", "u"}], "+", 
                RowBox[{"4", " ", "\[Chi]eff", " ", 
                 RowBox[{"(", 
                  RowBox[{"19", "+", 
                   RowBox[{"126", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"24", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"32", " ", 
              SuperscriptBox["t\[Chi]4", "3"], " ", 
              SuperscriptBox["u", "5"], " ", "\[Chi]eff"}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["t\[Chi]3", "3"], " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "5"}], "+", 
                RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"286", " ", 
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "5"}], "+", 
                RowBox[{"180", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"96", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox["t\[Chi]4", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "19"}], "+", 
                RowBox[{"330", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"432", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{"195", "-", 
                RowBox[{"2288", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"6006", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"3696", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"35", "+", 
                RowBox[{"1410", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"176", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                RowBox[{"8", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "3"}], "+", 
                   RowBox[{"4", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
               ")"}]}], "-", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"8", " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "4"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "2"}], "+", 
                   RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
                RowBox[{"4", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"44", "+", 
                   RowBox[{"489", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"48", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                RowBox[{"7", " ", 
                 RowBox[{"(", 
                  RowBox[{"13", "-", 
                   RowBox[{"286", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"1782", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                   RowBox[{"528", " ", 
                    SuperscriptBox["u", "3"], " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox["q", "16"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"24", " ", "t\[Chi]2", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "5"}], "+", 
             RowBox[{"t\[Chi]3", " ", 
              SuperscriptBox["u", "2"]}], "+", 
             RowBox[{"24", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"124", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"40", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
          RowBox[{"\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", "t\[Chi]1", " ", "u", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "t\[Chi]3"}], "+", "t\[Chi]4", "+", 
                RowBox[{"8", " ", 
                 SuperscriptBox["\[Chi]eff", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"3", "+", 
                   RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
               ")"}]}], "+", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "10"}], " ", 
                 SuperscriptBox["t\[Chi]3", "2"], " ", 
                 SuperscriptBox["u", "2"]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "2"]}], "+", 
                RowBox[{"8", " ", 
                 SuperscriptBox["\[Chi]eff", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"205", "+", 
                   RowBox[{"1120", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"364", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                RowBox[{"t\[Chi]3", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "77"}], "+", 
                   RowBox[{"8", " ", "t\[Chi]4", " ", 
                    SuperscriptBox["u", "2"]}], "-", 
                   RowBox[{"2240", " ", "u", " ", "\[Chi]eff"}], "-", 
                   RowBox[{"3088", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                   RowBox[{"672", " ", 
                    SuperscriptBox["u", "3"], " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
                RowBox[{"t\[Chi]4", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "5"}], "+", 
                   RowBox[{"2040", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"3072", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                   RowBox[{"672", " ", 
                    SuperscriptBox["u", "3"], " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox["q", "6"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           SuperscriptBox["t\[Chi]1", "4"], " ", 
           SuperscriptBox["u", "4"]}], "-", 
          RowBox[{"4", " ", 
           SuperscriptBox["t\[Chi]1", "3"], " ", 
           SuperscriptBox["u", "3"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", "t\[Chi]3", " ", "u", " ", 
              RowBox[{"(", 
               RowBox[{"5", "+", 
                RowBox[{"12", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "7"}], "+", 
                RowBox[{"22", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"]}], "+", 
                RowBox[{"568", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"120", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{"12", " ", "t\[Chi]2", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "5"}], "+", 
             RowBox[{"4", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "4"]}], "+", 
             RowBox[{"24", " ", "u", " ", "\[Chi]eff"}], "-", 
             RowBox[{"24", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "+", 
             RowBox[{"124", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"12", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "4"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"40", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}], "+", 
             RowBox[{"2", " ", 
              SuperscriptBox["t\[Chi]1", "2"], " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "5"}], "+", 
                RowBox[{"8", " ", "t\[Chi]3", " ", 
                 SuperscriptBox["u", "2"]}], "-", 
                RowBox[{"16", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"2", " ", "t\[Chi]3", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", 
                RowBox[{"6", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"14", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"2", " ", "t\[Chi]1", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "12"}], "+", 
                RowBox[{"65", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"12", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "+", 
                RowBox[{"8", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"4", " ", "t\[Chi]3", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "+", 
                   RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "16"}], " ", 
              SuperscriptBox["t\[Chi]3", "4"], " ", 
              SuperscriptBox["u", "6"]}], "-", 
             RowBox[{"256", " ", 
              SuperscriptBox["t\[Chi]4", "3"], " ", 
              SuperscriptBox["u", "5"], " ", "\[Chi]eff"}], "+", 
             RowBox[{"32", " ", 
              SuperscriptBox["t\[Chi]3", "3"], " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "5"}], "+", 
                RowBox[{"3", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"4", " ", 
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"205", "+", 
                RowBox[{"1120", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"364", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox["t\[Chi]4", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "39"}], "+", 
                RowBox[{"1760", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"804", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
             RowBox[{"4", " ", "t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "50"}], "+", 
                RowBox[{"213", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"2756", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"1136", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "125"}], "+", 
                RowBox[{"16", " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "4"]}], "+", 
                RowBox[{"920", " ", "u", " ", "\[Chi]eff"}], "-", 
                RowBox[{"572", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"16", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "7"}], "+", 
                   RowBox[{"16", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
               ")"}]}], "-", 
             RowBox[{"4", " ", "t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "41"}], "+", 
                RowBox[{"658", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"2876", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"1064", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}], "+", 
                RowBox[{"8", " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "4"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", 
                   RowBox[{"3", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
                
                RowBox[{"t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"11", "+", 
                   RowBox[{"722", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"54", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["t\[Chi]1", "2"], " ", 
           SuperscriptBox["u", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"8", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", 
                RowBox[{"36", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "261"}], "+", 
                RowBox[{"16", " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "4"]}], "+", 
                RowBox[{"16416", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"5152", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
             RowBox[{"4", " ", "t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{"7", "-", 
                RowBox[{"210", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"1090", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"13", " ", "t\[Chi]4", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"60", " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "+", 
          RowBox[{"2", " ", "t\[Chi]1", " ", "u", " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"16", " ", 
              SuperscriptBox["t\[Chi]3", "3"], " ", 
              SuperscriptBox["u", "4"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "3"}], "+", 
                RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["t\[Chi]4", "2"], " ", 
              SuperscriptBox["u", "3"], " ", "\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "22"}], "+", 
                RowBox[{"15", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["t\[Chi]3", "2"], " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"7", "+", 
                RowBox[{"131", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"4", " ", 
              SuperscriptBox["\[Chi]eff", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"91", "+", 
                RowBox[{"2442", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"880", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
             RowBox[{"t\[Chi]4", " ", 
              RowBox[{"(", 
               RowBox[{"33", "-", 
                RowBox[{"968", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"5094", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                RowBox[{"1920", " ", 
                 SuperscriptBox["u", "3"], " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]3", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "32"}], " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "5"], " ", "\[Chi]eff"}], "+", 
                RowBox[{"8", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"7", "+", 
                   RowBox[{"91", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"5", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
                RowBox[{"2", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "22"}], "+", 
                   RowBox[{"16", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"3225", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                   RowBox[{"832", " ", 
                    SuperscriptBox["u", "3"], " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{"4", " ", 
        SuperscriptBox["q", "15"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"12", " ", "t\[Chi]2", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             RowBox[{"t\[Chi]1", " ", 
              SuperscriptBox["u", "2"]}], "-", 
             RowBox[{"18", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"3", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "+", 
             RowBox[{"122", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"56", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}], "+", 
             RowBox[{"t\[Chi]3", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"2", "+", 
                RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}]}]}], ")"}]}], "+", 
          RowBox[{"\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"t\[Chi]1", " ", "u", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"t\[Chi]4", " ", 
                 RowBox[{"(", 
                  RowBox[{"3", "+", 
                   RowBox[{"11", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
                RowBox[{"t\[Chi]3", " ", 
                 RowBox[{"(", 
                  RowBox[{"4", "+", 
                   RowBox[{"11", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
                RowBox[{"8", " ", 
                 SuperscriptBox["\[Chi]eff", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"8", "+", 
                   RowBox[{"15", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
             "+", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", 
                  SuperscriptBox["t\[Chi]3", "2"]}], " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"19", "+", 
                   RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
                
                RowBox[{
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"3", "+", 
                   RowBox[{"22", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
                RowBox[{"28", " ", 
                 SuperscriptBox["\[Chi]eff", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"23", "+", 
                   RowBox[{"260", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"104", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                RowBox[{"t\[Chi]4", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "62"}], "+", 
                   RowBox[{"810", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"2510", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                   RowBox[{"800", " ", 
                    SuperscriptBox["u", "3"], " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "-", 
                RowBox[{"2", " ", "t\[Chi]3", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "11"}], "+", 
                   RowBox[{"544", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"1277", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                   RowBox[{"392", " ", 
                    SuperscriptBox["u", "3"], " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}], "+", 
                   RowBox[{"2", " ", "t\[Chi]4", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "4"}], "+", 
                    RowBox[{"5", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
                  ")"}]}]}], ")"}]}]}], ")"}]}]}], ")"}]}], "-", 
       RowBox[{"4", " ", 
        SuperscriptBox["q", "13"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"432", " ", 
           SuperscriptBox["t\[Chi]2", "2"], " ", 
           SuperscriptBox["u", "4"]}], "-", 
          RowBox[{"12", " ", "t\[Chi]2", " ", 
           RowBox[{"(", 
            RowBox[{"6", "-", 
             RowBox[{"44", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"9", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "-", 
             RowBox[{
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "-", 
             RowBox[{"36", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "4"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "-", 
             RowBox[{"120", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}], "+", 
             RowBox[{"2", " ", "t\[Chi]1", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", 
                RowBox[{"15", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]3", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "6"}], "+", 
                RowBox[{"13", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"64", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{"\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "2"}], " ", 
              SuperscriptBox["t\[Chi]1", "2"], " ", 
              SuperscriptBox["u", "2"], " ", "\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{"3", "+", 
                RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"t\[Chi]1", " ", "u", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "4"}], " ", 
                 SuperscriptBox["\[Chi]eff", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"91", "+", 
                   RowBox[{"860", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"264", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
                RowBox[{"t\[Chi]4", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "5"}], "+", 
                   RowBox[{"275", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"408", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                   RowBox[{"40", " ", 
                    SuperscriptBox["u", "3"], " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
                RowBox[{"t\[Chi]3", " ", 
                 RowBox[{"(", 
                  RowBox[{"12", "+", 
                   RowBox[{"361", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"416", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                   RowBox[{"40", " ", 
                    SuperscriptBox["u", "3"], " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}], 
             "+", 
             RowBox[{"2", " ", "\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 SuperscriptBox["t\[Chi]3", "2"], " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"47", "-", 
                   RowBox[{"34", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"66", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
                RowBox[{"52", " ", 
                 SuperscriptBox["\[Chi]eff", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", 
                   RowBox[{"308", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"154", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                RowBox[{
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   SuperscriptBox["u", "2"], "-", 
                   RowBox[{"286", " ", 
                    SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "-", 
                   RowBox[{"82", " ", 
                    SuperscriptBox["u", "4"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                RowBox[{"t\[Chi]4", " ", 
                 RowBox[{"(", 
                  RowBox[{"5", "+", 
                   RowBox[{"507", " ", "u", " ", "\[Chi]eff"}], "-", 
                   RowBox[{"4251", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                   RowBox[{"2112", " ", 
                    SuperscriptBox["u", "3"], " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
                RowBox[{"t\[Chi]3", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "47"}], "+", 
                   RowBox[{"364", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"4703", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                   RowBox[{"1848", " ", 
                    SuperscriptBox["u", "3"], " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}], "+", 
                   RowBox[{"8", " ", "t\[Chi]4", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "4"}], "+", 
                    RowBox[{"41", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
                  ")"}]}]}], ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox["q", "14"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "864"}], " ", 
           SuperscriptBox["t\[Chi]2", "2"], " ", 
           SuperscriptBox["u", "4"]}], "+", 
          RowBox[{"8", " ", "t\[Chi]2", " ", 
           RowBox[{"(", 
            RowBox[{"29", "-", 
             RowBox[{"360", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"72", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "+", 
             RowBox[{"732", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "-", 
             RowBox[{"36", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "4"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"280", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}], "+", 
             RowBox[{"6", " ", "t\[Chi]1", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"4", "+", 
                RowBox[{"5", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"6", " ", "t\[Chi]3", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", 
                RowBox[{"6", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"14", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{"2", " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              SuperscriptBox["t\[Chi]1", "2"], " ", 
              SuperscriptBox["u", "2"], " ", "\[Chi]eff"}], "+", 
             RowBox[{"2", " ", "t\[Chi]1", " ", "u", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "2"}], " ", "t\[Chi]3", " ", 
                 RowBox[{"(", 
                  RowBox[{"6", "+", 
                   RowBox[{"48", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"23", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                RowBox[{"t\[Chi]4", " ", 
                 RowBox[{"(", 
                  RowBox[{"5", "+", 
                   RowBox[{"88", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"46", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox["\[Chi]eff", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"49", "+", 
                   RowBox[{"206", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"48", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
             "+", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", 
                  SuperscriptBox["t\[Chi]3", "2"]}], " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"119", "+", 
                   RowBox[{"8", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"44", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                RowBox[{"364", " ", 
                 SuperscriptBox["\[Chi]eff", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"3", "+", 
                   RowBox[{"96", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"44", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                RowBox[{
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"11", "+", 
                   RowBox[{"352", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"52", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                RowBox[{"4", " ", "t\[Chi]4", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "50"}], "+", 
                   RowBox[{"213", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"2756", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                   RowBox[{"1136", " ", 
                    SuperscriptBox["u", "3"], " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "-", 
                RowBox[{"4", " ", "t\[Chi]3", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "41"}], "+", 
                   RowBox[{"658", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"2876", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                   RowBox[{"1064", " ", 
                    SuperscriptBox["u", "3"], " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}], "+", 
                   RowBox[{"t\[Chi]4", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "25"}], "+", 
                    RowBox[{"86", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
                  ")"}]}]}], ")"}]}]}], ")"}]}]}], ")"}]}]}], ")"}]}]}], ",", 
   
   RowBox[{"-", 
    RowBox[{
     FractionBox["1", 
      RowBox[{
       SuperscriptBox["q", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "10"]}]], 
     RowBox[{"4", " ", 
      SuperscriptBox["u", "2"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"4", " ", 
         SuperscriptBox["t\[Chi]1", "3"], " ", 
         SuperscriptBox["u", "3"]}], "-", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"t\[Chi]3", "+", "t\[Chi]4"}], ")"}], " ", "\[Chi]eff"}], 
        "+", 
        RowBox[{"6", " ", 
         SuperscriptBox["t\[Chi]1", "2"], " ", 
         SuperscriptBox["u", "2"], " ", "\[Chi]eff"}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox["q", "17"], " ", 
         SuperscriptBox["\[Chi]eff", "3"]}], "+", 
        RowBox[{"2", " ", "t\[Chi]1", " ", "u", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "t\[Chi]3"}], "+", 
           SuperscriptBox["\[Chi]eff", "2"]}], ")"}]}], "+", 
        RowBox[{
         SuperscriptBox["q", "16"], " ", "\[Chi]eff", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "t\[Chi]3"}], "+", "t\[Chi]4", "+", 
           RowBox[{"8", " ", 
            SuperscriptBox["\[Chi]eff", "2"], " ", 
            RowBox[{"(", 
             RowBox[{"3", "+", 
              RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], ")"}]}],
         "+", 
        RowBox[{"2", " ", 
         SuperscriptBox["q", "15"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"12", " ", "t\[Chi]2", " ", "u"}], "+", 
           RowBox[{"t\[Chi]4", " ", "\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{"3", "+", 
              RowBox[{"11", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
           RowBox[{"t\[Chi]3", " ", "\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{"4", "+", 
              RowBox[{"11", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
           RowBox[{"8", " ", 
            SuperscriptBox["\[Chi]eff", "3"], " ", 
            RowBox[{"(", 
             RowBox[{"8", "+", 
              RowBox[{"15", " ", "u", " ", "\[Chi]eff"}], "+", 
              RowBox[{"2", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], "+", 
        
        RowBox[{"2", " ", 
         SuperscriptBox["q", "12"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"12", " ", "t\[Chi]2", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "12"}], "+", 
              RowBox[{"65", " ", "u", " ", "\[Chi]eff"}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
           RowBox[{"t\[Chi]1", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "t\[Chi]3"}], "+", 
              RowBox[{
               SuperscriptBox["\[Chi]eff", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"59", "+", 
                 RowBox[{"144", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"16", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{"\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "4"}], " ", 
               SuperscriptBox["t\[Chi]3", "2"], " ", 
               SuperscriptBox["u", "2"]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["\[Chi]eff", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"91", "+", 
                 RowBox[{"2442", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"880", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{"2", " ", "t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{"9", "+", 
                 RowBox[{"2", " ", "t\[Chi]4", " ", 
                  SuperscriptBox["u", "2"]}], "-", 
                 RowBox[{"368", " ", "u", " ", "\[Chi]eff"}], "-", 
                 RowBox[{"838", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"144", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
              RowBox[{"t\[Chi]4", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "25"}], "+", 
                 RowBox[{"352", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"1608", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"320", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}]}], 
          ")"}]}], "+", 
        RowBox[{"2", " ", "q", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"40", " ", 
            SuperscriptBox["t\[Chi]1", "4"], " ", 
            SuperscriptBox["u", "5"]}], "+", 
           RowBox[{"4", " ", 
            SuperscriptBox["t\[Chi]1", "3"], " ", 
            SuperscriptBox["u", "3"], " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", 
              RowBox[{"20", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
           RowBox[{"6", " ", 
            SuperscriptBox["t\[Chi]1", "2"], " ", 
            SuperscriptBox["u", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"5", " ", "t\[Chi]3", " ", "u"}], "-", 
              RowBox[{"\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{"3", "+", 
                 RowBox[{"8", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
             ")"}]}], "+", 
           RowBox[{"\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{
              SuperscriptBox["\[Chi]eff", "2"], "-", 
              RowBox[{"t\[Chi]4", " ", 
               RowBox[{"(", 
                RowBox[{"3", "+", 
                 RowBox[{"11", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
              RowBox[{"t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{"4", "+", 
                 RowBox[{"11", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
             ")"}]}], "-", 
           RowBox[{"2", " ", "t\[Chi]1", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{"2", "+", 
                 RowBox[{"15", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"11", " ", "t\[Chi]4", " ", "u"}], "-", 
                 RowBox[{"2", " ", "\[Chi]eff", " ", 
                  RowBox[{"(", 
                   RowBox[{"3", "+", 
                    RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}]}], ")"}]}], "-", 
        RowBox[{"2", " ", 
         SuperscriptBox["q", "2"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"2", " ", 
            SuperscriptBox["t\[Chi]1", "3"], " ", 
            SuperscriptBox["u", "3"], " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", 
              RowBox[{"48", " ", "t\[Chi]3", " ", 
               SuperscriptBox["u", "2"]}], "-", 
              RowBox[{"160", " ", "u", " ", "\[Chi]eff"}], "-", 
              RowBox[{"16", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
           RowBox[{"3", " ", 
            SuperscriptBox["t\[Chi]1", "2"], " ", 
            SuperscriptBox["u", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"4", " ", "t\[Chi]3", " ", "u", " ", 
               RowBox[{"(", 
                RowBox[{"5", "+", 
                 RowBox[{"12", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
              RowBox[{"\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{"13", "-", 
                 RowBox[{"44", " ", "t\[Chi]4", " ", 
                  SuperscriptBox["u", "2"]}], "+", 
                 RowBox[{"144", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"16", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{"\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"4", " ", 
               SuperscriptBox["t\[Chi]3", "2"], " ", 
               SuperscriptBox["u", "2"]}], "-", 
              RowBox[{"4", " ", 
               SuperscriptBox["\[Chi]eff", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"3", "+", 
                 RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"2", " ", "t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{"6", "+", 
                 RowBox[{"2", " ", "t\[Chi]4", " ", 
                  SuperscriptBox["u", "2"]}], "+", 
                 RowBox[{"48", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"23", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{"t\[Chi]4", " ", 
               RowBox[{"(", 
                RowBox[{"5", "+", 
                 RowBox[{"88", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"46", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{"t\[Chi]1", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"8", " ", 
               SuperscriptBox["t\[Chi]3", "2"], " ", 
               SuperscriptBox["u", "2"]}], "+", 
              RowBox[{"2", " ", "t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", 
                 RowBox[{"90", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"70", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{"\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"44", " ", "t\[Chi]4", " ", "u", " ", 
                  RowBox[{"(", 
                   RowBox[{"2", "+", 
                    RowBox[{"3", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
                 RowBox[{"\[Chi]eff", " ", 
                  RowBox[{"(", 
                   RowBox[{"59", "+", 
                    RowBox[{"144", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"16", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox["q", "11"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"12", " ", "t\[Chi]2", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "17"}], "+", 
              RowBox[{"2", " ", "t\[Chi]3", " ", 
               SuperscriptBox["u", "2"]}], "+", 
              RowBox[{"40", " ", "u", " ", "\[Chi]eff"}], "-", 
              RowBox[{"32", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
           RowBox[{"\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"44", " ", 
               SuperscriptBox["t\[Chi]4", "2"], " ", 
               SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["t\[Chi]3", "2"], " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "3"}], "+", 
                 RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"88", " ", "u", " ", 
               SuperscriptBox["\[Chi]eff", "3"], " ", 
               RowBox[{"(", 
                RowBox[{"227", "+", 
                 RowBox[{"90", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"t\[Chi]4", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "17"}], "-", 
                 RowBox[{"121", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"3656", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"1080", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "-", 
              RowBox[{"t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "60"}], "+", 
                 RowBox[{"771", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"4000", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"904", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}], "+", 
                 RowBox[{"4", " ", "t\[Chi]4", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "4"}], "+", 
                    RowBox[{"13", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}], "+", 
           RowBox[{"2", " ", "t\[Chi]1", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "t\[Chi]3"}], " ", 
               RowBox[{"(", 
                RowBox[{"2", "+", 
                 RowBox[{"15", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"11", " ", "t\[Chi]4", " ", "u"}], "+", 
                 RowBox[{"4", " ", "\[Chi]eff", " ", 
                  RowBox[{"(", 
                   RowBox[{"19", "+", 
                    RowBox[{"126", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"24", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox["q", "14"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"12", " ", "t\[Chi]2", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{"4", "+", 
              RowBox[{"5", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
           RowBox[{"\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", "t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{"6", "+", 
                 RowBox[{"48", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"23", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{"t\[Chi]4", " ", 
               RowBox[{"(", 
                RowBox[{"5", "+", 
                 RowBox[{"88", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"46", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{"\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"t\[Chi]1", " ", "u"}], "+", 
                 RowBox[{"4", " ", "\[Chi]eff", " ", 
                  RowBox[{"(", 
                   RowBox[{"49", "+", 
                    RowBox[{"206", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"48", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox["q", "13"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"24", " ", "t\[Chi]2", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", 
              RowBox[{"15", " ", "u", " ", "\[Chi]eff"}], "+", 
              RowBox[{"2", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
           RowBox[{"\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"t\[Chi]4", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "5"}], "+", 
                 RowBox[{"275", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"408", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"40", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "-", 
              RowBox[{"t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{"12", "+", 
                 RowBox[{"361", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"416", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"40", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
              RowBox[{"4", " ", "\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"t\[Chi]1", " ", "u", " ", 
                  RowBox[{"(", 
                   RowBox[{"3", "+", 
                    RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
                 RowBox[{"\[Chi]eff", " ", 
                  RowBox[{"(", 
                   RowBox[{"91", "+", 
                    RowBox[{"860", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"264", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox["q", "5"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"4", " ", 
            SuperscriptBox["t\[Chi]1", "3"], " ", 
            SuperscriptBox["u", "3"], " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", 
              RowBox[{"20", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
           RowBox[{"24", " ", "t\[Chi]2", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "-", 
              RowBox[{"t\[Chi]3", " ", 
               SuperscriptBox["u", "2"]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["t\[Chi]1", "2"], " ", 
               SuperscriptBox["u", "4"]}], "-", 
              RowBox[{"15", " ", "u", " ", "\[Chi]eff"}], "-", 
              RowBox[{"2", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{"2", " ", "t\[Chi]1", " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "5"}], "+", 
                 RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
             ")"}]}], "-", 
           RowBox[{"6", " ", 
            SuperscriptBox["t\[Chi]1", "2"], " ", 
            SuperscriptBox["u", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"t\[Chi]3", " ", "u", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "5"}], "+", 
                 RowBox[{"96", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"8", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
              RowBox[{"2", " ", "\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "7"}], "+", 
                 RowBox[{"22", " ", "t\[Chi]4", " ", 
                  SuperscriptBox["u", "2"]}], "+", 
                 RowBox[{"360", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"80", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{"\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "16"}], " ", 
               SuperscriptBox["t\[Chi]3", "2"], " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"2", "+", 
                 RowBox[{"t\[Chi]4", " ", 
                  SuperscriptBox["u", "2"]}], "-", 
                 RowBox[{"29", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"16", " ", 
               SuperscriptBox["t\[Chi]3", "3"], " ", 
               SuperscriptBox["u", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", 
                 RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"16", " ", 
               SuperscriptBox["t\[Chi]4", "2"], " ", 
               SuperscriptBox["u", "3"], " ", "\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{"11", "+", 
                 RowBox[{"3", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["\[Chi]eff", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"91", "+", 
                 RowBox[{"860", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"264", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{"t\[Chi]4", " ", 
               RowBox[{"(", 
                RowBox[{"17", "+", 
                 RowBox[{"121", " ", "u", " ", "\[Chi]eff"}], "-", 
                 RowBox[{"3656", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"1080", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
              RowBox[{"t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{"60", "-", 
                 RowBox[{"771", " ", "u", " ", "\[Chi]eff"}], "-", 
                 RowBox[{"16", " ", 
                  SuperscriptBox["t\[Chi]4", "2"], " ", 
                  SuperscriptBox["u", "5"], " ", "\[Chi]eff"}], "-", 
                 RowBox[{"4000", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"904", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}], "+", 
                 RowBox[{"8", " ", "t\[Chi]4", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{"4", "+", 
                    RowBox[{"91", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}], "+", 
           RowBox[{"4", " ", "t\[Chi]1", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"4", " ", 
               SuperscriptBox["t\[Chi]3", "2"], " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"2", "+", 
                 RowBox[{"27", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
              RowBox[{"t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{"2", "-", 
                 RowBox[{"105", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"940", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"180", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
              RowBox[{"\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"8", " ", 
                  SuperscriptBox["t\[Chi]4", "2"], " ", 
                  SuperscriptBox["u", "4"], " ", "\[Chi]eff"}], "+", 
                 RowBox[{"t\[Chi]4", " ", "u", " ", 
                  RowBox[{"(", 
                   RowBox[{"77", "-", 
                    RowBox[{"462", " ", "u", " ", "\[Chi]eff"}], "-", 
                    RowBox[{"100", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                 RowBox[{"\[Chi]eff", " ", 
                  RowBox[{"(", 
                   RowBox[{"13", "+", 
                    RowBox[{"2430", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"704", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}]}], ")"}]}], "-", 
        RowBox[{"2", " ", 
         SuperscriptBox["q", "9"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", "6"}], " ", 
            SuperscriptBox["t\[Chi]1", "2"], " ", 
            SuperscriptBox["u", "2"], " ", "\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{"3", "+", 
              RowBox[{"8", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
           RowBox[{"24", " ", "t\[Chi]2", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", "t\[Chi]1", " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "5"}], "+", 
                 RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{
                 SuperscriptBox["u", "2"], "-", 
                 RowBox[{"16", " ", 
                  SuperscriptBox["u", "3"], " ", "\[Chi]eff"}]}], ")"}]}], 
              "+", 
              RowBox[{"2", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "7"}], "+", 
                 RowBox[{"35", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"6", " ", "t\[Chi]4", " ", 
                  SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "+", 
                 RowBox[{"50", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{"\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"16", " ", 
               SuperscriptBox["t\[Chi]3", "2"], " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"2", "-", 
                 RowBox[{"29", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
              RowBox[{"16", " ", 
               SuperscriptBox["t\[Chi]4", "2"], " ", 
               SuperscriptBox["u", "3"], " ", "\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{"11", "+", 
                 RowBox[{"3", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
              RowBox[{"66", " ", 
               SuperscriptBox["\[Chi]eff", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "13"}], "+", 
                 RowBox[{"528", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"224", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
              RowBox[{"t\[Chi]4", " ", 
               RowBox[{"(", 
                RowBox[{"55", "-", 
                 RowBox[{"1089", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"3888", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"1680", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
              RowBox[{"t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{"44", "-", 
                 RowBox[{"1143", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"7872", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"2128", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}], "+", 
                 RowBox[{"8", " ", "t\[Chi]4", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{"4", "+", 
                    RowBox[{"91", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}], "+", 
           RowBox[{"2", " ", "t\[Chi]1", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "6"}], "+", 
                 RowBox[{"195", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"424", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"24", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "-", 
              RowBox[{"\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"t\[Chi]4", " ", "u", " ", 
                  RowBox[{"(", 
                   RowBox[{"33", "+", 
                    RowBox[{"396", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"40", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                 RowBox[{"2", " ", "\[Chi]eff", " ", 
                  RowBox[{"(", 
                   RowBox[{"13", "+", 
                    RowBox[{"2430", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"704", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox["q", "10"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"3", " ", 
            SuperscriptBox["t\[Chi]1", "2"], " ", 
            SuperscriptBox["u", "2"], " ", "\[Chi]eff"}], "+", 
           RowBox[{"24", " ", "t\[Chi]2", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{"4", "+", 
              RowBox[{"5", " ", "t\[Chi]1", " ", 
               SuperscriptBox["u", "2"]}], "-", 
              RowBox[{"35", " ", "u", " ", "\[Chi]eff"}], "-", 
              RowBox[{"6", " ", "t\[Chi]4", " ", 
               SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "-", 
              RowBox[{"68", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{"2", " ", "t\[Chi]3", " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", 
                 RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
             ")"}]}], "+", 
           RowBox[{"\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"8", " ", 
               SuperscriptBox["t\[Chi]4", "2"], " ", 
               SuperscriptBox["u", "3"], " ", "\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{"22", "+", 
                 RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"52", " ", 
               SuperscriptBox["t\[Chi]3", "2"], " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", 
                 RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"44", " ", 
               SuperscriptBox["\[Chi]eff", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "13"}], "+", 
                 RowBox[{"690", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"288", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{"t\[Chi]4", " ", 
               RowBox[{"(", 
                RowBox[{"33", "-", 
                 RowBox[{"968", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"5094", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"1920", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "-", 
              RowBox[{"2", " ", "t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "22"}], "+", 
                 RowBox[{"16", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"3225", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"832", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}], "+", 
                 RowBox[{"2", " ", "t\[Chi]4", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    RowBox[{"78", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}], "+", 
           RowBox[{"t\[Chi]1", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", "t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", 
                 RowBox[{"90", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"70", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{"\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"44", " ", "t\[Chi]4", " ", "u", " ", 
                  RowBox[{"(", 
                   RowBox[{"2", "+", 
                    RowBox[{"3", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
                 RowBox[{"\[Chi]eff", " ", 
                  RowBox[{"(", 
                   RowBox[{"201", "+", 
                    RowBox[{"3920", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"976", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}]}], ")"}]}], "-", 
        RowBox[{"2", " ", 
         SuperscriptBox["q", "8"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", "3"}], " ", 
            SuperscriptBox["t\[Chi]1", "2"], " ", 
            SuperscriptBox["u", "2"], " ", "\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{"13", "+", 
              RowBox[{"144", " ", "u", " ", "\[Chi]eff"}], "+", 
              RowBox[{"16", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
           RowBox[{"24", " ", "t\[Chi]2", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "4"}], "+", 
              RowBox[{"35", " ", "u", " ", "\[Chi]eff"}], "+", 
              RowBox[{"68", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{"4", " ", "t\[Chi]3", " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"1", "-", 
                 RowBox[{"6", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"t\[Chi]1", " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"5", "+", 
                 RowBox[{"16", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
             ")"}]}], "+", 
           RowBox[{"\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"8", " ", 
               SuperscriptBox["t\[Chi]3", "3"], " ", 
               SuperscriptBox["u", "4"]}], "-", 
              RowBox[{"8", " ", 
               SuperscriptBox["t\[Chi]3", "2"], " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"7", "+", 
                 RowBox[{"t\[Chi]4", " ", 
                  SuperscriptBox["u", "2"]}], "+", 
                 RowBox[{"131", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              
              RowBox[{"t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{"99", "-", 
                 RowBox[{"1728", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"8232", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"2240", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}], "+", 
                 RowBox[{"8", " ", "t\[Chi]4", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{"7", "+", 
                    RowBox[{"91", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"5", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
              "-", 
              RowBox[{"4", " ", "\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"2", " ", 
                  SuperscriptBox["t\[Chi]4", "2"], " ", 
                  SuperscriptBox["u", "3"], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "22"}], "+", 
                    RowBox[{"15", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], 
                 "+", 
                 RowBox[{"11", " ", "\[Chi]eff", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "13"}], "+", 
                    RowBox[{"690", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"288", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}], "+", 
           RowBox[{"t\[Chi]1", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"8", " ", 
               SuperscriptBox["t\[Chi]3", "2"], " ", 
               SuperscriptBox["u", "2"]}], "+", 
              RowBox[{"t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "17"}], "+", 
                 RowBox[{"240", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"2288", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"288", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "-", 
              RowBox[{"\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"8", " ", "t\[Chi]4", " ", "u", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "22"}], "+", 
                    RowBox[{"231", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"40", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                 RowBox[{"\[Chi]eff", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "261"}], "+", 
                    RowBox[{"16416", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"5152", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox["q", "4"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"12", " ", "t\[Chi]2", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{"4", "+", 
              RowBox[{"10", " ", "t\[Chi]1", " ", 
               SuperscriptBox["u", "2"]}], "+", 
              RowBox[{"5", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
           RowBox[{"2", " ", 
            SuperscriptBox["t\[Chi]1", "3"], " ", 
            SuperscriptBox["u", "3"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "+", 
              RowBox[{"160", " ", "u", " ", "\[Chi]eff"}], "+", 
              RowBox[{"16", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
           RowBox[{"6", " ", 
            SuperscriptBox["t\[Chi]1", "2"], " ", 
            SuperscriptBox["u", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{"7", "-", 
                 RowBox[{"568", " ", "u", " ", "\[Chi]eff"}], "-", 
                 RowBox[{"120", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{"4", " ", "t\[Chi]3", " ", "u", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "5"}], "+", 
                 RowBox[{"36", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"4", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{"\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "8"}], " ", 
               SuperscriptBox["t\[Chi]3", "3"], " ", 
               SuperscriptBox["u", "4"]}], "-", 
              RowBox[{"4", " ", 
               SuperscriptBox["t\[Chi]3", "2"], " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"13", "+", 
                 RowBox[{"2", " ", "t\[Chi]4", " ", 
                  SuperscriptBox["u", "2"]}], "-", 
                 RowBox[{"26", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["t\[Chi]4", "2"], " ", 
               SuperscriptBox["u", "3"], " ", "\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{"22", "+", 
                 RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["\[Chi]eff", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"49", "+", 
                 RowBox[{"206", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"48", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
              RowBox[{"t\[Chi]4", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "25"}], "+", 
                 RowBox[{"352", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"1608", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"320", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
              RowBox[{"2", " ", "t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{"9", "-", 
                 RowBox[{"368", " ", "u", " ", "\[Chi]eff"}], "-", 
                 RowBox[{"838", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "-", 
                 RowBox[{"144", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}], "+", 
                 RowBox[{"2", " ", "t\[Chi]4", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "3"}], "+", 
                    RowBox[{"78", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}], "+", 
           RowBox[{"t\[Chi]1", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "16"}], " ", 
               SuperscriptBox["t\[Chi]3", "3"], " ", 
               SuperscriptBox["u", "4"]}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["t\[Chi]3", "2"], " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", 
                 RowBox[{"36", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"16", " ", 
                  SuperscriptBox["t\[Chi]4", "2"], " ", 
                  SuperscriptBox["u", "4"], " ", "\[Chi]eff"}], "-", 
                 RowBox[{"8", " ", "t\[Chi]4", " ", "u", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "22"}], "+", 
                    RowBox[{"231", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"40", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                 RowBox[{"\[Chi]eff", " ", 
                  RowBox[{"(", 
                   RowBox[{"201", "+", 
                    RowBox[{"3920", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"976", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
              "+", 
              RowBox[{"t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{"17", "-", 
                 RowBox[{"240", " ", "u", " ", "\[Chi]eff"}], "-", 
                 RowBox[{"2288", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"16", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"13", " ", "t\[Chi]4", " ", "\[Chi]eff"}], "-", 
                    RowBox[{"18", " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox["q", "3"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"12", " ", "t\[Chi]2", " ", "u"}], "+", 
           RowBox[{"8", " ", 
            SuperscriptBox["t\[Chi]1", "3"], " ", 
            SuperscriptBox["u", "3"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "+", 
              RowBox[{"60", " ", "u", " ", "\[Chi]eff"}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
           RowBox[{"6", " ", 
            SuperscriptBox["t\[Chi]1", "2"], " ", 
            SuperscriptBox["u", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"12", " ", 
               SuperscriptBox["t\[Chi]3", "2"], " ", 
               SuperscriptBox["u", "3"]}], "+", 
              RowBox[{"t\[Chi]3", " ", "u", " ", 
               RowBox[{"(", 
                RowBox[{"5", "-", 
                 RowBox[{"96", " ", "u", " ", "\[Chi]eff"}], "-", 
                 RowBox[{"8", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
              RowBox[{"4", " ", "\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{"1", "-", 
                 RowBox[{"11", " ", "t\[Chi]4", " ", 
                  SuperscriptBox["u", "2"]}], "+", 
                 RowBox[{"68", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"12", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{"\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"44", " ", 
               SuperscriptBox["t\[Chi]4", "2"], " ", 
               SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["t\[Chi]3", "2"], " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "3"}], "+", 
                 RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["\[Chi]eff", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"8", "+", 
                 RowBox[{"15", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"2", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
              RowBox[{"t\[Chi]4", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "5"}], "+", 
                 RowBox[{"275", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"408", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"40", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "-", 
              RowBox[{"t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{"12", "+", 
                 RowBox[{"361", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"416", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"40", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}], "+", 
                 RowBox[{"4", " ", "t\[Chi]4", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{"4", "-", 
                    RowBox[{"13", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}], "+", 
           RowBox[{"2", " ", "t\[Chi]1", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"4", " ", 
               SuperscriptBox["t\[Chi]3", "2"], " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", 
                 RowBox[{"9", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"4", " ", "\[Chi]eff", " ", 
                  RowBox[{"(", 
                   RowBox[{"19", "+", 
                    RowBox[{"126", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"24", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
                 RowBox[{"t\[Chi]4", " ", "u", " ", 
                  RowBox[{"(", 
                   RowBox[{"33", "+", 
                    RowBox[{"396", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"40", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
              "+", 
              RowBox[{"t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{"6", "-", 
                 RowBox[{"195", " ", "u", " ", "\[Chi]eff"}], "-", 
                 RowBox[{"424", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"4", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"13", " ", "t\[Chi]4", " ", "\[Chi]eff"}], "-", 
                    RowBox[{"6", " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}]}], ")"}]}], "-", 
        RowBox[{"2", " ", 
         SuperscriptBox["q", "7"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"12", " ", "t\[Chi]2", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{"17", "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["t\[Chi]3", "2"], " ", 
               SuperscriptBox["u", "4"]}], "-", 
              RowBox[{"40", " ", "u", " ", "\[Chi]eff"}], "-", 
              RowBox[{"24", " ", "t\[Chi]4", " ", 
               SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "+", 
              RowBox[{"32", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{"2", " ", "t\[Chi]3", " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"1", "-", 
                 RowBox[{"16", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"8", " ", "t\[Chi]1", " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"5", "+", 
                 RowBox[{"6", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
             ")"}]}], "-", 
           RowBox[{"6", " ", 
            SuperscriptBox["t\[Chi]1", "2"], " ", 
            SuperscriptBox["u", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "5"}], " ", "t\[Chi]3", " ", "u"}], "+", 
              RowBox[{"4", " ", "\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", 
                 RowBox[{"68", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"12", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{"\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "16"}], " ", 
               SuperscriptBox["t\[Chi]3", "3"], " ", 
               SuperscriptBox["u", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", 
                 RowBox[{"u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
              RowBox[{"40", " ", 
               SuperscriptBox["t\[Chi]4", "2"], " ", 
               SuperscriptBox["u", "3"], " ", "\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "11"}], "+", 
                 RowBox[{"4", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
              RowBox[{"16", " ", 
               SuperscriptBox["t\[Chi]3", "2"], " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"7", "+", 
                 RowBox[{"t\[Chi]4", " ", 
                  SuperscriptBox["u", "2"]}], "+", 
                 RowBox[{"85", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
              RowBox[{"88", " ", "u", " ", 
               SuperscriptBox["\[Chi]eff", "3"], " ", 
               RowBox[{"(", 
                RowBox[{"227", "+", 
                 RowBox[{"90", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"t\[Chi]4", " ", 
               RowBox[{"(", 
                RowBox[{"55", "-", 
                 RowBox[{"1089", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"3888", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"1680", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
              RowBox[{"t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{"44", "-", 
                 RowBox[{"1143", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"16", " ", 
                  SuperscriptBox["t\[Chi]4", "2"], " ", 
                  SuperscriptBox["u", "5"], " ", "\[Chi]eff"}], "+", 
                 RowBox[{"7872", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"2128", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}], 
           "-", 
           RowBox[{"4", " ", "t\[Chi]1", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", 
               SuperscriptBox["t\[Chi]3", "2"], " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", 
                 RowBox[{"9", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"t\[Chi]4", " ", "u", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "77"}], "+", 
                    RowBox[{"462", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"100", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                 RowBox[{"4", " ", "\[Chi]eff", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "27"}], "+", 
                    RowBox[{"1218", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"392", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
              "-", 
              RowBox[{"t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{"2", "-", 
                 RowBox[{"105", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"940", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"2", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"13", " ", "t\[Chi]4", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"90", " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox["q", "6"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"2", " ", 
            SuperscriptBox["t\[Chi]1", "3"], " ", 
            SuperscriptBox["u", "3"]}], "+", 
           RowBox[{"12", " ", "t\[Chi]2", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "12"}], "+", 
              RowBox[{"65", " ", "u", " ", "\[Chi]eff"}], "+", 
              RowBox[{"12", " ", "t\[Chi]4", " ", 
               SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}], "+", 
              RowBox[{"2", " ", "t\[Chi]1", " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "5"}], "+", 
                 RowBox[{"8", " ", "t\[Chi]3", " ", 
                  SuperscriptBox["u", "2"]}], "-", 
                 RowBox[{"16", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"4", " ", "t\[Chi]3", " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", 
                 RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
             ")"}]}], "-", 
           RowBox[{"6", " ", 
            SuperscriptBox["t\[Chi]1", "2"], " ", 
            SuperscriptBox["u", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", "t\[Chi]3", " ", "u", " ", 
               RowBox[{"(", 
                RowBox[{"5", "+", 
                 RowBox[{"12", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
              RowBox[{"\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "7"}], "+", 
                 RowBox[{"22", " ", "t\[Chi]4", " ", 
                  SuperscriptBox["u", "2"]}], "+", 
                 RowBox[{"568", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"120", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
           "+", 
           RowBox[{"t\[Chi]1", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"8", " ", 
               SuperscriptBox["t\[Chi]3", "2"], " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", 
                 RowBox[{"36", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["\[Chi]eff", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "261"}], "+", 
                 RowBox[{"16", " ", 
                  SuperscriptBox["t\[Chi]4", "2"], " ", 
                  SuperscriptBox["u", "4"]}], "+", 
                 RowBox[{"16416", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"5152", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
              RowBox[{"4", " ", "t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{"7", "-", 
                 RowBox[{"210", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"1090", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"4", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"13", " ", "t\[Chi]4", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"60", " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}], "+", 
           RowBox[{"\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"16", " ", 
               SuperscriptBox["t\[Chi]3", "3"], " ", 
               SuperscriptBox["u", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "3"}], "+", 
                 RowBox[{"2", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["t\[Chi]4", "2"], " ", 
               SuperscriptBox["u", "3"], " ", "\[Chi]eff", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "22"}], "+", 
                 RowBox[{"15", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["t\[Chi]3", "2"], " ", 
               SuperscriptBox["u", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"7", "+", 
                 RowBox[{"131", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
              
              RowBox[{"4", " ", 
               SuperscriptBox["\[Chi]eff", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"91", "+", 
                 RowBox[{"2442", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"880", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
              RowBox[{"t\[Chi]4", " ", 
               RowBox[{"(", 
                RowBox[{"33", "-", 
                 RowBox[{"968", " ", "u", " ", "\[Chi]eff"}], "+", 
                 RowBox[{"5094", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                 RowBox[{"1920", " ", 
                  SuperscriptBox["u", "3"], " ", 
                  SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
              RowBox[{"t\[Chi]3", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "32"}], " ", 
                  SuperscriptBox["t\[Chi]4", "2"], " ", 
                  SuperscriptBox["u", "5"], " ", "\[Chi]eff"}], "+", 
                 RowBox[{"8", " ", "t\[Chi]4", " ", 
                  SuperscriptBox["u", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{"7", "+", 
                    RowBox[{"91", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"5", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "22"}], "+", 
                    RowBox[{"16", " ", "u", " ", "\[Chi]eff"}], "+", 
                    RowBox[{"3225", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}], "+", 
                    RowBox[{"832", " ", 
                    SuperscriptBox["u", "3"], " ", 
                    SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], 
                ")"}]}]}], ")"}]}]}], ")"}]}]}], ")"}]}]}]}], ",", 
   RowBox[{
    FractionBox["1", 
     RowBox[{
      SuperscriptBox["q", "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "6"]}]], 
    RowBox[{"4", " ", 
     SuperscriptBox["u", "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"160", " ", "q", " ", 
        SuperscriptBox["t\[Chi]1", "3"], " ", 
        SuperscriptBox["u", "4"]}], "-", 
       RowBox[{"16", " ", 
        SuperscriptBox["q", "4"], " ", 
        SuperscriptBox["t\[Chi]3", "3"], " ", 
        SuperscriptBox["u", "4"]}], "-", 
       RowBox[{"8", " ", 
        SuperscriptBox["q", "2"], " ", 
        SuperscriptBox["t\[Chi]3", "2"], " ", 
        SuperscriptBox["u", "2"], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", 
          SuperscriptBox["q", "6"], "+", 
          RowBox[{
           SuperscriptBox["q", "5"], " ", 
           RowBox[{"(", 
            RowBox[{"2", "-", 
             RowBox[{"9", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
          RowBox[{"q", " ", 
           RowBox[{"(", 
            RowBox[{"2", "-", 
             RowBox[{"18", " ", "t\[Chi]1", " ", 
              SuperscriptBox["u", "2"]}], "-", 
             RowBox[{"9", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
          RowBox[{"2", " ", 
           SuperscriptBox["q", "3"], " ", 
           RowBox[{"(", 
            RowBox[{"2", "+", 
             RowBox[{"27", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
          RowBox[{
           SuperscriptBox["q", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"1", "+", 
             RowBox[{"36", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
          RowBox[{
           SuperscriptBox["q", "4"], " ", 
           RowBox[{"(", 
            RowBox[{"1", "+", 
             RowBox[{"36", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], ")"}]}],
        "+", 
       RowBox[{"6", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
        SuperscriptBox["t\[Chi]1", "2"], " ", 
        SuperscriptBox["u", "2"], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", 
          SuperscriptBox["q", "4"], "+", 
          RowBox[{"40", " ", "q", " ", "u", " ", "\[Chi]eff"}], "+", 
          RowBox[{"40", " ", 
           SuperscriptBox["q", "3"], " ", "u", " ", "\[Chi]eff"}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["q", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             RowBox[{"40", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{"6", " ", "t\[Chi]1", " ", "u", " ", 
        RowBox[{"(", 
         RowBox[{"\[Chi]eff", "+", 
          RowBox[{
           SuperscriptBox["q", "10"], " ", "\[Chi]eff"}], "+", 
          RowBox[{"2", " ", "q", " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{"3", "+", 
             RowBox[{"8", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["q", "9"], " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{"3", "+", 
             RowBox[{"8", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
          RowBox[{"8", " ", 
           SuperscriptBox["q", "7"], " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{"1", "+", 
             RowBox[{"68", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"12", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{"8", " ", 
           SuperscriptBox["q", "3"], " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"11", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "2"]}], "+", 
             RowBox[{"68", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"12", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["q", "8"], " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{"13", "+", 
             RowBox[{"144", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"16", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["q", "2"], " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{"13", "-", 
             RowBox[{"44", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "2"]}], "+", 
             RowBox[{"144", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"16", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["q", "4"], " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "7"}], "+", 
             RowBox[{"568", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"120", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["q", "6"], " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "7"}], "+", 
             RowBox[{"22", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "2"]}], "+", 
             RowBox[{"568", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"120", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
          RowBox[{"4", " ", 
           SuperscriptBox["q", "5"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"8", " ", "t\[Chi]2", " ", 
              SuperscriptBox["u", "3"]}], "-", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "7"}], "+", 
                RowBox[{"22", " ", "t\[Chi]4", " ", 
                 SuperscriptBox["u", "2"]}], "+", 
                RowBox[{"360", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"80", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}]}], 
         ")"}]}], "-", 
       RowBox[{"t\[Chi]3", " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", 
          SuperscriptBox["q", "12"], "+", 
          RowBox[{
           SuperscriptBox["q", "11"], " ", 
           RowBox[{"(", 
            RowBox[{"4", "+", 
             RowBox[{"30", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
          RowBox[{"q", " ", 
           RowBox[{"(", 
            RowBox[{"4", "+", 
             RowBox[{"60", " ", "t\[Chi]1", " ", 
              SuperscriptBox["u", "2"]}], "+", 
             RowBox[{"30", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["q", "10"], " ", 
           RowBox[{"(", 
            RowBox[{"1", "+", 
             RowBox[{"90", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"70", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["q", "9"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "6"}], "+", 
             RowBox[{"195", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"424", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"24", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["q", "8"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "17"}], "+", 
             RowBox[{"240", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"2288", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"288", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}], "+", 
          RowBox[{"4", " ", 
           SuperscriptBox["q", "6"], " ", 
           RowBox[{"(", 
            RowBox[{"7", "-", 
             RowBox[{"48", " ", "t\[Chi]2", " ", 
              SuperscriptBox["u", "4"]}], "-", 
             RowBox[{"210", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"52", " ", "t\[Chi]4", " ", 
              SuperscriptBox["u", "3"], " ", "\[Chi]eff"}], "+", 
             RowBox[{"1090", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"240", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}], "+", 
             RowBox[{"6", " ", "t\[Chi]1", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"5", "+", 
                RowBox[{"12", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
            ")"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["q", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"1", "+", 
             RowBox[{"144", " ", 
              SuperscriptBox["t\[Chi]1", "2"], " ", 
              SuperscriptBox["u", "4"]}], "+", 
             RowBox[{"90", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"70", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"12", " ", "t\[Chi]1", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"5", "+", 
                RowBox[{"12", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
            ")"}]}], "+", 
          RowBox[{"4", " ", 
           SuperscriptBox["q", "5"], " ", 
           RowBox[{"(", 
            RowBox[{"2", "-", 
             RowBox[{"105", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"940", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"180", " ", 
              SuperscriptBox["u", "3"], " ", 
              SuperscriptBox["\[Chi]eff", "3"]}], "+", 
             RowBox[{"3", " ", "t\[Chi]1", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "5"}], "+", 
                RowBox[{"96", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"8", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["q", "3"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "6"}], "+", 
             RowBox[{"195", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"424", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"6", " ", "t\[Chi]1", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "5"}], "+", 
                RowBox[{"96", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"8", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox["u", "3"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "52"}], " ", "t\[Chi]4", " ", "\[Chi]eff"}], 
                "+", 
                RowBox[{"24", " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{"4", " ", 
           SuperscriptBox["q", "7"], " ", 
           RowBox[{"(", 
            RowBox[{"2", "+", 
             RowBox[{"15", " ", "t\[Chi]1", " ", 
              SuperscriptBox["u", "2"]}], "-", 
             RowBox[{"105", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"940", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"2", " ", 
              SuperscriptBox["u", "3"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"13", " ", "t\[Chi]4", " ", "\[Chi]eff"}], "+", 
                RowBox[{"90", " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{
           SuperscriptBox["q", "4"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "17"}], "+", 
             RowBox[{"240", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"2288", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}], "+", 
             RowBox[{"48", " ", "t\[Chi]1", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "5"}], "+", 
                RowBox[{"36", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox["u", "3"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "208"}], " ", "t\[Chi]4", " ", "\[Chi]eff"}], 
                "+", 
                RowBox[{"288", " ", 
                 SuperscriptBox["\[Chi]eff", "3"]}]}], ")"}]}]}], ")"}]}]}], 
         ")"}]}], "+", 
       RowBox[{
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox["\[Chi]eff", "2"], "+", 
          RowBox[{
           SuperscriptBox["q", "12"], " ", 
           SuperscriptBox["\[Chi]eff", "2"]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["q", "11"], " ", 
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"5", "+", 
             RowBox[{"4", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["q", "10"], " ", 
           SuperscriptBox["\[Chi]eff", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"19", "+", 
             RowBox[{"64", " ", "u", " ", "\[Chi]eff"}], "+", 
             RowBox[{"8", " ", 
              SuperscriptBox["u", "2"], " ", 
              SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
          RowBox[{"2", " ", "q", " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "11"}], " ", "t\[Chi]4", " ", "u"}], "+", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{"5", "+", 
                RowBox[{"4", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], 
            ")"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["q", "2"], " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "22"}], " ", "t\[Chi]4", " ", "u", " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", 
                RowBox[{"3", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{"19", "+", 
                RowBox[{"64", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"8", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["q", "9"], " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"11", " ", "t\[Chi]4", " ", "u"}], "+", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{"33", "+", 
                RowBox[{"372", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"80", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["q", "3"], " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "2"}], " ", "t\[Chi]4", " ", "u", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "11"}], "+", 
                RowBox[{"132", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"20", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{"33", "+", 
                RowBox[{"372", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"80", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{"4", " ", 
           SuperscriptBox["q", "5"], " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "24"}], " ", "t\[Chi]2", " ", 
              SuperscriptBox["u", "3"]}], "+", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "19"}], "+", 
                RowBox[{"1092", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"344", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "+", 
          RowBox[{"4", " ", 
           SuperscriptBox["q", "7"], " ", "\[Chi]eff", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "24"}], " ", "t\[Chi]2", " ", 
              SuperscriptBox["u", "3"]}], "+", 
             RowBox[{"t\[Chi]4", " ", "u", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "11"}], "+", 
                RowBox[{"132", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"20", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "19"}], "+", 
                RowBox[{"1092", " ", "u", " ", "\[Chi]eff"}], "+", 
                RowBox[{"344", " ", 
                 SuperscriptBox["u", "2"], " ", 
                 SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], 
          "-", 
          RowBox[{"4", " ", 
           SuperscriptBox["q", "6"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"12", " ", "t\[Chi]2", " ", 
              SuperscriptBox["u", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"5", "+", 
                RowBox[{"4", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"t\[Chi]4", " ", "u", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "33"}], "+", 
                   RowBox[{"165", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"40", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                RowBox[{"\[Chi]eff", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "35"}], "+", 
                   RowBox[{"1344", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"440", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["q", "8"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"120", " ", "t\[Chi]2", " ", 
              SuperscriptBox["u", "2"]}], "+", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"44", " ", "t\[Chi]4", " ", "u", " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "+", 
                   RowBox[{"3", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
                
                RowBox[{"\[Chi]eff", " ", 
                 RowBox[{"(", 
                  RowBox[{"31", "+", 
                   RowBox[{"2304", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"640", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["q", "4"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"120", " ", "t\[Chi]2", " ", 
              SuperscriptBox["u", "2"]}], "+", 
             RowBox[{"\[Chi]eff", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"16", " ", 
                 SuperscriptBox["t\[Chi]4", "2"], " ", 
                 SuperscriptBox["u", "4"], " ", "\[Chi]eff"}], "-", 
                RowBox[{"4", " ", "t\[Chi]4", " ", "u", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "33"}], "+", 
                   RowBox[{"165", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"40", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
                RowBox[{"\[Chi]eff", " ", 
                 RowBox[{"(", 
                  RowBox[{"31", "+", 
                   RowBox[{"2304", " ", "u", " ", "\[Chi]eff"}], "+", 
                   RowBox[{"640", " ", 
                    SuperscriptBox["u", "2"], " ", 
                    SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], 
               ")"}]}]}], ")"}]}]}], ")"}]}]}], ")"}]}]}], ",", 
   RowBox[{"-", 
    RowBox[{
     FractionBox["1", 
      RowBox[{
       SuperscriptBox["q", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"]}]], 
     RowBox[{"8", " ", 
      SuperscriptBox["u", "2"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"80", " ", "q", " ", 
         SuperscriptBox["t\[Chi]1", "2"], " ", 
         SuperscriptBox["u", "3"]}], "+", "\[Chi]eff", "+", 
        RowBox[{
         SuperscriptBox["q", "10"], " ", "\[Chi]eff"}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox["q", "9"], " ", "\[Chi]eff", " ", 
         RowBox[{"(", 
          RowBox[{"3", "+", 
           RowBox[{"8", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
        RowBox[{
         SuperscriptBox["q", "8"], " ", "\[Chi]eff", " ", 
         RowBox[{"(", 
          RowBox[{"13", "+", 
           RowBox[{"144", " ", "u", " ", "\[Chi]eff"}], "+", 
           RowBox[{"16", " ", 
            SuperscriptBox["u", "2"], " ", 
            SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
        RowBox[{"q", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", "10"}], " ", "t\[Chi]3", " ", "u"}], "+", 
           RowBox[{"2", " ", "\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{"3", "+", 
              RowBox[{"8", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], ")"}]}],
         "-", 
        RowBox[{"2", " ", 
         SuperscriptBox["q", "4"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{"7", "-", 
              RowBox[{"568", " ", "u", " ", "\[Chi]eff"}], "-", 
              RowBox[{"120", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
           RowBox[{"4", " ", "t\[Chi]3", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "5"}], "+", 
              RowBox[{"36", " ", "u", " ", "\[Chi]eff"}], "+", 
              RowBox[{"4", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], "+", 
        
        RowBox[{
         SuperscriptBox["q", "7"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", "10"}], " ", "t\[Chi]3", " ", "u"}], "+", 
           RowBox[{"8", " ", "\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", 
              RowBox[{"68", " ", "u", " ", "\[Chi]eff"}], "+", 
              RowBox[{"12", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], "+", 
        
        RowBox[{"2", " ", 
         SuperscriptBox["q", "3"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"12", " ", 
            SuperscriptBox["t\[Chi]3", "2"], " ", 
            SuperscriptBox["u", "3"]}], "+", 
           RowBox[{"t\[Chi]3", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{"5", "-", 
              RowBox[{"96", " ", "u", " ", "\[Chi]eff"}], "-", 
              RowBox[{"8", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
           RowBox[{"4", " ", "\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{"1", "-", 
              RowBox[{"11", " ", "t\[Chi]4", " ", 
               SuperscriptBox["u", "2"]}], "+", 
              RowBox[{"68", " ", "u", " ", "\[Chi]eff"}], "+", 
              RowBox[{"12", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], "+", 
        
        RowBox[{
         SuperscriptBox["q", "2"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", "4"}], " ", "t\[Chi]3", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{"5", "+", 
              RowBox[{"12", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
           RowBox[{"\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{"13", "-", 
              RowBox[{"44", " ", "t\[Chi]4", " ", 
               SuperscriptBox["u", "2"]}], "+", 
              RowBox[{"144", " ", "u", " ", "\[Chi]eff"}], "+", 
              RowBox[{"16", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], "+", 
        
        RowBox[{"2", " ", "t\[Chi]1", " ", "u", " ", 
         RowBox[{"(", 
          RowBox[{"1", "+", 
           SuperscriptBox["q", "6"], "+", 
           RowBox[{"q", " ", 
            RowBox[{"(", 
             RowBox[{"2", "+", 
              RowBox[{"40", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
           RowBox[{
            SuperscriptBox["q", "5"], " ", 
            RowBox[{"(", 
             RowBox[{"2", "+", 
              RowBox[{"40", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
           RowBox[{"4", " ", 
            SuperscriptBox["q", "3"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "+", 
              RowBox[{"60", " ", "u", " ", "\[Chi]eff"}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
           RowBox[{
            SuperscriptBox["q", "4"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "+", 
              RowBox[{"160", " ", "u", " ", "\[Chi]eff"}], "+", 
              RowBox[{"16", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
           RowBox[{
            SuperscriptBox["q", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "-", 
              RowBox[{"48", " ", "t\[Chi]3", " ", 
               SuperscriptBox["u", "2"]}], "+", 
              RowBox[{"160", " ", "u", " ", "\[Chi]eff"}], "+", 
              RowBox[{"16", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], "-", 
        
        RowBox[{"2", " ", 
         SuperscriptBox["q", "5"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"16", " ", "t\[Chi]2", " ", 
            SuperscriptBox["u", "3"]}], "+", 
           RowBox[{"t\[Chi]3", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "5"}], "+", 
              RowBox[{"96", " ", "u", " ", "\[Chi]eff"}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "-", 
           RowBox[{"2", " ", "\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "7"}], "+", 
              RowBox[{"22", " ", "t\[Chi]4", " ", 
               SuperscriptBox["u", "2"]}], "+", 
              RowBox[{"360", " ", "u", " ", "\[Chi]eff"}], "+", 
              RowBox[{"80", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}], "-", 
        
        RowBox[{"2", " ", 
         SuperscriptBox["q", "6"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"2", " ", "t\[Chi]3", " ", "u", " ", 
            RowBox[{"(", 
             RowBox[{"5", "+", 
              RowBox[{"12", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "-", 
           RowBox[{"\[Chi]eff", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "7"}], "+", 
              RowBox[{"22", " ", "t\[Chi]4", " ", 
               SuperscriptBox["u", "2"]}], "+", 
              RowBox[{"568", " ", "u", " ", "\[Chi]eff"}], "+", 
              RowBox[{"120", " ", 
               SuperscriptBox["u", "2"], " ", 
               SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}]}], 
       ")"}]}]}]}], ",", 
   RowBox[{
    FractionBox["1", 
     SuperscriptBox["q", "2"]], 
    RowBox[{"4", " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
     SuperscriptBox["u", "2"], " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", 
       SuperscriptBox["q", "6"], "+", 
       RowBox[{
        SuperscriptBox["q", "5"], " ", 
        RowBox[{"(", 
         RowBox[{"2", "+", 
          RowBox[{"40", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
       RowBox[{"q", " ", 
        RowBox[{"(", 
         RowBox[{"2", "+", 
          RowBox[{"80", " ", "t\[Chi]1", " ", 
           SuperscriptBox["u", "2"]}], "+", 
          RowBox[{"40", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}], "+", 
       RowBox[{"4", " ", 
        SuperscriptBox["q", "3"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", 
          RowBox[{"60", " ", "u", " ", "\[Chi]eff"}], "+", 
          RowBox[{"8", " ", 
           SuperscriptBox["u", "2"], " ", 
           SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox["q", "4"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", 
          RowBox[{"160", " ", "u", " ", "\[Chi]eff"}], "+", 
          RowBox[{"16", " ", 
           SuperscriptBox["u", "2"], " ", 
           SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox["q", "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "-", 
          RowBox[{"48", " ", "t\[Chi]3", " ", 
           SuperscriptBox["u", "2"]}], "+", 
          RowBox[{"160", " ", "u", " ", "\[Chi]eff"}], "+", 
          RowBox[{"16", " ", 
           SuperscriptBox["u", "2"], " ", 
           SuperscriptBox["\[Chi]eff", "2"]}]}], ")"}]}]}], ")"}]}]}], ",", 
   RowBox[{"-", 
    FractionBox[
     RowBox[{"64", " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "6"], " ", 
      SuperscriptBox["u", "3"]}], "q"]}]}], "}"}]], "Output",
 CellLabel->
  "Out[99]=",ExpressionUUID->"b18c03a4-ca6a-4e19-9020-2dc4e73baba2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"myc", "[", 
    RowBox[{"[", "6", "]"}], "]"}], "/", "overalldiscr"}], "//", 
  "Simplify"}]], "Input",
 CellLabel->
  "In[100]:=",ExpressionUUID->"7e796557-c103-44ab-956e-588fe8dd6f45"],

Cell[BoxData[
 RowBox[{"-", "u"}]], "Output",
 CellLabel->"Out[100]=",ExpressionUUID->"2d025f83-eea5-479f-8dba-c05f45095b15"]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]56", "=", "0"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[Delta]55", "=", "0"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[Delta]54", "=", "0"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[Delta]53", "=", "0"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[Delta]52", "=", "0"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[Delta]51", "=", 
   RowBox[{"-", "1"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[Delta]50", "=", "0"}], ";"}]}], "Input",
 CellLabel->
  "In[101]:=",ExpressionUUID->"fac4133f-0179-4aef-88ad-c9d5482f3a6f"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]46", "=", "0"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[Delta]45", "=", "0"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[Delta]44", "=", "0"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[Delta]43", "=", "0"}], ";"}]}], "Input",
 CellLabel->
  "In[108]:=",ExpressionUUID->"f2ea867f-df0b-4edd-931a-bc7a16cdf85e"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]42", "=", 
   RowBox[{
    RowBox[{
     FractionBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{"5", "-", 
         RowBox[{"3", " ", "q"}]}], ")"}], " "}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "4"]], 
     SuperscriptBox["\[Chi]1", "2"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "3"], " ", 
       RowBox[{"(", 
        RowBox[{"3", "-", 
         RowBox[{"5", " ", "q"}]}], ")"}], " "}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "4"]], 
     SuperscriptBox["\[Chi]2", "2"]}], "+", 
    RowBox[{
     FractionBox["q", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"]], "  ", 
     SuperscriptBox["\[Chi]eff", "2"]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]42", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "5", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "3", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[112]:=",ExpressionUUID->"a6da1893-5238-423c-8314-12ea05be349d"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[113]=",ExpressionUUID->"64adf3d9-9896-4eb1-9dc5-0b83355d343a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]41", "=", 
   FractionBox[
    RowBox[{"5", " ", "\[Chi]eff"}], "2"]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]41", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "5", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "2", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[114]:=",ExpressionUUID->"2c364196-11da-481e-b637-78efd911ed8e"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[115]=",ExpressionUUID->"714fa5f1-10f7-48e7-861a-93002fe661f3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]40", "=", 
   FractionBox[
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"1", "-", "q"}], ")"}], "2"], 
    RowBox[{"16", " ", "q"}]]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]40", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "5", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "1", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[116]:=",ExpressionUUID->"a107cf88-c7b4-4145-a5d0-1f8f8fad1733"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[117]=",ExpressionUUID->"43159c21-c5a1-474d-835e-90879b6c1730"]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]36", "=", "0"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[Delta]35", "=", "0"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[Delta]34", "=", "0"}], ";"}]}], "Input",
 CellLabel->
  "In[118]:=",ExpressionUUID->"51643b50-9206-4184-bbc6-29def066f90a"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]33", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{"(", 
        RowBox[{"10", "-", 
         RowBox[{"12", " ", "q"}], "+", 
         RowBox[{"3", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "8"]]}], " ", 
     SuperscriptBox["\[Chi]1", "4"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{"2", " ", 
       SuperscriptBox["q", "3"], " ", 
       RowBox[{"(", 
        RowBox[{"6", "-", 
         RowBox[{"11", " ", "q"}], "+", 
         RowBox[{"6", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}], " "}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "8"]], 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "2"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "6"], " ", 
       RowBox[{"(", 
        RowBox[{"3", "-", 
         RowBox[{"12", " ", "q"}], "+", 
         RowBox[{"10", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "8"]], " ", 
     SuperscriptBox["\[Chi]2", "4"]}], "+", " ", 
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{"2", " ", "q", " ", 
        RowBox[{"(", 
         RowBox[{"2", "-", "q"}], ")"}]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "6"]]}], " ", 
     SuperscriptBox["\[Chi]1", "2"], 
     SuperscriptBox["\[Chi]eff", "2"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{"2", " ", 
       SuperscriptBox["q", "4"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"2", " ", "q"}]}], ")"}], " "}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "6"]], " ", 
     SuperscriptBox["\[Chi]2", "2"], 
     SuperscriptBox["\[Chi]eff", "2"]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]33", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "4", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "4", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[121]:=",ExpressionUUID->"3e496d32-5a07-4d66-b130-c96833a67f0e"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[122]=",ExpressionUUID->"d3e65d0f-ea41-4f93-8be7-897d356dcb3c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]32", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{"20", "-", 
        RowBox[{"3", " ", "q"}], "-", 
        SuperscriptBox["q", "2"]}], 
       RowBox[{"2", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "5"]}]]}], " ", 
     SuperscriptBox["\[Chi]1", "2"], "\[Chi]eff"}], " ", "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "3"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", 
         RowBox[{"3", " ", "q"}], "-", 
         RowBox[{"20", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "5"]}]], " ", 
     SuperscriptBox["\[Chi]2", "2"], "\[Chi]eff"}], " ", "-", 
    RowBox[{
     FractionBox[
      RowBox[{"2", " ", "q", " "}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"]], " ", 
     SuperscriptBox["\[Chi]eff", "3"]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]32", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "4", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "3", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[123]:=",ExpressionUUID->"f56020c2-b525-41ae-bcdf-648dc26839bb"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[124]=",ExpressionUUID->"74fabecc-3fa7-4bbf-822b-956ac2f76a06"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]31", "=", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"5", " ", "q"}]}], ")"}], " ", 
       SuperscriptBox["\[Chi]1", "2"]}], 
      RowBox[{"4", " ", "q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "4"]}]]}], "+", " ", 
    FractionBox[
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
      RowBox[{"(", 
       RowBox[{"5", "-", "q"}], ")"}], " ", 
      SuperscriptBox["q", "2"], " ", 
      SuperscriptBox["\[Chi]2", "2"]}], 
     RowBox[{"4", "  ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "4"]}]], "-", 
    FractionBox[
     RowBox[{"2", " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", 
        RowBox[{"3", "q"}], " ", "+", 
        SuperscriptBox["q", "2"]}], ")"}], " ", 
      SuperscriptBox["\[Chi]eff", "2"]}], 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "2"]]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]31", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "4", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "2", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[168]:=",ExpressionUUID->"03a0427f-ed5b-4309-8731-236e6217b8e9"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[169]=",ExpressionUUID->"953e2047-b75a-4003-b2c0-0356a1db63c5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]30", "=", 
   RowBox[{"-", 
    FractionBox[
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], "2"], " ", "\[Chi]eff"}], 
     RowBox[{"8", " ", "q"}]]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]30", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "4", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "1", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[127]:=",ExpressionUUID->"5b97dcc4-0f76-4637-8eff-4e0cddcb0349"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[128]=",ExpressionUUID->"4483814e-4b57-4590-aea4-6d019066c005"]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]26", "=", "0"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[Delta]25", "=", "0"}], ";"}]}], "Input",
 CellLabel->
  "In[129]:=",ExpressionUUID->"50ccc355-aab3-42b1-bc47-4475fed74777"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]24", "=", " ", 
   RowBox[{
    RowBox[{
     FractionBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], 
       RowBox[{"(", 
        RowBox[{"10", "-", 
         RowBox[{"8", " ", "q"}], "+", 
         SuperscriptBox["q", "2"]}], ")"}], " "}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "12"]], 
     SuperscriptBox["\[Chi]1", "6"]}], "-", " ", 
    RowBox[{
     FractionBox[
      RowBox[{"9", " ", 
       SuperscriptBox["q", "3"], 
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], 
       RowBox[{"(", 
        RowBox[{"2", "-", 
         RowBox[{"2", " ", "q"}], "+", 
         SuperscriptBox["q", "2"]}], ")"}], " "}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "12"]], 
     SuperscriptBox["\[Chi]1", "4"], " ", 
     SuperscriptBox["\[Chi]2", "2"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{"9", " ", 
       SuperscriptBox["q", "6"], 
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"2", " ", "q"}], "+", 
         RowBox[{"2", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "12"]], "  ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "4"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "9"], 
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"8", " ", "q"}], "+", 
         RowBox[{"10", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "12"]], "  ", 
     SuperscriptBox["\[Chi]2", "6"]}], "+", "\[IndentingNewLine]", " ", 
    RowBox[{
     FractionBox[
      RowBox[{"q", 
       RowBox[{"(", 
        RowBox[{"6", "-", 
         RowBox[{"6", " ", "q"}], "+", 
         SuperscriptBox["q", "2"]}], ")"}], " "}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "10"]], 
     SuperscriptBox["\[Chi]1", "4"], 
     SuperscriptBox["\[Chi]eff", "2"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{"2", " ", 
       SuperscriptBox["q", "4"], " ", 
       RowBox[{"(", 
        RowBox[{"3", "-", 
         RowBox[{"5", " ", "q"}], "+", 
         RowBox[{"3", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "10"]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "2"], 
     SuperscriptBox["\[Chi]eff", "2"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "7"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"6", " ", "q"}], "+", 
         RowBox[{"6", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}], " "}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "10"]], 
     SuperscriptBox["\[Chi]2", "4"], 
     SuperscriptBox["\[Chi]eff", "2"]}]}]}], " ", 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]24", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "3", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "5", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[131]:=",ExpressionUUID->"4f92b936-d1a5-490c-a8a9-cc95878b1f66"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[132]=",ExpressionUUID->"4502125a-6060-42b3-a76c-eb9681f6f63d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]23", "=", " ", 
   RowBox[{
    RowBox[{
     FractionBox[
      RowBox[{"(", 
       RowBox[{"30", "-", 
        RowBox[{"39", " ", "q"}], "+", 
        RowBox[{"19", " ", 
         SuperscriptBox["q", "2"]}], "-", 
        RowBox[{"4", " ", 
         SuperscriptBox["q", "3"]}]}], ")"}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "9"]}]], " ", 
     SuperscriptBox["\[Chi]1", "4"], "\[Chi]eff"}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{"3", " ", 
       SuperscriptBox["q", "3"], "  ", 
       RowBox[{"(", 
        RowBox[{"1", "+", 
         SuperscriptBox["q", "2"]}], ")"}], " "}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "8"]}]], 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "2"], "\[Chi]eff"}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "6"], " ", 
       RowBox[{"(", 
        RowBox[{"4", "-", 
         RowBox[{"19", " ", "q"}], "+", 
         RowBox[{"39", " ", 
          SuperscriptBox["q", "2"]}], "-", 
         RowBox[{"30", " ", 
          SuperscriptBox["q", "3"]}]}], ")"}], " "}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "9"]}]], 
     SuperscriptBox["\[Chi]2", "4"], "\[Chi]eff"}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{"2", "q", " ", 
       RowBox[{"(", 
        RowBox[{"3", "-", "q", "+", 
         SuperscriptBox["q", "2"]}], ")"}], " "}], 
      RowBox[{" ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "7"]}]], " ", 
     SuperscriptBox["\[Chi]1", "2"], 
     SuperscriptBox["\[Chi]eff", "3"]}], " ", "+", 
    RowBox[{
     FractionBox[
      RowBox[{"2", 
       SuperscriptBox["q", "4"], "  ", 
       RowBox[{"(", 
        RowBox[{"1", "-", "q", "+", 
         RowBox[{"3", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}], " "}], 
      RowBox[{" ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "7"]}]], " ", 
     SuperscriptBox["\[Chi]2", "2"], 
     SuperscriptBox["\[Chi]eff", "3"]}]}]}], "   ", 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]23", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "3", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "4", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"(*", 
   RowBox[{
    RowBox[{
     RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
    "TeXForm"}], "*)"}]}]}], "Input",
 CellLabel->
  "In[133]:=",ExpressionUUID->"80e0cd68-243d-4abc-94c1-dc2203912b86"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[134]=",ExpressionUUID->"63123de9-bc10-492e-aa2b-32519ee3e53f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]22", "=", " ", 
   RowBox[{
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], 
       RowBox[{"(", 
        RowBox[{"3", "-", 
         RowBox[{"30", " ", "q"}], "-", 
         RowBox[{"4", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      RowBox[{"8", " ", "q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "8"]}]], " ", 
     SuperscriptBox["\[Chi]1", "4"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{" ", 
       RowBox[{
        SuperscriptBox["q", "2"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "-", "q"}], ")"}], "2"], 
        RowBox[{"(", 
         RowBox[{"15", "-", 
          RowBox[{"29", " ", "q"}], "+", 
          RowBox[{"15", " ", 
           SuperscriptBox["q", "2"]}]}], ")"}]}]}], 
      RowBox[{"4", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "8"]}]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "2"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "5"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], 
       RowBox[{"(", 
        RowBox[{"4", "+", 
         RowBox[{"30", " ", "q"}], "-", 
         RowBox[{"3", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}], " "}], 
      RowBox[{"8", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "8"]}]], 
     SuperscriptBox["\[Chi]2", "4"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{"12", "+", 
       RowBox[{"2", " ", "q"}], "+", 
       RowBox[{"11", " ", 
        SuperscriptBox["q", "2"]}], "-", " ", 
       SuperscriptBox["q", "3"]}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "6"]}]], " ", 
     SuperscriptBox["\[Chi]1", "2"], 
     SuperscriptBox["\[Chi]eff", "2"]}], "-", " ", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "3"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"11", " ", "q"}], "-", 
         RowBox[{"2", " ", 
          SuperscriptBox["q", "2"]}], "-", 
         RowBox[{"12", " ", 
          SuperscriptBox["q", "3"]}]}], ")"}]}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "6"]}]], " ", 
     SuperscriptBox["\[Chi]2", "2"], 
     SuperscriptBox["\[Chi]eff", "2"]}], " ", "+", 
    RowBox[{
     FractionBox[
      RowBox[{" ", 
       RowBox[{"q", " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", 
          RowBox[{"4", "q"}], "+", " ", 
          SuperscriptBox["q", "2"]}], ")"}], " "}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "4"]], " ", 
     SuperscriptBox["\[Chi]eff", "4"]}]}]}], " ", 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]22", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "3", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "3", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[172]:=",ExpressionUUID->"e682939d-8a7d-4a4e-9bf0-c9ecec996627"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[173]=",ExpressionUUID->"e39d39bb-86ec-46a2-a1a9-0335d81eae1d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]21", "=", " ", 
   RowBox[{
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], 
       RowBox[{"(", 
        RowBox[{"3", "-", 
         RowBox[{"23", " ", "q"}], "-", 
         RowBox[{"4", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      RowBox[{"8", " ", "q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "5"]}]], " ", 
     SuperscriptBox["\[Chi]1", "2"], "\[Chi]eff"}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], 
       RowBox[{"(", 
        RowBox[{"4", "+", 
         RowBox[{"23", " ", "q"}], "-", 
         RowBox[{"3", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      RowBox[{"8", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "5"]}]], " ", 
     SuperscriptBox["\[Chi]2", "2"], "\[Chi]eff"}], " ", "+", 
    RowBox[{
     FractionBox[
      RowBox[{" ", 
       RowBox[{"1", "+", 
        RowBox[{"8", "q"}], " ", "+", 
        SuperscriptBox["q", "2"]}]}], 
      RowBox[{"2", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"]}]], " ", 
     SuperscriptBox["\[Chi]eff", "3"]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]21", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "3", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "2", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[137]:=",ExpressionUUID->"fb98d237-dc01-4c99-ad5b-836803035407"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[138]=",ExpressionUUID->"a5972143-5288-40c4-ab57-391178848a1d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]20", "=", " ", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "-", "q"}], ")"}], "4"], " "}], 
       RowBox[{"16", " ", "q", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "4"]}]]}], 
     SuperscriptBox["\[Chi]1", "2"]}], "-", " ", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "2"], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "4"], "  "}], 
      RowBox[{"16", " ", "q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "4"]}]], 
     SuperscriptBox["\[Chi]2", "2"]}], "+", " ", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", 
         RowBox[{"4", "q"}], "+", " ", 
         SuperscriptBox["q", "2"]}], " ", ")"}]}], 
      RowBox[{"16", " ", "q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"]}]], " ", 
     SuperscriptBox["\[Chi]eff", "2"]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]20", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "3", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "1", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[139]:=",ExpressionUUID->"0455ca35-e8b7-47ff-bb09-41e460d79896"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[140]=",ExpressionUUID->"a08e18ec-23e6-4aac-961e-b2acedd166ea"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"\[Delta]16", "=", "0"}], ";"}]], "Input",
 CellLabel->
  "In[141]:=",ExpressionUUID->"90a0cbd8-1404-4961-93fc-d19d642fa54c"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]15", "=", " ", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
        RowBox[{"(", 
         RowBox[{"5", "-", 
          RowBox[{"2", " ", "q"}]}], ")"}]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "16"]]}], " ", 
     SuperscriptBox["\[Chi]1", "8"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{"2", " ", 
       SuperscriptBox["q", "3"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{"6", "-", "q", "+", 
         SuperscriptBox["q", "2"]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "16"]], " ", 
     SuperscriptBox["\[Chi]1", "6"], " ", 
     SuperscriptBox["\[Chi]2", "2"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{"9", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       SuperscriptBox["q", "6"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", 
         SuperscriptBox["q", "2"]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "16"]], " ", 
     SuperscriptBox["\[Chi]1", "4"], " ", 
     SuperscriptBox["\[Chi]2", "4"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       SuperscriptBox["q", "9"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", "q", "+", 
         RowBox[{"6", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "16"]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "6"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       SuperscriptBox["q", "13"], " ", 
       RowBox[{"(", 
        RowBox[{"2", "-", 
         RowBox[{"5", " ", "q"}]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "16"]], " ", 
     SuperscriptBox["\[Chi]2", "8"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{" ", 
       RowBox[{"2", " ", "q", " ", 
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], 
        RowBox[{"(", 
         RowBox[{"2", "-", "q"}], ")"}]}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "14"]], " ", 
     SuperscriptBox["\[Chi]1", "6"], 
     SuperscriptBox["\[Chi]eff", "2"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{"2", 
       SuperscriptBox["q", "4"], 
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], "  ", 
       RowBox[{"(", 
        RowBox[{"3", "-", "q", "+", 
         SuperscriptBox["q", "2"]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "14"]], " ", 
     SuperscriptBox["\[Chi]1", "4"], " ", 
     SuperscriptBox["\[Chi]2", "2"], 
     SuperscriptBox["\[Chi]eff", "2"]}], "-", " ", 
    RowBox[{
     FractionBox[
      RowBox[{"2", 
       SuperscriptBox["q", "7"], 
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", "q", "+", 
         RowBox[{"3", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "14"]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "4"], 
     SuperscriptBox["\[Chi]eff", "2"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{"2", 
       SuperscriptBox["q", "11"], 
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], "  ", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"2", " ", "q"}]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "14"]], " ", 
     SuperscriptBox["\[Chi]2", "6"], 
     SuperscriptBox["\[Chi]eff", "2"]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]15", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "2", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "6", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[174]:=",ExpressionUUID->"fdf4fd76-78fb-4afd-9aa8-331134b98c97"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[175]=",ExpressionUUID->"cc00a3ed-d298-4500-8aa5-bf80c321679c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]14", "=", "  ", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], 
        RowBox[{"(", 
         RowBox[{"20", "-", 
          RowBox[{"29", " ", "q"}], "+", 
          RowBox[{"12", " ", 
           SuperscriptBox["q", "2"]}]}], ")"}]}], 
       RowBox[{"2", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "13"]}]]}], " ", 
     SuperscriptBox["\[Chi]1", "6"], "\[Chi]eff"}], " ", "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "3"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], 
       RowBox[{"(", 
        RowBox[{"3", "+", 
         RowBox[{"14", " ", 
          SuperscriptBox["q", "2"]}], "-", 
         RowBox[{"8", " ", 
          SuperscriptBox["q", "3"]}]}], ")"}]}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "13"]}]], " ", 
     SuperscriptBox["\[Chi]1", "4"], " ", 
     SuperscriptBox["\[Chi]2", "2"], "\[Chi]eff"}], " ", "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "6"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], 
       RowBox[{"(", 
        RowBox[{"8", "-", 
         RowBox[{"14", " ", "q"}], "-", 
         RowBox[{"3", " ", 
          SuperscriptBox["q", "3"]}]}], ")"}]}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "13"]}]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "4"], "\[Chi]eff"}], " ", "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], 
       SuperscriptBox["q", "10"], " ", 
       RowBox[{"(", 
        RowBox[{"12", "-", 
         RowBox[{"29", " ", "q"}], "+", 
         RowBox[{"20", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "13"]}]], " ", 
     SuperscriptBox["\[Chi]2", "6"], "\[Chi]eff"}], "  ", "-", 
    RowBox[{
     FractionBox[
      RowBox[{"2", "q", "  ", 
       RowBox[{"(", 
        RowBox[{"3", "-", 
         RowBox[{"5", " ", "q"}], "+", 
         RowBox[{"3", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "11"]], " ", 
     SuperscriptBox["\[Chi]1", "4"], 
     SuperscriptBox["\[Chi]eff", "3"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{"2", "  ", 
       SuperscriptBox["q", "4"], " ", 
       RowBox[{"(", 
        RowBox[{"2", "-", "q"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"2", " ", "q"}]}], ")"}]}], 
      RowBox[{" ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "10"]}]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "2"], 
     SuperscriptBox["\[Chi]eff", "3"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{"2", "  ", 
       SuperscriptBox["q", "8"], " ", 
       RowBox[{"(", 
        RowBox[{"3", "-", 
         RowBox[{"5", " ", "q"}], "+", 
         RowBox[{"3", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "11"]], " ", 
     SuperscriptBox["\[Chi]2", "4"], 
     SuperscriptBox["\[Chi]eff", "3"]}]}]}], "  ", 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]14", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "2", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "5", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[144]:=",ExpressionUUID->"c59eda44-9d76-4b77-b5d0-475b9e613fcc"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[145]=",ExpressionUUID->"30af6266-dd16-4af9-915e-c946b1a6abf6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]13", "=", "   ", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "-", "q"}], ")"}], "2"], 
        RowBox[{"(", 
         RowBox[{"1", "-", 
          RowBox[{"15", " ", "q"}], "-", 
          RowBox[{"4", " ", 
           SuperscriptBox["q", "2"]}]}], ")"}]}], 
       RowBox[{"4", " ", "q", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "12"]}]]}], " ", 
     SuperscriptBox["\[Chi]1", "6"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], 
       RowBox[{"(", 
        RowBox[{"15", "-", 
         RowBox[{"55", " ", "q"}], "+", 
         RowBox[{"18", " ", 
          SuperscriptBox["q", "2"]}], "+", 
         RowBox[{"4", " ", 
          SuperscriptBox["q", "3"]}]}], ")"}]}], 
      RowBox[{"4", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "12"]}]], " ", 
     SuperscriptBox["\[Chi]1", "4"], " ", 
     SuperscriptBox["\[Chi]2", "2"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "5"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], 
       RowBox[{"(", 
        RowBox[{"4", "+", 
         RowBox[{"18", " ", "q"}], "-", 
         RowBox[{"55", " ", 
          SuperscriptBox["q", "2"]}], "+", 
         RowBox[{"15", " ", 
          SuperscriptBox["q", "3"]}]}], ")"}]}], 
      RowBox[{"4", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "12"]}]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "4"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "9"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], 
       RowBox[{"(", 
        RowBox[{"4", "+", 
         RowBox[{"15", " ", "q"}], "-", 
         SuperscriptBox["q", "2"]}], ")"}], " "}], 
      RowBox[{"4", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "12"]}]], 
     SuperscriptBox["\[Chi]2", "6"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{"6", "-", 
         RowBox[{"16", " ", "q"}], "+", 
         RowBox[{"18", " ", 
          SuperscriptBox["q", "2"]}], "-", 
         RowBox[{"5", " ", 
          SuperscriptBox["q", "3"]}]}], ")"}], " "}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "10"]], 
     SuperscriptBox["\[Chi]1", "4"], " ", 
     SuperscriptBox["\[Chi]eff", "2"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "3"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"8", " ", "q"}], "+", 
         RowBox[{"20", " ", 
          SuperscriptBox["q", "2"]}], "-", 
         RowBox[{"8", " ", 
          SuperscriptBox["q", "3"]}], "+", 
         SuperscriptBox["q", "4"]}], ")"}], " "}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "10"]], 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "2"], " ", 
     SuperscriptBox["\[Chi]eff", "2"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "7"], " ", 
       RowBox[{"(", 
        RowBox[{"5", "-", 
         RowBox[{"18", " ", "q"}], "+", 
         RowBox[{"16", " ", 
          SuperscriptBox["q", "2"]}], "-", 
         RowBox[{"6", " ", 
          SuperscriptBox["q", "3"]}]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "10"]], " ", 
     SuperscriptBox["\[Chi]2", "4"], " ", 
     SuperscriptBox["\[Chi]eff", "2"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{"2", "q", 
       RowBox[{"(", 
        RowBox[{"1", "-", "q", "+", 
         RowBox[{"3", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "8"]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]eff", "4"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{"2", 
       SuperscriptBox["q", "5"], " ", 
       RowBox[{"(", 
        RowBox[{"3", "-", "q", "+", 
         SuperscriptBox["q", "2"]}], ")"}], " "}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "8"]], 
     SuperscriptBox["\[Chi]2", "2"], " ", 
     SuperscriptBox["\[Chi]eff", "4"]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]13", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "2", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "4", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[146]:=",ExpressionUUID->"6441b0da-d7ea-47bb-9cd1-8ab798ac3609"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[147]=",ExpressionUUID->"faf02378-8152-4e21-b7d8-a8e0d417bb05"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]12", "=", "    ", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "-", "q"}], ")"}], "2"], 
        RowBox[{"(", 
         RowBox[{"3", "-", 
          RowBox[{"49", " ", "q"}], "-", 
          RowBox[{"16", " ", 
           SuperscriptBox["q", "2"]}]}], ")"}], " "}], 
       RowBox[{"8", " ", "q", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "9"]}]]}], 
     SuperscriptBox["\[Chi]1", "4"], "\[Chi]eff"}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{" ", 
       RowBox[{
        SuperscriptBox["q", "2"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
        RowBox[{"(", 
         RowBox[{"4", "-", 
          RowBox[{"7", " ", "q"}], "+", 
          RowBox[{"4", " ", 
           SuperscriptBox["q", "2"]}]}], ")"}]}]}], 
      RowBox[{"4", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "8"]}]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "2"], "\[Chi]eff"}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "6"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], 
       RowBox[{"(", 
        RowBox[{"16", "+", 
         RowBox[{"49", " ", "q"}], "-", 
         RowBox[{"3", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      RowBox[{"8", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "9"]}]], " ", 
     SuperscriptBox["\[Chi]2", "4"], "\[Chi]eff"}], " ", "-", 
    RowBox[{
     FractionBox[
      RowBox[{"2", "-", 
       RowBox[{"5", " ", "q"}], "+", 
       RowBox[{"19", " ", 
        SuperscriptBox["q", "2"]}]}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "7"]}]], " ", 
     SuperscriptBox["\[Chi]1", "2"], 
     SuperscriptBox["\[Chi]eff", "3"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "5"], " ", 
       RowBox[{"(", 
        RowBox[{"19", "-", 
         RowBox[{"5", " ", "q"}], "+", 
         RowBox[{"2", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "7"]}]], " ", 
     SuperscriptBox["\[Chi]2", "2"], 
     SuperscriptBox["\[Chi]eff", "3"]}], " ", "-", 
    RowBox[{
     FractionBox[
      RowBox[{"2", " ", 
       SuperscriptBox["q", "2"], " "}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "4"]], " ", 
     SuperscriptBox["\[Chi]eff", "5"]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]12", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "2", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "3", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[176]:=",ExpressionUUID->"4f01651e-3907-434c-9ab1-eb7e1197f668"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[177]=",ExpressionUUID->"0f78e7f3-a037-432f-9b80-4ffdae414ceb"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]11", "=", "     ", 
   RowBox[{
    RowBox[{
     FractionBox[
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], "4"], 
      RowBox[{"8", " ", "q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "8"]}]], " ", 
     SuperscriptBox["\[Chi]1", "4"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{"q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "4"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"12", " ", "q"}], "+", 
         SuperscriptBox["q", "2"]}], ")"}]}], 
      RowBox[{"8", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "8"]}]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "2"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "5"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "4"]}], 
      RowBox[{"8", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "8"]}]], " ", 
     SuperscriptBox["\[Chi]2", "4"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"18", " ", "q"}], "-", 
         RowBox[{"7", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      RowBox[{"8", " ", "q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "6"]}]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]eff", "2"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "3"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], 
       RowBox[{"(", 
        RowBox[{"7", "+", 
         RowBox[{"18", " ", "q"}], "-", 
         SuperscriptBox["q", "2"]}], ")"}], " "}], 
      RowBox[{"8", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "6"]}]], 
     SuperscriptBox["\[Chi]2", "2"], " ", 
     SuperscriptBox["\[Chi]eff", "2"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{" ", 
       RowBox[{"q", " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", 
          RowBox[{"3", " ", "q"}], "+", 
          SuperscriptBox["q", "2"]}], ")"}]}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "4"]], " ", 
     SuperscriptBox["\[Chi]eff", "4"]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]11", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "2", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "2", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[150]:=",ExpressionUUID->"d8d1082f-dec4-4a54-b26f-42fd1595ff2c"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[151]=",ExpressionUUID->"50c881a0-c91e-4ad6-bdc8-a979536a98ca"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]10", "=", "     ", 
   RowBox[{
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "4"], "  "}], 
      RowBox[{"8", " ", "q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "5"]}]], 
     SuperscriptBox["\[Chi]1", "2"], "\[Chi]eff"}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "2"], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "4"], "  "}], 
      RowBox[{"8", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "5"]}]], 
     SuperscriptBox["\[Chi]2", "2"], "\[Chi]eff"}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{"  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " "}], 
      RowBox[{"8", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"]}]], 
     SuperscriptBox["\[Chi]eff", "3"]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]10", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "2", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "1", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[152]:=",ExpressionUUID->"5629b012-9c0f-4194-9d5e-836327b8ed9a"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[153]=",ExpressionUUID->"fd1af4aa-92bb-450b-8cc8-1424776aa848"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]06", "=", "     ", 
   RowBox[{
    FractionBox[
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], "3"], " ", 
      SuperscriptBox["\[Chi]1", "10"]}], 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "20"]], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "3"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "3"], " ", 
       RowBox[{"(", 
        RowBox[{"3", "+", 
         RowBox[{"2", " ", "q"}]}], ")"}], " "}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "20"]], 
     SuperscriptBox["\[Chi]1", "8"], " ", 
     SuperscriptBox["\[Chi]2", "2"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "6"], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "3"], "  ", 
       RowBox[{"(", 
        RowBox[{"3", "+", 
         RowBox[{"6", " ", "q"}], "+", 
         SuperscriptBox["q", "2"]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "20"]], " ", 
     SuperscriptBox["\[Chi]1", "6"], " ", 
     SuperscriptBox["\[Chi]2", "4"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "9"], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "3"], "  ", 
       RowBox[{"(", 
        RowBox[{"1", "+", 
         RowBox[{"6", " ", "q"}], "+", 
         RowBox[{"3", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "20"]], 
     SuperscriptBox["\[Chi]1", "4"], " ", 
     SuperscriptBox["\[Chi]2", "6"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "13"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "3"], " ", 
       RowBox[{"(", 
        RowBox[{"2", "+", 
         RowBox[{"3", " ", "q"}]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "20"]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "8"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "17"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "3"], " "}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "20"]], 
     SuperscriptBox["\[Chi]2", "10"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{"q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " "}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "18"]], 
     SuperscriptBox["\[Chi]1", "8"], 
     SuperscriptBox["\[Chi]eff", "2"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{" ", 
       RowBox[{"2", " ", 
        SuperscriptBox["q", "4"], "  ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "-", "q"}], ")"}], "2"], " "}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "17"]], 
     SuperscriptBox["\[Chi]1", "6"], " ", 
     SuperscriptBox["\[Chi]2", "2"], 
     SuperscriptBox["\[Chi]eff", "2"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "7"], " ", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", 
         RowBox[{"4", " ", "q"}], "+", 
         SuperscriptBox["q", "2"]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "18"]], " ", 
     SuperscriptBox["\[Chi]1", "4"], " ", 
     SuperscriptBox["\[Chi]2", "4"], 
     SuperscriptBox["\[Chi]eff", "2"]}], "-", " ", 
    RowBox[{
     FractionBox[
      RowBox[{"2", " ", 
       SuperscriptBox["q", "11"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " "}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "17"]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "6"], 
     SuperscriptBox["\[Chi]eff", "2"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "15"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " "}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "18"]], " ", 
     SuperscriptBox["\[Chi]2", "8"], 
     SuperscriptBox["\[Chi]eff", "2"]}]}]}], " ", 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]06", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "1", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "7", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[154]:=",ExpressionUUID->"50ea5fdb-2459-4a79-b163-b55e70fd421e"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[155]=",ExpressionUUID->"bd7be8f4-2ae5-495f-8d51-85e1d3ce7ce1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]05", "=", " ", 
   RowBox[{
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{"5", "-", 
         RowBox[{"8", " ", "q"}]}], ")"}]}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "17"]}]], " ", 
     SuperscriptBox["\[Chi]1", "8"], "\[Chi]eff"}], "-", " ", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "3"], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", 
         RowBox[{"3", " ", "q"}]}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"4", " ", "q"}]}], ")"}]}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "17"]}]], " ", 
     SuperscriptBox["\[Chi]1", "6"], " ", 
     SuperscriptBox["\[Chi]2", "2"], "\[Chi]eff"}], "-", " ", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "6"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{"4", "+", "q", "+", 
         RowBox[{"4", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "16"]}]], " ", 
     SuperscriptBox["\[Chi]1", "4"], " ", 
     SuperscriptBox["\[Chi]2", "4"], "\[Chi]eff"}], "+", " ", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "10"], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{"4", "-", "q"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"3", "+", "q"}], ")"}]}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "17"]}]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "6"], "\[Chi]eff"}], "-", " ", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "14"], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{"8", "-", 
         RowBox[{"5", " ", "q"}]}], ")"}]}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "17"]}]], " ", 
     SuperscriptBox["\[Chi]2", "8"], "\[Chi]eff"}], " ", "+", " ", 
    RowBox[{
     FractionBox[
      RowBox[{"2", " ", "q", 
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"2", " ", "q"}]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "15"]], " ", 
     SuperscriptBox["\[Chi]1", "6"], 
     SuperscriptBox["\[Chi]eff", "3"]}], " ", "+", 
    RowBox[{
     FractionBox[
      RowBox[{"2", 
       SuperscriptBox["q", "4"], "  ", 
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", "q", "+", 
         RowBox[{"3", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "15"]], " ", 
     SuperscriptBox["\[Chi]1", "4"], " ", 
     SuperscriptBox["\[Chi]2", "2"], 
     SuperscriptBox["\[Chi]eff", "3"]}], " ", "-", 
    RowBox[{
     FractionBox[
      RowBox[{"2", " ", 
       SuperscriptBox["q", "8"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"3", "-", "q", "+", 
         SuperscriptBox["q", "2"]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "15"]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "4"], 
     SuperscriptBox["\[Chi]eff", "3"]}], " ", "+", 
    RowBox[{
     FractionBox[
      RowBox[{"2", "  ", 
       SuperscriptBox["q", "12"], 
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"2", "-", "q"}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "15"]], " ", 
     SuperscriptBox["\[Chi]2", "6"], 
     SuperscriptBox["\[Chi]eff", "3"]}]}]}], " ", 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]05", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "1", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "6", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[156]:=",ExpressionUUID->"5c5856d2-1a64-41f2-8600-79592d81cf1c"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[157]=",ExpressionUUID->"110ebe7c-69cd-45ab-8fa4-bd98e4a22979"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]04", "=", " ", 
   RowBox[{
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"20", " ", "q"}], "-", 
         RowBox[{"8", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      RowBox[{"16", " ", "q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "16"]}]], " ", 
     SuperscriptBox["\[Chi]1", "8"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{" ", 
       RowBox[{
        SuperscriptBox["q", "2"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
        RowBox[{"(", 
         RowBox[{"5", "-", 
          RowBox[{"27", " ", "q"}], "+", 
          RowBox[{"3", " ", 
           SuperscriptBox["q", "2"]}], "-", 
          RowBox[{"8", " ", 
           SuperscriptBox["q", "3"]}]}], ")"}], " "}]}], 
      RowBox[{"4", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "16"]}]], 
     SuperscriptBox["\[Chi]1", "6"], " ", 
     SuperscriptBox["\[Chi]2", "2"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{" ", 
       RowBox[{
        SuperscriptBox["q", "5"], 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "-", "q"}], ")"}], "2"], " ", " ", 
        RowBox[{"(", 
         RowBox[{"4", "+", 
          RowBox[{"6", " ", "q"}], "+", 
          RowBox[{"61", " ", 
           SuperscriptBox["q", "2"]}], "+", 
          RowBox[{"6", " ", 
           SuperscriptBox["q", "3"]}], "+", 
          RowBox[{"4", " ", 
           SuperscriptBox["q", "4"]}]}], ")"}]}]}], 
      RowBox[{"8", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "16"]}]], " ", 
     SuperscriptBox["\[Chi]1", "4"], " ", 
     SuperscriptBox["\[Chi]2", "4"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{" ", 
       RowBox[{
        SuperscriptBox["q", "9"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
        RowBox[{"(", 
         RowBox[{"8", "-", 
          RowBox[{"3", " ", "q"}], "+", 
          RowBox[{"27", " ", 
           SuperscriptBox["q", "2"]}], "-", 
          RowBox[{"5", " ", 
           SuperscriptBox["q", "3"]}]}], ")"}], " "}]}], 
      RowBox[{"4", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "16"]}]], 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "6"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "13"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{"8", "+", 
         RowBox[{"20", " ", "q"}], "-", 
         SuperscriptBox["q", "2"]}], ")"}]}], 
      RowBox[{"16", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "16"]}]], " ", 
     SuperscriptBox["\[Chi]2", "8"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"4", "-", 
         RowBox[{"18", " ", "q"}], "+", 
         RowBox[{"11", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      RowBox[{"2", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "14"]}]], " ", 
     SuperscriptBox["\[Chi]1", "6"], 
     SuperscriptBox["\[Chi]eff", "2"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "3"], "  ", 
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"4", " ", "q"}], "-", 
         RowBox[{"14", " ", 
          SuperscriptBox["q", "2"]}], "+", 
         RowBox[{"8", " ", 
          SuperscriptBox["q", "3"]}]}], ")"}], " "}], 
      RowBox[{"2", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "14"]}]], 
     SuperscriptBox["\[Chi]1", "4"], " ", 
     SuperscriptBox["\[Chi]2", "2"], 
     SuperscriptBox["\[Chi]eff", "2"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "7"], "  ", 
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"8", "-", 
         RowBox[{"14", " ", "q"}], "-", 
         RowBox[{"4", " ", 
          SuperscriptBox["q", "2"]}], "+", 
         SuperscriptBox["q", "3"]}], ")"}]}], 
      RowBox[{"2", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "14"]}]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "4"], 
     SuperscriptBox["\[Chi]eff", "2"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "11"], "  ", 
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"11", "-", 
         RowBox[{"18", " ", "q"}], "+", 
         RowBox[{"4", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      RowBox[{"2", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "14"]}]], " ", 
     SuperscriptBox["\[Chi]2", "6"], 
     SuperscriptBox["\[Chi]eff", "2"]}], " ", "+", " ", 
    RowBox[{
     FractionBox[
      RowBox[{"q", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"6", " ", "q"}], "+", 
         RowBox[{"6", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "12"]], " ", 
     SuperscriptBox["\[Chi]1", "4"], 
     SuperscriptBox["\[Chi]eff", "4"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{" ", 
       RowBox[{"2", " ", 
        SuperscriptBox["q", "5"], " ", 
        RowBox[{"(", 
         RowBox[{"3", "-", 
          RowBox[{"5", " ", "q"}], "+", 
          RowBox[{"3", " ", 
           SuperscriptBox["q", "2"]}]}], ")"}]}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "12"]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "2"], 
     SuperscriptBox["\[Chi]eff", "4"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "9"], " ", 
       RowBox[{"(", 
        RowBox[{"6", "-", 
         RowBox[{"6", " ", "q"}], "+", 
         SuperscriptBox["q", "2"]}], ")"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "12"]], " ", 
     SuperscriptBox["\[Chi]2", "4"], 
     SuperscriptBox["\[Chi]eff", "4"]}]}]}], " ", 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]04", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "1", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "5", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[180]:=",ExpressionUUID->"858b54d9-fd24-4553-87a6-36fe1b1a3b1f"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[181]=",ExpressionUUID->"7263aed5-7dca-4e8d-acc8-6a04049b3168"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]03", "=", " ", 
   RowBox[{
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"25", " ", "q"}], "-", 
         RowBox[{"12", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      RowBox[{"8", " ", "q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "13"]}]], " ", 
     SuperscriptBox["\[Chi]1", "6"], "\[Chi]eff"}], "-", " ", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "2"], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{"4", "-", 
         RowBox[{"29", " ", "q"}], "+", 
         RowBox[{"21", " ", 
          SuperscriptBox["q", "2"]}], "-", 
         RowBox[{"32", " ", 
          SuperscriptBox["q", "3"]}]}], ")"}]}], 
      RowBox[{"8", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "13"]}]], " ", 
     SuperscriptBox["\[Chi]1", "4"], " ", 
     SuperscriptBox["\[Chi]2", "2"], "\[Chi]eff"}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "6"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{"32", "-", 
         RowBox[{"21", " ", "q"}], "+", 
         RowBox[{"29", " ", 
          SuperscriptBox["q", "2"]}], "-", 
         RowBox[{"4", " ", 
          SuperscriptBox["q", "3"]}]}], ")"}]}], 
      RowBox[{"8", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "13"]}]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "4"], "\[Chi]eff"}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "10"], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{"12", "+", 
         RowBox[{"25", " ", "q"}], "-", 
         SuperscriptBox["q", "2"]}], ")"}], " ", " "}], 
      RowBox[{"8", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "13"]}]], 
     SuperscriptBox["\[Chi]2", "6"], "\[Chi]eff"}], " ", "+", " ", 
    RowBox[{
     FractionBox[
      RowBox[{"1", "-", 
       RowBox[{"14", " ", "q"}], "+", 
       RowBox[{"20", " ", 
        SuperscriptBox["q", "2"]}], "-", 
       RowBox[{"5", " ", 
        SuperscriptBox["q", "3"]}]}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "11"]}]], " ", 
     SuperscriptBox["\[Chi]1", "4"], 
     SuperscriptBox["\[Chi]eff", "3"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{" ", 
       RowBox[{
        SuperscriptBox["q", "4"], "  ", 
        RowBox[{"(", 
         RowBox[{"2", "-", 
          RowBox[{"3", " ", "q"}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["q", "2"]}]}], ")"}]}]}], 
      RowBox[{" ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "10"]}]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "2"], 
     SuperscriptBox["\[Chi]eff", "3"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "8"], " ", 
       RowBox[{"(", 
        RowBox[{"5", "-", 
         RowBox[{"20", " ", "q"}], "+", 
         RowBox[{"14", " ", 
          SuperscriptBox["q", "2"]}], "-", 
         SuperscriptBox["q", "3"]}], ")"}]}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "11"]}]], " ", 
     SuperscriptBox["\[Chi]2", "4"], 
     SuperscriptBox["\[Chi]eff", "3"]}], "  ", "-", 
    RowBox[{
     FractionBox[
      RowBox[{"2", " ", 
       SuperscriptBox["q", "2"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"2", " ", "q"}]}], ")"}], " "}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "9"]], " ", 
     SuperscriptBox["\[Chi]1", "2"], 
     SuperscriptBox["\[Chi]eff", "5"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{"2", " ", 
       SuperscriptBox["q", "6"], " ", 
       RowBox[{"(", 
        RowBox[{"2", "-", "q"}], ")"}], " "}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "9"]], " ", 
     SuperscriptBox["\[Chi]2", "2"], 
     SuperscriptBox["\[Chi]eff", "5"]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]03", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "1", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "4", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[184]:=",ExpressionUUID->"6baf57b1-6d56-4fe8-8543-4c142fbdf23c"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[185]=",ExpressionUUID->"39af8bfc-f803-462c-a0ba-3b442f8f8f23"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]02", "=", " ", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "-", "q"}], ")"}], "4"], " "}], 
       RowBox[{"16", " ", "q", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "12"]}]]}], 
     SuperscriptBox["\[Chi]1", "6"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{"q", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "4"], " ", " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"24", " ", "q"}], "-", 
         RowBox[{"10", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}], " "}], 
      RowBox[{"16", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "12"]}]], 
     SuperscriptBox["\[Chi]1", "4"], " ", 
     SuperscriptBox["\[Chi]2", "2"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "5"], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "4"], "  ", 
       RowBox[{"(", 
        RowBox[{"10", "+", 
         RowBox[{"24", " ", "q"}], "-", 
         SuperscriptBox["q", "2"]}], ")"}]}], 
      RowBox[{"16", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "12"]}]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "4"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "9"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "4"], " "}], 
      RowBox[{"16", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "12"]}]], 
     SuperscriptBox["\[Chi]2", "6"]}], "+", " ", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"40", " ", "q"}], "-", 
         RowBox[{"23", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      RowBox[{"16", " ", "q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "10"]}]], " ", 
     SuperscriptBox["\[Chi]1", "4"], 
     SuperscriptBox["\[Chi]eff", "2"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{" ", 
       RowBox[{
        SuperscriptBox["q", "3"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "-", "q"}], ")"}], "2"], 
        RowBox[{"(", 
         RowBox[{"11", "-", 
          RowBox[{"24", " ", "q"}], "+", 
          RowBox[{"11", " ", 
           SuperscriptBox["q", "2"]}]}], ")"}]}]}], 
      RowBox[{"8", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "10"]}]], " ", 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "2"], 
     SuperscriptBox["\[Chi]eff", "2"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "7"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], 
       RowBox[{"(", 
        RowBox[{"23", "+", 
         RowBox[{"40", " ", "q"}], "-", 
         SuperscriptBox["q", "2"]}], ")"}]}], 
      RowBox[{"16", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "10"]}]], " ", 
     SuperscriptBox["\[Chi]2", "4"], 
     SuperscriptBox["\[Chi]eff", "2"]}], " ", "-", 
    RowBox[{
     FractionBox[
      RowBox[{"q", " ", 
       RowBox[{"(", 
        RowBox[{"4", "-", 
         RowBox[{"7", " ", "q"}], "-", 
         SuperscriptBox["q", "2"]}], ")"}]}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "8"]}]], " ", 
     SuperscriptBox["\[Chi]1", "2"], 
     SuperscriptBox["\[Chi]eff", "4"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "5"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", 
         RowBox[{"7", " ", "q"}], "-", 
         RowBox[{"4", " ", 
          SuperscriptBox["q", "2"]}]}], ")"}]}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "8"]}]], " ", 
     SuperscriptBox["\[Chi]2", "2"], 
     SuperscriptBox["\[Chi]eff", "4"]}], " ", "+", 
    RowBox[{
     FractionBox[
      RowBox[{" ", 
       SuperscriptBox["q", "3"]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "6"]], "  ", 
     SuperscriptBox["\[Chi]eff", "6"]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]02", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "1", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "3", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[190]:=",ExpressionUUID->"23a7e014-567b-48d0-8f22-ab93206b3c6f"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[191]=",ExpressionUUID->"3a66f0b8-dee3-472c-849a-b96c4cac08a5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]01", "=", "  ", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "4"], 
       RowBox[{"8", " ", "q", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "9"]}]]}], 
     SuperscriptBox["\[Chi]1", "4"], "\[Chi]eff"}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{"5", " ", 
       SuperscriptBox["q", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "4"], " "}], 
      RowBox[{"8", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "8"]}]], 
     SuperscriptBox["\[Chi]1", "2"], " ", 
     SuperscriptBox["\[Chi]2", "2"], "\[Chi]eff"}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "6"], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "4"], " "}], 
      RowBox[{"8", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "9"]}]], 
     SuperscriptBox["\[Chi]2", "4"], "\[Chi]eff"}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{"5", "+", 
         RowBox[{"3", " ", "q"}]}], ")"}], " "}], 
      RowBox[{"8", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "7"]}]], 
     SuperscriptBox["\[Chi]1", "2"], 
     SuperscriptBox["\[Chi]eff", "3"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "4"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{"3", "+", 
         RowBox[{"5", " ", "q"}]}], ")"}]}], 
      RowBox[{"8", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "7"]}]], " ", 
     SuperscriptBox["\[Chi]2", "2"], 
     SuperscriptBox["\[Chi]eff", "3"]}], " ", "+", 
    RowBox[{
     FractionBox[
      RowBox[{" ", 
       SuperscriptBox["q", "2"], " "}], 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "4"]}]], " ", 
     SuperscriptBox["\[Chi]eff", "5"]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]01", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "1", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "2", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[164]:=",ExpressionUUID->"2f978e8c-0b2f-48d9-84f7-f42b96f21d33"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[165]=",ExpressionUUID->"67935b59-b8b2-4d85-a16b-141c8845d807"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Delta]00", "=", "  ", 
   RowBox[{
    FractionBox[
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], "6"], " ", "q", " ", 
      SuperscriptBox["\[Chi]1", "2"], " ", 
      SuperscriptBox["\[Chi]2", "2"]}], 
     RowBox[{"16", " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "8"]}]], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "4"], " "}], 
      RowBox[{"16", " ", "q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "6"]}]], 
     SuperscriptBox["\[Chi]1", "2"], 
     SuperscriptBox["\[Chi]eff", "2"]}], "-", 
    RowBox[{
     FractionBox[
      RowBox[{
       SuperscriptBox["q", "3"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "4"]}], 
      RowBox[{"16", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "6"]}]], " ", 
     SuperscriptBox["\[Chi]2", "2"], 
     SuperscriptBox["\[Chi]eff", "2"]}], "+", 
    RowBox[{
     FractionBox[
      RowBox[{"q", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], "2"], "  "}], 
      RowBox[{"16", "  ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "4"]}]], 
     SuperscriptBox["\[Chi]eff", "4"]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Delta]00", "-", " ", 
    RowBox[{
     RowBox[{"CoefficientList", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"coeffs", "[", 
         RowBox[{"[", "1", "]"}], "]"}], "/", "overalldiscr"}], ",", "u"}], 
      "]"}], "[", 
     RowBox[{"[", "1", "]"}], "]"}]}], "/.", "\[Chi]tr"}], "//", 
  "FullSimplify"}], "\[IndentingNewLine]", 
 RowBox[{
  StyleBox[
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
      RowBox[{"Evaluate", "[", "%%", "]"}], "//", "HoldForm"}], "//", 
     "TeXForm"}], "*)"}],
   FontSize->14]}]}], "Input",
 CellLabel->
  "In[166]:=",ExpressionUUID->"32269cb9-98f5-4c97-8dfa-c4783402e4ed"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[167]=",ExpressionUUID->"7c8405ae-3e88-4cbd-977e-9720f77d88c1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"precessionform", "[", 
  RowBox[{"\"\<coeff4\>\"", ",", 
   RowBox[{"Simplify", "[", 
    RowBox[{
     RowBox[{"coeffs", "[", 
      RowBox[{"[", "5", "]"}], "]"}], "/", "overalldiscr"}], "]"}]}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"precessionform", "[", 
  RowBox[{"\"\<coeff3\>\"", ",", 
   RowBox[{"Simplify", "[", 
    RowBox[{
     RowBox[{"coeffs", "[", 
      RowBox[{"[", "4", "]"}], "]"}], "/", "overalldiscr"}], "]"}]}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"precessionform", "[", 
  RowBox[{"\"\<coeff2\>\"", ",", 
   RowBox[{"Simplify", "[", 
    RowBox[{
     RowBox[{"coeffs", "[", 
      RowBox[{"[", "3", "]"}], "]"}], "/", "overalldiscr"}], "]"}]}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"precessionform", "[", 
  RowBox[{"\"\<coeff1\>\"", ",", 
   RowBox[{"Simplify", "[", 
    RowBox[{
     RowBox[{"coeffs", "[", 
      RowBox[{"[", "2", "]"}], "]"}], "/", "overalldiscr"}], "]"}]}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"precessionform", "[", 
  RowBox[{"\"\<coeff0\>\"", ",", 
   RowBox[{"Simplify", "[", 
    RowBox[{
     RowBox[{"coeffs", "[", 
      RowBox[{"[", "1", "]"}], "]"}], "/", "overalldiscr"}], "]"}]}], 
  "]"}]}], "Input",
 CellLabel->
  "In[192]:=",ExpressionUUID->"447ab176-437f-47df-b6ab-8af410db6d70"],

Cell[BoxData["\<\"# Machine generated with eq_generator.nb\\ncoeff4 = 1/16 * \
q**(-1) * ((1 + q))**(-4) * (1 + (q**6 + (q * (2 + (80 * u**2 * chi1**2 + 40 \
* u * chieff)) + (q**5 * (2 + (80 * u**2 * chi2**2 + 40 * u * chieff)) + (4 * \
q**3 * (-1 + (60 * u * chieff + 8 * u**2 * chieff**2)) + (q**2 * (-1 + (160 * \
u * chieff + 16 * u**2 * (-3 * chi1**2 + chieff**2))) + q**4 * (-1 + (160 * u \
* chieff + 16 * u**2 * (-3 * chi2**2 + chieff**2)))))))))\"\>"], "Output",
 CellLabel->
  "Out[192]=",ExpressionUUID->"70cf60ed-5fb5-47c0-8b9a-dac4b75ca2a5"],

Cell[BoxData["\<\"# Machine generated with eq_generator.nb\\ncoeff3 = -1/8 * \
q**(-1) * ((1 + q))**(-8) * (((-1 + q))**2 * ((1 + q))**8 * chieff + (8 * q * \
u**3 * ((10 + (-12 * q + 3 * q**2)) * chi1**4 + (-2 * q * chi1**2 * (6 * q**2 \
* chi2**2 + (6 * q**4 * chi2**2 + (-2 * chieff**2 + (-3 * q * chieff**2 + \
q**3 * (-11 * chi2**2 + chieff**2))))) + q**4 * chi2**2 * (10 * q**4 * \
chi2**2 + (-2 * chieff**2 + (4 * q**3 * (-3 * chi2**2 + chieff**2) + 3 * q**2 \
* (chi2**2 + 2 * chieff**2)))))) + (4 * q * ((1 + q))**3 * u**2 * chieff * \
(-1 * (-20 + (3 * q + q**2)) * chi1**2 + q * (20 * q**4 * chi2**2 + (4 * \
chieff**2 + (12 * q * chieff**2 + (-1 * q**2 * (chi2**2 + -12 * chieff**2) + \
q**3 * (-3 * chi2**2 + 4 * chieff**2)))))) + 2 * ((1 + q))**4 * u * (-1 * \
((-1 + q))**2 * (-1 + 5 * q) * chi1**2 + q * (q**5 * chi2**2 + (8 * chieff**2 \
+ (40 * q * chieff**2 + (q**4 * (-7 * chi2**2 + 8 * chieff**2) + (q**3 * (11 \
* chi2**2 + 40 * chieff**2) + q**2 * (-5 * chi2**2 + 64 * chieff**2))))))))))\
\"\>"], "Output",
 CellLabel->
  "Out[193]=",ExpressionUUID->"835e0b08-d019-49a0-97ed-e802df527fd9"],

Cell[BoxData["\<\"# Machine generated with eq_generator.nb\\ncoeff2 = 1/16 * \
q**(-1) * ((1 + q))**(-12) * (-16 * q * (-10 + (18 * q + (-9 * q**2 + q**3))) \
* u**4 * chi1**6 + (chieff**2 + (4 * q * chieff**2 * (3 + 2 * u * chieff) + \
(q**(14) * (6 * u**2 * chi2**4 + (chieff**2 + chi2**2 * (-1 + 6 * u * \
chieff))) + (q**2 * (-1 * chi2**2 + chieff**2 * (59 + (144 * u * chieff + 16 \
* u**2 * chieff**2))) + (4 * q**(13) * (40 * u**4 * chi2**6 + (chieff**2 * (3 \
+ 2 * u * chieff) + (12 * u**2 * chi2**4 * (-1 + 5 * u * chieff) + chi2**2 * \
(-1 + (-4 * u * chieff + 24 * u**2 * chieff**2))))) + (-4 * q**3 * (chi2**2 * \
(1 + 2 * u * chieff) + -2 * chieff**2 * (19 + (126 * u * chieff + 24 * u**2 * \
chieff**2))) + (q**4 * (-2 * chi2**2 * (1 + (43 * u * chieff + 4 * u**2 * \
chieff**2)) + chieff**2 * (201 + (3920 * u * chieff + 976 * u**2 * \
chieff**2))) + (4 * q**5 * (chieff**2 * (13 + (2430 * u * chieff + 704 * u**2 \
* chieff**2)) + chi2**2 * (3 + (-72 * u * chieff + (10 * u**2 * chieff**2 + 8 \
* u**3 * chieff**3)))) + (-4 * q**7 * (u**2 * chi2**4 * (19 + 8 * u * chieff) \
+ (-4 * chieff**2 * (-27 + (1218 * u * chieff + 392 * u**2 * chieff**2)) + -2 \
* chi2**2 * (-1 + (20 * u * chieff + (169 * u**2 * chieff**2 + 32 * u**3 * \
chieff**3))))) + (q**(12) * (-288 * u**4 * chi2**6 + (chieff**2 * (59 + (144 \
* u * chieff + 16 * u**2 * chieff**2)) + (2 * u**2 * chi2**4 * (-67 + (204 * \
u * chieff + 48 * u**2 * chieff**2)) + 2 * chi2**2 * (-1 + (-95 * u * chieff \
+ (296 * u**2 * chieff**2 + 48 * u**3 * chieff**3)))))) + (4 * q**9 * (2 * \
u**2 * chi2**4 * (13 + (6 * u * chieff + -8 * u**2 * chieff**2)) + (chieff**2 \
* (13 + (2430 * u * chieff + 704 * u**2 * chieff**2)) + 2 * chi2**2 * (-1 + \
(70 * u * chieff + (379 * u**2 * chieff**2 + 100 * u**3 * chieff**3))))) + (4 \
* q**(11) * (36 * u**4 * chi2**6 + (u**2 * chi2**4 * (5 + (-16 * u * chieff + \
24 * u**2 * chieff**2)) + (2 * chieff**2 * (19 + (126 * u * chieff + 24 * \
u**2 * chieff**2)) + chi2**2 * (3 + (-102 * u * chieff + (406 * u**2 * \
chieff**2 + 112 * u**3 * chieff**3)))))) + (q**8 * (2 * u**2 * chi2**4 * (-53 \
+ (28 * u * chieff + 8 * u**2 * chieff**2)) + (chieff**2 * (-261 + (16416 * u \
* chieff + 5152 * u**2 * chieff**2)) + 4 * chi2**2 * (-7 + (189 * u * chieff \
+ (614 * u**2 * chieff**2 + 120 * u**3 * chieff**3))))) + (q**6 * (-8 * u**2 \
* chi2**4 + (chieff**2 * (-261 + (16416 * u * chieff + 5152 * u**2 * \
chieff**2)) + chi2**2 * (17 + (-338 * u * chieff + (424 * u**2 * chieff**2 + \
128 * u**3 * chieff**3))))) + (q**10 * (-16 * u**4 * chi2**6 + (-2 * u**2 * \
chi2**4 * (-121 + (136 * u * chieff + 40 * u**2 * chieff**2)) + (chieff**2 * \
(201 + (3920 * u * chieff + 976 * u**2 * chieff**2)) + chi2**2 * (17 + (-148 \
* u * chieff + (2680 * u**2 * chieff**2 + 832 * u**3 * chieff**3)))))) + (2 * \
u**2 * chi1**4 * (3 + (-4 * q**8 + (2 * q**7 * (-19 + (36 * u**2 * chi2**2 + \
-8 * u * chieff)) + (24 * q * (-1 + 5 * u * chieff) + (2 * q**3 * (5 + (-16 * \
u * chieff + 24 * u**2 * chieff**2)) + (q**2 * (-67 + (204 * u * chieff + 48 \
* u**2 * chieff**2)) + (4 * q**5 * (13 + (6 * u * chieff + 8 * u**2 * (9 * \
chi2**2 + -1 * chieff**2))) + (q**6 * (-53 + (28 * u * chieff + 8 * u**2 * \
(-27 * chi2**2 + chieff**2))) + -1 * q**4 * (-121 + (136 * u * chieff + 8 * \
u**2 * (18 * chi2**2 + 5 * chieff**2))))))))))) + -1 * chi1**2 * (1 + \
(q**(12) + (-6 * u * chieff + (q**(11) * (4 + (60 * u**2 * chi2**2 + 8 * u * \
chieff)) + (q * (4 + (16 * u * chieff + -96 * u**2 * chieff**2)) + (q**2 * (2 \
+ (190 * u * chieff + (-592 * u**2 * chieff**2 + -96 * u**3 * chieff**3))) + \
(q**10 * (2 + (288 * u**4 * chi2**4 + (86 * u * chieff + (24 * u**3 * chi2**2 \
* chieff + 4 * u**2 * (chi2**2 + 2 * chieff**2))))) + (-4 * q**3 * (3 + (-102 \
* u * chieff + (112 * u**3 * chieff**3 + u**2 * (-15 * chi2**2 + 406 * \
chieff**2)))) + (-4 * q**9 * (3 + (-72 * u * chieff + (8 * u**3 * chieff * \
(-3 * chi2**2 + chieff**2) + (2 * u**2 * (29 * chi2**2 + 5 * chieff**2) + 24 \
* u**4 * (6 * chi2**4 + -1 * chi2**2 * chieff**2))))) + (q**4 * (-17 + (148 * \
u * chieff + (4 * u**2 * (chi2**2 + -670 * chieff**2) + 8 * u**3 * (3 * \
chi2**2 * chieff + -104 * chieff**3)))) + (8 * q**5 * (1 + (-70 * u * chieff \
+ (12 * u**4 * chi2**2 * chieff**2 + (-1 * u**2 * (29 * chi2**2 + 379 * \
chieff**2) + 4 * u**3 * (3 * chi2**2 * chieff + -25 * chieff**3))))) + (4 * \
q**6 * (7 + (-189 * u * chieff + (8 * u**4 * chi2**2 * chieff**2 + (-1 * u**2 \
* (chi2**2 + 614 * chieff**2) + 6 * u**3 * (7 * chi2**2 * chieff + -20 * \
chieff**3))))) + (q**8 * (-17 + (338 * u * chieff + (-4 * u**2 * (chi2**2 + \
106 * chieff**2) + (16 * u**4 * (27 * chi2**4 + 2 * chi2**2 * chieff**2) + 8 \
* u**3 * (21 * chi2**2 * chieff + -16 * chieff**3))))) + -8 * q**7 * (-1 + \
(20 * u * chieff + (u**2 * (-43 * chi2**2 + 169 * chieff**2) + (2 * u**4 * (9 \
* chi2**4 + 8 * chi2**2 * chieff**2) + -8 * u**3 * (3 * chi2**2 * chieff + -4 \
* chieff**3)))))))))))))))))))))))))))))))))))\"\>"], "Output",
 CellLabel->
  "Out[194]=",ExpressionUUID->"9961231d-334c-4e5b-a131-d288a432bed3"],

Cell[BoxData["\<\"# Machine generated with eq_generator.nb\\ncoeff1 = -1/8 * \
q**(-1) * ((1 + q))**(-16) * (-1 * ((-1 + q))**2 * ((1 + q))**(11) * chieff * \
(((-1 + q))**2 * chi1**2 + q * (q**4 * chi2**2 + (-1 * chieff**2 + (-3 * q * \
chieff**2 + (q**2 * (chi2**2 + -3 * chieff**2) + -1 * q**3 * (2 * chi2**2 + \
chieff**2)))))) + (8 * (-1 + q) * q * u**5 * (-1 * chi1**2 + q**3 * chi2**2) \
* ((5 + (-7 * q + 2 * q**2)) * chi1**6 + (q * chi1**4 * (-7 * q**2 * chi2**2 \
+ (-2 * q**4 * chi2**2 + (2 * q**5 * chi2**2 + (4 * chieff**2 + (6 * q * \
chieff**2 + q**3 * (7 * chi2**2 + -2 * chieff**2)))))) + (-1 * q**4 * chi1**2 \
* chi2**2 * (7 * q**5 * chi2**2 + (2 * chieff**2 + (4 * q * chieff**2 + (-2 * \
q**2 * (chi2**2 + -2 * chieff**2) + (q**4 * (-7 * chi2**2 + 2 * chieff**2) + \
2 * q**3 * (chi2**2 + 2 * chieff**2)))))) + q**8 * chi2**4 * (5 * q**4 * \
chi2**2 + (-2 * chieff**2 + (2 * q**2 * (chi2**2 + 3 * chieff**2) + q**3 * \
(-7 * chi2**2 + 4 * chieff**2))))))) + (4 * q * ((1 + q))**3 * u**4 * chieff \
* ((20 + (-49 * q + (41 * q**2 + -12 * q**3))) * chi1**6 + (q**4 * chi1**2 * \
chi2**2 * (-3 * q**6 * chi2**2 + (8 * chieff**2 + (4 * q * chieff**2 + (q**3 \
* (22 * chi2**2 + -28 * chieff**2) + (q**4 * (-14 * chi2**2 + 4 * chieff**2) \
+ (-4 * q**2 * (2 * chi2**2 + 7 * chieff**2) + q**5 * (3 * chi2**2 + 8 * \
chieff**2))))))) + (q**8 * chi2**4 * (20 * q**5 * chi2**2 + (12 * chieff**2 + \
(4 * q * chieff**2 + (-4 * q**2 * (3 * chi2**2 + 4 * chieff**2) + (q**3 * (41 \
* chi2**2 + 4 * chieff**2) + q**4 * (-49 * chi2**2 + 12 * chieff**2)))))) + q \
* chi1**4 * (22 * q**5 * chi2**2 + (-8 * q**6 * chi2**2 + (12 * chieff**2 + \
(4 * q * chieff**2 + (-2 * q**4 * (7 * chi2**2 + -6 * chieff**2) + (q**3 * (3 \
* chi2**2 + 4 * chieff**2) + -1 * q**2 * (3 * chi2**2 + 16 * \
chieff**2)))))))))) + (-1 * ((1 + q))**8 * u * (((-1 + q))**4 * chi1**4 + \
(((-1 + q))**2 * chi1**2 * (-14 * q**5 * chi2**2 + (q**6 * chi2**2 + (-1 * \
chieff**2 + (16 * q * chieff**2 + (q**4 * (26 * chi2**2 + 7 * chieff**2) + \
(q**3 * (-14 * chi2**2 + 32 * chieff**2) + q**2 * (chi2**2 + 42 * \
chieff**2))))))) + q**2 * (-8 * chieff**4 + (-56 * q * chieff**4 + (q**8 * \
(chi2**4 + -1 * chi2**2 * chieff**2) + (q**7 * (-4 * chi2**4 + 18 * chi2**2 * \
chieff**2) + (q**4 * (chi2**4 + (-15 * chi2**2 * chieff**2 + -152 * \
chieff**4)) + (q**2 * (7 * chi2**2 * chieff**2 + -152 * chieff**4) + (2 * \
q**3 * (9 * chi2**2 * chieff**2 + -104 * chieff**4) + (q**6 * (6 * chi2**4 + \
(9 * chi2**2 * chieff**2 + -8 * chieff**4)) + -4 * q**5 * (chi2**4 + (9 * \
chi2**2 * chieff**2 + 14 * chieff**4)))))))))))) + (2 * ((1 + q))**4 * u**3 * \
(-1 * ((-1 + q))**2 * (-1 + (15 * q + 4 * q**2)) * chi1**6 + (-1 * q * \
chi1**4 * (10 * q**6 * chi2**2 + (4 * q**7 * chi2**2 + (-24 * chieff**2 + (16 \
* q * chieff**2 + (q**4 * (143 * chi2**2 + -32 * chieff**2) + (-5 * q**3 * \
(17 * chi2**2 + 12 * chieff**2) + (q**5 * (-87 * chi2**2 + 20 * chieff**2) + \
q**2 * (15 * chi2**2 + 32 * chieff**2)))))))) + (q**2 * chi1**2 * (-15 * q**9 \
* chi2**4 + (8 * chieff**4 + (24 * q * chieff**4 + (q**8 * (85 * chi2**4 + -4 \
* chi2**2 * chieff**2) + (q**7 * (-143 * chi2**4 + 24 * chi2**2 * chieff**2) \
+ (-2 * q**5 * (5 * chi2**4 + (48 * chi2**2 * chieff**2 + -44 * chieff**4)) + \
(-4 * q**4 * (chi2**4 + (5 * chi2**2 * chieff**2 + -30 * chieff**4)) + (8 * \
q**3 * (3 * chi2**2 * chieff**2 + 10 * chieff**4) + (q**6 * (87 * chi2**4 + \
(-20 * chi2**2 * chieff**2 + 24 * chieff**4)) + q**2 * (-4 * chi2**2 * \
chieff**2 + 40 * chieff**4)))))))))) + q**6 * chi2**2 * (q**8 * chi2**4 + (24 \
* chieff**4 + (88 * q * chieff**4 + (-20 * q**2 * chieff**2 * (chi2**2 + -6 * \
chieff**2) + (q**7 * (-17 * chi2**4 + 24 * chi2**2 * chieff**2) + (16 * q**3 \
* (2 * chi2**2 * chieff**2 + 5 * chieff**4) + (q**6 * (27 * chi2**4 + (-16 * \
chi2**2 * chieff**2 + 8 * chieff**4)) + (q**5 * (-7 * chi2**4 + (-32 * \
chi2**2 * chieff**2 + 24 * chieff**4)) + q**4 * (-4 * chi2**4 + (60 * chi2**2 \
* chieff**2 + 40 * chieff**4))))))))))))) + ((1 + q))**7 * u**2 * chieff * \
(-1 * ((-1 + q))**2 * (-3 + (49 * q + 16 * q**2)) * chi1**4 + (-2 * q * (1 + \
q) * chi1**2 * (22 * q**4 * chi2**2 + (-15 * q**5 * chi2**2 + (4 * q**6 * \
chi2**2 + (-4 * chieff**2 + (6 * q * chieff**2 + (4 * q**2 * (chi2**2 + -7 * \
chieff**2) + -1 * q**3 * (15 * chi2**2 + 38 * chieff**2))))))) + q**3 * (3 * \
q**8 * chi2**4 + (16 * chieff**4 + (80 * q * chieff**4 + (160 * q**2 * \
chieff**4 + (q**6 * (85 * chi2**4 + -4 * chi2**2 * chieff**2) + (q**7 * (-55 \
* chi2**4 + 8 * chi2**2 * chieff**2) + (q**5 * (-17 * chi2**4 + (44 * chi2**2 \
* chieff**2 + 16 * chieff**4)) + (4 * q**3 * (19 * chi2**2 * chieff**2 + 40 * \
chieff**4) + q**4 * (-16 * chi2**4 + (132 * chi2**2 * chieff**2 + 80 * \
chieff**4)))))))))))))))))\"\>"], "Output",
 CellLabel->
  "Out[195]=",ExpressionUUID->"34dbe160-4607-4ff9-b5d9-0e483ba689e5"],

Cell[BoxData["\<\"# Machine generated with eq_generator.nb\\ncoeff0 = -1/16 * \
q**(-1) * ((1 + q))**(-20) * (((-1 + q))**2 * ((1 + q))**(12) * (-1 * ((-1 + \
q))**2 * chi1**2 + q**2 * ((1 + q))**2 * chieff**2) * (-2 * q**3 * chi2**2 + \
(q**4 * chi2**2 + (-1 * chieff**2 + (-2 * q * chieff**2 + q**2 * (chi2**2 + \
-1 * chieff**2))))) + (-16 * ((-1 + q))**2 * q * u**6 * ((chi1**4 + (-1 * \
q**3 * (1 + q) * chi1**2 * chi2**2 + q**7 * chi2**4)))**2 * (-1 * (-1 + q) * \
chi1**2 + q * (q**3 * chi2**2 + (chieff**2 + (2 * q * chieff**2 + q**2 * (-1 \
* chi2**2 + chieff**2))))) + (-8 * (-1 + q) * q * ((1 + q))**3 * u**5 * (-1 * \
chi1**2 + q**4 * chi2**2) * chieff * ((5 + (-13 * q + 8 * q**2)) * chi1**6 + \
(q**8 * chi2**4 * (8 * q**2 * chi2**2 + (5 * q**4 * chi2**2 + (-8 * chieff**2 \
+ (-12 * q * chieff**2 + q**3 * (-13 * chi2**2 + 4 * chieff**2))))) + (q**4 * \
chi1**2 * chi2**2 * (-1 * q**5 * chi2**2 + (4 * chieff**2 + (8 * q * \
chieff**2 + (-2 * q**3 * (chi2**2 + -4 * chieff**2) + (-4 * q**2 * (chi2**2 + \
-2 * chieff**2) + q**4 * (7 * chi2**2 + 4 * chieff**2)))))) + -1 * chi1**4 * \
(2 * q**5 * chi2**2 + (4 * q**6 * chi2**2 + (-4 * q * chieff**2 + (q**4 * (-7 \
* chi2**2 + 8 * chieff**2) + q**3 * (chi2**2 + 12 * chieff**2)))))))) + (2 * \
((1 + q))**(11) * u * chieff * (((-1 + q))**4 * chi1**4 + (-1 * ((-1 + q))**2 \
* q * (1 + q) * chi1**2 * (-10 * q**3 * chi2**2 + (5 * q**4 * chi2**2 + (-5 * \
chieff**2 + (-8 * q * chieff**2 + q**2 * (5 * chi2**2 + -3 * chieff**2))))) + \
q**3 * (q**8 * chi2**4 + (-4 * chieff**4 + (-20 * q * chieff**4 + (5 * q**3 * \
chieff**2 * (chi2**2 + -8 * chieff**2) + (3 * q**6 * chi2**2 * (2 * chi2**2 + \
chieff**2) + (q**7 * (-4 * chi2**4 + 5 * chi2**2 * chieff**2) + (q**2 * (3 * \
chi2**2 * chieff**2 + -40 * chieff**4) + (q**4 * (chi2**4 + (-6 * chi2**2 * \
chieff**2 + -20 * chieff**4)) + -2 * q**5 * (2 * chi2**4 + (5 * chi2**2 * \
chieff**2 + 2 * chieff**4)))))))))))) + (2 * ((1 + q))**7 * u**3 * chieff * \
(((-1 + q))**2 * (-1 + (25 * q + 12 * q**2)) * chi1**6 + (q * chi1**4 * (85 * \
q**6 * chi2**2 + (-32 * q**7 * chi2**2 + (-4 * chieff**2 + (48 * q * \
chieff**2 + (q**4 * (83 * chi2**2 + -40 * chieff**2) + (4 * q**2 * (chi2**2 + \
7 * chieff**2) + (q**5 * (-103 * chi2**2 + 20 * chieff**2) + -1 * q**3 * (37 \
* chi2**2 + 84 * chieff**2)))))))) + (q**3 * chi1**2 * (-37 * q**8 * chi2**4 \
+ (4 * q**9 * chi2**4 + (16 * chieff**4 + (32 * q * chieff**4 + (16 * q**2 * \
chieff**2 * (chi2**2 + -2 * chieff**2) + (q**7 * (83 * chi2**4 + 16 * chi2**2 \
* chieff**2) + (q**6 * (-103 * chi2**4 + 24 * chi2**2 * chieff**2) + (q**5 * \
(85 * chi2**4 + (-8 * chi2**2 * chieff**2 + -32 * chieff**4)) + (8 * q**3 * \
(3 * chi2**2 * chieff**2 + -16 * chieff**4) + -8 * q**4 * (4 * chi2**4 + \
(chi2**2 * chieff**2 + 14 * chieff**4))))))))))) + q**7 * chi2**2 * (-1 * \
q**8 * chi2**4 + (-32 * chieff**4 + (-112 * q * chieff**4 + (q**7 * (27 * \
chi2**4 + -4 * chi2**2 * chieff**2) + (q**6 * (-39 * chi2**4 + 48 * chi2**2 * \
chieff**2) + (4 * q**2 * (5 * chi2**2 * chieff**2 + -32 * chieff**4) + (-8 * \
q**3 * (5 * chi2**2 * chieff**2 + 4 * chieff**4) + (4 * q**4 * (3 * chi2**4 + \
(-21 * chi2**2 * chieff**2 + 8 * chieff**4)) + q**5 * (chi2**4 + (28 * \
chi2**2 * chieff**2 + 16 * chieff**4))))))))))))) + (((1 + q))**4 * u**4 * \
(((-1 + q))**2 * (-1 + (20 * q + 8 * q**2)) * chi1**8 + (-4 * (-1 + q) * q * \
chi1**6 * (-11 * q**5 * chi2**2 + (8 * q**6 * chi2**2 + (-8 * chieff**2 + (20 \
* q * chieff**2 + (q**4 * (30 * chi2**2 + -22 * chieff**2) + (-8 * q**3 * (4 \
* chi2**2 + chieff**2) + q**2 * (5 * chi2**2 + 42 * chieff**2))))))) + (q**10 \
* chi2**4 * (-1 * q**8 * chi2**4 + (-96 * chieff**4 + (-288 * q * chieff**4 + \
(q**7 * (22 * chi2**4 + -32 * chi2**2 * chieff**2) + (8 * q**2 * (11 * \
chi2**2 * chieff**2 + -26 * chieff**4) + (-8 * q**3 * (7 * chi2**2 * \
chieff**2 + -16 * chieff**4) + (q**6 * (-33 * chi2**4 + (112 * chi2**2 * \
chieff**2 + -16 * chieff**4)) + (4 * q**5 * (chi2**4 + (22 * chi2**2 * \
chieff**2 + 8 * chieff**4)) + 8 * q**4 * (chi2**4 + (-25 * chi2**2 * \
chieff**2 + 24 * chieff**4)))))))))) + (4 * q**6 * chi1**2 * chi2**2 * (5 * \
q**9 * chi2**4 + (24 * chieff**4 + (56 * q * chieff**4 + (12 * q**3 * \
chieff**2 * (chi2**2 + -4 * chieff**2) + (8 * q**2 * chieff**2 * (-2 * \
chi2**2 + chieff**2) + (q**7 * (62 * chi2**4 + -6 * chi2**2 * chieff**2) + \
(q**8 * (-37 * chi2**4 + 2 * chi2**2 * chieff**2) + (q**4 * (-8 * chi2**4 + \
(52 * chi2**2 * chieff**2 + 8 * chieff**4)) + (q**6 * (-41 * chi2**4 + (-38 * \
chi2**2 * chieff**2 + 24 * chieff**4)) + q**5 * (19 * chi2**4 + (-6 * chi2**2 \
* chieff**2 + 56 * chieff**4))))))))))) + 2 * q**2 * chi1**4 * (-2 * q**9 * \
chi2**4 + (4 * q**10 * chi2**4 + (-8 * chieff**4 + (16 * q * chieff**4 + (4 * \
q**2 * chieff**2 * (chi2**2 + 24 * chieff**2) + (q**8 * (53 * chi2**4 + -32 * \
chi2**2 * chieff**2) + (q**7 * (-110 * chi2**4 + 24 * chi2**2 * chieff**2) + \
(q**6 * (53 * chi2**4 + (104 * chi2**2 * chieff**2 + -48 * chieff**4)) + (4 * \
q**4 * (chi2**4 + (-19 * chi2**2 * chieff**2 + -26 * chieff**4)) + (q**3 * \
(-12 * chi2**2 * chieff**2 + 64 * chieff**4) + -2 * q**5 * (chi2**4 + (6 * \
chi2**2 * chieff**2 + 72 * chieff**4)))))))))))))))) + ((1 + q))**8 * u**2 * \
(((-1 + q))**4 * chi1**6 + (-1 * ((-1 + q))**2 * chi1**4 * (4 * q**5 * \
chi2**2 + (10 * q**6 * chi2**2 + (chieff**2 + (-38 * q * chieff**2 + (q**3 * \
(26 * chi2**2 + -86 * chieff**2) + (-1 * q**4 * (39 * chi2**2 + 23 * \
chieff**2) + -1 * q**2 * (chi2**2 + 102 * chieff**2))))))) + (q**2 * chi1**2 \
* (-28 * q**9 * chi2**4 + (q**10 * chi2**4 + (32 * chieff**4 + (72 * q * \
chieff**4 + (48 * q**3 * chieff**2 * (chi2**2 + -5 * chieff**2) + (q**8 * (92 \
* chi2**4 + -22 * chi2**2 * chieff**2) + (-12 * q**7 * (9 * chi2**4 + -4 * \
chi2**2 * chieff**2) + (8 * q**5 * (2 * chi2**4 + (-12 * chi2**2 * chieff**2 \
+ -11 * chieff**4)) + (q**6 * (37 * chi2**4 + (22 * chi2**2 * chieff**2 + -8 \
* chieff**4)) + (-2 * q**2 * (11 * chi2**2 * chieff**2 + 20 * chieff**4) + -2 \
* q**4 * (5 * chi2**4 + (-11 * chi2**2 * chieff**2 + 120 * \
chieff**4)))))))))))) + q**4 * (-16 * chieff**6 + (-96 * q * chieff**6 + (-8 \
* q**2 * chieff**4 * (chi2**2 + 30 * chieff**2) + (-4 * q**9 * (chi2**6 + -10 \
* chi2**4 * chieff**2) + (q**10 * (chi2**6 + -1 * chi2**4 * chieff**2) + (-4 \
* q**7 * (chi2**6 + (20 * chi2**4 * chieff**2 + -18 * chi2**2 * chieff**4)) + \
(q**8 * (6 * chi2**6 + (25 * chi2**4 * chieff**2 + 32 * chi2**2 * chieff**4)) \
+ (q**4 * (23 * chi2**4 * chieff**2 + (-240 * chi2**2 * chieff**4 + -240 * \
chieff**6)) + (q**6 * (chi2**6 + (-47 * chi2**4 * chieff**2 + (-40 * chi2**2 \
* chieff**4 + -16 * chieff**6))) + (8 * q**5 * (5 * chi2**4 * chieff**2 + \
(-30 * chi2**2 * chieff**4 + -12 * chieff**6)) + -8 * q**3 * (11 * chi2**2 * \
chieff**4 + 40 * chieff**6))))))))))))))))))))\"\>"], "Output",
 CellLabel->"Out[196]=",ExpressionUUID->"ef5040d2-f19e-46b8-90a9-a803f58ccce8"]
}, Open  ]],

Cell["Large separation limit", \
"Text",ExpressionUUID->"52ed52ad-25ee-4b31-b3af-5637e30185f8"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Limit", "[", 
   RowBox[{
    RowBox[{"coeffs", "/", "overalldiscr"}], ",", 
    RowBox[{"u", "->", "0"}]}], "]"}], "//", "FullSimplify"}]], "Input",
 CellLabel->
  "In[1536]:=",ExpressionUUID->"2b663d0d-abea-4cf3-bc6c-a88849f38c10"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    FractionBox["1", 
     RowBox[{"16", " ", "q", " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "8"]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "q"}], ")"}], "6"], " ", 
       SuperscriptBox["q", "2"], " ", 
       SuperscriptBox["\[Chi]1", "2"], " ", 
       SuperscriptBox["\[Chi]2", "2"]}], "-", 
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "q"}], ")"}], "4"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox["\[Chi]1", "2"], "+", 
         RowBox[{
          SuperscriptBox["q", "4"], " ", 
          SuperscriptBox["\[Chi]2", "2"]}]}], ")"}], " ", 
       SuperscriptBox["\[Chi]eff", "2"]}], "+", 
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "q"}], ")"}], "2"], " ", 
       SuperscriptBox["q", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "4"], " ", 
       SuperscriptBox["\[Chi]eff", "4"]}]}], ")"}]}], ",", 
   FractionBox[
    RowBox[{
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "q"}], ")"}], "4"], " ", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["\[Chi]1", "2"], "+", 
        RowBox[{
         SuperscriptBox["q", "3"], " ", 
         SuperscriptBox["\[Chi]2", "2"]}]}], ")"}], " ", "\[Chi]eff"}], "-", 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "q"}], ")"}], "2"], " ", "q", " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "3"], " ", 
      SuperscriptBox["\[Chi]eff", "3"]}]}], 
    RowBox[{"8", " ", "q", " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "5"]}]], ",", 
   FractionBox[
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "q"}], ")"}], "4"]}], " ", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["\[Chi]1", "2"], "+", 
        RowBox[{
         SuperscriptBox["q", "2"], " ", 
         SuperscriptBox["\[Chi]2", "2"]}]}], ")"}]}], "+", 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", 
         SuperscriptBox["q", "2"]}], ")"}], "2"], " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", 
        RowBox[{"q", " ", 
         RowBox[{"(", 
          RowBox[{"4", "+", "q"}], ")"}]}]}], ")"}], " ", 
      SuperscriptBox["\[Chi]eff", "2"]}]}], 
    RowBox[{"16", " ", "q", " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "4"]}]], ",", 
   RowBox[{"-", 
    FractionBox[
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "q"}], ")"}], "2"], " ", "\[Chi]eff"}], 
     RowBox[{"8", " ", "q"}]]}], ",", 
   FractionBox[
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", "q"}], ")"}], "2"], 
    RowBox[{"16", " ", "q"}]], ",", "0"}], "}"}]], "Output",
 CellLabel->
  "Out[1536]=",ExpressionUUID->"b1aeffcd-9298-4cc8-ad33-d8f26b57629a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Solve", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Limit", "[", 
      RowBox[{
       RowBox[{"discr", "/", "overalldiscr"}], ",", 
       RowBox[{"u", "->", "0"}]}], "]"}], "==", "0"}], ",", "\[Kappa]"}], 
   "]"}], "//", "FullSimplify"}]], "Input",
 CellLabel->
  "In[1537]:=",ExpressionUUID->"496647fd-3dfb-46b3-8be1-ac52da6f30dc"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"\[Kappa]", "\[Rule]", 
     FractionBox[
      RowBox[{"\[Chi]eff", "+", 
       RowBox[{"q", " ", 
        RowBox[{"(", 
         RowBox[{"\[Chi]2", "-", 
          RowBox[{"q", " ", "\[Chi]2"}], "+", "\[Chi]eff"}], ")"}]}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"]]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"\[Kappa]", "\[Rule]", 
     FractionBox[
      RowBox[{"\[Chi]eff", "+", 
       RowBox[{"q", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", "\[Chi]2"}], "+", 
          "\[Chi]eff"}], ")"}]}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"]]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"\[Kappa]", "\[Rule]", 
     FractionBox[
      RowBox[{"\[Chi]1", "-", 
       RowBox[{"q", " ", "\[Chi]1"}], "+", 
       RowBox[{"q", " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], " ", "\[Chi]eff"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"]]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"\[Kappa]", "\[Rule]", 
     FractionBox[
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", "\[Chi]1"}], "+", 
       RowBox[{"q", " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], " ", "\[Chi]eff"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"]]}], "}"}]}], "}"}]], "Output",
 CellLabel->
  "Out[1537]=",ExpressionUUID->"180c641e-900d-41e3-b732-070cdfc7fa05"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["(v) Limits", \
"Section",ExpressionUUID->"a71f7e8e-8c82-449b-86f6-9fb478fba622"],

Cell[BoxData[
 RowBox[{"Clear", "@@", 
  RowBox[{"DeleteCases", "[", 
   RowBox[{
    RowBox[{"Names", "@", "\"\<`*\>\""}], ",", 
    RowBox[{"\"\<ToPython\>\"", "|", "\"\<precessionform\>\""}]}], 
   "]"}]}]], "Input",
 CellLabel->
  "In[1538]:=",ExpressionUUID->"da6425a3-5b5f-4fa7-be4f-93539f8c5d6f"],

Cell["Large separation limit", \
"Text",ExpressionUUID->"d1d412ce-19ad-4b12-8543-ac6425f8a02d"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"chieff", " ", "=", " ", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"\[Chi]1", " ", "cos\[Theta]1x"}], " ", "+", " ", 
      RowBox[{"q", " ", "\[Chi]2", " ", "cos\[Theta]2x"}]}], ")"}], "/", 
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}]}], "//", 
   "Simplify"}]}], "\[IndentingNewLine]", 
 RowBox[{"kappa", " ", "=", " ", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"\[Chi]1", " ", "cos\[Theta]1x"}], " ", "+", " ", 
      RowBox[{
       RowBox[{"q", "^", "2"}], " ", "\[Chi]2", " ", "cos\[Theta]2x"}]}], 
     ")"}], "/", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}], "//", 
   "Simplify"}]}]}], "Input",
 CellLabel->
  "In[197]:=",ExpressionUUID->"b4500df8-3162-40a4-a33d-a530f3735daa"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{"cos\[Theta]1x", " ", "\[Chi]1"}], "+", 
   RowBox[{"cos\[Theta]2x", " ", "q", " ", "\[Chi]2"}]}], 
  RowBox[{"1", "+", "q"}]]], "Output",
 CellLabel->
  "Out[197]=",ExpressionUUID->"12be9da3-a68e-4c4e-93b4-537d2821bdfe"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{"cos\[Theta]1x", " ", "\[Chi]1"}], "+", 
   RowBox[{"cos\[Theta]2x", " ", 
    SuperscriptBox["q", "2"], " ", "\[Chi]2"}]}], 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{"1", "+", "q"}], ")"}], "2"]]], "Output",
 CellLabel->"Out[198]=",ExpressionUUID->"84a46134-9a50-457f-8956-bc5be54a7bee"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Solve", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"chieff", "==", "\[Chi]eff"}], ",", 
        RowBox[{"kappa", "==", "\[Kappa]"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"cos\[Theta]1x", ",", "cos\[Theta]2x"}], "}"}]}], "]"}], "[", 
    RowBox[{"[", "1", "]"}], "]"}], "//", "Simplify"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"cos\[Theta]1sol", "=", " ", 
  RowBox[{"cos\[Theta]1x", "/.", 
   RowBox[{"Out", "[", 
    RowBox[{"-", "1"}], "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{"cos\[Theta]2sol", " ", "=", 
  RowBox[{"cos\[Theta]2x", "/.", 
   RowBox[{"Out", "[", 
    RowBox[{"-", "2"}], "]"}]}]}]}], "Input",
 CellLabel->
  "In[199]:=",ExpressionUUID->"ded1d79c-d89a-41d0-83fc-a8f8f2a618ee"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"\[Kappa]", "+", 
      RowBox[{"q", " ", "\[Kappa]"}], "-", 
      RowBox[{"q", " ", "\[Chi]eff"}]}], ")"}]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", "\[Chi]1"}]]}]], "Output",
 CellLabel->
  "Out[200]=",ExpressionUUID->"52647810-84e6-48a2-8ec0-f03d5ecd7ef1"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{"(", 
    RowBox[{"1", "+", "q"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{"\[Kappa]", "+", 
     RowBox[{"q", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}], 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", "q", " ", 
   "\[Chi]2"}]]], "Output",
 CellLabel->"Out[201]=",ExpressionUUID->"7b4697df-61c2-47a6-976f-aa2e127d58e3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"deltachi", " ", "=", " ", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"\[Chi]1", " ", "cos\[Theta]1sol"}], " ", "-", " ", 
      RowBox[{"q", " ", "\[Chi]2", " ", "cos\[Theta]2sol"}]}], ")"}], "/", 
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}]}], "//", "Simplify"}]}]], "Input",
 CellLabel->
  "In[202]:=",ExpressionUUID->"11af1178-2d80-4949-8266-8e6aa529de92"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}], 
   RowBox[{
    RowBox[{"-", "1"}], "+", "q"}]]}]], "Output",
 CellLabel->"Out[202]=",ExpressionUUID->"ec75af49-971d-4451-a589-4a3b82ff9b11"]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"S1", " ", "=", " ", 
   RowBox[{"\[Chi]1", " ", "/", " ", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"S2", " ", "=", " ", 
   RowBox[{"\[Chi]2", " ", 
    RowBox[{
     RowBox[{"q", "^", "2"}], " ", "/", " ", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"L", "=", 
   RowBox[{
    RowBox[{"q", " ", "/", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}], " ", 
    RowBox[{"Sqrt", "[", "r", "]"}]}]}], ";"}]}], "Input",
 CellLabel->
  "In[203]:=",ExpressionUUID->"e71664c5-d2f7-411d-8c53-f0d4414ea27b"],

Cell["Deriving the limits on \[Kappa] from the aligned configurations", \
"Text",ExpressionUUID->"a762e971-31d7-4e1e-9a62-8499bea78085"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"J", "=", 
   RowBox[{"L", "+", "S1", "+", "S2"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[Kappa]", "=", 
   RowBox[{"Simplify", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"J", "^", "2"}], 
         RowBox[{"Sign", "[", "J", "]"}]}], "-", " ", 
        RowBox[{"L", "^", "2"}]}], ")"}], "/", 
      RowBox[{"(", 
       RowBox[{"2", "L"}], ")"}]}], ",", 
     RowBox[{
      RowBox[{"q", ">", "0"}], "&&", 
      RowBox[{"q", "<", "1"}], "&&", 
      RowBox[{"r", ">", "0"}], "&&", 
      RowBox[{"\[Chi]1", ">", "0"}], "&&", 
      RowBox[{"\[Chi]2", ">", "0"}]}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"my\[Kappa]0", " ", "=", " ", 
   RowBox[{
    FractionBox[
     RowBox[{"\[Chi]1", "+", 
      RowBox[{
       SuperscriptBox["q", "2"], " ", "\[Chi]2"}]}], 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"], " "}]], 
    RowBox[{"(", 
     RowBox[{
      FractionBox[
       RowBox[{"\[Chi]1", "+", 
        RowBox[{
         SuperscriptBox["q", "2"], " ", "\[Chi]2"}]}], 
       RowBox[{"2", "q", " ", 
        SqrtBox["r"]}]], "+", "1"}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"Simplify", "[", 
  RowBox[{
   RowBox[{"\[Kappa]", "-", "my\[Kappa]0"}], ",", 
   RowBox[{
    RowBox[{"q", ">", "0"}], "&&", 
    RowBox[{"q", "<", "1"}], "&&", 
    RowBox[{"r", ">", "0"}], "&&", 
    RowBox[{"\[Chi]1", ">", "0"}], "&&", 
    RowBox[{"\[Chi]2", ">", "0"}]}]}], "]"}], "\[IndentingNewLine]"}], "Input",
 CellLabel->
  "In[206]:=",ExpressionUUID->"df3c994f-f8d1-4d25-a246-f12ae8087d05"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[209]=",ExpressionUUID->"b2392a27-8a0b-4914-9368-9a5d42c96c87"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"J", "=", 
   RowBox[{"L", "-", "S1", "-", "S2"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Kappa]", "=", 
    RowBox[{"Simplify", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"J", "^", "2"}], 
          RowBox[{"Sign", "[", "J", "]"}]}], "-", " ", 
         RowBox[{"L", "^", "2"}]}], ")"}], "/", 
       RowBox[{"(", 
        RowBox[{"2", "L"}], ")"}]}], ",", 
      RowBox[{
       RowBox[{"q", ">", "0"}], "&&", 
       RowBox[{"q", "<", "1"}], "&&", 
       RowBox[{"r", ">", "0"}], "&&", 
       RowBox[{"\[Chi]1", ">", "0"}], "&&", 
       RowBox[{"\[Chi]2", ">", "0"}]}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]", 
  RowBox[{"(*", 
   RowBox[{"Limit", "[", 
    RowBox[{"\[Kappa]", ",", 
     RowBox[{"r", "->", 
      RowBox[{"+", "\[Infinity]"}]}], ",", 
     RowBox[{"Assumptions", "->", 
      RowBox[{
       RowBox[{"q", ">", "0"}], "&&", 
       RowBox[{"q", "<", "1"}], "&&", 
       RowBox[{"\[Chi]1", ">", "0"}], "&&", 
       RowBox[{"\[Chi]2", ">", "0"}]}]}]}], "]"}], 
   "*)"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"my\[Kappa]1", " ", "=", " ", 
   RowBox[{
    FractionBox[
     RowBox[{"q", 
      SqrtBox["r"]}], 
     RowBox[{"2", " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"], " "}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "-", 
          FractionBox[
           RowBox[{"\[Chi]1", "+", 
            RowBox[{
             SuperscriptBox["q", "2"], " ", "\[Chi]2"}]}], 
           RowBox[{"q", " ", 
            SqrtBox["r"]}]]}], ")"}], "2"], " ", 
       RowBox[{"Sign", "[", 
        RowBox[{"1", "-", 
         FractionBox[
          RowBox[{"\[Chi]1", "+", 
           RowBox[{
            SuperscriptBox["q", "2"], " ", "\[Chi]2"}]}], 
          RowBox[{"q", " ", 
           SqrtBox["r"]}]]}], "]"}]}], "-", "1"}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"Simplify", "[", 
  RowBox[{
   RowBox[{"\[Kappa]", "-", "my\[Kappa]1"}], ",", 
   RowBox[{
    RowBox[{"q", ">", "0"}], "&&", 
    RowBox[{"q", "<", "1"}], "&&", 
    RowBox[{"r", ">", "0"}], "&&", 
    RowBox[{"\[Chi]1", ">", "0"}], "&&", 
    RowBox[{"\[Chi]2", ">", "0"}]}]}], "]"}], "\[IndentingNewLine]"}], "Input",
 CellLabel->
  "In[210]:=",ExpressionUUID->"f8e06607-675a-450c-8788-7903d5c8ce92"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[213]=",ExpressionUUID->"939257da-30dc-464a-8ff5-c884d70cbef6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"J", "=", 
   RowBox[{
    RowBox[{"Abs", "[", 
     RowBox[{"S1", "-", "S2"}], "]"}], "-", "L"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"\[Kappa]", "=", 
    RowBox[{"Simplify", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"J", "^", "2"}], 
          RowBox[{"Sign", "[", "J", "]"}]}], "-", " ", 
         RowBox[{"L", "^", "2"}]}], ")"}], "/", 
       RowBox[{"(", 
        RowBox[{"2", "L"}], ")"}]}], ",", 
      RowBox[{
       RowBox[{"q", ">", "0"}], "&&", 
       RowBox[{"q", "<", "1"}], "&&", 
       RowBox[{"r", ">", "0"}], "&&", 
       RowBox[{"\[Chi]1", ">", "0"}], "&&", 
       RowBox[{"\[Chi]2", ">", "0"}]}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]", 
  RowBox[{"(*", 
   RowBox[{"Limit", "[", 
    RowBox[{"\[Kappa]", ",", 
     RowBox[{"r", "->", 
      RowBox[{"+", "\[Infinity]"}]}], ",", 
     RowBox[{"Assumptions", "->", 
      RowBox[{
       RowBox[{"q", ">", "0"}], "&&", 
       RowBox[{"q", "<", "1"}], "&&", 
       RowBox[{"\[Chi]1", ">", "0"}], "&&", 
       RowBox[{"\[Chi]2", ">", "0"}]}]}]}], "]"}], 
   "*)"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"my\[Kappa]2", " ", "=", " ", 
   RowBox[{
    FractionBox[
     RowBox[{"q", 
      SqrtBox["r"]}], 
     RowBox[{"2", " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"], " "}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          FractionBox[
           RowBox[{"Abs", "[", 
            RowBox[{"\[Chi]1", "-", 
             RowBox[{
              SuperscriptBox["q", "2"], " ", "\[Chi]2"}]}], "]"}], 
           RowBox[{"q", " ", 
            SqrtBox["r"]}]], "-", "1"}], ")"}], "2"], " ", 
       RowBox[{"Sign", "[", 
        RowBox[{
         FractionBox[
          RowBox[{"Abs", "[", 
           RowBox[{"\[Chi]1", "-", 
            RowBox[{
             SuperscriptBox["q", "2"], " ", "\[Chi]2"}]}], "]"}], 
          RowBox[{"q", " ", 
           SqrtBox["r"]}]], "-", "1"}], "]"}]}], "-", "1"}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"Simplify", "[", 
  RowBox[{
   RowBox[{"\[Kappa]", "-", "my\[Kappa]2"}], ",", 
   RowBox[{
    RowBox[{"q", ">", "0"}], "&&", 
    RowBox[{"q", "<", "1"}], "&&", 
    RowBox[{"r", ">", "0"}], "&&", 
    RowBox[{"\[Chi]1", ">", "0"}], "&&", 
    RowBox[{"\[Chi]2", ">", "0"}]}]}], "]"}]}], "Input",
 CellLabel->
  "In[214]:=",ExpressionUUID->"23486271-39f0-4cfc-ae44-31b8ffd28a70"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[217]=",ExpressionUUID->"a70800b4-2c0a-4e0d-8f09-7333c88c7ba0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"J", "=", "0"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[Kappa]3", "=", 
   RowBox[{"Simplify", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"J", "^", "2"}], 
         RowBox[{"Sign", "[", "J", "]"}]}], "-", " ", 
        RowBox[{"L", "^", "2"}]}], ")"}], "/", 
      RowBox[{"(", 
       RowBox[{"2", "L"}], ")"}]}], ",", 
     RowBox[{
      RowBox[{"q", ">", "0"}], "&&", 
      RowBox[{"q", "<", "1"}], "&&", 
      RowBox[{"r", ">", "0"}], "&&", 
      RowBox[{"\[Chi]1", ">", "0"}], "&&", 
      RowBox[{"\[Chi]2", ">", "0"}]}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Limit", "[", 
   RowBox[{"\[Kappa]", ",", 
    RowBox[{"r", "->", 
     RowBox[{"+", "\[Infinity]"}]}], ",", 
    RowBox[{"Assumptions", "->", 
     RowBox[{
      RowBox[{"q", ">", "0"}], "&&", 
      RowBox[{"q", "<", "1"}], "&&", 
      RowBox[{"\[Chi]1", ">", "0"}], "&&", 
      RowBox[{"\[Chi]2", ">", "0"}]}]}]}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"my\[Kappa]3", "  ", "=", " ", 
   RowBox[{"-", 
    FractionBox[
     RowBox[{"q", 
      SqrtBox["r"]}], 
     RowBox[{"2", " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"], " "}]]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"Simplify", "[", 
  RowBox[{
   RowBox[{"\[Kappa]", "-", "my\[Kappa]3"}], " ", ",", 
   RowBox[{
    RowBox[{"q", ">", "0"}], "&&", 
    RowBox[{"q", "<", "1"}], "&&", 
    RowBox[{"r", ">", "0"}], "&&", 
    RowBox[{"\[Chi]1", ">", "0"}], "&&", 
    RowBox[{"\[Chi]2", ">", "0"}]}]}], "]"}]}], "Input",
 CellLabel->
  "In[218]:=",ExpressionUUID->"e567413d-1264-40e3-a77a-585e12837055"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "q"}], " ", 
       SqrtBox["r"]}], "+", 
      RowBox[{"Abs", "[", 
       RowBox[{"\[Chi]1", "-", 
        RowBox[{
         SuperscriptBox["q", "2"], " ", "\[Chi]2"}]}], "]"}]}], ")"}], "2"], 
   " ", 
   RowBox[{"Sign", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "q"}], " ", 
      SqrtBox["r"]}], "+", 
     RowBox[{"Abs", "[", 
      RowBox[{"\[Chi]1", "-", 
       RowBox[{
        SuperscriptBox["q", "2"], " ", "\[Chi]2"}]}], "]"}]}], "]"}]}], 
  RowBox[{"2", " ", "q", " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
   SqrtBox["r"]}]]], "Output",
 CellLabel->"Out[222]=",ExpressionUUID->"0ea8055b-60bd-4342-a8e8-f15c7e1bcd52"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Limit", " ", 
  FractionBox[
   RowBox[{"q", 
    SqrtBox["r"]}], 
   RowBox[{"2", " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], "2"], " "}]], 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "-", 
        FractionBox[
         RowBox[{"\[Chi]1", "+", 
          RowBox[{
           SuperscriptBox["q", "2"], " ", "\[Chi]2"}]}], 
         RowBox[{"q", " ", 
          SqrtBox["r"]}]]}], ")"}], "2"], " ", 
     RowBox[{"Sign", "[", 
      RowBox[{"1", "-", 
       FractionBox[
        RowBox[{"\[Chi]1", "+", 
         RowBox[{
          SuperscriptBox["q", "2"], " ", "\[Chi]2"}]}], 
        RowBox[{"q", " ", 
         SqrtBox["r"]}]]}], "]"}]}], "-", "1"}], ")"}]}]], "Input",
 CellLabel->
  "In[223]:=",ExpressionUUID->"04eb11b9-9755-4a42-a9a7-df99e92dbeff"],

Cell[BoxData[
 FractionBox[
  RowBox[{"Limit", " ", "q", " ", 
   SqrtBox["r"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "-", 
         FractionBox[
          RowBox[{"\[Chi]1", "+", 
           RowBox[{
            SuperscriptBox["q", "2"], " ", "\[Chi]2"}]}], 
          RowBox[{"q", " ", 
           SqrtBox["r"]}]]}], ")"}], "2"], " ", 
      RowBox[{"Sign", "[", 
       RowBox[{"1", "-", 
        FractionBox[
         RowBox[{"\[Chi]1", "+", 
          RowBox[{
           SuperscriptBox["q", "2"], " ", "\[Chi]2"}]}], 
         RowBox[{"q", " ", 
          SqrtBox["r"]}]]}], "]"}]}]}], ")"}]}], 
  RowBox[{"2", " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], "2"]}]]], "Output",
 CellLabel->"Out[223]=",ExpressionUUID->"aca640be-78e9-4ede-af0a-15a17eeb8ac2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"ass", " ", "=", " ", 
   RowBox[{
    RowBox[{"q", ">", "0"}], "&&", 
    RowBox[{"q", "<", "1"}], "&&", 
    RowBox[{"\[Chi]1", ">", "0"}], "&&", 
    RowBox[{"\[Chi]2", ">", "0"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"Simplify", "[", 
  RowBox[{
   RowBox[{"Max", "[", " ", 
    RowBox[{
     RowBox[{"Limit", "[", 
      RowBox[{"my\[Kappa]1", ",", 
       RowBox[{"r", "->", "\[Infinity]"}], ",", 
       RowBox[{"Assumptions", "->", "ass"}]}], "]"}], ",", " ", 
     RowBox[{"Limit", "[", 
      RowBox[{"my\[Kappa]2", ",", 
       RowBox[{"r", "->", "\[Infinity]"}], ",", 
       RowBox[{"Assumptions", "->", "ass"}]}], "]"}], ",", 
     RowBox[{"Limit", "[", 
      RowBox[{"my\[Kappa]3", ",", 
       RowBox[{"r", "->", "\[Infinity]"}], ",", 
       RowBox[{"Assumptions", "->", "ass"}]}], "]"}]}], "]"}], ",", "ass"}], 
  "]"}]}], "Input",
 CellLabel->
  "In[224]:=",ExpressionUUID->"d0e63910-0a18-4bae-9578-138419c8ed22"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox[
   RowBox[{"\[Chi]1", "+", 
    RowBox[{
     SuperscriptBox["q", "2"], " ", "\[Chi]2"}]}], 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], "2"]]}]], "Output",
 CellLabel->"Out[225]=",ExpressionUUID->"1dd93c85-e3e0-4cec-adcd-18042d496cf6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["(vi) Some elliptic integrals", \
"Section",ExpressionUUID->"bfebcdd1-f771-4e86-a2c4-77d50b5eb5b5"],

Cell[BoxData[
 RowBox[{"Clear", "@@", 
  RowBox[{"DeleteCases", "[", 
   RowBox[{
    RowBox[{"Names", "@", "\"\<`*\>\""}], ",", 
    RowBox[{"\"\<ToPython\>\"", "|", "\"\<precessionform\>\""}]}], 
   "]"}]}]], "Input",
 CellLabel->
  "In[226]:=",ExpressionUUID->"35d9ad59-fdd6-437f-94dd-aa89a6370e94"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"integral1", "=", " ", 
   RowBox[{
    FractionBox[
     RowBox[{"2", " "}], 
     RowBox[{" ", 
      SqrtBox[
       RowBox[{"c", "-", "a"}]]}]], " ", 
    RowBox[{"EllipticF", "[", 
     RowBox[{
      RowBox[{"ArcSin", "[", 
       SqrtBox[
        FractionBox[
         RowBox[{"x", "-", "a"}], 
         RowBox[{"b", "-", "a"}]]], "]"}], ",", 
      FractionBox[
       RowBox[{"b", "-", "a"}], 
       RowBox[{"c", "-", "a"}]]}], "]"}]}]}], " ", ";"}], "\n", 
 RowBox[{
  RowBox[{"Simplify", "[", 
   RowBox[{
    RowBox[{"D", "[", 
     RowBox[{"integral1", ",", "x"}], "]"}], ",", 
    RowBox[{"a", "<", "x", "<", "b", "<", "c"}]}], "]"}], "//", 
  "Simplify"}], "\[IndentingNewLine]", 
 RowBox[{"integral1", "/.", 
  RowBox[{"x", "->", "a"}]}], "\[IndentingNewLine]", 
 RowBox[{"integral1", "/.", 
  RowBox[{"x", "->", "b"}]}]}], "Input",
 CellLabel->
  "In[227]:=",ExpressionUUID->"4a0f0e7d-9435-4ad4-8a83-ff90a8be3662"],

Cell[BoxData[
 FractionBox["1", 
  SqrtBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"b", "-", "x"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"c", "-", "x"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "a"}], "+", "x"}], ")"}]}]]]], "Output",
 CellLabel->
  "Out[228]=",ExpressionUUID->"9f537aad-8322-45de-a3ac-963d70cfdc5d"],

Cell[BoxData["0"], "Output",
 CellLabel->
  "Out[229]=",ExpressionUUID->"91f69816-9543-475e-9193-a64256c3959d"],

Cell[BoxData[
 FractionBox[
  RowBox[{"2", " ", 
   RowBox[{"EllipticK", "[", 
    FractionBox[
     RowBox[{
      RowBox[{"-", "a"}], "+", "b"}], 
     RowBox[{
      RowBox[{"-", "a"}], "+", "c"}]], "]"}]}], 
  SqrtBox[
   RowBox[{
    RowBox[{"-", "a"}], "+", "c"}]]]], "Output",
 CellLabel->"Out[230]=",ExpressionUUID->"eca9ccdf-0143-4700-9579-21b526be8432"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"integral2", "=", 
   RowBox[{
    FractionBox[
     RowBox[{"2", " "}], 
     SqrtBox[
      RowBox[{"c", "-", "a"}]]], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"c", " ", 
       RowBox[{"EllipticF", "[", 
        RowBox[{
         RowBox[{"ArcSin", "[", 
          SqrtBox[
           FractionBox[
            RowBox[{"x", "-", "a"}], 
            RowBox[{"b", "-", "a"}]]], "]"}], ",", 
         FractionBox[
          RowBox[{"b", "-", "a"}], 
          RowBox[{"c", "-", "a"}]]}], "]"}]}], "-", " ", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"c", "-", "a"}], ")"}], " ", 
       RowBox[{"EllipticE", "[", 
        RowBox[{
         RowBox[{"ArcSin", "[", 
          SqrtBox[
           FractionBox[
            RowBox[{"x", "-", "a"}], 
            RowBox[{"b", "-", "a"}]]], "]"}], ",", 
         FractionBox[
          RowBox[{"b", "-", "a"}], 
          RowBox[{"c", "-", "a"}]]}], "]"}]}]}], ")"}]}]}], " ", 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Simplify", "[", 
   RowBox[{
    RowBox[{"D", "[", 
     RowBox[{"integral2", ",", "x"}], "]"}], ",", 
    RowBox[{"a", "<", "x", "<", "b", "<", "c"}]}], "]"}], "//", 
  "Simplify"}], "\[IndentingNewLine]", 
 RowBox[{"integral2", "/.", 
  RowBox[{"x", "->", "a"}]}], "\[IndentingNewLine]", 
 RowBox[{"integral2", "/.", 
  RowBox[{"x", "->", "b"}]}]}], "Input",
 CellLabel->
  "In[235]:=",ExpressionUUID->"0d4b6e11-b27a-4e14-a431-6121082a56fa"],

Cell[BoxData[
 FractionBox["x", 
  SqrtBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "a"}], "+", "x"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "b"}], "+", "x"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "c"}], "+", "x"}], ")"}]}]]]], "Output",
 CellLabel->
  "Out[236]=",ExpressionUUID->"d8a18626-ed70-4b8a-a829-03136a6f5f16"],

Cell[BoxData["0"], "Output",
 CellLabel->
  "Out[237]=",ExpressionUUID->"501a8719-d4ce-4388-aa12-f5da74b20e82"],

Cell[BoxData[
 FractionBox[
  RowBox[{"2", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "a"}], "+", "c"}], ")"}], " ", 
        RowBox[{"EllipticE", "[", 
         FractionBox[
          RowBox[{
           RowBox[{"-", "a"}], "+", "b"}], 
          RowBox[{
           RowBox[{"-", "a"}], "+", "c"}]], "]"}]}], ")"}]}], "+", 
     RowBox[{"c", " ", 
      RowBox[{"EllipticK", "[", 
       FractionBox[
        RowBox[{
         RowBox[{"-", "a"}], "+", "b"}], 
        RowBox[{
         RowBox[{"-", "a"}], "+", "c"}]], "]"}]}]}], ")"}]}], 
  SqrtBox[
   RowBox[{
    RowBox[{"-", "a"}], "+", "c"}]]]], "Output",
 CellLabel->"Out[238]=",ExpressionUUID->"307ffbaa-9377-40cb-9a26-19ed8fff5dfb"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"ratio1", " ", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"integral2", "/.", 
       RowBox[{"x", "->", "b"}]}], ")"}], "/", 
     RowBox[{"(", 
      RowBox[{"integral1", "/.", 
       RowBox[{"x", "->", "b"}]}], ")"}]}], "//", "Simplify"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"mtr", " ", "=", " ", 
   RowBox[{
    RowBox[{"Solve", " ", "[", 
     RowBox[{
      RowBox[{
       FractionBox[
        RowBox[{"b", "-", "a"}], 
        RowBox[{"c", "-", "a"}]], "==", "m"}], ",", "c"}], "]"}], " ", "[", 
    RowBox[{"[", "1", "]"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"myratio1", " ", "=", " ", 
   RowBox[{"a", "+", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"b", "-", "a"}], ")"}], 
     RowBox[{"(", 
      RowBox[{
       FractionBox["1", "m"], 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         FractionBox[
          RowBox[{"EllipticE", "[", "m", "]"}], 
          RowBox[{"EllipticK", "[", "m", "]"}]]}], ")"}]}], ")"}]}]}]}], " ", 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"myratio1", "==", " ", "ratio1"}], "/.", "mtr"}], "//", 
  "Simplify"}]}], "Input",
 CellLabel->
  "In[239]:=",ExpressionUUID->"5466e55b-8cc8-4a5d-8b89-c0093cd702bf"],

Cell[BoxData["True"], "Output",
 CellLabel->"Out[242]=",ExpressionUUID->"6ed37dab-bfbd-4558-bcd1-78310fe51f8b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"integral3", "=", " ", 
   RowBox[{
    FractionBox["2", 
     RowBox[{
      SqrtBox[
       RowBox[{"c", "-", "a"}]], 
      RowBox[{"(", 
       RowBox[{"k", "-", " ", "a"}], ")"}]}]], 
    RowBox[{"EllipticPi", "[", 
     RowBox[{
      FractionBox[
       RowBox[{"b", "-", "a"}], 
       RowBox[{"k", "-", " ", "a"}]], ",", 
      RowBox[{"ArcSin", "[", 
       SqrtBox[
        FractionBox[
         RowBox[{" ", 
          RowBox[{"x", "-", "a"}]}], 
         RowBox[{"b", "-", "a"}]]], "]"}], ",", 
      FractionBox[
       RowBox[{"b", "-", "a"}], 
       RowBox[{"c", "-", "a"}]]}], "]"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Simplify", "[", 
   RowBox[{
    RowBox[{"D", "[", 
     RowBox[{"integral3", ",", "x"}], "]"}], ",", 
    RowBox[{"a", "<", "x", "<", "b", "<", "c"}]}], "]"}], "//", 
  "Simplify"}], "\[IndentingNewLine]", 
 RowBox[{"integral3", "/.", 
  RowBox[{"x", "->", "a"}]}], "\[IndentingNewLine]", 
 RowBox[{"integral3", "/.", 
  RowBox[{"x", "->", "b"}]}]}], "Input",
 CellLabel->
  "In[243]:=",ExpressionUUID->"c92dc7cc-03b1-4861-a910-897f74126b57"],

Cell[BoxData[
 FractionBox["1", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{"k", "-", "x"}], ")"}], " ", 
   SqrtBox[
    RowBox[{
     RowBox[{"(", 
      RowBox[{"b", "-", "x"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"c", "-", "x"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "a"}], "+", "x"}], ")"}]}]]}]]], "Output",
 CellLabel->
  "Out[244]=",ExpressionUUID->"6e86ab0a-eeb9-457c-bc81-1a31fc4cadf2"],

Cell[BoxData["0"], "Output",
 CellLabel->
  "Out[245]=",ExpressionUUID->"647f7139-0e00-4100-82ab-f601ff4885fc"],

Cell[BoxData[
 FractionBox[
  RowBox[{"2", " ", 
   RowBox[{"EllipticPi", "[", 
    RowBox[{
     FractionBox[
      RowBox[{
       RowBox[{"-", "a"}], "+", "b"}], 
      RowBox[{
       RowBox[{"-", "a"}], "+", "k"}]], ",", 
     FractionBox[
      RowBox[{
       RowBox[{"-", "a"}], "+", "b"}], 
      RowBox[{
       RowBox[{"-", "a"}], "+", "c"}]]}], "]"}]}], 
  RowBox[{
   SqrtBox[
    RowBox[{
     RowBox[{"-", "a"}], "+", "c"}]], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "a"}], "+", "k"}], ")"}]}]]], "Output",
 CellLabel->"Out[246]=",ExpressionUUID->"3604d341-d4a3-4978-a750-952c739a52a6"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"integral4", " ", "=", " ", 
   RowBox[{
    FractionBox["2", 
     RowBox[{"3", 
      SqrtBox[
       RowBox[{"c", "-", "a"}]]}]], " ", 
    RowBox[{"(", 
     RowBox[{
      SqrtBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "a"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"b", "-", "x"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"c", "-", "x"}], ")"}], 
        RowBox[{"(", 
         RowBox[{"c", "-", "a"}], ")"}]}]], "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"b", 
          RowBox[{"(", 
           RowBox[{"c", "-", "a"}], ")"}]}], "+", 
         RowBox[{"c", 
          RowBox[{"(", 
           RowBox[{"a", "+", 
            RowBox[{"2", "c"}]}], ")"}]}]}], " ", ")"}], 
       RowBox[{"EllipticF", "[", 
        RowBox[{
         RowBox[{"ArcSin", "[", 
          SqrtBox[
           FractionBox[
            RowBox[{" ", 
             RowBox[{"x", "-", "a"}]}], 
            RowBox[{"b", "-", "a"}]]], "]"}], ",", 
         FractionBox[
          RowBox[{"b", "-", "a"}], 
          RowBox[{"c", "-", "a"}]]}], "]"}]}], "-", 
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{"c", "-", "a"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"a", "+", "b", "+", "c"}], ")"}], " ", 
       RowBox[{"EllipticE", "[", 
        RowBox[{
         RowBox[{"ArcSin", "[", 
          SqrtBox[
           FractionBox[
            RowBox[{" ", 
             RowBox[{"x", "-", "a"}]}], 
            RowBox[{"b", "-", "a"}]]], "]"}], ",", 
         FractionBox[
          RowBox[{"b", "-", "a"}], 
          RowBox[{"c", "-", "a"}]]}], "]"}]}]}], ")"}]}]}], ";"}]], "Input",
 CellLabel->
  "In[247]:=",ExpressionUUID->"d99561af-4138-45fb-99c6-4ea05bae2710"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Simplify", "[", 
     RowBox[{
      RowBox[{"D", "[", 
       RowBox[{"integral4", ",", "x"}], "]"}], ",", 
      RowBox[{"a", "<", "x", "<", "b", "<", "c"}]}], "]"}], "//", "Expand"}], 
   "//", "FullSimplify"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"Simplify", "[", 
  RowBox[{
   RowBox[{"%", " ", "==", "  ", 
    FractionBox[
     RowBox[{
      SuperscriptBox["x", "2"], " "}], 
     SqrtBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{"x", "-", "a"}], ")"}], 
       RowBox[{"(", 
        RowBox[{"b", "-", "x"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"c", "-", "x"}], ")"}], " "}]]]}], ",", 
   RowBox[{"a", "<", "x", "<", "b", "<", "c"}]}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"integral4", "/.", 
  RowBox[{"x", "->", "a"}]}], "\[IndentingNewLine]", 
 RowBox[{"integral4", "/.", 
  RowBox[{"x", "->", "b"}]}]}], "Input",
 CellLabel->
  "In[248]:=",ExpressionUUID->"8f823d5a-b206-4cbc-8108-6c3a4da198a3"],

Cell[BoxData["True"], "Output",
 CellLabel->
  "Out[249]=",ExpressionUUID->"d14737ad-f866-46df-88ba-f22d8790f3dc"],

Cell[BoxData["0"], "Output",
 CellLabel->
  "Out[250]=",ExpressionUUID->"38bb16e1-9724-4db1-a1d9-2f5c9e7a5bdb"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", 
   RowBox[{"3", " ", 
    SqrtBox[
     RowBox[{
      RowBox[{"-", "a"}], "+", "c"}]]}]], 
  RowBox[{"2", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "a"}], "+", "c"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"a", "+", "b", "+", "c"}], ")"}], " ", 
      RowBox[{"EllipticE", "[", 
       FractionBox[
        RowBox[{
         RowBox[{"-", "a"}], "+", "b"}], 
        RowBox[{
         RowBox[{"-", "a"}], "+", "c"}]], "]"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"b", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "a"}], "+", "c"}], ")"}]}], "+", 
        RowBox[{"c", " ", 
         RowBox[{"(", 
          RowBox[{"a", "+", 
           RowBox[{"2", " ", "c"}]}], ")"}]}]}], ")"}], " ", 
      RowBox[{"EllipticK", "[", 
       FractionBox[
        RowBox[{
         RowBox[{"-", "a"}], "+", "b"}], 
        RowBox[{
         RowBox[{"-", "a"}], "+", "c"}]], "]"}]}]}], ")"}]}]}]], "Output",
 CellLabel->"Out[251]=",ExpressionUUID->"375a3fc6-ea33-4cab-a8fc-ce80b5362951"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"ratio2", " ", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"integral4", "/.", 
       RowBox[{"x", "->", "b"}]}], ")"}], "/", 
     RowBox[{"(", 
      RowBox[{"integral1", "/.", 
       RowBox[{"x", "->", "b"}]}], ")"}]}], "//", "Simplify"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"mtr", " ", "=", " ", 
   RowBox[{
    RowBox[{"Solve", " ", "[", 
     RowBox[{
      RowBox[{
       FractionBox[
        RowBox[{"b", "-", "a"}], 
        RowBox[{"c", "-", "a"}]], "==", "m"}], ",", "c"}], "]"}], " ", "[", 
    RowBox[{"[", "1", "]"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"myratio2", " ", "=", " ", 
   RowBox[{
    FractionBox["1", 
     RowBox[{"3", " ", 
      SuperscriptBox["m", "2"], " "}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"2", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"b", "-", "a"}], ")"}], "2"]}], "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"b", "-", "a"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"5", " ", "a"}], "+", "b"}], ")"}], " ", "m"}], "+", 
        RowBox[{"3", " ", 
         SuperscriptBox["a", "2"], " ", 
         SuperscriptBox["m", "2"]}]}], ")"}], "-", 
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{"b", "-", "a"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"b", "-", "a", "+", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"2", " ", "a"}], "+", "b"}], ")"}], " ", "m"}]}], ")"}], 
       FractionBox[
        RowBox[{" ", 
         RowBox[{"EllipticE", "[", "m", "]"}]}], 
        RowBox[{"EllipticK", "[", "m", "]"}]]}]}], ")"}]}]}], " ", 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"myratio2", "==", " ", "ratio2"}], "/.", "mtr"}], "//", 
  "Simplify"}]}], "Input",
 CellLabel->
  "In[252]:=",ExpressionUUID->"13db11ea-c5c2-4332-a018-ae3ef5a29cb8"],

Cell[BoxData["True"], "Output",
 CellLabel->"Out[255]=",ExpressionUUID->"c1acfc30-3669-4188-a3f1-d53f79ec54a5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{"a", " ", "+", " ", 
      RowBox[{"dct", " ", 
       RowBox[{"(", 
        RowBox[{"b", "-", "a"}], ")"}]}]}], ")"}], "^", "2"}], "//", 
   "Expand"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"%", "/.", 
     RowBox[{
      SuperscriptBox["dct", "2"], "->", "dct2"}]}], "/.", " ", 
    RowBox[{"dct", "->", 
     RowBox[{
      FractionBox["1", "m"], 
      RowBox[{"(", 
       RowBox[{"1", "-", 
        FractionBox[
         RowBox[{"EllipticE", "[", "m", "]"}], 
         RowBox[{"EllipticK", "[", "m", "]"}]]}], ")"}]}]}]}], "//", 
   "Simplify"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Solve", "[", 
   RowBox[{
    RowBox[{"%", "==", "myratio2"}], ",", "dct2"}], "]"}], "//", 
  "Simplify"}]}], "Input",
 CellLabel->
  "In[256]:=",ExpressionUUID->"2fb344cd-9dc6-4c39-811a-3575ca03534a"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"{", 
   RowBox[{"dct2", "\[Rule]", 
    FractionBox[
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "m"}], ")"}], " ", 
       RowBox[{"EllipticE", "[", "m", "]"}]}], "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"2", "+", "m"}], ")"}], " ", 
       RowBox[{"EllipticK", "[", "m", "]"}]}]}], 
     RowBox[{"3", " ", 
      SuperscriptBox["m", "2"], " ", 
      RowBox[{"EllipticK", "[", "m", "]"}]}]]}], "}"}], "}"}]], "Output",
 CellLabel->"Out[258]=",ExpressionUUID->"f2a30458-4ab7-49fc-9c9a-0589c32b5d42"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Limit", "[", 
  RowBox[{
   FractionBox[
    RowBox[{"2", "+", "m", "-", 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "m"}], ")"}], " ", 
       RowBox[{"EllipticE", "[", "m", "]"}]}], 
      RowBox[{"EllipticK", "[", "m", "]"}]]}], 
    RowBox[{"3", " ", 
     SuperscriptBox["m", "2"]}]], ",", 
   RowBox[{"m", "->", "0"}]}], "]"}]], "Input",
 CellLabel->
  "In[259]:=",ExpressionUUID->"c8ce1de9-6496-48f8-b05f-8e144f3f5790"],

Cell[BoxData[
 FractionBox["3", "8"]], "Output",
 CellLabel->"Out[259]=",ExpressionUUID->"9bbff9f3-6954-453b-9c23-df4680567d32"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Limit", "[", 
  RowBox[{
   FractionBox[
    RowBox[{"2", "+", "m", "-", 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "m"}], ")"}], " ", 
       RowBox[{"EllipticE", "[", "m", "]"}]}], 
      RowBox[{"EllipticK", "[", "m", "]"}]]}], 
    RowBox[{"3", " ", 
     SuperscriptBox["m", "2"]}]], ",", 
   RowBox[{"m", "->", "1"}]}], "]"}]], "Input",
 CellLabel->
  "In[260]:=",ExpressionUUID->"e8b77680-7248-4cac-a7e1-b7d0e40e85e8"],

Cell[BoxData["1"], "Output",
 CellLabel->"Out[260]=",ExpressionUUID->"a4c65a4d-0a58-4bae-989e-1d3d4cb5f7b4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["(vii) Precession-average ODE", \
"Section",ExpressionUUID->"e8b15b02-0303-4ba9-b831-2f1df02e63f8"],

Cell[BoxData[
 RowBox[{"Clear", "@@", 
  RowBox[{"DeleteCases", "[", 
   RowBox[{
    RowBox[{"Names", "@", "\"\<`*\>\""}], ",", 
    RowBox[{"\"\<ToPython\>\"", "|", "\"\<precessionform\>\""}]}], 
   "]"}]}]], "Input",
 CellLabel->
  "In[262]:=",ExpressionUUID->"1f0d843e-f5fc-46b4-948a-f7300b5b6244"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"DSolve", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"k", "'"}], "[", "u", "]"}], "\[Equal]", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"1", "/", "u"}], ")"}], 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"k", "[", "u", "]"}], "-", " ", 
           RowBox[{"chieff", " ", "/", "2"}]}], ")"}]}]}], ",", 
       RowBox[{
        RowBox[{"k", "[", "u0", "]"}], "==", "k0"}]}], "}"}], ",", 
     RowBox[{"k", "[", "u", "]"}], ",", "u"}], "]"}], "//", "Simplify"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"k", "[", "u", "]"}], "/.", 
     RowBox[{"%", "[", 
      RowBox[{"[", "1", "]"}], "]"}]}], "/.", 
    RowBox[{"u", "->", " ", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "^", "2"}], "/", " ", 
       RowBox[{"(", 
        RowBox[{"2", " ", "q"}], ")"}]}], " ", "/", 
      RowBox[{"Sqrt", "[", "r", "]"}]}]}]}], "/.", 
   RowBox[{"u0", "->", " ", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "^", "2"}], "/", " ", 
      RowBox[{"(", 
       RowBox[{"2", " ", "q"}], ")"}]}], " ", "/", 
     RowBox[{"Sqrt", "[", "r0", "]"}]}]}]}], "//", "Simplify"}]}], "Input",
 CellLabel->
  "In[263]:=",ExpressionUUID->"002c79fa-98b9-4719-9ad1-18abce27c996"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{"chieff", " ", 
    RowBox[{"(", 
     RowBox[{
      SqrtBox["r"], "-", 
      SqrtBox["r0"]}], ")"}]}], "+", 
   RowBox[{"2", " ", "k0", " ", 
    SqrtBox["r0"]}]}], 
  RowBox[{"2", " ", 
   SqrtBox["r"]}]]], "Output",
 CellLabel->"Out[264]=",ExpressionUUID->"adc2fcbf-c371-412e-b73a-846e82d0fe07"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["(viii) Inertial frame frequency", \
"Section",ExpressionUUID->"0eff2c7c-9c71-41cb-85c3-6200b6d4f8b9"],

Cell[BoxData[
 RowBox[{"Clear", "@@", 
  RowBox[{"DeleteCases", "[", 
   RowBox[{
    RowBox[{"Names", "@", "\"\<`*\>\""}], ",", 
    RowBox[{"\"\<ToPython\>\"", "|", "\"\<precessionform\>\""}]}], 
   "]"}]}]], "Input",
 CellLabel->
  "In[265]:=",ExpressionUUID->"29deabff-f6c3-4661-849b-b3e03ac4965c"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"sal", " ", "=", " ", 
   RowBox[{"Simplify", "[", 
    RowBox[{"SolveAlways", "[", 
     RowBox[{
      RowBox[{"{", " ", 
       RowBox[{
        FractionBox[
         RowBox[{
          SuperscriptBox["x", 
           RowBox[{"2", " "}]], "+", 
          RowBox[{"n1", " ", "x"}], " ", "+", " ", "n0"}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{"x", "-", "rp"}], ")"}], 
          RowBox[{"(", 
           RowBox[{"x", "-", "rm"}], ")"}]}]], "==", " ", 
        RowBox[{"1", "+", " ", 
         FractionBox["cp", 
          RowBox[{"x", "-", "rp"}]], "+", 
         FractionBox["cm", 
          RowBox[{" ", 
           RowBox[{"x", "-", "rm"}]}]]}]}], "}"}], " ", ",", "x"}], "]"}], 
    "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"{", 
   RowBox[{"cp", ",", "cm"}], "}"}], "=", 
  RowBox[{
   RowBox[{
    RowBox[{"{", 
     RowBox[{"cp", ",", " ", "cm"}], "}"}], " ", "/.", " ", 
    RowBox[{
     RowBox[{"Solve", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"n1", " ", "==", " ", 
          RowBox[{"Evaluate", "[", 
           RowBox[{"n1", " ", "/.", " ", "sal"}], "]"}]}], ",", " ", 
         RowBox[{"n0", " ", "==", " ", 
          RowBox[{"Evaluate", "[", 
           RowBox[{"n0", " ", "/.", " ", "sal"}], "]"}]}]}], "}"}], ",", " ", 
       
       RowBox[{"{", 
        RowBox[{"cp", ",", " ", "cm"}], "}"}]}], "]"}], "[", 
     RowBox[{"[", "1", "]"}], "]"}]}], " ", "//", " ", 
   "Simplify"}]}]}], "Input",
 CellLabel->
  "In[266]:=",ExpressionUUID->"dd81f9b7-44f3-41d9-8d42-cfe4adef2aa0"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", 
    FractionBox[
     RowBox[{"n0", "+", 
      RowBox[{"rp", " ", 
       RowBox[{"(", 
        RowBox[{"n1", "+", "rp"}], ")"}]}]}], 
     RowBox[{"rm", "-", "rp"}]]}], ",", 
   FractionBox[
    RowBox[{"n0", "+", 
     RowBox[{"rm", " ", 
      RowBox[{"(", 
       RowBox[{"n1", "+", "rm"}], ")"}]}]}], 
    RowBox[{"rm", "-", "rp"}]]}], "}"}]], "Output",
 CellLabel->"Out[267]=",ExpressionUUID->"160fe840-36e5-4002-80e6-3a37022735d9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"A1A2sq", " ", "=", " ", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"J", " ", "^", "2"}], "-", " ", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"L", "-", "S"}], ")"}], "^", "2"}]}], ")"}], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{"L", "+", "S"}], ")"}], "^", "2"}], " ", "-", " ", 
       RowBox[{"J", "^", "2"}]}], ")"}]}], "//", "Expand"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"A1A2sq", "=", " ", 
  RowBox[{
   RowBox[{
    RowBox[{"A1A2sq", " ", "//.", 
     RowBox[{
      RowBox[{"S", "^", "2"}], "\[Rule]", "Ss"}]}], "//.", 
    RowBox[{
     RowBox[{"S", "^", "4"}], "\[Rule]", 
     RowBox[{"Ss", "^", "2"}]}]}], "//", "Simplify"}]}]}], "Input",
 CellLabel->
  "In[268]:=",ExpressionUUID->"144f7aed-2d41-4b90-bcb0-6b501e197e80"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   SuperscriptBox["J", "4"]}], "-", 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{
     SuperscriptBox["L", "2"], "-", "Ss"}], ")"}], "2"], "+", 
  RowBox[{"2", " ", 
   SuperscriptBox["J", "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     SuperscriptBox["L", "2"], "+", "Ss"}], ")"}]}]}]], "Output",
 CellLabel->"Out[269]=",ExpressionUUID->"64fd059a-da8c-4871-b497-25cfcd01418c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"omega", " ", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{"J", "/", "2"}], ")"}], 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"eta", "^", "2"}], "/", 
       RowBox[{"L", "^", "2"}]}], ")"}], "^", "3"}], "  ", 
    RowBox[{"(", 
     RowBox[{"1", "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"eta", " ", 
          RowBox[{"\[Chi]eff", "/", " ", "L"}]}]}], ")"}], "*", 
       RowBox[{"3", "/", 
        RowBox[{"(", 
         RowBox[{"2", " ", "eta"}], ")"}]}]}], " ", "-", "   ", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"3", "/", "2"}], ")"}], 
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         RowBox[{"eta", "  ", 
          RowBox[{"\[Chi]eff", " ", "/", " ", "L"}]}]}], ")"}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"4", 
           RowBox[{"(", 
            RowBox[{"1", "-", "q"}], ")"}], 
           RowBox[{"L", "^", "2"}], 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"S1", "^", "2"}], "-", 
             RowBox[{"S2", "^", "2"}]}], ")"}]}], "-", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"1", "+", "q"}], ")"}], 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"J", "^", "2"}], "-", 
             RowBox[{"L", "^", "2"}], "-", "Ss"}], ")"}], 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"J", "^", "2"}], "-", 
             RowBox[{"L", "^", "2"}], "-", "Ss", " ", "-", 
             RowBox[{"4", " ", "eta", " ", "L", " ", "\[Chi]eff"}]}], 
            ")"}]}]}], ")"}], "/", 
        RowBox[{"(", 
         RowBox[{"q", "  ", "A1A2sq"}], ")"}]}]}]}], ")"}]}], "//", 
   "FullSimplify"}]}]], "Input",
 CellLabel->
  "In[270]:=",ExpressionUUID->"e95cc54b-f03e-4008-bfba-02f3ff71cbb1"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", 
   RowBox[{"2", " ", 
    SuperscriptBox["L", "6"]}]], 
  RowBox[{
   SuperscriptBox["eta", "6"], " ", "J", " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", 
     RowBox[{
      FractionBox["3", "2"], " ", 
      RowBox[{"(", 
       RowBox[{
        FractionBox["1", "eta"], "-", 
        FractionBox["\[Chi]eff", "L"]}], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"3", " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"1", "-", 
          FractionBox[
           RowBox[{"eta", " ", "\[Chi]eff"}], "L"]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "4"}], " ", 
           SuperscriptBox["L", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{"S1", "-", "S2"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{"S1", "+", "S2"}], ")"}]}], "-", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"1", "+", "q"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox["J", "2"], "-", 
             SuperscriptBox["L", "2"], "-", "Ss"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox["J", "2"], "-", "Ss", "-", 
             RowBox[{"L", " ", 
              RowBox[{"(", 
               RowBox[{"L", "+", 
                RowBox[{"4", " ", "eta", " ", "\[Chi]eff"}]}], ")"}]}]}], 
            ")"}]}]}], ")"}]}], ")"}], "/", 
      RowBox[{"(", 
       RowBox[{"2", " ", "q", " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox["J", "4"], "+", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox["L", "2"], "-", "Ss"}], ")"}], "2"], "-", 
          RowBox[{"2", " ", 
           SuperscriptBox["J", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox["L", "2"], "+", "Ss"}], ")"}]}]}], ")"}]}], 
       ")"}]}]}], ")"}]}]}]], "Output",
 CellLabel->"Out[270]=",ExpressionUUID->"919ed884-203d-4500-b465-38e39e51a32e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"tr", " ", "=", " ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"sps", "->", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "-", "q"}], ")"}], 
       RowBox[{"L", "^", "2"}], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"S1", "^", "2"}], "-", 
         RowBox[{"S2", "^", "2"}]}], ")"}]}]}], ",", 
     RowBox[{"ecl", " ", "->", 
      RowBox[{"eta", " ", "L", " ", "\[Chi]eff"}]}], ",", " ", 
     RowBox[{"Ss", "->", 
      RowBox[{"dc0", " ", "+", 
       RowBox[{"dc1", " ", "\[Delta]\[Chi]t"}]}]}], ",", 
     RowBox[{"jml", " ", "->", 
      RowBox[{
       RowBox[{"J", "^", "2"}], "-", 
       RowBox[{"L", "^", "2"}]}]}], ",", 
     RowBox[{"jpl", " ", "->", 
      RowBox[{
       RowBox[{"J", "^", "2"}], "+", 
       RowBox[{"L", "^", "2"}]}]}]}], "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"a", "=", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"J", "/", "2"}], ")"}], 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"eta", "^", "2"}], "/", 
       RowBox[{"L", "^", "2"}]}], ")"}], "^", "3"}]}]}], "  ", 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"b", " ", "=", " ", 
   RowBox[{"1", "+", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "-", 
       RowBox[{"ecl", "/", 
        RowBox[{"L", "^", "2"}]}]}], ")"}], "*", 
     RowBox[{"3", "/", 
      RowBox[{"(", 
       RowBox[{"2", " ", "eta"}], ")"}]}]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"c", "=", " ", 
   RowBox[{
    RowBox[{"-", "   ", 
     RowBox[{"(", 
      RowBox[{"3", "/", "2"}], ")"}]}], 
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], 
    RowBox[{"(", 
     RowBox[{"1", "-", " ", 
      RowBox[{"ecl", "/", 
       RowBox[{"L", "^", "2"}]}]}], " ", ")"}], 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"4", "sps"}], "-", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], 
        RowBox[{"(", 
         RowBox[{"jml", "-", 
          RowBox[{"(", 
           RowBox[{"dc0", " ", "+", 
            RowBox[{"dc1", " ", "\[Delta]\[Chi]t"}]}], ")"}]}], ")"}], 
        RowBox[{"(", 
         RowBox[{"jml", "-", 
          RowBox[{"(", 
           RowBox[{"dc0", " ", "+", 
            RowBox[{"dc1", " ", "\[Delta]\[Chi]t"}]}], ")"}], " ", "-", 
          RowBox[{"4", "ecl"}]}], ")"}]}]}], ")"}], "/", 
     RowBox[{"(", "q", ")"}]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"d", "=", "  ", 
   RowBox[{
    RowBox[{"-", 
     SuperscriptBox[
      RowBox[{"(", 
       FractionBox[
        RowBox[{"jml", "+", "jpl"}], "2"], ")"}], "2"]}], "-", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{"(", 
        FractionBox[
         RowBox[{
          RowBox[{"-", "jml"}], "+", "jpl"}], "2"], ")"}], "-", 
       RowBox[{"(", 
        RowBox[{"dc0", " ", "+", 
         RowBox[{"dc1", " ", "\[Delta]\[Chi]t"}]}], ")"}]}], ")"}], "2"], "+", 
    RowBox[{"2", " ", 
     FractionBox[
      RowBox[{"jml", "+", "jpl"}], "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       FractionBox[
        RowBox[{
         RowBox[{"-", "jml"}], "+", "jpl"}], "2"], "+", 
       RowBox[{"(", 
        RowBox[{"dc0", " ", "+", 
         RowBox[{"dc1", " ", "\[Delta]\[Chi]t"}]}], ")"}]}], ")"}]}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"num", " ", "=", " ", 
   RowBox[{"a", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"b", " ", "d"}], "+", "c"}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"den", " ", "=", " ", "d"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"omega", " ", "==", " ", 
    RowBox[{"(", 
     RowBox[{"num", "/", "den"}], ")"}]}], " ", "/.", "tr"}], "//", 
  "Simplify"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"(*", 
   RowBox[{"\[Delta]\[Chi]t", " ", "=", " ", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "-", "q"}], ")"}], " ", "\[Delta]\[Chi]"}]}], 
   "*)"}]}]}], "Input",
 CellLabel->
  "In[271]:=",ExpressionUUID->"0319fb0d-d2fd-44ff-bdd6-6e72d2cf2c08"],

Cell[BoxData["True"], "Output",
 CellLabel->"Out[278]=",ExpressionUUID->"37c0e08d-8a1e-4df8-843d-beb0ab2dafcb"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{"num0", ",", "num1", ",", "num2"}], "}"}], "=", 
  RowBox[{
   RowBox[{
    RowBox[{"CoefficientList", "[", 
     RowBox[{"num", ",", "\[Delta]\[Chi]t"}], "]"}], "/.", 
    RowBox[{"eta", " ", "->", 
     RowBox[{"q", " ", "/", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], "//", 
   "Simplify"}]}]], "Input",
 CellLabel->
  "In[279]:=",ExpressionUUID->"0f12fef4-cc49-4456-87f9-11828868f6b3"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", 
    RowBox[{
     FractionBox["1", 
      RowBox[{"2", " ", 
       SuperscriptBox["L", "8"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "12"]}]], 
     RowBox[{"J", " ", 
      SuperscriptBox["q", "5"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         SuperscriptBox["dc0", "2"], " ", 
         SuperscriptBox["L", "2"], " ", "q"}], "-", 
        RowBox[{"6", " ", 
         SuperscriptBox["ecl", "2"], " ", "jml", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "2"]}], "+", 
        RowBox[{"dc0", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"6", " ", 
            SuperscriptBox["ecl", "2"], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"1", "+", "q"}], ")"}], "2"]}], "-", 
           RowBox[{"3", " ", "ecl", " ", 
            RowBox[{"(", 
             RowBox[{"jml", "-", "jpl", "+", 
              RowBox[{"2", " ", 
               SuperscriptBox["L", "2"]}]}], ")"}], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"1", "+", "q"}], ")"}], "2"]}], "+", 
           RowBox[{
            SuperscriptBox["L", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"3", " ", "jml", " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{"1", "+", "q"}], ")"}], "2"]}], "-", 
              RowBox[{"jpl", " ", 
               RowBox[{"(", 
                RowBox[{"3", "+", 
                 RowBox[{"8", " ", "q"}], "+", 
                 RowBox[{"3", " ", 
                  SuperscriptBox["q", "2"]}]}], ")"}]}]}], ")"}]}]}], ")"}]}],
         "+", 
        RowBox[{"6", " ", "ecl", " ", 
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"jml", " ", 
            SuperscriptBox["L", "2"], " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}]}], "-", "sps"}], ")"}]}], "+", 
        RowBox[{
         SuperscriptBox["L", "2"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            SuperscriptBox["jml", "2"], " ", "q"}], "+", 
           RowBox[{"6", " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], " ", "sps"}]}], ")"}]}]}], 
       ")"}]}]}]}], ",", 
   RowBox[{
    FractionBox["1", 
     RowBox[{"2", " ", 
      SuperscriptBox["L", "8"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "12"]}]], 
    RowBox[{"dc1", " ", "J", " ", 
     SuperscriptBox["q", "5"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "6"}], " ", 
        SuperscriptBox["ecl", "2"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "2"]}], "+", 
       RowBox[{"3", " ", "ecl", " ", 
        RowBox[{"(", 
         RowBox[{"jml", "-", "jpl", "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["L", "2"]}]}], ")"}], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "2"]}], "+", 
       RowBox[{
        SuperscriptBox["L", "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "2"}], " ", "dc0", " ", "q"}], "-", 
          RowBox[{"3", " ", "jml", " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "2"]}], "+", 
          RowBox[{"jpl", " ", 
           RowBox[{"(", 
            RowBox[{"3", "+", 
             RowBox[{"8", " ", "q"}], "+", 
             RowBox[{"3", " ", 
              SuperscriptBox["q", "2"]}]}], ")"}]}]}], ")"}]}]}], ")"}]}]}], 
   ",", 
   RowBox[{"-", 
    FractionBox[
     RowBox[{
      SuperscriptBox["dc1", "2"], " ", "J", " ", 
      SuperscriptBox["q", "6"]}], 
     RowBox[{"2", " ", 
      SuperscriptBox["L", "6"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "12"]}]]}]}], "}"}]], "Output",
 CellLabel->"Out[279]=",ExpressionUUID->"88a2ea5f-0690-4d08-9afe-33afc70b1d62"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{"den0", ",", "den1", ",", "den2"}], "}"}], "=", 
  RowBox[{
   RowBox[{
    RowBox[{"CoefficientList", "[", 
     RowBox[{"den", ",", "\[Delta]\[Chi]t"}], "]"}], "/.", 
    RowBox[{"eta", " ", "->", 
     RowBox[{"q", " ", "/", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], "//", 
   "Simplify"}]}]], "Input",
 CellLabel->
  "In[280]:=",ExpressionUUID->"39350e7e-76ae-476c-a839-9a817f52c012"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     SuperscriptBox["dc0", "2"]}], "-", 
    SuperscriptBox["jml", "2"], "+", 
    RowBox[{"2", " ", "dc0", " ", "jpl"}]}], ",", 
   RowBox[{"2", " ", "dc1", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "dc0"}], "+", "jpl"}], ")"}]}], ",", 
   RowBox[{"-", 
    SuperscriptBox["dc1", "2"]}]}], "}"}]], "Output",
 CellLabel->"Out[280]=",ExpressionUUID->"9b2a10f5-cc5e-4d85-b624-49882d88898e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{"Rm", ",", "Rp"}], "}"}], "=", " ", 
  RowBox[{
   RowBox[{"\[Delta]\[Chi]t", "/.", 
    RowBox[{"Solve", "[", 
     RowBox[{
      RowBox[{"den", "==", "0"}], ",", "\[Delta]\[Chi]t"}], "]"}]}], "//", 
   "FullSimplify"}]}]], "Input",
 CellLabel->
  "In[281]:=",ExpressionUUID->"fd391246-9e98-41b5-bf9b-882213b9ef67"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", 
    FractionBox[
     RowBox[{
      RowBox[{"dc0", " ", "dc1"}], "-", 
      RowBox[{"dc1", " ", "jpl"}], "+", 
      SqrtBox[
       RowBox[{
        SuperscriptBox["dc1", "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["jml", "2"]}], "+", 
          SuperscriptBox["jpl", "2"]}], ")"}]}]]}], 
     SuperscriptBox["dc1", "2"]]}], ",", 
   FractionBox[
    RowBox[{
     RowBox[{
      RowBox[{"-", "dc0"}], " ", "dc1"}], "+", 
     RowBox[{"dc1", " ", "jpl"}], "+", 
     SqrtBox[
      RowBox[{
       SuperscriptBox["dc1", "2"], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", 
          SuperscriptBox["jml", "2"]}], "+", 
         SuperscriptBox["jpl", "2"]}], ")"}]}]]}], 
    SuperscriptBox["dc1", "2"]]}], "}"}]], "Output",
 CellLabel->"Out[281]=",ExpressionUUID->"b1ee4b40-5e0f-4ece-8277-6768769b6bbd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       FractionBox["num2", "den2"], 
       FractionBox[
        RowBox[{
         SuperscriptBox["\[Delta]\[Chi]t", "2"], "+", 
         RowBox[{
          FractionBox["num1", "num2"], " ", "\[Delta]\[Chi]t"}], " ", "+", 
         " ", 
         FractionBox["num0", "num2"]}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"\[Delta]\[Chi]t", "-", "Rm"}], ")"}], 
         RowBox[{"(", 
          RowBox[{"\[Delta]\[Chi]t", "-", "Rp"}], ")"}]}]]}], "-", "omega"}], 
     "/.", "tr"}], "/.", 
    RowBox[{"eta", " ", "->", 
     RowBox[{"q", " ", "/", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], "//", "Simplify"}], 
  "\[IndentingNewLine]"}]], "Input",
 CellLabel->
  "In[282]:=",ExpressionUUID->"cb0b8064-3652-4a55-8bc7-44e5f1e578b0"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[282]=",ExpressionUUID->"2a340e08-8298-4fa0-bfd9-941548fff694"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"(*", 
  RowBox[{
   FractionBox[
    RowBox[{
     SuperscriptBox["x", 
      RowBox[{"2", " "}]], "+", 
     RowBox[{"n1", " ", "x"}], " ", "+", " ", "n0"}], 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"rp", "-", "x"}], ")"}], 
     RowBox[{"(", 
      RowBox[{"rm", "-", "x"}], ")"}]}]], "==", " ", 
   RowBox[{"1", "+", " ", 
    FractionBox["cp", 
     RowBox[{"rp", " ", "-", " ", "x"}]], "+", 
    FractionBox["cm", 
     RowBox[{"rm", " ", "-", " ", "x"}]]}]}], "*)"}]], "Input",
 CellLabel->
  "In[543]:=",ExpressionUUID->"d30064f9-9d2f-4f8e-965b-317c0f18cb3a"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"C0", "=", " ", 
   FractionBox["num2", "den2"]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Cp", "=", 
   RowBox[{
    RowBox[{"cp", "/.", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"n1", "->", 
        RowBox[{"num1", "/", "num2"}]}], ",", 
       RowBox[{"n0", "->", 
        RowBox[{"num0", "/", "num2"}]}], ",", 
       RowBox[{"rm", "->", "Rm"}], ",", 
       RowBox[{"rp", "->", "Rp"}]}], "}"}]}], "//", "Simplify"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Cm", "=", 
   RowBox[{
    RowBox[{"cm", "/.", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"n1", "->", 
        RowBox[{"num1", "/", "num2"}]}], ",", 
       RowBox[{"n0", "->", 
        RowBox[{"num0", "/", "num2"}]}], ",", 
       RowBox[{"rm", "->", "Rm"}], ",", 
       RowBox[{"rp", "->", "Rp"}]}], "}"}]}], "//", "Simplify"}]}], 
  ";"}]}], "Input",
 CellLabel->
  "In[283]:=",ExpressionUUID->"565a325f-545b-463a-855a-33459629cfb6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"C0", 
      RowBox[{"(", 
       RowBox[{"1", "+", 
        FractionBox["Cp", 
         RowBox[{"\[Delta]\[Chi]t", "-", "Rp"}]], "+", 
        FractionBox["Cm", 
         RowBox[{"\[Delta]\[Chi]t", "-", "Rm"}]]}], ")"}]}], "==", "omega"}], 
    "/.", "tr"}], "/.", 
   RowBox[{"eta", " ", "->", 
    RowBox[{"q", " ", "/", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], "//", 
  "Simplify"}]], "Input",
 CellLabel->
  "In[286]:=",ExpressionUUID->"e8fbac8e-ea59-4991-8ff0-d6dc5b6d58c2"],

Cell[BoxData["True"], "Output",
 CellLabel->"Out[286]=",ExpressionUUID->"f32cfc86-a7f5-4438-925a-1e9eb0c6916e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Ssofdchi", " ", "=", " ", 
  RowBox[{
   RowBox[{"q", " ", "/", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}], " ", 
   RowBox[{"Sqrt", "[", "r", "]"}], 
   RowBox[{"(", " ", 
    RowBox[{
     RowBox[{"2", " ", "\[Kappa]"}], " ", "-", "\[Chi]eff", " ", "-", 
     RowBox[{"\[Delta]\[Chi]t", "/", 
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}]}]}], ")"}]}]}]], "Input",
 CellLabel->
  "In[287]:=",ExpressionUUID->"1b76da9a-9f5d-4249-9d2d-b23b2f60f100"],

Cell[BoxData[
 FractionBox[
  RowBox[{"q", " ", 
   SqrtBox["r"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      FractionBox["\[Delta]\[Chi]t", 
       RowBox[{"1", "+", "q"}]]}], "+", 
     RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}], 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{"1", "+", "q"}], ")"}], "2"]]], "Output",
 CellLabel->"Out[287]=",ExpressionUUID->"80acffa1-f41d-475b-870f-0d8390b36727"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Rpsim", "=", 
   RowBox[{
    RowBox[{"-", 
     SqrtBox["r"]}], 
    RowBox[{"(", " ", 
     RowBox[{
      RowBox[{
       FractionBox[
        RowBox[{"2", " ", "q"}], 
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}]], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         SqrtBox[
          RowBox[{"1", "+", 
           RowBox[{
            FractionBox[
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{"1", "+", "q"}], ")"}], "2"], "q"], 
            FractionBox[
             RowBox[{"2", " ", "\[Kappa]"}], 
             SqrtBox["r"]]}]}]]}], ")"}]}], "+", 
      RowBox[{
       FractionBox["\[Chi]eff", 
        SqrtBox["r"]], 
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}]}]}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Rpsim", "-", 
   RowBox[{"Simplify", "[", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{"Rp", "/.", "tr"}], "/.", 
            RowBox[{"eta", " ", "->", 
             RowBox[{"q", " ", "/", 
              RowBox[{
               RowBox[{"(", 
                RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], "/.", 
           RowBox[{"S1", " ", "->", " ", 
            RowBox[{"\[Chi]1", " ", "/", " ", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], "/.", 
          RowBox[{"S2", " ", "->", " ", 
           RowBox[{"\[Chi]2", " ", 
            RowBox[{
             RowBox[{"q", "^", "2"}], " ", "/", " ", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}]}], "/.", 
         RowBox[{"J", "->", 
          RowBox[{"L", " ", 
           RowBox[{"Sqrt", "[", 
            RowBox[{"1", "+", " ", 
             RowBox[{"2", 
              RowBox[{"\[Kappa]", "/", "L"}]}]}], "]"}]}]}]}], "/.", 
        RowBox[{"L", "->", 
         RowBox[{
          RowBox[{"q", " ", "/", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}], " ", 
          RowBox[{"Sqrt", "[", "r", "]"}]}]}]}], "/.", 
       RowBox[{"dc1", " ", "->", " ", 
        RowBox[{"Coefficient", "[", 
         RowBox[{"Ssofdchi", ",", "\[Delta]\[Chi]t", ",", "1"}], "]"}]}]}], "/.", 
      RowBox[{"dc0", " ", "->", " ", 
       RowBox[{"Coefficient", "[", 
        RowBox[{"Ssofdchi", ",", "\[Delta]\[Chi]t", ",", "0"}], "]"}]}]}], 
     ",", 
     RowBox[{
      RowBox[{"J", ">", "0"}], "&&", 
      RowBox[{"q", ">", "0"}], "&&", 
      RowBox[{"r", ">", "0"}], "&&", 
      RowBox[{"q", "<", "1"}]}]}], "]"}]}], "//", 
  "Simplify"}], "\[IndentingNewLine]"}], "Input",
 CellLabel->
  "In[291]:=",ExpressionUUID->"724cf56e-25ce-444d-8c9e-383be6230b33"],

Cell[BoxData[
 FractionBox[
  RowBox[{"2", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       SuperscriptBox["r", 
        RowBox[{"1", "/", "4"}]]}], " ", 
      SqrtBox[
       RowBox[{"q", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"q", " ", 
           SqrtBox["r"]}], "+", 
          RowBox[{"2", " ", "\[Kappa]"}], "+", 
          RowBox[{"4", " ", "q", " ", "\[Kappa]"}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["q", "2"], " ", "\[Kappa]"}]}], ")"}]}]]}], "+", 
     RowBox[{"q", " ", 
      SqrtBox["r"], " ", 
      SqrtBox[
       RowBox[{"1", "+", 
        FractionBox[
         RowBox[{"2", " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}], 
         RowBox[{"q", " ", 
          SqrtBox["r"]}]]}]]}]}], ")"}]}], 
  RowBox[{"1", "+", "q"}]]], "Output",
 CellLabel->"Out[292]=",ExpressionUUID->"3ca1ae39-641f-4882-ac69-0726f242bbdc"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", 
   RowBox[{
    RowBox[{
    "Not", " ", "sure", " ", "why", " ", "Mathemathica", " ", "is", " ", 
     "not", " ", "giving", " ", "zero", " ", "there"}], ",", " ", 
    RowBox[{
     RowBox[{"but", " ", 
      RowBox[{"that", "'"}], "s", " ", "a", " ", "zero"}], "..."}]}], "*)"}], 
  "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", 
        SuperscriptBox["r", 
         RowBox[{"1", "/", "4"}]]}], " ", 
       SqrtBox[
        RowBox[{"q", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"q", " ", 
            SqrtBox["r"]}], "+", 
           RowBox[{"2", " ", "\[Kappa]"}], "+", 
           RowBox[{"4", " ", "q", " ", "\[Kappa]"}], "+", 
           RowBox[{"2", " ", 
            SuperscriptBox["q", "2"], " ", "\[Kappa]"}]}], ")"}]}]]}], ")"}], 
     "^", "2"}], "-", 
    RowBox[{
     RowBox[{"(", " ", 
      RowBox[{"q", " ", 
       SqrtBox["r"], " ", 
       SqrtBox[
        RowBox[{"1", "+", 
         FractionBox[
          RowBox[{"2", " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}], 
          RowBox[{"q", " ", 
           SqrtBox["r"]}]]}]]}], ")"}], "^", "2"}]}], "//", 
   "Simplify"}]}]], "Input",
 CellLabel->
  "In[293]:=",ExpressionUUID->"c803d871-d8c0-4904-8cad-b9650fefede6"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[293]=",ExpressionUUID->"da3b20f3-e701-4865-989c-b19d7472bf8f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Rmsim", "=", 
   RowBox[{
    RowBox[{"-", 
     SqrtBox["r"]}], 
    RowBox[{"(", " ", 
     RowBox[{
      RowBox[{
       FractionBox[
        RowBox[{"2", " ", "q"}], 
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}]], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", 
         SqrtBox[
          RowBox[{"1", "+", 
           RowBox[{
            FractionBox[
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{"1", "+", "q"}], ")"}], "2"], "q"], 
            FractionBox[
             RowBox[{"2", " ", "\[Kappa]"}], 
             SqrtBox["r"]]}]}]]}], ")"}]}], "+", 
      RowBox[{
       FractionBox["\[Chi]eff", 
        SqrtBox["r"]], 
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}]}]}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Rmsim", "-", " ", 
   RowBox[{"Simplify", "[", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{"Rm", "/.", "tr"}], "/.", 
            RowBox[{"eta", " ", "->", 
             RowBox[{"q", " ", "/", 
              RowBox[{
               RowBox[{"(", 
                RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], "/.", 
           RowBox[{"S1", " ", "->", " ", 
            RowBox[{"\[Chi]1", " ", "/", " ", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], "/.", 
          RowBox[{"S2", " ", "->", " ", 
           RowBox[{"\[Chi]2", " ", 
            RowBox[{
             RowBox[{"q", "^", "2"}], " ", "/", " ", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}]}], "/.", 
         RowBox[{"J", "->", 
          RowBox[{"L", " ", 
           RowBox[{"Sqrt", "[", 
            RowBox[{"1", "+", " ", 
             RowBox[{"2", 
              RowBox[{"\[Kappa]", "/", "L"}]}]}], "]"}]}]}]}], "/.", 
        RowBox[{"L", "->", 
         RowBox[{
          RowBox[{"q", " ", "/", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}], " ", 
          RowBox[{"Sqrt", "[", "r", "]"}]}]}]}], "/.", 
       RowBox[{"dc1", " ", "->", " ", 
        RowBox[{"Coefficient", "[", 
         RowBox[{"Ssofdchi", ",", "\[Delta]\[Chi]t", ",", "1"}], "]"}]}]}], "/.", 
      RowBox[{"dc0", " ", "->", " ", 
       RowBox[{"Coefficient", "[", 
        RowBox[{"Ssofdchi", ",", "\[Delta]\[Chi]t", ",", "0"}], "]"}]}]}], 
     ",", 
     RowBox[{
      RowBox[{"J", ">", "0"}], "&&", 
      RowBox[{"q", ">", "0"}], "&&", 
      RowBox[{"r", ">", "0"}], "&&", 
      RowBox[{"q", "<", "1"}]}]}], "]"}]}], "//", "FullSimplify"}]}], "Input",
 
 CellLabel->
  "In[294]:=",ExpressionUUID->"eafa5d15-0db3-4d31-84d2-db4667585230"],

Cell[BoxData[
 FractionBox[
  RowBox[{"2", " ", 
   SuperscriptBox["r", 
    RowBox[{"1", "/", "4"}]], " ", 
   RowBox[{"(", 
    RowBox[{
     SqrtBox[
      RowBox[{"q", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"q", " ", 
          SqrtBox["r"]}], "+", 
         RowBox[{"2", " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}]}], 
        ")"}]}]], "-", 
     RowBox[{"q", " ", 
      SuperscriptBox["r", 
       RowBox[{"1", "/", "4"}]], " ", 
      SqrtBox[
       RowBox[{"1", "+", 
        FractionBox[
         RowBox[{"2", " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}], 
         RowBox[{"q", " ", 
          SqrtBox["r"]}]]}]]}]}], ")"}]}], 
  RowBox[{"1", "+", "q"}]]], "Output",
 CellLabel->"Out[295]=",ExpressionUUID->"b065ff94-695e-4560-b387-d84dd8bbdca9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", 
   RowBox[{
    RowBox[{
    "Not", " ", "sure", " ", "why", " ", "Mathemathica", " ", "is", " ", 
     "not", " ", "giving", " ", "zero", " ", "there"}], ",", " ", 
    RowBox[{
     RowBox[{"but", " ", 
      RowBox[{"that", "'"}], "s", " ", "a", " ", "zero"}], "..."}]}], "*)"}], 
  
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      SqrtBox[
       RowBox[{"q", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"q", " ", 
           SqrtBox["r"]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}]}], 
         ")"}]}]], ")"}], "^", "2"}], "==", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"q", " ", 
       SuperscriptBox["r", 
        RowBox[{"1", "/", "4"}]], " ", 
       SqrtBox[
        RowBox[{"1", "+", 
         FractionBox[
          RowBox[{"2", " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}], 
          RowBox[{"q", " ", 
           SqrtBox["r"]}]]}]]}], ")"}], "^", "2"}]}], "//", 
   "Simplify"}]}]], "Input",
 CellLabel->
  "In[296]:=",ExpressionUUID->"e43c9d72-2a31-45fd-8b24-d8fc073820ef"],

Cell[BoxData["True"], "Output",
 CellLabel->"Out[296]=",ExpressionUUID->"60a8b8ae-06c0-4850-a125-45768192330a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"C0sim", "=", 
   RowBox[{
    FractionBox[
     RowBox[{"q", " "}], 
     RowBox[{"2", " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
      SuperscriptBox["r", 
       RowBox[{"5", "/", "2"}]]}]], 
    SqrtBox[
     RowBox[{"1", "+", 
      RowBox[{
       FractionBox[
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "2"], "q"], 
       FractionBox[
        RowBox[{"2", " ", "\[Kappa]"}], 
        SqrtBox["r"]]}]}]]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"C0sim", "-", 
  RowBox[{"Simplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{
           RowBox[{"C0", "/.", "tr"}], "/.", 
           RowBox[{"eta", " ", "->", 
            RowBox[{"q", " ", "/", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], "/.", 
          RowBox[{"S1", " ", "->", " ", 
           RowBox[{"\[Chi]1", " ", "/", " ", 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], "/.", 
         RowBox[{"S2", " ", "->", " ", 
          RowBox[{"\[Chi]2", " ", 
           RowBox[{
            RowBox[{"q", "^", "2"}], " ", "/", " ", 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}]}], "/.", 
        RowBox[{"J", "->", 
         RowBox[{"L", " ", 
          RowBox[{"Sqrt", "[", 
           RowBox[{"1", "+", " ", 
            RowBox[{"2", 
             RowBox[{"\[Kappa]", "/", "L"}]}]}], "]"}]}]}]}], "/.", 
       RowBox[{"L", "->", 
        RowBox[{
         RowBox[{"q", " ", "/", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}], " ", 
         RowBox[{"Sqrt", "[", "r", "]"}]}]}]}], "/.", 
      RowBox[{"dc1", " ", "->", " ", 
       RowBox[{"Coefficient", "[", 
        RowBox[{"Ssofdchi", ",", "\[Delta]\[Chi]t", ",", "1"}], "]"}]}]}], "/.", 
     RowBox[{"dc0", " ", "->", " ", 
      RowBox[{"Coefficient", "[", 
       RowBox[{"Ssofdchi", ",", "\[Delta]\[Chi]t", ",", "0"}], "]"}]}]}], ",", 
    RowBox[{
     RowBox[{"J", ">", "0"}], "&&", 
     RowBox[{"q", ">", "0"}], "&&", 
     RowBox[{"r", ">", "0"}], "&&", 
     RowBox[{"q", "<", "1"}]}]}], "]"}]}]}], "Input",
 CellLabel->
  "In[297]:=",ExpressionUUID->"e0d0418f-fd5b-4a2a-a178-054d20af7817"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[298]=",ExpressionUUID->"fa5cdf99-fcac-4e30-addd-67f1aeab9077"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Cpsim", "=", 
   RowBox[{
    FractionBox[
     RowBox[{" ", 
      RowBox[{"3", " ", 
       SqrtBox["r"]}]}], 
     RowBox[{"2", "  ", 
      SuperscriptBox["q", "2"], " "}]], 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      FractionBox["\[Chi]eff", 
       SqrtBox["r"]]}], ")"}], " ", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        FractionBox[
         RowBox[{
          RowBox[{"(", 
           RowBox[{"1", "-", "q"}], ")"}], " "}], "r"], 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox["\[Chi]1", "2"], "-", 
          RowBox[{
           SuperscriptBox["q", "4"], " ", 
           SuperscriptBox["\[Chi]2", "2"]}]}], ")"}]}], "-", "  ", 
       RowBox[{"2", "q", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "3"], " ", " ", 
        FractionBox[
         RowBox[{" ", "\[Kappa]"}], 
         SqrtBox["r"]]}], "-", 
       RowBox[{"2", " ", 
        SuperscriptBox["q", "2"], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "  ", 
        RowBox[{"(", 
         RowBox[{"1", "-", 
          SqrtBox[
           RowBox[{"1", "+", 
            RowBox[{
             FractionBox[
              RowBox[{"2", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{"1", "+", "q"}], ")"}], "2"]}], 
              RowBox[{"q", " "}]], 
             FractionBox[
              RowBox[{" ", "\[Kappa]"}], 
              SqrtBox["r"]]}]}]]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", 
          FractionBox["\[Chi]eff", 
           SqrtBox["r"]]}], ")"}]}]}], ")"}], "/", 
     RowBox[{"(", 
      SqrtBox[
       RowBox[{"1", "+", 
        RowBox[{
         FractionBox[
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"1", "+", "q"}], ")"}], "2"], "q"], 
         FractionBox[
          RowBox[{"2", " ", "\[Kappa]"}], 
          SqrtBox["r"]]}]}]], ")"}]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Cpsim", "-", " ", 
    RowBox[{"Simplify", "[", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{
             RowBox[{"Cp", "/.", "tr"}], "/.", 
             RowBox[{"eta", " ", "->", 
              RowBox[{"q", " ", "/", 
               RowBox[{
                RowBox[{"(", 
                 RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], "/.", 
            RowBox[{"S1", " ", "->", " ", 
             RowBox[{"\[Chi]1", " ", "/", " ", 
              RowBox[{
               RowBox[{"(", 
                RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], "/.", 
           RowBox[{"S2", " ", "->", " ", 
            RowBox[{"\[Chi]2", " ", 
             RowBox[{
              RowBox[{"q", "^", "2"}], " ", "/", " ", 
              RowBox[{
               RowBox[{"(", 
                RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}]}], "/.", 
          RowBox[{"J", "->", 
           RowBox[{"L", " ", 
            RowBox[{"Sqrt", "[", 
             RowBox[{"1", "+", " ", 
              RowBox[{"2", 
               RowBox[{"\[Kappa]", "/", "L"}]}]}], "]"}]}]}]}], "/.", 
         RowBox[{"L", "->", 
          RowBox[{
           RowBox[{"q", " ", "/", 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}], " ", 
           RowBox[{"Sqrt", "[", "r", "]"}]}]}]}], "/.", 
        RowBox[{"dc1", " ", "->", " ", 
         RowBox[{"Coefficient", "[", 
          RowBox[{"Ssofdchi", ",", "\[Delta]\[Chi]t", ",", "1"}], "]"}]}]}], "/.", 
       RowBox[{"dc0", " ", "->", " ", 
        RowBox[{"Coefficient", "[", 
         RowBox[{"Ssofdchi", ",", "\[Delta]\[Chi]", ",", "0"}], "]"}]}]}], 
      ",", 
      RowBox[{
       RowBox[{"J", ">", "0"}], "&&", 
       RowBox[{"q", ">", "0"}], "&&", 
       RowBox[{"r", ">", "0"}], "&&", 
       RowBox[{"q", "<", "1"}], "&&", 
       RowBox[{"\[Kappa]", "\[Element]", "Reals"}]}]}], "]"}]}], "//", 
   "FullSimplify"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]"}], "Input",
 CellLabel->
  "In[299]:=",ExpressionUUID->"0f54411b-9f3a-4a61-82d9-e8aa10c51cfe"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"3", " ", 
    RowBox[{"(", 
     RowBox[{
      SqrtBox[
       RowBox[{
        RowBox[{"q", " ", 
         SqrtBox["r"]}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}]}]], "-", 
      RowBox[{
       SqrtBox["q"], " ", 
       SuperscriptBox["r", 
        RowBox[{"1", "/", "4"}]], " ", 
       SqrtBox[
        RowBox[{"1", "+", 
         FractionBox[
          RowBox[{"2", " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}], 
          RowBox[{"q", " ", 
           SqrtBox["r"]}]]}]]}]}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      SqrtBox["r"], "-", "\[Chi]eff"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      SuperscriptBox["\[Chi]1", "2"], "-", 
      RowBox[{"q", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", 
          SqrtBox["r"], " ", "\[Kappa]"}], "+", 
         SuperscriptBox["\[Chi]1", "2"]}], ")"}]}], "+", 
      RowBox[{
       SuperscriptBox["q", "5"], " ", 
       SuperscriptBox["\[Chi]2", "2"]}], "-", 
      RowBox[{
       SuperscriptBox["q", "4"], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", 
          SqrtBox["r"], " ", "\[Kappa]"}], "+", 
         SuperscriptBox["\[Chi]2", "2"]}], ")"}]}], "-", 
      RowBox[{"2", " ", 
       SuperscriptBox["q", "2"], " ", 
       SqrtBox["r"], " ", 
       RowBox[{"(", 
        RowBox[{
         SqrtBox["r"], "+", 
         RowBox[{"3", " ", "\[Kappa]"}], "+", "\[Chi]eff"}], ")"}]}], "-", 
      RowBox[{"2", " ", 
       SuperscriptBox["q", "3"], " ", 
       SqrtBox["r"], " ", 
       RowBox[{"(", 
        RowBox[{
         SqrtBox["r"], "+", 
         RowBox[{"3", " ", "\[Kappa]"}], "+", "\[Chi]eff"}], ")"}]}]}], 
     ")"}]}], ")"}], "/", 
  RowBox[{"(", 
   RowBox[{"2", " ", 
    SuperscriptBox["q", "2"], " ", "r", " ", 
    SqrtBox[
     RowBox[{
      RowBox[{"q", " ", 
       SqrtBox["r"]}], "+", 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}]}]], " ", 
    SqrtBox[
     RowBox[{"1", "+", 
      FractionBox[
       RowBox[{"2", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}], 
       RowBox[{"q", " ", 
        SqrtBox["r"]}]]}]]}], ")"}]}]], "Output",
 CellLabel->"Out[300]=",ExpressionUUID->"e0b6b6a3-2ddd-461e-98cb-052cb62f72fe"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", 
   RowBox[{
    RowBox[{
    "Not", " ", "sure", " ", "why", " ", "Mathemathica", " ", "is", " ", 
     "not", " ", "giving", " ", "zero", " ", "there"}], ",", " ", 
    RowBox[{
     RowBox[{"but", " ", 
      RowBox[{"that", "'"}], "s", " ", "a", " ", "zero"}], "..."}]}], "*)"}], 
  "\[IndentingNewLine]", 
  RowBox[{"Simplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      SqrtBox[
       RowBox[{
        RowBox[{"q", " ", 
         SqrtBox["r"]}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}]}]], ")"}], 
     "^", "2"}], " ", "-", " ", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       SqrtBox["q"], " ", 
       SuperscriptBox["r", 
        RowBox[{"1", "/", "4"}]], " ", 
       SqrtBox[
        RowBox[{"1", "+", 
         FractionBox[
          RowBox[{"2", " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}], 
          RowBox[{"q", " ", 
           SqrtBox["r"]}]]}]]}], ")"}], "^", "2"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[301]:=",ExpressionUUID->"7b2defea-d39e-480b-acdd-e8d8a51081fd"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[301]=",ExpressionUUID->"b4e2dc41-7ae3-4335-bffc-2ea0995c9e4c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", 
   RowBox[{
    RowBox[{
    "Another", " ", "check", " ", "with", " ", "some", " ", "numbers"}], 
    "..."}], "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"Cpsim", "/.", 
         RowBox[{"r", "->", "12.56"}]}], "/.", 
        RowBox[{"q", "->", "0.5676"}]}], "/.", 
       RowBox[{"\[Kappa]", "->", "0.321"}]}], "/.", 
      RowBox[{"\[Chi]1", "->", "0.12981"}]}], "/.", 
     RowBox[{"\[Chi]2", "->", "0.21879"}]}], "/.", 
    RowBox[{"\[Chi]eff", "->", "0.2891"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"Simplify", "[", 
          RowBox[{
           RowBox[{
            RowBox[{
             RowBox[{
              RowBox[{
               RowBox[{
                RowBox[{
                 RowBox[{
                  RowBox[{"Cp", "/.", "tr"}], "/.", 
                  RowBox[{"eta", " ", "->", 
                   RowBox[{"q", " ", "/", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], "/.", 
                 RowBox[{"S1", " ", "->", " ", 
                  RowBox[{"\[Chi]1", " ", "/", " ", 
                   RowBox[{
                    RowBox[{"(", 
                    RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], "/.", 
                RowBox[{"S2", " ", "->", " ", 
                 RowBox[{"\[Chi]2", " ", 
                  RowBox[{
                   RowBox[{"q", "^", "2"}], " ", "/", " ", 
                   RowBox[{
                    RowBox[{"(", 
                    RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}]}], "/.", 
               
               RowBox[{"J", "->", 
                RowBox[{"L", " ", 
                 RowBox[{"Sqrt", "[", 
                  RowBox[{"1", "+", " ", 
                   RowBox[{"2", 
                    RowBox[{"\[Kappa]", "/", "L"}]}]}], "]"}]}]}]}], "/.", 
              RowBox[{"L", "->", 
               RowBox[{
                RowBox[{"q", " ", "/", 
                 RowBox[{
                  RowBox[{"(", 
                   RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}], " ", 
                RowBox[{"Sqrt", "[", "r", "]"}]}]}]}], "/.", 
             RowBox[{"dc1", " ", "->", " ", 
              RowBox[{"Coefficient", "[", 
               RowBox[{"Ssofdchi", ",", "\[Delta]\[Chi]t", ",", "1"}], 
               "]"}]}]}], "/.", 
            RowBox[{"dc0", " ", "->", " ", 
             RowBox[{"Coefficient", "[", 
              RowBox[{"Ssofdchi", ",", "\[Delta]\[Chi]t", ",", "0"}], 
              "]"}]}]}], ",", 
           RowBox[{
            RowBox[{"J", ">", "0"}], "&&", 
            RowBox[{"q", ">", "0"}], "&&", 
            RowBox[{"r", ">", "0"}], "&&", 
            RowBox[{"q", "<", "1"}], "&&", 
            RowBox[{"\[Kappa]", "\[Element]", "Reals"}]}]}], "]"}], "/.", 
         RowBox[{"r", "->", "12.56"}]}], "/.", 
        RowBox[{"q", "->", "0.5676"}]}], "/.", 
       RowBox[{"\[Kappa]", "->", "0.321"}]}], "/.", 
      RowBox[{"\[Chi]1", "->", "0.12981"}]}], "/.", 
     RowBox[{"\[Chi]2", "->", "0.21879"}]}], "/.", 
    RowBox[{"\[Chi]eff", "->", "0.2891"}]}]}]}]], "Input",
 CellLabel->
  "In[302]:=",ExpressionUUID->"c6783d58-9a58-42ca-ab50-da25989a8569"],

Cell[BoxData[
 RowBox[{"-", "0.32745359851666767`"}]], "Output",
 CellLabel->
  "Out[302]=",ExpressionUUID->"fd82427b-ad2b-4e4b-bb24-9d3ee7dff589"],

Cell[BoxData[
 RowBox[{"-", "0.3274535985166711`"}]], "Output",
 CellLabel->"Out[303]=",ExpressionUUID->"2268f595-7b95-4f2b-8d79-9c90782b468b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Cmsim", "=", 
   RowBox[{
    FractionBox[
     RowBox[{" ", 
      RowBox[{"3", " ", 
       SqrtBox["r"]}]}], 
     RowBox[{"2", "  ", 
      SuperscriptBox["q", "2"], " "}]], 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      FractionBox["\[Chi]eff", 
       SqrtBox["r"]]}], ")"}], " ", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        FractionBox[
         RowBox[{
          RowBox[{"(", 
           RowBox[{"1", "-", "q"}], ")"}], " "}], "r"], 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox["\[Chi]1", "2"], "-", 
          RowBox[{
           SuperscriptBox["q", "4"], " ", 
           SuperscriptBox["\[Chi]2", "2"]}]}], ")"}]}], "-", "  ", 
       RowBox[{"2", "q", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "3"], " ", " ", 
        FractionBox[
         RowBox[{" ", "\[Kappa]"}], 
         SqrtBox["r"]]}], "-", 
       RowBox[{"2", " ", 
        SuperscriptBox["q", "2"], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "  ", 
        RowBox[{"(", 
         RowBox[{"1", "+", 
          SqrtBox[
           RowBox[{"1", "+", 
            RowBox[{
             FractionBox[
              RowBox[{"2", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{"1", "+", "q"}], ")"}], "2"]}], 
              RowBox[{"q", " "}]], 
             FractionBox[
              RowBox[{" ", "\[Kappa]"}], 
              SqrtBox["r"]]}]}]]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", 
          FractionBox["\[Chi]eff", 
           SqrtBox["r"]]}], ")"}]}]}], ")"}], "/", 
     RowBox[{"(", 
      RowBox[{"-", 
       SqrtBox[
        RowBox[{"1", "+", 
         RowBox[{
          FractionBox[
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "2"], "q"], 
          FractionBox[
           RowBox[{"2", " ", "\[Kappa]"}], 
           SqrtBox["r"]]}]}]]}], ")"}]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Cmsim", "-", " ", 
    RowBox[{"Simplify", "[", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{
             RowBox[{"Cm", "/.", "tr"}], "/.", 
             RowBox[{"eta", " ", "->", 
              RowBox[{"q", " ", "/", 
               RowBox[{
                RowBox[{"(", 
                 RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], "/.", 
            RowBox[{"S1", " ", "->", " ", 
             RowBox[{"\[Chi]1", " ", "/", " ", 
              RowBox[{
               RowBox[{"(", 
                RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], "/.", 
           RowBox[{"S2", " ", "->", " ", 
            RowBox[{"\[Chi]2", " ", 
             RowBox[{
              RowBox[{"q", "^", "2"}], " ", "/", " ", 
              RowBox[{
               RowBox[{"(", 
                RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}]}], "/.", 
          RowBox[{"J", "->", 
           RowBox[{"L", " ", 
            RowBox[{"Sqrt", "[", 
             RowBox[{"1", "+", " ", 
              RowBox[{"2", 
               RowBox[{"\[Kappa]", "/", "L"}]}]}], "]"}]}]}]}], "/.", 
         RowBox[{"L", "->", 
          RowBox[{
           RowBox[{"q", " ", "/", 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}], " ", 
           RowBox[{"Sqrt", "[", "r", "]"}]}]}]}], "/.", 
        RowBox[{"dc1", " ", "->", " ", 
         RowBox[{"Coefficient", "[", 
          RowBox[{"Ssofdchi", ",", "\[Delta]\[Chi]t", ",", "1"}], "]"}]}]}], "/.", 
       RowBox[{"dc0", " ", "->", " ", 
        RowBox[{"Coefficient", "[", 
         RowBox[{"Ssofdchi", ",", "\[Delta]\[Chi]t", ",", "0"}], "]"}]}]}], 
      ",", 
      RowBox[{
       RowBox[{"J", ">", "0"}], "&&", 
       RowBox[{"q", ">", "0"}], "&&", 
       RowBox[{"r", ">", "0"}], "&&", 
       RowBox[{"q", "<", "1"}], "&&", 
       RowBox[{"\[Kappa]", "\[Element]", "Reals"}]}]}], "]"}]}], "//", 
   "FullSimplify"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]"}], "Input",
 CellLabel->
  "In[304]:=",ExpressionUUID->"0cd3dcdb-45e8-4e35-8bca-bc77c6ff2975"],

Cell[BoxData[
 RowBox[{"-", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       RowBox[{
        SqrtBox[
         RowBox[{
          RowBox[{"q", " ", 
           SqrtBox["r"]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}]}]], "-", 
        
        RowBox[{
         SqrtBox["q"], " ", 
         SuperscriptBox["r", 
          RowBox[{"1", "/", "4"}]], " ", 
         SqrtBox[
          RowBox[{"1", "+", 
           FractionBox[
            RowBox[{"2", " ", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}], 
            RowBox[{"q", " ", 
             SqrtBox["r"]}]]}]]}]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        SqrtBox["r"], "-", "\[Chi]eff"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["\[Chi]1", "2"], "-", 
        RowBox[{"q", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"2", " ", 
            SqrtBox["r"], " ", "\[Kappa]"}], "+", 
           SuperscriptBox["\[Chi]1", "2"]}], ")"}]}], "+", 
        RowBox[{
         SuperscriptBox["q", "5"], " ", 
         SuperscriptBox["\[Chi]2", "2"]}], "-", 
        RowBox[{
         SuperscriptBox["q", "4"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"2", " ", 
            SqrtBox["r"], " ", "\[Kappa]"}], "+", 
           SuperscriptBox["\[Chi]2", "2"]}], ")"}]}], "-", 
        RowBox[{"2", " ", 
         SuperscriptBox["q", "2"], " ", 
         SqrtBox["r"], " ", 
         RowBox[{"(", 
          RowBox[{
           SqrtBox["r"], "+", 
           RowBox[{"3", " ", "\[Kappa]"}], "+", "\[Chi]eff"}], ")"}]}], "-", 
        RowBox[{"2", " ", 
         SuperscriptBox["q", "3"], " ", 
         SqrtBox["r"], " ", 
         RowBox[{"(", 
          RowBox[{
           SqrtBox["r"], "+", 
           RowBox[{"3", " ", "\[Kappa]"}], "+", "\[Chi]eff"}], ")"}]}]}], 
       ")"}]}], ")"}], "/", 
    RowBox[{"(", 
     RowBox[{"2", " ", 
      SuperscriptBox["q", "2"], " ", "r", " ", 
      SqrtBox[
       RowBox[{
        RowBox[{"q", " ", 
         SqrtBox["r"]}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}]}]], " ", 
      SqrtBox[
       RowBox[{"1", "+", 
        FractionBox[
         RowBox[{"2", " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}], 
         RowBox[{"q", " ", 
          SqrtBox["r"]}]]}]]}], ")"}]}], ")"}]}]], "Output",
 CellLabel->"Out[305]=",ExpressionUUID->"93fc7194-fea6-4b77-bc8f-50dfe3e91c29"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", 
   RowBox[{
    RowBox[{
    "Not", " ", "sure", " ", "why", " ", "Mathemathica", " ", "is", " ", 
     "not", " ", "giving", " ", "zero", " ", "there"}], ",", " ", 
    RowBox[{
     RowBox[{"but", " ", 
      RowBox[{"that", "'"}], "s", " ", "a", " ", "zero"}], "..."}]}], "*)"}], 
  "\[IndentingNewLine]", 
  RowBox[{"Simplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      SqrtBox[
       RowBox[{
        RowBox[{"q", " ", 
         SqrtBox["r"]}], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}]}]], ")"}], 
     "^", "2"}], " ", "-", " ", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       SqrtBox["q"], " ", 
       SuperscriptBox["r", 
        RowBox[{"1", "/", "4"}]], " ", 
       SqrtBox[
        RowBox[{"1", "+", 
         FractionBox[
          RowBox[{"2", " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "q"}], ")"}], "2"], " ", "\[Kappa]"}], 
          RowBox[{"q", " ", 
           SqrtBox["r"]}]]}]]}], ")"}], "^", "2"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[306]:=",ExpressionUUID->"6c2cab04-30f2-46c2-9bec-57ce37140444"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[306]=",ExpressionUUID->"6731d84c-280e-4896-ac35-0cec04c8ea7c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", 
   RowBox[{
    RowBox[{
    "Another", " ", "check", " ", "with", " ", "some", " ", "numbers"}], 
    "..."}], "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"Cmsim", "/.", 
         RowBox[{"r", "->", "12.56"}]}], "/.", 
        RowBox[{"q", "->", "0.5676"}]}], "/.", 
       RowBox[{"\[Kappa]", "->", "0.321"}]}], "/.", 
      RowBox[{"\[Chi]1", "->", "0.12981"}]}], "/.", 
     RowBox[{"\[Chi]2", "->", "0.21879"}]}], "/.", 
    RowBox[{"\[Chi]eff", "->", "0.2891"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"Simplify", "[", 
          RowBox[{
           RowBox[{
            RowBox[{
             RowBox[{
              RowBox[{
               RowBox[{
                RowBox[{
                 RowBox[{
                  RowBox[{"Cm", "/.", "tr"}], "/.", 
                  RowBox[{"eta", " ", "->", 
                   RowBox[{"q", " ", "/", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], "/.", 
                 RowBox[{"S1", " ", "->", " ", 
                  RowBox[{"\[Chi]1", " ", "/", " ", 
                   RowBox[{
                    RowBox[{"(", 
                    RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}], "/.", 
                RowBox[{"S2", " ", "->", " ", 
                 RowBox[{"\[Chi]2", " ", 
                  RowBox[{
                   RowBox[{"q", "^", "2"}], " ", "/", " ", 
                   RowBox[{
                    RowBox[{"(", 
                    RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}]}]}]}], "/.", 
               
               RowBox[{"J", "->", 
                RowBox[{"L", " ", 
                 RowBox[{"Sqrt", "[", 
                  RowBox[{"1", "+", " ", 
                   RowBox[{"2", 
                    RowBox[{"\[Kappa]", "/", "L"}]}]}], "]"}]}]}]}], "/.", 
              RowBox[{"L", "->", 
               RowBox[{
                RowBox[{"q", " ", "/", 
                 RowBox[{
                  RowBox[{"(", 
                   RowBox[{"1", "+", "q"}], ")"}], "^", "2"}]}], " ", 
                RowBox[{"Sqrt", "[", "r", "]"}]}]}]}], "/.", 
             RowBox[{"dc1", " ", "->", " ", 
              RowBox[{"Coefficient", "[", 
               RowBox[{"Ssofdchi", ",", "\[Delta]\[Chi]t", ",", "1"}], 
               "]"}]}]}], "/.", 
            RowBox[{"dc0", " ", "->", " ", 
             RowBox[{"Coefficient", "[", 
              RowBox[{"Ssofdchi", ",", "\[Delta]\[Chi]t", ",", "0"}], 
              "]"}]}]}], ",", 
           RowBox[{
            RowBox[{"J", ">", "0"}], "&&", 
            RowBox[{"q", ">", "0"}], "&&", 
            RowBox[{"r", ">", "0"}], "&&", 
            RowBox[{"q", "<", "1"}], "&&", 
            RowBox[{"\[Kappa]", "\[Element]", "Reals"}]}]}], "]"}], "/.", 
         RowBox[{"r", "->", "12.56"}]}], "/.", 
        RowBox[{"q", "->", "0.5676"}]}], "/.", 
       RowBox[{"\[Kappa]", "->", "0.321"}]}], "/.", 
      RowBox[{"\[Chi]1", "->", "0.12981"}]}], "/.", 
     RowBox[{"\[Chi]2", "->", "0.21879"}]}], "/.", 
    RowBox[{"\[Chi]eff", "->", "0.2891"}]}]}]}]], "Input",
 CellLabel->
  "In[307]:=",ExpressionUUID->"b2156938-58f5-43f1-8408-c290d6fdca2f"],

Cell[BoxData["33.43917147885196`"], "Output",
 CellLabel->
  "Out[307]=",ExpressionUUID->"50abd381-3b7d-4cc2-819f-98f1fc2e0d39"],

Cell[BoxData["33.43917147885193`"], "Output",
 CellLabel->"Out[308]=",ExpressionUUID->"43f44f5c-866a-4a70-8105-04a13068a91c"]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"CC0", "=", 
   RowBox[{
    FractionBox["q", 
     RowBox[{" ", 
      RowBox[{"2", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "q"}], ")"}], "2"], 
       SuperscriptBox["r", 
        RowBox[{"5", "/", "2"}]]}]}]], 
    SqrtBox[
     RowBox[{"1", "+", 
      RowBox[{
       FractionBox[
        RowBox[{"2", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "2"]}], "q"], 
       FractionBox[
        RowBox[{" ", "\[Kappa]"}], 
        SqrtBox["r"]]}]}]]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"CCm", "=", 
   RowBox[{
    RowBox[{"-", "\[IndentingNewLine]", "3"}], 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      FractionBox["\[Chi]eff", 
       SqrtBox["r"]]}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "  ", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         SqrtBox[
          RowBox[{"1", "+", 
           RowBox[{
            FractionBox[
             RowBox[{"2", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"1", "+", "q"}], ")"}], "2"]}], 
             RowBox[{"q", " "}]], 
            FractionBox[
             RowBox[{" ", "\[Kappa]"}], 
             SqrtBox["r"]]}]}]]}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", 
         FractionBox["\[Chi]eff", 
          SqrtBox["r"]]}], ")"}]}], "+", "  ", 
      RowBox[{
       FractionBox[
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "3"], "q"], " ", 
       FractionBox[
        RowBox[{" ", "\[Kappa]"}], 
        SqrtBox["r"]]}], "-", 
      RowBox[{
       FractionBox[
        RowBox[{
         RowBox[{"(", 
          RowBox[{"1", "-", "q"}], ")"}], " "}], 
        RowBox[{"2", "  ", 
         SuperscriptBox["q", "2"], "  "}]], 
       FractionBox[
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox["\[Chi]1", "2"], "-", 
          RowBox[{
           SuperscriptBox["q", "4"], " ", 
           SuperscriptBox["\[Chi]2", "2"]}]}], ")"}], "r"]}]}], ")"}], 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"1", "+", 
       RowBox[{
        FractionBox[
         RowBox[{"2", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"1", "+", "q"}], ")"}], "2"]}], 
         RowBox[{"q", " "}]], 
        FractionBox[
         RowBox[{" ", "\[Kappa]"}], 
         SqrtBox["r"]]}]}], ")"}], 
     RowBox[{
      RowBox[{"-", "1"}], "/", "2"}]]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"CCp", "=", 
   RowBox[{
    RowBox[{"+", "\[IndentingNewLine]", "3"}], 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      FractionBox["\[Chi]eff", 
       SqrtBox["r"]]}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "  ", 
       RowBox[{"(", 
        RowBox[{"1", "+", 
         SqrtBox[
          RowBox[{"1", "+", 
           RowBox[{
            FractionBox[
             RowBox[{"2", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"1", "+", "q"}], ")"}], "2"]}], 
             RowBox[{"q", " "}]], 
            FractionBox[
             RowBox[{" ", "\[Kappa]"}], 
             SqrtBox["r"]]}]}]]}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", 
         FractionBox["\[Chi]eff", 
          SqrtBox["r"]]}], ")"}]}], "+", "  ", 
      RowBox[{
       FractionBox[
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "3"], "q"], " ", 
       FractionBox[
        RowBox[{" ", "\[Kappa]"}], 
        SqrtBox["r"]]}], "-", 
      RowBox[{
       FractionBox[
        RowBox[{
         RowBox[{"(", 
          RowBox[{"1", "-", "q"}], ")"}], " "}], 
        RowBox[{"2", "  ", 
         SuperscriptBox["q", "2"], "  "}]], 
       FractionBox[
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox["\[Chi]1", "2"], "-", 
          RowBox[{
           SuperscriptBox["q", "4"], " ", 
           SuperscriptBox["\[Chi]2", "2"]}]}], ")"}], "r"]}]}], ")"}], 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"1", "+", 
       RowBox[{
        FractionBox[
         RowBox[{"2", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"1", "+", "q"}], ")"}], "2"]}], 
         RowBox[{"q", " "}]], 
        FractionBox[
         RowBox[{" ", "\[Kappa]"}], 
         SqrtBox["r"]]}]}], ")"}], 
     RowBox[{
      RowBox[{"-", "1"}], "/", "2"}]]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"RCm", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"-", " ", 
      FractionBox[
       RowBox[{"2", " ", "q"}], 
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}]]}], " ", 
     RowBox[{"(", 
      RowBox[{"1", "-", 
       SqrtBox[
        RowBox[{"1", "+", 
         RowBox[{
          FractionBox[
           RowBox[{"2", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"1", "+", "q"}], ")"}], "2"]}], "q"], 
          FractionBox[
           RowBox[{" ", "\[Kappa]"}], 
           SqrtBox["r"]]}]}]]}], ")"}]}], "-", 
    RowBox[{
     FractionBox["\[Chi]eff", 
      SqrtBox["r"]], 
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"RCp", "=", " ", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{"2", " ", "q"}], 
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}]]}], " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", 
       SqrtBox[
        RowBox[{"1", "+", 
         RowBox[{
          FractionBox[
           RowBox[{"2", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"1", "+", "q"}], ")"}], "2"]}], "q"], 
          FractionBox[
           RowBox[{" ", "\[Kappa]"}], 
           SqrtBox["r"]]}]}]]}], ")"}]}], "-", 
    RowBox[{
     FractionBox["\[Chi]eff", 
      SqrtBox["r"]], 
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}]}]}]}], ";"}]}], "Input",
 CellLabel->
  "In[309]:=",ExpressionUUID->"d47b9ba1-dfc3-482c-9c5f-01062dc1b34b"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"precessionform", "[", 
  RowBox[{"\"\<bigC0\>\"", ",", "CC0"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"precessionform", "[", 
  RowBox[{"\"\<bigCplus\>\"", ",", "CCp"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"precessionform", "[", 
  RowBox[{"\"\<bigCminus\>\"", ",", "CCm"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"precessionform", "[", 
  RowBox[{"\"\<bigRplus\>\"", ",", "RCp"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"precessionform", "[", 
  RowBox[{"\"\<bigRminus\>\"", ",", "RCm"}], "]"}]}], "Input",
 CellLabel->
  "In[314]:=",ExpressionUUID->"59c80e5a-2c30-4092-a620-19e2f84d4f97"],

Cell[BoxData["\<\"# Machine generated with eq_generator.nb\\nbigC0 = 1/2 * q \
* ((1 + q))**(-2) * (r)**(-5/2) * ((1 + 2 * q**(-1) * ((1 + q))**2 * \
(r)**(-1/2) * kappa))**(1/2)\"\>"], "Output",
 CellLabel->
  "Out[314]=",ExpressionUUID->"2d4119ec-3eec-4a78-8182-407a05996be2"],

Cell[BoxData["\<\"# Machine generated with eq_generator.nb\\nbigCplus = 3 * \
((1 + 2 * q**(-1) * ((1 + q))**2 * (r)**(-1/2) * kappa))**(-1/2) * (1 + -1 * \
(r)**(-1/2) * chieff) * (q**(-1) * ((1 + q))**3 * (r)**(-1/2) * kappa + (-1/2 \
* (1 + -1 * q) * q**(-2) * (r)**(-1) * (chi1**2 + -1 * q**4 * chi2**2) + (1 + \
q) * (1 + ((1 + 2 * q**(-1) * ((1 + q))**2 * (r)**(-1/2) * kappa))**(1/2)) * \
(1 + (r)**(-1/2) * chieff)))\"\>"], "Output",
 CellLabel->
  "Out[315]=",ExpressionUUID->"0c9378ef-a4a8-4b42-a363-bd710c1cd047"],

Cell[BoxData["\<\"# Machine generated with eq_generator.nb\\nbigCminus = -3 * \
((1 + 2 * q**(-1) * ((1 + q))**2 * (r)**(-1/2) * kappa))**(-1/2) * (1 + -1 * \
(r)**(-1/2) * chieff) * (q**(-1) * ((1 + q))**3 * (r)**(-1/2) * kappa + (-1/2 \
* (1 + -1 * q) * q**(-2) * (r)**(-1) * (chi1**2 + -1 * q**4 * chi2**2) + (1 + \
q) * (1 + -1 * ((1 + 2 * q**(-1) * ((1 + q))**2 * (r)**(-1/2) * \
kappa))**(1/2)) * (1 + (r)**(-1/2) * chieff)))\"\>"], "Output",
 CellLabel->
  "Out[316]=",ExpressionUUID->"af6dba81-7c2d-41f2-b293-f50601eb43f7"],

Cell[BoxData["\<\"# Machine generated with eq_generator.nb\\nbigRplus = (-2 * \
q * ((1 + q))**(-1) * (1 + ((1 + 2 * q**(-1) * ((1 + q))**2 * (r)**(-1/2) * \
kappa))**(1/2)) + -1 * (1 + q) * (r)**(-1/2) * chieff)\"\>"], "Output",
 CellLabel->
  "Out[317]=",ExpressionUUID->"5bc6d86f-51aa-40af-88b1-1b7495464ac9"],

Cell[BoxData["\<\"# Machine generated with eq_generator.nb\\nbigRminus = (-2 \
* q * ((1 + q))**(-1) * (1 + -1 * ((1 + 2 * q**(-1) * ((1 + q))**2 * \
(r)**(-1/2) * kappa))**(1/2)) + -1 * (1 + q) * (r)**(-1/2) * chieff)\"\>"], \
"Output",
 CellLabel->"Out[318]=",ExpressionUUID->"7925981f-f224-4997-b6ba-5d5a8b02b2ce"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"C0sim", 
    RowBox[{"(", 
     RowBox[{"1", "+", 
      FractionBox["Cpsim", 
       RowBox[{
        RowBox[{"\[Delta]\[Chi]", 
         RowBox[{"(", 
          RowBox[{"1", "-", "q"}], ")"}]}], "-", "Rpsim"}]], "+", 
      FractionBox["Cmsim", 
       RowBox[{
        RowBox[{"\[Delta]\[Chi]", 
         RowBox[{"(", 
          RowBox[{"1", "-", "q"}], ")"}]}], "-", "Rmsim"}]]}], ")"}]}], " ", 
   "-", " ", 
   RowBox[{"CC0", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      FractionBox["CCp", 
       RowBox[{"RCp", "-", 
        FractionBox[
         RowBox[{"\[Delta]\[Chi]", 
          RowBox[{"(", 
           RowBox[{"1", "-", "q"}], ")"}]}], 
         SqrtBox["r"]]}]], "-", 
      FractionBox["CCm", 
       RowBox[{"RCm", "-", 
        FractionBox[
         RowBox[{"\[Delta]\[Chi]", 
          RowBox[{"(", 
           RowBox[{"1", "-", "q"}], ")"}]}], 
         SqrtBox["r"]]}]]}], ")"}]}]}], "//", "Simplify"}]], "Input",
 CellLabel->
  "In[319]:=",ExpressionUUID->"d4f229a1-6d1b-4aac-8f08-9159e2c4a9ef"],

Cell[BoxData["0"], "Output",
 CellLabel->"Out[319]=",ExpressionUUID->"9a282e71-0252-4925-8aac-fda5da7a2dca"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"(*", 
  RowBox[{
  "This", " ", "is", " ", "the", " ", "form", " ", "we", " ", "actually", " ",
    "want", " ", "to", " ", "use"}], "*)"}]], "Input",
 CellLabel->
  "In[582]:=",ExpressionUUID->"078bcc13-cdcc-4cb8-b860-8bea309fa0b1"],

Cell[BoxData[
 RowBox[{
  RowBox[{"omega", " ", "=", " ", 
   RowBox[{"CC0", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      FractionBox["CCp", 
       RowBox[{"RCp", "-", 
        FractionBox[
         RowBox[{"\[Delta]\[Chi]", 
          RowBox[{"(", 
           RowBox[{"1", "-", "q"}], ")"}]}], 
         SqrtBox["r"]]}]], "-", 
      FractionBox["CCm", 
       RowBox[{"RCm", "-", 
        FractionBox[
         RowBox[{"\[Delta]\[Chi]", 
          RowBox[{"(", 
           RowBox[{"1", "-", "q"}], ")"}]}], 
         SqrtBox["r"]]}]]}], ")"}]}]}], ";"}]], "Input",
 CellLabel->
  "In[326]:=",ExpressionUUID->"105b8acc-424e-445e-a984-ab148e2ef65f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", 
  RowBox[{"tauq1", " ", "=", " ", 
   FractionBox[
    RowBox[{"4", " ", "\[Pi]", " ", 
     SuperscriptBox["r", 
      RowBox[{"13", "/", "4"}]]}], 
    RowBox[{"3", " ", 
     SqrtBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          SqrtBox["r"], "-", "\[Chi]eff"}], ")"}], "2"]}]]}]]}]}]], "Input",
 CellLabel->
  "In[327]:=",ExpressionUUID->"a384e834-bf69-437d-ba2b-6dfbf1c2e43e"],

Cell[BoxData[
 FractionBox[
  RowBox[{"4", " ", "\[Pi]", " ", 
   SuperscriptBox["r", 
    RowBox[{"13", "/", "4"}]]}], 
  RowBox[{"3", " ", 
   SqrtBox[
    RowBox[{
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        SqrtBox["r"], "-", "\[Chi]eff"}], ")"}], "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], 
      ")"}]}]]}]]], "Output",
 CellLabel->"Out[327]=",ExpressionUUID->"3d5f5a92-6c1e-47fe-a826-ce56716c0a76"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"omegaq1", "=", 
  RowBox[{
   RowBox[{"Limit", "[", 
    RowBox[{"omega", ",", 
     RowBox[{"q", "->", "1"}]}], "]"}], "//", "Simplify"}]}]], "Input",
 CellLabel->
  "In[328]:=",ExpressionUUID->"822353c1-e180-42f8-a4cf-4f9fb809be54"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   SqrtBox[
    RowBox[{"1", "+", 
     FractionBox[
      RowBox[{"8", " ", "\[Kappa]"}], 
      SqrtBox["r"]]}]], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"7", " ", 
      SqrtBox["r"]}], "-", 
     RowBox[{"6", " ", "\[Chi]eff"}]}], ")"}]}], 
  RowBox[{"8", " ", 
   SuperscriptBox["r", "3"]}]]], "Output",
 CellLabel->"Out[328]=",ExpressionUUID->"7c0025ed-ef22-489a-bb7e-dc5d409a9368"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"alphaq1", "=", 
  RowBox[{"tauq1", "*", "omegaq1"}]}]], "Input",
 CellLabel->
  "In[329]:=",ExpressionUUID->"8f838ca1-f267-4e5b-a4e2-b0b08c88542c"],

Cell[BoxData[
 FractionBox[
  RowBox[{"\[Pi]", " ", 
   SuperscriptBox["r", 
    RowBox[{"1", "/", "4"}]], " ", 
   SqrtBox[
    RowBox[{"1", "+", 
     FractionBox[
      RowBox[{"8", " ", "\[Kappa]"}], 
      SqrtBox["r"]]}]], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"7", " ", 
      SqrtBox["r"]}], "-", 
     RowBox[{"6", " ", "\[Chi]eff"}]}], ")"}]}], 
  RowBox[{"6", " ", 
   SqrtBox[
    RowBox[{
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        SqrtBox["r"], "-", "\[Chi]eff"}], ")"}], "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], 
      ")"}]}]]}]]], "Output",
 CellLabel->"Out[329]=",ExpressionUUID->"78556446-5418-446e-bee9-7c416925470a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["(ix) chi_p estimator", \
"Section",ExpressionUUID->"13be2cb5-8e9e-496b-8330-489956715222"],

Cell[BoxData[
 RowBox[{"Clear", "@@", 
  RowBox[{"DeleteCases", "[", 
   RowBox[{
    RowBox[{"Names", "@", "\"\<`*\>\""}], ",", 
    RowBox[{"\"\<ToPython\>\"", "|", "\"\<precessionform\>\""}]}], 
   "]"}]}]], "Input",
 CellLabel->
  "In[330]:=",ExpressionUUID->"d414464f-e2b3-4bca-ac24-9226f986e938"],

Cell[BoxData[
 RowBox[{
  RowBox[{"\[CapitalOmega]t", " ", "=", " ", 
   RowBox[{"q", 
    FractionBox[
     RowBox[{
      RowBox[{"4", "q"}], "+", "3"}], 
     RowBox[{"4", "+", 
      RowBox[{"3", "q"}]}]]}]}], ";"}]], "Input",
 CellLabel->
  "In[331]:=",ExpressionUUID->"16a732eb-32ff-4866-bb36-5388675a9228"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"cos\[Theta]1", "=", 
   FractionBox[
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"\[Chi]eff", "+", "\[Delta]\[Chi]"}], ")"}]}], 
    RowBox[{"2", " ", "\[Chi]1"}]]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"cos\[Theta]2", "=", 
   FractionBox[
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"\[Chi]eff", "-", "\[Delta]\[Chi]"}], ")"}]}], 
    RowBox[{"2", " ", "q", " ", "\[Chi]2"}]]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"cos\[Theta]12", " ", "=", 
   RowBox[{
    FractionBox["1", 
     RowBox[{"2", " ", "\[Chi]1", " ", "\[Chi]2", " ", 
      SuperscriptBox["q", "2"]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"q", " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", "\[Kappa]", " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}]}], "-", " ", 
         RowBox[{"\[Chi]eff", 
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}]}], "-", 
         RowBox[{"\[Delta]\[Chi]", 
          RowBox[{"(", 
           RowBox[{"1", "-", "q"}], ")"}]}]}], ")"}], " ", 
       SqrtBox["r"]}], " ", "-", " ", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["\[Chi]1", "2"], "+", 
        RowBox[{
         SuperscriptBox["\[Chi]2", "2"], "  ", 
         SuperscriptBox["q", "4"]}]}], ")"}]}], ")"}]}]}], " ", 
  ";"}]}], "Input",
 CellLabel->
  "In[332]:=",ExpressionUUID->"1215bcd5-e04c-4fbd-87cd-b996b8572372"],

Cell[BoxData[
 RowBox[{
  RowBox[{"r", " ", "=", " ", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "+", "q"}], ")"}], "^", "2"}], " ", "/", " ", 
      RowBox[{"(", 
       RowBox[{"2", " ", "u", " ", "q"}], " ", ")"}]}], ")"}], "^", "2"}]}], 
  ";"}]], "Input",
 CellLabel->
  "In[335]:=",ExpressionUUID->"75570627-b6d3-4243-a324-0c1bb5585e30"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"chips", " ", "=", " ", 
  RowBox[{"Simplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      SuperscriptBox["\[Chi]1", "2"], 
      RowBox[{"(", 
       RowBox[{"1", "-", 
        SuperscriptBox["cos\[Theta]1", "2"]}], " ", ")"}]}], "+", 
     RowBox[{
      SuperscriptBox["\[CapitalOmega]t", "2"], " ", 
      SuperscriptBox["\[Chi]2", "2"], 
      RowBox[{"(", 
       RowBox[{"1", "-", 
        SuperscriptBox["cos\[Theta]2", "2"]}], " ", ")"}]}], " ", "+", " ", 
     RowBox[{"2", " ", "\[CapitalOmega]t", " ", "\[Chi]1", " ", "\[Chi]2", 
      RowBox[{"(", 
       RowBox[{"cos\[Theta]12", " ", "-", " ", 
        RowBox[{"cos\[Theta]1", " ", "cos\[Theta]2"}]}], ")"}]}]}], ",", 
    RowBox[{
     RowBox[{"0", "<", "q", "<", "1"}], "&&", 
     RowBox[{"u", ">", "0"}]}]}], "]"}]}]], "Input",
 CellLabel->
  "In[336]:=",ExpressionUUID->"39fbf7ec-2400-4883-a77f-07c41413979d"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "4"], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"4", " ", 
     SuperscriptBox["\[Chi]1", "2"]}], "+", 
    FractionBox[
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"3", "+", 
         RowBox[{"4", " ", "q"}]}], ")"}], "2"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"4", " ", 
         SuperscriptBox["q", "2"], " ", 
         SuperscriptBox["\[Chi]2", "2"]}], "-", 
        RowBox[{
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"\[Delta]\[Chi]", "-", "\[Chi]eff"}], ")"}], "2"]}]}], 
       ")"}]}], 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"4", "+", 
        RowBox[{"3", " ", "q"}]}], ")"}], "2"]], "-", 
    RowBox[{
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"\[Delta]\[Chi]", "+", "\[Chi]eff"}], ")"}], "2"]}], "+", 
    RowBox[{
     FractionBox["1", 
      RowBox[{"q", " ", 
       RowBox[{"(", 
        RowBox[{"4", "+", 
         RowBox[{"3", " ", "q"}]}], ")"}]}]], 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       RowBox[{"3", "+", 
        RowBox[{"4", " ", "q"}]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         SuperscriptBox["\[Chi]1", "2"]}], "-", 
        RowBox[{"2", " ", 
         SuperscriptBox["q", "4"], " ", 
         SuperscriptBox["\[Chi]2", "2"]}], "+", 
        FractionBox[
         RowBox[{
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"1", "+", "q"}], ")"}], "3"], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", "\[Delta]\[Chi]"}],
             "+", 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"1", "+", "q"}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"2", " ", "\[Kappa]"}], "-", "\[Chi]eff"}], ")"}]}]}], 
           ")"}]}], "u"], "+", 
        RowBox[{"q", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
         RowBox[{"(", 
          RowBox[{"\[Delta]\[Chi]", "-", "\[Chi]eff"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"\[Delta]\[Chi]", "+", "\[Chi]eff"}], ")"}]}]}], 
       ")"}]}]}]}], ")"}]}]], "Output",
 CellLabel->"Out[336]=",ExpressionUUID->"9ddc70b4-b9db-415e-9b88-71c33bf1ef05"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"overallchip", " ", "=", " ", 
  RowBox[{
   FractionBox[
    RowBox[{"1", "+", "q"}], 
    RowBox[{"4", " ", "q", " ", "u", " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"4", "+", 
        RowBox[{"3", " ", "q"}]}], ")"}], "2"]}]], "//", 
   "FullSimplify"}]}]], "Input",
 CellLabel->
  "In[337]:=",ExpressionUUID->"9847d45e-a9c9-4d57-a195-69f53e2dec08"],

Cell[BoxData[
 FractionBox[
  RowBox[{"1", "+", "q"}], 
  RowBox[{"4", " ", "q", " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"4", "+", 
      RowBox[{"3", " ", "q"}]}], ")"}], "2"], " ", "u"}]]], "Output",
 CellLabel->"Out[337]=",ExpressionUUID->"9ba121b1-a906-435b-a4bc-a36bd27dbeda"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"coeffs", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"CoefficientList", "[", 
     RowBox[{
      RowBox[{"chips", "/", "overallchip"}], ",", "\[Delta]\[Chi]"}], "]"}], "//",
     "Expand"}], "//", "FullSimplify"}]}]], "Input",
 CellLabel->
  "In[338]:=",ExpressionUUID->"09e31e37-d179-4922-855e-da0a2774fde9"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"4", " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "3"], " ", 
     RowBox[{"(", 
      RowBox[{"4", "+", 
       RowBox[{"3", " ", "q"}]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"3", "+", 
       RowBox[{"4", " ", "q"}]}], ")"}], " ", "\[Kappa]"}], "+", 
    RowBox[{"12", " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", "u", " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{"4", "+", 
          RowBox[{"3", " ", "q"}]}], ")"}], " ", 
        SuperscriptBox["\[Chi]1", "2"]}], "-", 
       RowBox[{
        SuperscriptBox["q", "3"], " ", 
        RowBox[{"(", 
         RowBox[{"3", "+", 
          RowBox[{"4", " ", "q"}]}], ")"}], " ", 
        SuperscriptBox["\[Chi]2", "2"]}]}], ")"}]}], "-", 
    RowBox[{"2", " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "3"], " ", 
     RowBox[{"(", 
      RowBox[{"4", "+", 
       RowBox[{"3", " ", "q"}]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"3", "+", 
       RowBox[{"4", " ", "q"}]}], ")"}], " ", "\[Chi]eff"}], "-", 
    RowBox[{"49", " ", "q", " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "3"], " ", "u", " ", 
     SuperscriptBox["\[Chi]eff", "2"]}]}], ",", 
   RowBox[{"2", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "q"}], ")"}], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
    RowBox[{"(", 
     RowBox[{"12", "+", 
      RowBox[{"q", " ", 
       RowBox[{"(", 
        RowBox[{"25", "+", 
         RowBox[{"12", " ", "q"}], "+", 
         RowBox[{"7", " ", "u", " ", "\[Chi]eff"}]}], ")"}]}]}], ")"}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "q"}], ")"}], "2"]}], " ", "q", " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], " ", "u"}]}], "}"}]], "Output",
 CellLabel->"Out[338]=",ExpressionUUID->"a47ce3d6-7fed-46e4-ba9b-34a7fde2dc16"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Lambda]2", " ", "=", 
   RowBox[{
    RowBox[{"-", " ", "q"}], "  ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"1", "-", "q"}], ")"}], "2"], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "q"}], ")"}], " ", "u"}]}], " ", 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[Lambda]2", "  ", "==", " ", 
   RowBox[{"coeffs", "[", 
    RowBox[{"[", "3", "]"}], "]"}]}], "//", "Simplify"}]}], "Input",
 CellLabel->
  "In[339]:=",ExpressionUUID->"c08b3b0e-6a1a-454c-98f2-6f98ec8d5f0d"],

Cell[BoxData["True"], "Output",
 CellLabel->"Out[340]=",ExpressionUUID->"540ca1ef-0562-4591-8172-e969d0a32480"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Lambda]1", " ", "=", 
   RowBox[{
    RowBox[{"-", " ", "2"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "-", "q"}], ")"}], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"1", "+", "q"}], ")"}], "2"], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{"4", "+", 
         RowBox[{"3", " ", "q"}]}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"3", "+", 
         RowBox[{"4", " ", "q"}]}], ")"}]}], " ", "+", 
      RowBox[{"7", "  ", "q", " ", "u", " ", "\[Chi]eff"}]}], " ", ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[Lambda]1", "==", 
   RowBox[{"coeffs", "[", 
    RowBox[{"[", "2", "]"}], "]"}]}], "//", "Simplify"}]}], "Input",
 CellLabel->
  "In[341]:=",ExpressionUUID->"349efc9b-9d4e-4502-a937-f9688c0c864e"],

Cell[BoxData["True"], "Output",
 CellLabel->"Out[342]=",ExpressionUUID->"61d5b1a2-715c-4a21-ac91-8b8c7561f9d9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Lambda]0", "=", 
   RowBox[{
    RowBox[{"2", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", "q"}], ")"}], "3"], 
     RowBox[{"(", 
      RowBox[{"4", "+", 
       RowBox[{"3", " ", "q"}]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"3", "+", 
       RowBox[{"4", " ", "q"}]}], ")"}], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"2", " ", "\[Kappa]"}], "-", " ", "\[Chi]eff"}], " ", ")"}]}], 
    " ", "-", "  ", 
    RowBox[{"u", " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"12", " ", 
        RowBox[{"(", 
         RowBox[{"1", "-", "q"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"(", 
            RowBox[{"4", "+", 
             RowBox[{"3", " ", "q"}]}], ")"}], " ", 
           SuperscriptBox["\[Chi]1", "2"]}], "-", 
          RowBox[{
           SuperscriptBox["q", "3"], " ", 
           RowBox[{"(", 
            RowBox[{"3", "+", 
             RowBox[{"4", " ", "q"}]}], ")"}], " ", 
           SuperscriptBox["\[Chi]2", "2"]}]}], ")"}]}], "+", 
       RowBox[{"49", " ", "q", "  ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "q"}], ")"}], "3"], " ", 
        SuperscriptBox[
         RowBox[{"\[Chi]eff", " "}], "2"]}]}], ")"}]}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[Lambda]0", "==", 
   RowBox[{"coeffs", "[", 
    RowBox[{"[", "1", "]"}], "]"}]}], "//", "Simplify"}]}], "Input",
 CellLabel->
  "In[343]:=",ExpressionUUID->"7221ddb1-4e04-4a19-8e3f-977a88f0e899"],

Cell[BoxData["True"], "Output",
 CellLabel->"Out[344]=",ExpressionUUID->"7042ae28-221d-4527-b22b-0d5d8847613a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"precessionform", "[", 
  RowBox[{"\"\<lambdabar\>\"", ",", "overallchip"}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"precessionform", "[", 
  RowBox[{"\"\<lambda2\>\"", ",", "\[Lambda]2"}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"precessionform", "[", 
  RowBox[{"\"\<lambda1\>\"", ",", "\[Lambda]1"}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"precessionform", "[", 
  RowBox[{"\"\<lambda0\>\"", ",", "\[Lambda]0"}], "]"}]}], "Input",
 CellLabel->
  "In[345]:=",ExpressionUUID->"9865ffab-dc54-40a1-a50c-05e32983e9c6"],

Cell[BoxData["\<\"# Machine generated with eq_generator.nb\\nlambdabar = 1/4 \
* q**(-1) * (1 + q) * ((4 + 3 * q))**(-2) * u**(-1)\"\>"], "Output",
 CellLabel->
  "Out[345]=",ExpressionUUID->"e85bf654-2dab-4f3c-b4bb-e641beb94a10"],

Cell[BoxData["\<\"# Machine generated with eq_generator.nb\\nlambda2 = -1 * \
((1 + -1 * q))**2 * q * (1 + q) * u\"\>"], "Output",
 CellLabel->
  "Out[346]=",ExpressionUUID->"1cd2600e-da0f-44c5-80a6-32092eb6bdd5"],

Cell[BoxData["\<\"# Machine generated with eq_generator.nb\\nlambda1 = -2 * \
(1 + -1 * q) * ((1 + q))**2 * ((4 + 3 * q) * (3 + 4 * q) + 7 * q * u * \
chieff)\"\>"], "Output",
 CellLabel->
  "Out[347]=",ExpressionUUID->"4a4ff141-5a10-4dd8-ae95-2f2823a3d0ea"],

Cell[BoxData["\<\"# Machine generated with eq_generator.nb\\nlambda0 = (2 * \
((1 + q))**3 * (4 + 3 * q) * (3 + 4 * q) * (2 * kappa + -1 * chieff) + -1 * u \
* (12 * (1 + -1 * q) * ((4 + 3 * q) * chi1**2 + -1 * q**3 * (3 + 4 * q) * \
chi2**2) + 49 * q * ((1 + q))**3 * chieff**2))\"\>"], "Output",
 CellLabel->"Out[348]=",ExpressionUUID->"d53cc594-7210-4049-bf18-f9b90b2e3938"]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Clean this notebook for inclusion in git repository. ",
 StyleBox["Credits: http://bhptoolkit.org/cleannb.html", "Subsection"],
 " "
}], "Subsubsection",ExpressionUUID->"674ab233-80a7-47cd-9472-8c1e1cbebfa1"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CleanNotebook", "[", "file_", "]"}], ":=", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"nb", ",", "contents", ",", "newcontents"}], "}"}], ",", 
     RowBox[{
      RowBox[{"nb", "=", 
       RowBox[{"NotebookOpen", "[", "file", "]"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"SetOptions", "[", 
       RowBox[{"nb", ",", 
        RowBox[{"\"\<TrackCellChangeTimes\>\"", "->", "False"}], ",", 
        RowBox[{"PrivateNotebookOptions", "->", 
         RowBox[{"{", 
          RowBox[{"\"\<FileOutlineCache\>\"", "->", "False"}], "}"}]}]}], 
       "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"contents", "=", 
       RowBox[{"NotebookGet", "[", "nb", "]"}]}], ";", "\[IndentingNewLine]", 
      
      RowBox[{"newcontents", "=", 
       RowBox[{"contents", "/.", " ", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{
           RowBox[{"(", 
            RowBox[{"CellChangeTimes", "->", "_"}], ")"}], "->", 
           RowBox[{"Sequence", "[", "]"}]}], ",", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"CellTags", "->", "_"}], ")"}], "->", 
           RowBox[{"Sequence", "[", "]"}]}]}], "}"}]}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"NotebookPut", "[", 
       RowBox[{"newcontents", ",", "nb"}], "]"}], ";", "\[IndentingNewLine]", 
      
      RowBox[{"NotebookSave", "[", "nb", "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"NotebookClose", "[", "nb", "]"}], ";"}]}], "]"}]}], 
  ";"}]], "Input",
 CellLabel->
  "In[352]:=",ExpressionUUID->"b4b7b5b3-0764-43b0-8e8a-5bc54510647b"],

Cell[BoxData[
 RowBox[{
  RowBox[{"CleanNotebook", "[", 
   RowBox[{"NotebookFileName", "[", "]"}], "]"}], ";"}]], "Input",
 CellLabel->"In[353]:=",ExpressionUUID->"63c91bf2-4c13-481a-b9d9-e15836f91ec7"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
},
WindowSize->{904, 847},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
PrivateNotebookOptions->{"FileOutlineCache"->False},
TaggingRules-><|"TryRealOnly" -> False|>,
TrackCellChangeTimes->False,
FrontEndVersion->"13.1 for Mac OS X x86 (64-bit) (June 16, 2022)",
StyleDefinitions->"Default.nb",
ExpressionUUID->"ceb27b6a-17ef-40ed-91d0-092d78b52fe3"
]

